/*
 * wkseqm.c    REVISION HISTORY:
 * Generated by SRVWIZARD on Tue Apr 25 22:23:12 2006
*/

#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "errors.h"
#include "rpctypes.h"
#include "swaplib.h"
#include "almsrv.h"
#include "toolkit.h"
#include "wkseqm.h"

float g_temperatureBuffer[PRP_TEMPERATURE_SIZE];
float g_amplitudeBuffer[PRP_AMPLITUDE_SIZE];
float g_amplitudeIn;
float g_sineBuffer[NUM_WKSEQM_DEVICES][PRP_SINE_SIZE];

void wkseqm_bkg(void)
{
  int i, n;
  /* TODO: put your IO or other background activity here         */

  /* Clear alarms at start of IO task (see if they come back)    */
  ClearAlarm(WKSEQM_TAG,-1);

  /* TODO: replace SimulateData() with your own data acquisition       */
  /*       and manipulation routines                                   */
  for (i=0; i<PRP_TEMPERATURE_SIZE; i++)
  {
     g_temperatureBuffer[i] = (float)SimulateData(0,100);
  }
  /* TODO: replace SimulateData() with your own data acquisition       */
  /*       and manipulation routines                                   */
  for (n=0; n<NUM_WKSEQM_DEVICES; n++)
    for (i=0; i<PRP_SINE_SIZE; i++)
  {
     g_sineBuffer[n][i] = (float)SimulateData(-500,1000);
  }

  /* make use of SetAlarm() as needed :

  SetAlarm(WKSEQM_TAG,devNr,almCode,almFlag);

  */
}

/* Alternative to 'exports.csv': You can call 'RegisterWkseqmProperties()'   */
/*   -> This might be preferable in cases where the server does not have a   */
/*   -> file system.  You should also call RegisterServerName() in lieu of   */
/*   -> the 'fecid.csv' file.                                                */

int RegisterWkseqmProperties(void)
{
  int cc = 0;
  DTYPE dout,din;

  dout.dArrayLength = PRP_TEMPERATURE_SIZE;
  dout.dFormat = CF_FLOAT;
  din.dArrayLength = 0;
  din.dFormat = CF_NULL;
  if ((cc=RegisterPropertyInformation(WKSEQM_TAG,"Temperature",&dout,&din,CA_READ,AT_CHANNEL,0,"[0:100 'C]Temperature of the Sine generato",PRP_TEMPERATURE,"")) < 0) goto err;

  dout.dArrayLength = PRP_AMPLITUDE_SIZE;
  dout.dFormat = CF_FLOAT;
  din.dArrayLength = PRP_AMPLITUDE_INSIZE;
  din.dFormat = CF_FLOAT;
  if ((cc=RegisterPropertyInformation(WKSEQM_TAG,"Amplitude",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"[0:500 V]Sine Curve Amplitude",PRP_AMPLITUDE,"")) < 0) goto err;

  dout.dArrayLength = PRP_SINE_SIZE;
  dout.dFormat = CF_FLOAT;
  din.dArrayLength = 0;
  din.dFormat = CF_NULL;
  if ((cc=RegisterPropertyInformation(WKSEQM_TAG,"Sine",&dout,&din,CA_READ,AT_SPECTRUM,0,"[-500:500 V]Sine Curve",PRP_SINE,"")) < 0) goto err;

err:
  if (cc < 0)
  {
    printf("Can't register property : %s\n>",erlst[-cc]);
  }
  return cc;
}


void wkseqm_ini(void)
{
  char devnam[16];
  int i;
  /* TODO: put your initialization here */
  /* If you plan to use Property registration via API call rather than          */
  /* the exports.csv startup file, uncomment the following line:                */

  /* RegisterWkseqmProperties(); */

  /* TODO: register any device names here */
  /*       below is only an example ...   */
  for (i=0; i<NUM_WKSEQM_DEVICES; i++)
  {
    sprintf(devnam,"device_%d",i);
    RegisterDeviceName(WKSEQM_TAG,devnam,i);
  }

  /* TODO: register any global receive routines as in this example  */
  /*

  glbID = recvNetGlobal("HPMAGEN",&dout,rcvGlobal);

  */

  /* TODO: start any links you might need as in this example  */
  /*

  lnkID = AttachLinkEx("/HERA/BPM/WL197X","POSITIONS.X",&dout,&din,CA_READ,1000,showlink,CM_POLL,myid);

  */
}

void wkseqm_exi(void)
{
  /* TODO: put your shutdown routines here */
}

int wkseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,i,cc;
  float l_amplitude;

  /* get device number from device name                                                     */
  devnr = GetDeviceNumberEx(WKSEQM_TAG,devName,devProperty);
  /* you may want to make the following check:                                              */
  if (devnr < 0) return illegal_equipment_number;

  /* TODO: If READ properties take input data, include code to examine the contents of din. */
  /*       If different actions need to be taken at the start or end of a link, examine the */
  /*       'access' parameter against CA_FIRST or CA_LAST.                                  */
  /*       If allow format overloading (you return different data according to the request  */
  /*       format), then replace calls to putDataFromShort() etc with the desired code.     */

  prpid = GetPropertyId(WKSEQM_TAG,devProperty);

  switch (prpid)
  {
    case PRP_TEMPERATURE:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_TEMPERATURE_SIZE) return dimension_error;
          if ((cc=putValuesFromFloatEx(dout,g_temperatureBuffer,PRP_TEMPERATURE_SIZE,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_AMPLITUDE:
        if (access&CA_WRITE)
        {
          if (din->dArrayLength > 0)
          {
            if (din->dArrayLength > PRP_AMPLITUDE_INSIZE) return dimension_error;
            if ((cc=getValuesAsFloat(din,&l_amplitude,PRP_AMPLITUDE_INSIZE)) != 0) return cc;
            if (l_amplitude > PRP_AMPLITUDE_UPR_LIMIT) return out_of_range;
            if (l_amplitude < PRP_AMPLITUDE_LWR_LIMIT) return out_of_range;
            g_amplitudeIn = l_amplitude;
          }
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_AMPLITUDE_SIZE) return dimension_error;
          if ((cc=putValuesFromFloatEx(dout,g_amplitudeBuffer,PRP_AMPLITUDE_SIZE,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_SINE:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_SINE_SIZE) return dimension_error;
          if ((cc=putValuesFromFloat(dout,g_sineBuffer[devnr],PRP_SINE_SIZE)) != 0) return cc;
        }
        return 0;
    default:
        return illegal_property;
  }
}
