#include "prolog.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "errors.h"
#include "rpctypes.h"
#include "database.h"
#include "toolkit.h"
#include "swaplib.h"
#include "archive.h"
#if defined(UNIX)
# include "unixcmd.h"
#elif defined(WINNT)
# include "ntcmd.h"
#endif

extern char ArchiveDBpath[];
int totalNrFECs, nfecs;
char cbuff[256];

#define MAX_ADBFILES 20
char adbfile[MAX_ADBFILES][32];
int nadbfiles = 0;
int archdebug = 0,archPending = 0;
int g_isTerminating = 0;
int AppendProblemsLog(char *s) { return 0;}

#define DATABLOB_SIZE 16384
BYTE g_datablob[DATABLOB_SIZE];
#define TRACE_BUFFER_SIZE DATABLOB_SIZE
BYTE *traceDataBuffer = g_datablob;
#define MAXNAMES (DATABLOB_SIZE/16)
NAME16 *name16blob = (NAME16 *)g_datablob;

int getNumLines(FILE *fp);
int getdata(char *fname, int ntotal, int offset);
int opendata(int mon, int day, int hour);
long loadData(int n, int p, int index, time_t starttime,time_t endtime,
             void *datbuf,int datfmt,int bufsize,int sampleFlag,
             float upr,float lwr);
int getEntryFromRectyp(int rectyp);
int getArchivedEntryFromKeyword(char *keyword,int *rectyp,int *param);
int fetch(char *keyword,int index,time_t starttime,
          time_t stoptime,int arraysize,BYTE *data,int datafmt,int sampleFlag,
          float upr, float lwr);
int fetchTrace(char *keyword,time_t tgttime,
               int arraysize,BYTE *data,int datafmt,int index,int sample);
long locateRecord(int i,time_t t);
void initHDA(void);
int hdaeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
int OpenDataFile(int i);
time_t getArchiveStartTimeFromDepth(char *keyword,int depth);

FEC *fec;
int ntags;

#define addNumberFecs(f) {\
  sprintf(fn,"%s%s",ArchiveDBpath,f);\
  if ((fp = fopen(fn,"rt")) == NULL) return;\
  totalNrFECs += getNumLines(fp);\
  fclose(fp); }

#define ccerr(i) { cc = i; goto err; }
#define cctxt(txt) { printf("HDA: %s\n>",txt); ccerr(database_not_loaded); }
#define quit(i) { printf("HDA: %s\n>",erlst[i]); exit(1); }
int getDatabaseInfo(void)
{
  FILE *fp=NULL;
  char str[256],scratch[170];
  int done = 0,cc = 0;
  int database_col;

  if ((fp=fopen("database.csv","rt")) == NULL)
  {
    printf("\n Can't open data base file : database.csv\n");
    exit(1);
  }
  while (fgets(str,255,fp) != NULL )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((database_col=findcol(str,"DATABASE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(database_col,str),170);
    addADBtoList(scratch);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int addADBtoList(char *fn)
{
  int i;

  for (i=0; i<nadbfiles && strncmp(adbfile[i],fn,16); i++);
  if (i < nadbfiles) return 0;
  if (nadbfiles == MAX_ADBFILES) return out_of_local_memory;
  strncpy(adbfile[i],fn,16);
  nadbfiles++;
  return 0;
}
void PostSystemInit(void)
{
  int cc;
  signal(SIGFPE,SIG_IGN); /* catch floating point domain errors */
  if ((cc=RegisterEquipmentModule(NULL,"HDAEQM",0,hdaeqm,initHDA,NULL,0,NULL)) != 0)
    printf("Can't register HDAEQM: %s\n",erlst[cc]);
}
void RegisterStructs(void)
{
  if (fillstruct("ARS",OFFSETIN(ARS,starttime),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ARS",OFFSETIN(ARS,index),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("ARS",OFFSETIN(ARS,minutes),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("ARS",OFFSETIN(ARS,keyword),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ARS",sizeof(ARS),1,CF_NULL)) quit(out_of_local_memory);

  RegisterUserCommand("ArchiveDebug",NULL,&archdebug,NULL,CA_READ+CA_WRITE);

}
void initHDA(void)
{
  int id,i,n,np,off=0,cc=0;
  static char fn[80],desc[256];
  FECDATA *fecdata;
  int p, rectyp;
  FILE *fp;
  DTYPE dout;

  getDatabaseInfo();
  RegisterStructs();
  totalNrFECs = 0;
  for (i=0; i<nadbfiles; i++)
  {
    addNumberFecs(adbfile[i]);
  }
  if ((fec=(FEC *)calloc(totalNrFECs,sizeof(FEC))) == NULL) ccerr(out_of_server_memory);
  if ((prpAlias=(PRP_ALIAS *)calloc(totalNrFECs,sizeof(PRP_ALIAS))) == NULL) ccerr(out_of_server_memory);
  if ((subSystems=(SUB_SYS *)calloc(totalNrFECs,sizeof(SUB_SYS))) == NULL) ccerr(out_of_server_memory);

  for (i=0; i<nadbfiles; i++)
  {
    if (!(n=archLoad(adbfile[i],off))) cctxt("No Servers loaded!");
    if (!(np=paramLoad(n,off))) cctxt("No Parameters loaded!");
    off += n;
  }
  nfecs = totalNrFECs;
  fillinArrayTypes(nfecs);
  for (i=0,ntags=0; i<totalNrFECs; i++)       /* register properties */
  {
    if (fec[i].option&NEVER) continue; /* it's not being archived */
    if (!fec[i].active) continue;      /* legacy entry, don't count it! */
    ntags += fec[i].nparam;
    for (n=0; n<fec[i].nparam; n++)
    {
      sprintf(desc,"[%g:%g %s!%s]%s",fec[i].fecdata[n].minimum,
              fec[i].fecdata[n].maximum,fec[i].fecdata[n].units,
              fec[i].fecdata[n].plot,fec[i].fecdata[n].desc);
      dout.dFormat = fec[i].fecdata[n].format;
      dout.dArrayLength = fec[i].fecdata[n].size;
      dout.dTag[0] = 0;
      if ((id=RegisterPropertyInformation("HDAEQM",fec[i].fecdata[n].param,
           &dout,NULL,CA_READ,fec[i].fecdata[n].arrayType,0,desc,-1,NULL)) < 0)
      {
        cc = -id;
        printf("Couldn't register property %s: %s\n",fec[i].fecdata[n].param,erlst[cc]);
      }
    }
  }
  for (n=0; n<nPrpAlias; n++)
  {
    if ((i=getArchivedEntryFromKeyword(prpAlias[n].alias,&rectyp,&p)) < 0) continue;
    fecdata = fec[i].fecdata;
    dout.dFormat = fecdata[p].format;
    dout.dArrayLength = fecdata[p].size;
    dout.dTag[0] = 0;
    if ((id=RegisterPropertyInformation("HDAEQM",prpAlias[n].name,
         &dout,NULL,CA_READ,fecdata[p].arrayType,0,desc,-1,NULL)) < 0)
    {
      cc = -id;
      printf("Couldn't register property %s: %s\n",prpAlias[n].name,erlst[cc]);
    }
  }

  RegisterProperty("HDAEQM","NSUBSYSTEMS",1,CF_LONG,CA_READ,"number of archive subsystems");
  RegisterProperty("HDAEQM","SUBSYSTEMS",nSubSystems,CF_NAME16,CA_READ,"archive subsystems");
  RegisterProperty("HDAEQM","NTAGS",1,CF_LONG,CA_READ,"number of archive subsystems");
  RegisterProperty("HDAEQM","TAGS",ntags,CF_NAME16,CA_READ,"archive subsystems");
  RegisterProperty("HDAEQM","RECORDSIZE",1,CF_LONG,CA_READ,"size of record at timestamp");
  RegisterProperty("HDAEQM","RECORDSOURCE",3,CF_NAME32,CA_READ,"source of record");

err:if (cc)
  {
    printf("HDA: %s\n>",erlst[cc]);
    if (fec != NULL) free(fec);
    exit(1);
  }
  else printf("HDA: Data Access server initialized\n>");
  return;
}

#define getDataFilename(i,yy,mm) \
    sprintf(fec[i].fname,"ar%02d%02d01.%0x",yy,mm,fec[i].rectyp);\
    fec[i].archYear = 1900 + yy;\
    fec[i].archMonth = mm;

int fileExists(int i,time_t *mtime)
{
  int hFile, found = 0;
  struct stat sbuf;
  if ((hFile=OpenDataFile(i)) != -1) 
  {
    found = -1;
    fstat(hFile,&sbuf);
    *mtime = sbuf.st_mtime;
    close(hFile);
  }
  return found;
}
char *months[] = {"Jan","Feb","Mar","Apr","May","Jun",
                  "Jul","Aug","Sep","Oct","Nov","Dec"};
int getmonth(char *m)
{
  int i;
  for (i=0; i<12; i++) if (!stricmp(m,months[i])) return i+1;
  return 0;
}

int getEntryFromRectyp(int rectyp)
{
  int i;

  for (i=0; i<totalNrFECs; i++)
    if (fec[i].rectyp == rectyp) return i;
  return -1;
}
int getLegacyEntryFromKeyword(char *keyword,int *rectyp,int *param)
{
  FECDATA *fecdata;
  int i,j;

  for (i=0; i<totalNrFECs; i++)
  {
    if (fec[i].active && !(fec[i].option&NEVER)) continue;
    fecdata = fec[i].fecdata;
    for (j=0; j<fec[i].nparam; j++)
    {
      if (!strcmp(fecdata[j].param,keyword))
      {
        *rectyp = fec[i].rectyp;
        *param  = j;
        return i;
      }
    }
  }
  return -1;
}
int getArchivedEntryFromKeyword(char *keyword,int *rectyp,int *param)
{
  FECDATA *fecdata;
  int i,j;

  for (i=0; i<totalNrFECs; i++)
  {
    if (!fec[i].active) continue;
    if (fec[i].option&NEVER) continue;
    fecdata = fec[i].fecdata;
    for (j=0; j<fec[i].nparam; j++)
    {
      if (!strcmp(fecdata[j].param,keyword))
      {
        *rectyp = fec[i].rectyp;
        *param  = j;
        return i;
      }
    }
  }
  return -illegal_property;
}
int getArchivedEntryFromKeywordAndTime(char *keyword,int *rectyp,int *param,time_t tgt)
{
  FECDATA *fecdata;
  int i,j,best=-1,keyExists=FALSE;
  time_t mtime;
  struct tm *lt=localtime((time_t *)&tgt);

  for (i=0; i<totalNrFECs; i++)
  {
    if (fec[i].option&NEVER) continue;
    fecdata = fec[i].fecdata;
    for (j=0; j<fec[i].nparam; j++)
    {
      if (!strcmp(fecdata[j].param,keyword))
      {
        keyExists = TRUE;
        *rectyp = fec[i].rectyp;
        *param  = j;
        getDataFilename(i,lt->tm_year,lt->tm_mon+1);
        if (fileExists(i,&mtime)) 
        {
          if (mtime > tgt) return i;
          if (fec[i].active) return i;
          best = i;
        }
      } 
    }
  }
  if (best >= 0) return best;
  return keyExists ? -out_of_range : -illegal_property;
}

int getIndexFromDeviceName(char *devName,char *devProperty)
{
  FECDATA *fecdata;
  DTYPE ddin,ddout;
  int cc, i, siz, n, p, rectyp;
  char namProperty[32] = "";

  if ((n=getArchivedEntryFromKeyword(devProperty,&rectyp,&p)) < 0) return n;
  fecdata = fec[n].fecdata;
  if ((siz=fecdata[p].size) > MAXNAMES) siz = MAXNAMES; 
  ddout.dArrayLength = siz;
  ddout.dFormat = CF_NAME16;
  ddout.data.vptr = name16blob;
  ddin.dArrayLength = 0;
  ddin.dFormat = CF_NULL;
  ddin.data.vptr = NULL;
  strncpy(namProperty,devProperty,27);
  strcat(namProperty,".NAM");
  if ((cc=hdaeqm("#0",namProperty,&ddout,&ddin,CA_READ)) != 0) return -cc;
  for (i=0; i<siz && strncmp(devName,((NAME16 *)ddout.data.vptr)[i].name,16); i++);
  if (i == siz) return -illegal_equipment_number;
  return i; 
}
int getIndexFromDeviceNameEx(char *devName,char *devProperty,DTYPE *din)
{
  FECDATA *fecdata;
  DTYPE ddout,ddin;
  int cc, i, siz, n, p, rectyp;
  int startstopArray[2];
  char namProperty[32] = "";

  if ((n=getArchivedEntryFromKeyword(devProperty,&rectyp,&p)) < 0) return n;
  fecdata = fec[n].fecdata;
  if ((siz=fecdata[p].size) > MAXNAMES) siz = MAXNAMES; 
  ddout.dArrayLength = siz;
  ddout.dFormat = CF_NAME16;
  ddout.data.vptr = name16blob;
  ddin.dArrayLength = 2;
  ddin.dFormat = CF_LONG;
  if (din == NULL || din->data.lptr == NULL || din->data.lptr[0] == 0)
  {
    startstopArray[1] = time(NULL);
    startstopArray[0] = startstopArray[1]-3600;
  }
  else
  {
    startstopArray[0] = din->data.lptr[0];
    startstopArray[1] = din->data.lptr[1];
  }
  ddin.data.lptr = (UINT32 *)startstopArray;
  strncpy(namProperty,devProperty,27);
  strcat(namProperty,".NAM");
  if ((cc=hdaeqm("#0",namProperty,&ddout,&ddin,CA_READ)) != 0) return -cc;
  for (i=0; i<siz && strncmp(devName,((NAME16 *)ddout.data.vptr)[i].name,16); i++);
  if (i == siz) return -illegal_equipment_number;
  return i; 
}
int srchcmp(FECDATA *f,char *s)
{
  char str[80],*c=str,*ss,substr[48],keystr[96];
  int i,found=FALSE,orflg=FALSE,andflg=FALSE;
  strncpy(str,s,80);
  while (c != NULL)
  {
    ss = c; found = FALSE;
    if ((c=strchr(ss,',')) != NULL || (c=strchr(ss,'|')) != NULL)
    {
      *c++ = 0; orflg = TRUE;
    }   
    if ((c=strchr(ss,'+')) != NULL || (c=strchr(ss,'&')) != NULL)
    {
      *c++ = 0; andflg = TRUE;
    }
    for (i=0; i<48 && ss[i] == ' '; i++);   
    strncpy(substr,&ss[i],48); strupr(substr);
    if ((ss=strchr(substr,' ')) != NULL) *ss = 0; 
    strcpy(keystr,f->param); strcat(keystr,f->subsys);
    strcat(keystr,f->desc); strupr(keystr);
    if (strstr(keystr,substr) != NULL) found = TRUE;
    if (andflg && !found) return FALSE;
    if (!andflg && found) return TRUE;
  }
  return found ? TRUE : FALSE;
}
void assignNrOfPoints(BYTE *data,int fmt,int npoints)
{
  if (data == NULL) return;
  switch (fmt)
  {
    case CF_FLTINT:
      ((FLTINT *)data)[0].ival = (UINT32)(npoints);
      ((FLTINT *)data)[0].fval = (float)(npoints);
      break;
    case CF_FLTINTINT:
      ((FLTINTINT *)data)[0].i1val = (UINT32)(npoints);
      ((FLTINTINT *)data)[0].i2val = (UINT32)(npoints);
      ((FLTINTINT *)data)[0].fval = (float)(npoints);
      break;
    case CF_INTINT:
      ((INTINT *)data)[0].i1val = (UINT32)(npoints);
      ((INTINT *)data)[0].i2val = (UINT32)(npoints);
      break;
    case CF_DBLDBL:
      ((DBLDBL *)data)[0].d1val = (double)(npoints);
      ((DBLDBL *)data)[0].d2val = (double)(npoints);
      break;
    case CF_NAME16I:
      ((NAME16I *)data)[0].ival = (UINT32)(npoints);
      break;
    case CF_NAME32I:
      ((NAME32I *)data)[0].ival = (UINT32)(npoints);
      break;
    case CF_NAME48I:
      ((NAME48I *)data)[0].ival = (UINT32)(npoints);
      break;
    case CF_NAME64I:
      ((NAME64I *)data)[0].ival = (UINT32)(npoints);
      break;
    case CF_LONG:
      *(UINT32 *)data = (UINT32)(npoints);
      break;
    case CF_FLOAT:
      *(float *)data = (float)(npoints);
      break;
    case CF_INTFLTINT:
      ((INTFLTINT *)data)[0].i1val = (UINT32)(npoints);
      ((INTFLTINT *)data)[0].i2val = (UINT32)(npoints);
      ((INTFLTINT *)data)[0].fval = (float)(npoints);
      break;
    case CF_DOUBLE:
      *(double *)data = (double)(npoints);
      break;
  }
}
int hdaeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int fmt,fmtsize,i,j,n,len,index=0,sample=0,startpos=0;
  int recsreadFlag=0,inputgiven=0;
  float lwrlimit=0,uprlimit=0;
  ARS *ars;
  time_t tnow=time(NULL),starttime=tnow,stoptime=tnow;
  char keyword[36],subs[16],sstr[128];
  int cc,rectyp,p;
  UINT32 tgt;

  if (!strcmp(devProperty,"NTAGS"))
  {
    return putValuesFromLong(dout,(UINT32 *)&ntags,1);
  }
  if (!strcmp(devProperty,"TAGS"))
  {
    subs[0] = 0; sstr[0] = 0;
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_NAME32:
        case CF_NAME16:
          strncpy(subs,din->data.cptr,16);
          break;
        case CF_TEXT:
          if (din->dArrayLength > 80) din->dArrayLength = 80;
          strncpy(sstr,din->data.cptr,din->dArrayLength);
          sstr[din->dArrayLength] = 0;
          break;
        default: 
          return illegal_format;
      }
    }
    switch (dout->dFormat)
    {
      case CF_NAME8: len = 8; break;
      case CF_NAME16: len = 16; break;
      case CF_NAME32: len = 32; break;
      default: return illegal_format;
    }
    for (i=0,n=0; i<totalNrFECs; i++)
    {
      if (!fec[i].active) continue;
      if (fec[i].option&NEVER) continue; /* it's not being archived */
      if (dout->dArrayLength < (UINT32)n) break;
      for (j=0; j<fec[i].nparam; j++)
      {
        if (dout->dArrayLength < (UINT32)n) break;
/*        if (strlen(subs) > 0 && strncmp(fec[i].fecdata[j].subsys,subs,16)) continue; */
        if (strlen(subs) > 0 && !srchcmp(&fec[i].fecdata[j],subs)) continue;
        if (strlen(sstr) > 0 && !srchcmp(&fec[i].fecdata[j],sstr)) continue;
        strncpy((char *)(dout->data.bptr+n*len),fec[i].fecdata[j].param,len);
        n++;
      }
    }
    dout->dArrayLength = n;
    return 0;
  }
  if (!strcmp(devProperty,"NSUBSYSTEMS"))
  {
    return putValuesFromLong(dout,(UINT32 *)&nSubSystems,1);
  }
  if (!strcmp(devProperty,"SUBSYSTEMS"))
  {
    switch (dout->dFormat)
    {
      case CF_NAME8: len = 8; break;
      case CF_NAME16: len = 16; break;
      case CF_NAME32: len = 32; break;
      default: return illegal_format;
    }
    for (i=0; i<(int)dout->dArrayLength && i<nSubSystems; i++)
    {
      strncpy((char *)(dout->data.bptr+i*len),subSystems[i].name,len);
    }
    return 0;
  }
  if (!strcmp(devProperty,"RECORDSOURCE"))
  {
    if (din->dArrayLength == 0) return argument_list_error;
    tgt = time(NULL);
    switch(din->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
        len = fmtsizeof(din->dFormat);
        break;
      case CF_NAME16I:
      case CF_NAME32I:
        len = fmtsizeof(din->dFormat) - 4;
        tgt = *(UINT32 *)(din->data.cptr+len);
        break;
      case CF_TEXT:
        len = din->dArrayLength;
        break;
      default: 
        return illegal_format;
    }
    if (len > 32) len = 32;
    strncpy(keyword,din->data.cptr,len); keyword[len] = 0;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
        break;
      default:
        return illegal_format;
    }
    len = fmtsizeof(dout->dFormat);
    if ((i=getArchivedEntryFromKeywordAndTime(keyword,&rectyp,&p,tgt)) < 0)
      return -i;
    switch (dout->dArrayLength)
    {
      case 3:
        strncpy(dout->data.cptr+2*len,fec[i].devname,len);
      case 2:
        strncpy(dout->data.cptr+1*len,fec[i].property,len);
      case 1:
        strncpy(dout->data.cptr,fec[i].eqm,len);
        break;
      default:
        return dimension_error;
    }
    return 0;
  }
  if (!strcmp(devProperty,"RECORDSIZE"))
  {
    tgt = time(NULL);
    switch (din->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
        len = fmtsizeof(din->dFormat);
        strncpy(keyword,din->data.cptr,len); keyword[len] = 0;
        break;
      case CF_NAME16I:
      case CF_NAME32I:
        len = fmtsizeof(din->dFormat) - 4;
        tgt = *(UINT32 *)(din->data.cptr+len);
        strncpy(keyword,din->data.cptr,len); keyword[len] = 0;
        break;
      case CF_TEXT:
        len = din->dArrayLength;
        if (len > 32) len = 32;
        strncpy(keyword,din->data.cptr,len); keyword[len] = 0;
        break;
      default: 
        if ((cc=getValuesAsLong(din,&tgt,1)) != 0) return cc;
        strncpy(keyword,devName,32);
    }
    if ((i=getArchivedEntryFromKeywordAndTime(keyword,&rectyp,&p,tgt)) < 0) 
      return -i;
    len = fec[i].fecdata[p].size;
    return putValuesFromLong(dout,(UINT32 *)&len,1);
  }
  /* generally true below: */
  if (dout->dArrayLength < 1) return illegal_read_write;
  fmt = dout->dFormat == CF_SI      ? CF_SHORT :
        dout->dFormat == CF_II      ? CF_LONG :
        dout->dFormat == CF_FI      ? CF_FLOAT :
        dout->dFormat == CF_TDS     ? CF_TDS :
        dout->dFormat == CF_DBLDBL  ? CF_DOUBLE :
        dout->dFormat == CF_CHAR8I  ? CF_CHAR8 :
        dout->dFormat == CF_CHAR16I ? CF_CHAR16 :
        dout->dFormat == CF_CHAR32I ? CF_CHAR32 :
        dout->dFormat == CF_CHAR48I ? CF_CHAR48 :
        dout->dFormat == CF_CHAR64I ? CF_CHAR64 : -1;
  fmtsize = fmtsizeof(fmt);
  
  if (!strncmp(devProperty,"FETCH",5))
  {
    if (fmt == -1) return illegal_format;
    if (din->dArrayLength != 1) return illegal_read_write;
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if ((ars = (ARS *)din->data.vptr) == NULL) return out_of_range;
    recsreadFlag = 1; /* for legacy calls out of the old dll */
    starttime = ars->starttime;
    stoptime = starttime + ars->minutes*60L;
    startpos = recsreadFlag * (fmtsize+sizeof(UINT32));
    n = fetch(ars->keyword,ars->index,ars->starttime,stoptime,
              dout->dArrayLength-recsreadFlag,dout->data.bptr+startpos,fmt,sample,uprlimit,lwrlimit);
    if (n < 0) return -n;
    if (recsreadFlag) /* put number in 1st int slot */
       *(UINT32 *)(dout->data.bptr+fmtsize) = n;  
    if (n < (int)dout->dArrayLength) dout->dArrayLength = n + recsreadFlag;
    return 0;
  }
  if (!strncmp(devProperty,"SAMPLE",6))
  {
    if (fmt == -1) return illegal_format;
    if (din->dArrayLength != 1) return illegal_read_write;
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if ((ars = (ARS *)din->data.vptr) == NULL) return out_of_range;
    recsreadFlag = 1; /* for legacy calls out of the old dll */
    starttime = ars->starttime;
    stoptime = starttime + ars->minutes*60L;
    startpos = recsreadFlag * (fmtsize+sizeof(UINT32));
    n = fetch(ars->keyword,ars->index,ars->starttime,stoptime,
              dout->dArrayLength-recsreadFlag,dout->data.bptr+startpos,fmt,1,0,0);
    if (n < 0) return -n;
    if (recsreadFlag) /* put number in 1st int slot */
       *(UINT32 *)(dout->data.bptr+fmtsize) = n;  
    if (n < (int)dout->dArrayLength) dout->dArrayLength = n+recsreadFlag;
    return 0;
  }

  /* else devProperty is the Keyword */
  strncpy(keyword,devProperty,32);
  for (n=0; n<nPrpAlias; n++)
  {
    if (!strncmp(prpAlias[n].name,devProperty,32))
    {
      strncpy(keyword,prpAlias[n].alias,32);
      break;
    }
  }

  if ((index=GetDeviceNumber("HDAEQM",devName)) < 0 &&
      (index=getIndexFromDeviceNameEx(devName,keyword,din)) < 0)
  { 
    if (index == -illegal_equipment_number) return -index;
    if ((index=getIndexFromDeviceName(devName,keyword)) < 0) 
      return illegal_equipment_number;
  }
  if (din->dArrayLength > 0)
  {
    switch (din->dFormat)
    {
      case CF_LONG:
        switch (din->dArrayLength)
        {
          default:
          case 5: recsreadFlag = din->data.lptr[4];
          case 4: sample    = din->data.lptr[3];
          case 3: if (index == 0) index = din->data.lptr[2];
          case 2: stoptime  = din->data.lptr[1];
                  inputgiven = TRUE;
          case 1: starttime = din->data.lptr[0];
          case 0: break;
        }
        break;
      case CF_FWINDOW:
        starttime = ((FWINDOW *)din->data.vptr)[0].i1val;
        stoptime  = ((FWINDOW *)din->data.vptr)[0].i2val;
        lwrlimit  = ((FWINDOW *)din->data.vptr)[0].f1val;
        uprlimit  = ((FWINDOW *)din->data.vptr)[0].f2val;
        if (lwrlimit < uprlimit) sample    = -1;
        break;
      case CF_WINDOW:
        starttime = ((DWINDOW *)din->data.vptr)[0].t1;
        stoptime  = ((DWINDOW *)din->data.vptr)[0].t2;
        sample    = ((DWINDOW *)din->data.vptr)[0].i2val;
        break;
      default:
        return illegal_format;
    }
  }
  else
  {
    recsreadFlag = 0;
    starttime = getArchiveStartTimeFromDepth(keyword,dout->dArrayLength);
  }
  if (stoptime > tnow) stoptime = tnow;
  if (dout->dArrayLength == 1 && inputgiven) recsreadFlag = 1;
  if (recsreadFlag && fmt == -1 && stoptime > starttime) fmt = dout->dFormat;
  if (fmt != -1) /* assume request for pair-wise data + timestamp */
  {
    startpos = recsreadFlag * (fmtsize+sizeof(UINT32));
    n = fetch(keyword,index,starttime,stoptime,
              dout->dArrayLength-recsreadFlag,dout->data.bptr+startpos,fmt,sample,uprlimit,lwrlimit);
    if (n < 0) return -n;
    if (recsreadFlag) assignNrOfPoints(dout->data.bptr,dout->dFormat,n);
    if (n < (int)dout->dArrayLength) dout->dArrayLength = n + recsreadFlag;
    return 0;
  }
  fmt = dout->dFormat;
  fmtsize = fmtsizeof(fmt);
  return fetchTrace(keyword,starttime,dout->dArrayLength,(BYTE *)dout->data.vptr,fmt,index,sample);
}
#define NULLVALUE 0.05
int isTransition(BYTE *d1, BYTE *d0, int fmt)
{
  switch (fmt)
  {
    case CF_FLOAT:
      if (*(float *)d1 > NULLVALUE || *(float *)d0 <= NULLVALUE) return 0;
      return -1;
    case CF_DOUBLE:
      if (*(double *)d1 > NULLVALUE || *(double *)d0 <= NULLVALUE) return 0;
      return -1;
    default: return 0;
  }
}
long loadData(int n, int p, int index, time_t starttime,time_t endtime,
              void *datbuf,int datfmt,int bufsize,int sampleFlag,
              float upr,float lwr)
{
  int i,hFile,nrecs,raster = 1,cc = 0;
  long timebin = 1,pstart,pstop,pnow,ppos,p0,number = 0,cnt = 0,rlen;
  char s[80];
  DTYPE d;
  int fmtsize,dsize,datlen;
  time_t timenull,timestamp,datatime = 0;
  struct tm *lt=localtime((time_t *)&starttime);
  float fval,fvalnull;

  getDataFilename(n,lt->tm_year,lt->tm_mon+1);

  if ((pstart=locateRecord(n,starttime)) < 0) goto err;
  if ((pstop=locateRecord(n,endtime)) < 0) goto err;
  rlen = fec[n].reclen + sizeof(UINT32);
  nrecs = (pstop-pstart)/rlen;
  if (bufsize == 0) return nrecs;
  if (sampleFlag > 0) raster = sampleFlag;
  else if (bufsize < nrecs) raster += nrecs/(bufsize+1);

  if ((hFile=OpenDataFile(n)) == -1) ccerr(file_error);

/*startscan: */
  number = 0; cnt = 0; datatime = 0;
  pnow = lseek(hFile,pstart,SEEK_SET);

  d.dFormat = fec[n].fecdata[p].format;
  d.dArrayLength = 1;
  fmtsize = fmtsizeof(fec[n].fecdata[p].format);
  dsize = fec[n].fecdata[p].size * fmtsize;
  if (archdebug)
  {
    sprintf(s,"HDA: rastering at %ld records\n>",raster);
    ttyoutput(s);
  }
  while(read(hFile,&timestamp,4) == 4)   /* read through file */
  {
    cnt++;
    if (index < 0 || index >= fec[n].fecdata[p].size) goto err;
    if (timestamp > endtime) goto err;
    if (fec[n].reclen > TRACE_BUFFER_SIZE) goto err;
    if (read(hFile,traceDataBuffer,fec[n].reclen) != (int)fec[n].reclen) goto err;

    datatime = timestamp;
    d.data.vptr = traceDataBuffer + fec[n].fecdata[p].offset + index*fmtsize;

    switch (datfmt)
    {
      case CF_LONG: 
        ((INTINT *)datbuf)[number].i2val = timestamp;
        getValuesAsLong(&d,&((INTINT *)datbuf)[number].i1val,1);
        if (sampleFlag < 0 && 
            (d.data.lptr[0] < lwr || d.data.lptr[0] > upr))
        {
          ((INTINT *)datbuf)[number].i1val = 0; /* flag the boundary */
          if (number > 0 && ((INTINT *)datbuf)[number-1].i1val == 0) continue;
        }
        break;
      case CF_FLOAT: 
        if (number > 1 && timestamp == (int)((FLTINT *)datbuf)[number-1].ival)
        {
          number--;
          break;
        }
        ((FLTINT *)datbuf)[number].ival = timestamp;
        getValuesAsFloat(&d,&((FLTINT *)datbuf)[number].fval,1);
        if (raster > 1 && number > 1 &&
            (fec[n].option & PROTONS || fec[n].option & ELECTRONS))
        {
          fval = ((FLTINT *)datbuf)[number].fval; timenull = 0;
          for (i=0; i<raster; i++)
          {
            if (!isTransition((BYTE *)&fval,(BYTE *)&((FLTINT *)datbuf)[number-1].fval,datfmt)) break;
            if (i == raster-1) break; /* stop right here */
            if (i > 0) { fvalnull = fval; timenull = timestamp; }
            ppos = pnow - rlen*(1+i);
            if ((p0=lseek(hFile,ppos,SEEK_SET)) < 0) goto err;
            if (read(hFile,&timestamp,4) != 4) goto err;
            if (read(hFile,traceDataBuffer,fec[n].reclen) != (int)fec[n].reclen) goto err;
            getValuesAsFloat(&d,&fval,1);
          }
          if (timenull > 0)
          {
            ((FLTINT *)datbuf)[number].ival = timenull;
            ((FLTINT *)datbuf)[number].fval = fvalnull;
          }
          if (lseek(hFile,pnow+rlen,SEEK_SET) < 0) goto err; /* back to where we were */
        }
        if (sampleFlag < 0 &&
            (d.data.fptr[0] < lwr || d.data.fptr[0] > upr))
        {
          ((FLTINT *)datbuf)[number].ival = 0; /* flag the boundary */
          if (number > 0 && ((FLTINT *)datbuf)[number-1].ival == 0) continue;
        }
        break;
      case CF_TDS: /* special case where the timestamp is already part of the format */
        ((INTFLTINT *)datbuf)[number].i1val = timestamp;
        getValuesAsFloat(&d,&((INTFLTINT *)datbuf)[number].fval,1);
        if (sampleFlag < 0 && 
            (d.data.fptr[0] < lwr || d.data.fptr[0] > upr))
        {
          ((INTFLTINT *)datbuf)[number].i1val = 0; /* flag the boundary */
          if (number > 0 && ((INTFLTINT *)datbuf)[number-1].i1val == 0) continue;
        }
        ((INTFLTINT *)datbuf)[number].i2val = 0;
        break;
      case CF_DOUBLE: 
        ((DBLDBL *)datbuf)[number].d2val = (double)timestamp;
        getValuesAsDouble(&d,&((DBLDBL *)datbuf)[number].d1val,1);
        if (sampleFlag < 0 && 
            (d.data.dptr[0] < lwr || d.data.dptr[0] > upr))
        {
          ((DBLDBL *)datbuf)[number].d2val = 0; /* flag the boundary */
          if (number > 0 && ((DBLDBL *)datbuf)[number-1].d1val == 0) continue;
        }
        break;
      case CF_NAME16: 
        switch (fec[n].fecdata[p].format)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          default: ccerr(illegal_format);
        }
        datlen = fmtsizeof(datfmt);
        if (dsize > datlen) dsize = datlen;
        strncpy(((NAME16I *)datbuf)[number].name,d.data.cptr,dsize);
        ((NAME16I *)datbuf)[number].ival = timestamp;
        break;
      case CF_NAME32: 
        switch (fec[n].fecdata[p].format)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          default: ccerr(illegal_format);
        }
        datlen = fmtsizeof(datfmt);
        if (dsize > datlen) dsize = datlen;
        strncpy(((NAME32I *)datbuf)[number].name,d.data.cptr,dsize);
        ((NAME32I *)datbuf)[number].ival = timestamp;
        break;
      case CF_NAME48: 
        switch (fec[n].fecdata[p].format)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          default: ccerr(illegal_format);
        }
        datlen = fmtsizeof(datfmt);
        if (dsize > datlen) dsize = datlen;
        strncpy(((NAME48I *)datbuf)[number].name,d.data.cptr,dsize);
        ((NAME48I *)datbuf)[number].ival = timestamp;
        break;
      case CF_NAME64:
        switch (fec[n].fecdata[p].format)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          default: ccerr(illegal_format);
        }
        datlen = fmtsizeof(datfmt);
        if (dsize > datlen) dsize = datlen;
        strncpy(((NAME64I *)datbuf)[number].name,d.data.cptr,dsize);
        ((NAME64I *)datbuf)[number].ival = timestamp;
        break;
      default: goto err;
    }
    if (!(++number%100) && archdebug) 
    {
      ttyoutput(".");
    }
    if (number >= bufsize) goto err;
    if (raster > 1) /* move along through file */
    {
      if ((pnow=lseek(hFile,rlen*(raster-1),SEEK_CUR)) < 0) goto err;
    }
    else pnow += rlen;
  }
err:
  if (hFile != -1) close(hFile);
  if (archdebug) 
  {
    sprintf(s,"HDA: scanned %d records\n>",cnt);
    ttyoutput(s);
  }

  return cc > 0 ? -cc : number;
}
time_t getArchiveStartTimeFromDepth(char *keyword,int depth)
{
  time_t endtime = time(NULL), stime = endtime-depth;
  struct tm *lt=localtime((time_t *)&endtime);
  int rlen, fsiz;
  static int rectyp,p;
  int hFile=-1,cc=0,recsPerDay,n,yr,mo;
  char s[128];

  if ((n=getArchivedEntryFromKeywordAndTime(keyword,&rectyp,&p,endtime)) < 0) 
    ccerr(-n);

  rlen = fec[n].reclen + sizeof(UINT32);

  /* see how big last month's file was ... */
  yr = lt->tm_year; mo = lt->tm_mon;
  if (mo == 0) yr--;
  mo = (mo - 1 + 12)%12;  
  getDataFilename(n,yr,mo+1);
  if ((hFile=OpenDataFile(n)) == -1) ccerr(no_such_file);
  if ((fsiz=lseek(hFile,0,SEEK_END)) <= 0) ccerr(file_error);
  recsPerDay = fsiz / (rlen * 30);  /* approximately ... */
  if (recsPerDay < 48) recsPerDay = 48; /* be reasonable ... */
  stime = endtime - (depth * 86400)/recsPerDay;
err:
  if (hFile != -1) close(hFile);
  if (archdebug)
  {
    sprintf(s,"getArchiveStartTimeFromDepth %d pts, recs/day %d, file size %d: %s>",
      depth,recsPerDay,fsiz,erlst[cc]);
    ttyoutput(s);
  }
  return stime;
}
int fetch(char *keyword,int index,time_t starttime,
          time_t endtime,int arraysize,BYTE *data,int datafmt,int sampleFlag,
          float upr,float lwr)
{
  FECDATA *fecdata;
  static int rectyp,p;
  long ps;
  int recsread=0,ptr=0,recsfound=0,recsize,n,searchCycle=0;
  int asize,daystotal,daysinmonth;
  int finalPass=FALSE,nquery=FALSE;
  long offset = sizeof(UINT32);
  struct tm st,et,it;
  time_t itime;
  char s[128];

  if (archdebug)
  {
    sprintf(s,"HDA: searching for %s <%d> %d entries beginning %s>",
            keyword,index,arraysize,ctime(&starttime));
    ttyoutput(s);
  }

  if ((n=getArchivedEntryFromKeywordAndTime(keyword,&rectyp,&p,endtime)) < 0) 
    return n;
tof:
  fecdata = fec[n].fecdata;
  if (index > fecdata[p].size && searchCycle == 0) goto bof;
  if (index > fecdata[p].size || index < 0) return -out_of_range;
  if (arraysize < 0) return -buffer_too_small;
  if (arraysize == 0) nquery = TRUE;
  recsize = fmtsizeof(datafmt);
  switch (datafmt)
  {
    case CF_DOUBLE:
    case CF_TDS:
      recsize += 8;
      break;
    default:
      recsize += sizeof(UINT32);
      break;
  }
  memcpy(&et,localtime(&endtime), sizeof(struct tm));
  memcpy(&st,localtime(&starttime),sizeof(struct tm));
  if (et.tm_mon == st.tm_mon)
  {
    asize = arraysize;
    finalPass = TRUE;
  }
  else
  {
    asize = sampleFlag > 0 ? arraysize - 1 : 0;
  }
  daystotal = (endtime-starttime)/(3600*24);
  if (asize == 0 && daystotal == 0) daystotal++;
  if ((endtime-starttime)%(3600*24) > 0) daystotal++;

  while (recsread < arraysize || nquery)
  {
    if (asize < arraysize || nquery)
    {
      it = st;
      if (it.tm_year == et.tm_year && it.tm_mon == et.tm_mon)
      {
        asize = arraysize;
        finalPass = TRUE;
      }
      else
      {
        it.tm_mon = (it.tm_mon+1)%12; if (it.tm_mon == 0) it.tm_year++;
        it.tm_mday = 1; it.tm_hour = 0; it.tm_min = 0; it.tm_sec = 0;
        itime = mktime(&it);
        daysinmonth = MAX((itime-starttime)/(3600*24),1);
        if (!nquery && !sampleFlag)
        {
          asize += MAX((arraysize * daysinmonth)/daystotal,1);
          if (asize >= arraysize) finalPass = TRUE; 
        }
      }
    }
    if (asize > arraysize) asize = arraysize;
    ps = loadData(n,p,index,starttime,endtime,&data[ptr],datafmt,asize-recsread,sampleFlag,upr,lwr);
    if (ps < 0) return ps;
    if (nquery) recsfound += ps; else recsread += ps;
    ptr = recsread * recsize;
    if (finalPass) break;
    starttime = itime;
    memcpy(&st,localtime(&starttime),sizeof(struct tm));
  }
bof:
  if (recsread == 0 && searchCycle == 0 && recsfound == 0)
  {
    searchCycle++;
    if ((n=getLegacyEntryFromKeyword(keyword,&rectyp,&p)) < 0) return 0;
    goto tof;    
  }
  if (archdebug)
  {
    sprintf(s,"HDA: returning %d records\n>",recsread);
    ttyoutput(s);
  }
  return nquery ? recsfound : recsread;
}
int loadTraceData(int n, int p, time_t tgttime,
                   void *datbuf,int datfmt,int bufsize,int index,int sample)
{
  int cc = 0,hFile;
  long pstart,number = 0, cnt = 0;
  int i,pos,fmtsize,tsize;
  time_t timestamp,datatime = 0;
  struct tm *lt=localtime((time_t *)&tgttime);
  DTYPE dout;
  BYTE *bptr;

  if (index > fec[n].fecdata[p].size) return invalid_index;
  if (sample > fec[n].fecdata[p].size) return illegal_delimiter;
  if (fec[n].reclen > TRACE_BUFFER_SIZE) return resources_exhausted;
  getDataFilename(n,lt->tm_year,lt->tm_mon+1);
  if ((pstart=locateRecord(n,tgttime)) < 0) ccerr(invalid_data);
  if ((hFile=OpenDataFile(n)) == -1) ccerr(no_such_file);
  lseek(hFile,pstart,SEEK_SET);
  fmtsize = fmtsizeof(fec[n].fecdata[p].format);
  if (read(hFile,&timestamp,4) != 4) ccerr(file_error);
  if (read(hFile,traceDataBuffer,fec[n].reclen) != (int)fec[n].reclen) ccerr(file_error);
  setDataTimeStamp((double)timestamp);
  if (index == 0 && sample == 0) 
  {
    tsize = MIN(fec[n].fecdata[p].size,bufsize);
    if (datfmt == fec[n].fecdata[p].format)
    {
      memcpy(datbuf,traceDataBuffer+fec[n].fecdata[p].offset,tsize*fmtsize);
    }
    else
    {
      dout.dArrayLength = bufsize;
      dout.dFormat = datfmt;
      dout.data.vptr = datbuf;
      switch (fec[n].fecdata[p].format)
      {
        case CF_SHORT:
          if ((cc=putValuesFromShort(&dout,(short *)(traceDataBuffer+fec[n].fecdata[p].offset),(short)tsize)) != 0) return cc;
          break;
        case CF_LONG:
          if ((cc=putValuesFromLong(&dout,(UINT32 *)(traceDataBuffer+fec[n].fecdata[p].offset),(short)tsize)) != 0) return cc;
          break;
        case CF_FLOAT:
          if ((cc=putValuesFromFloat(&dout,(float *)(traceDataBuffer+fec[n].fecdata[p].offset),(short)tsize)) != 0) return cc;
          break;
        case CF_DOUBLE:
          if ((cc=putValuesFromDouble(&dout,(double *)(traceDataBuffer+fec[n].fecdata[p].offset),(short)tsize)) != 0) return cc;
          break;
        default:
          return illegal_format;
      }
    }
  }
  else
  {
    for (i=0; i<bufsize; i++)
    {
      dout.dArrayLength = 1;
      dout.dFormat = datfmt;
      dout.data.vptr = &((BYTE *)datbuf)[i*fmtsizeof(datfmt)];
      pos = (index+sample*i)%fec[n].fecdata[p].size;
      bptr = traceDataBuffer + fec[n].fecdata[p].offset + pos*fmtsize;
      switch (fec[n].fecdata[p].format)
      {
        case CF_SHORT:
          if ((cc=putValuesFromShort(&dout,(short *)bptr,1)) != 0) return cc;
          break;
        case CF_LONG:
          if ((cc=putValuesFromLong(&dout,(UINT32 *)bptr,1)) != 0) return cc;
          break;
        case CF_FLOAT:
          if ((cc=putValuesFromFloat(&dout,(float *)bptr,1)) != 0) return cc;
          break;
        case CF_DOUBLE:
          if ((cc=putValuesFromDouble(&dout,(double *)bptr,1)) != 0) return cc;
          break;
        default:
          if (datfmt != fec[n].fecdata[p].format) return illegal_format;
          memcpy(&((BYTE *)datbuf)[i*fmtsize],bptr,fmtsize);
          break;
      }
    }
  }
err: if (hFile != -1) close(hFile);
  return cc;
}
int fetchTrace(char *keyword,time_t tgttime,
               int arraysize,BYTE *data,int datafmt,int index,int sample)
{
  FECDATA *fecdata;
  static int rectyp,p;
  int recsread=0,ptr=0,i,n,searchCycle=0;
  int asize = arraysize/4;
  int cc = 0, N32toN16 = FALSE;
  long offset = sizeof(UINT32);
  char s[128];

  if (archdebug)
  {
    sprintf(s,"HDA: searching for %s trace %d entries at %s>",
            keyword,arraysize,ctime(&tgttime));
    ttyoutput(s);
  }

  if ((n=getArchivedEntryFromKeywordAndTime(keyword,&rectyp,&p,tgttime)) < 0) 
    return -n;
tof:
  fecdata = fec[n].fecdata;
  if (fecdata[p].format == CF_NAME32 &&
      datafmt == CF_NAME16)
  {
    datafmt = CF_NAME32;
    N32toN16 = TRUE;
  }
  cc = loadTraceData(n,p,tgttime,data,datafmt,arraysize,index,sample);
  if (cc == file_error && searchCycle == 0)
  {
    searchCycle++;
    if ((n=getLegacyEntryFromKeyword(keyword,&rectyp,&p)) < 0) return 0;
    goto tof;    
  }
  if (N32toN16)
  {
    for (i=0; i<arraysize; i++) 
     strncpy(((NAME16 *)data)[i].name,((NAME32 *)data)[i].name,16);
  }
  return cc;
}


