/* ens.c - Equipment Name Server */

#include "prolog.h"
#include  <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include "rpcdefs.h"
#include "servdef.h"
#include "srvip.h"
#include "errors.h"
#include "database.h"
#include "brdcast.h"
#include "toolkit.h"
#if defined(WINNT)
#include "ntcmd.h"
#elif defined(UNIX)
#include "unixcmd.h"
#endif

extern int NameServerLoaded;
int findcol(char *str,char *tag);
char *colptr(int col,char *str);
void HexToBytes(char *s,BYTE *b,int nibs);
extern int RPCNumRecsRead,RPCNumFecsRead,RPCNumEqpsRead;
extern int RPCallocNum,FECallocNum;
extern char ControlDBpath[80];
extern char controlFS[48];
extern int CommandLevel;
char eqpfname[80],fecfname[80];
RPCFecStruct rpcfec;
RPCDataStruct rpcdata;
short *fecIndex;
int debug = 0;
int checkExportFlag = 0;
int dumpFecFileFlag = 0;
int dumpEqpFileFlag = 0;
int scanFlag = 0;
int fixFecProtocol(RPCFecStruct *fec);
void initENS(void);
void taskENS(void);
int pollInit(void);
int enseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);  
int getContexts(void);

char logbuf[212];

#ifdef WINNT
int unsetenv(char *env)
{
  char envstr[128];
  strncpy(envstr,env,120);
  strcat(envstr,"=");
  return _putenv(envstr);
}
#endif
void PreSystemInit(void)
{
  char *c=NULL;
  int keepEnvSettings=FALSE;
  if ((c=getenv("USE_TINE_HOME")) != NULL && strcmp(c,"TRUE") == 0)
     keepEnvSettings = TRUE;
  if (!keepEnvSettings)
  {
    if (unsetenv("TINE_HOME") != 0) perror("unsetenv TINE_HOME");
    if (unsetenv("FECDB") != 0) perror("unsetenv FECDB");
    /* legacy env variable */
    if (unsetenv("CONTROLDB") != 0) perror("unsetenv CONTROLDB"); 
  }
  MaxRPCTransportSize = 0x40000;
  MaxNumContracts = 1000;
  gRespondToServiceRequests = TRUE;
}
void PostSystemInit(void)
{
  RegisterEquipmentModule(NULL,"ENSEQM",0,enseqm,initENS,taskENS,200,NULL);
  RegisterLocalAlias("ENSEQM","HERANS");
  RegisterDeviceName("ENSEQM","DEFAULT",0);
  RegisterDeviceName("ENSEQM","FEC",1);
  RegisterDeviceName("ENSEQM","EQP",2);
  RegisterDeviceName("ENSEQM","FEC.EXT",3);

  RegisterUserCommand("ensdebug",NULL,&debug,NULL,CA_WRITE|CA_READ);
  RegisterUserCommand("dumpfecs",NULL,&dumpFecFileFlag,NULL,CA_WRITE|CA_READ);
  RegisterUserCommand("dumpeqps",NULL,&dumpEqpFileFlag,NULL,CA_WRITE|CA_READ);
}
#define ENSLOGDEPTH 5000
void nslog(char *txt)
{
  FILE *fp;
  time_t t = time(NULL);
  char lclbuf[256];
  long flen;

  if ((fp=fopen("ens.log","a")) == NULL) return;
  fseek(fp, 0L, SEEK_END); flen = ftell(fp);
  if (flen > ENSLOGDEPTH*80)
  {
    fclose(fp);
    remove("ens-bak.log");
	rename("ens.log","ens-bak.log");
    if ((fp=fopen("ens.log","a")) == NULL) return;
  }
  sprintf(lclbuf,"%s @ %s",txt,ctime(&t));
  fprintf(fp,lclbuf);
  fclose(fp);

  return;
}

void resolveFecAddress(int i)
{
  DWORD haddr = 0;
  struct hostent *host;

  if (TCPloaded && (host=gethostbyname(RPCFec[i].Name)) != NULL)
  {
    if (strlen(RPCFec[i].IP)) haddr = inet_addr(RPCFec[i].IP);
    if (haddr != *(DWORD *)host->h_addr)
    {
      printf("ENS: %s using new IP address\n>",RPCFec[i].Name);
      haddr = *(DWORD *)host->h_addr;
      strncpy(RPCFec[i].IP,inet_ntoa(*(struct in_addr *)host->h_addr),16);
    }
    memcpy(RPCFec[i].IPh_addr,&haddr,4);
  }
  return;
}
void nsSend(char *txt)
{
  /* this is a bit of a kluge, but since I'm the name server */
  /* I know that NameServer is not used */
  memcpy(NameServer,&rpcfec,sizeof(RPCFecStruct));
  tExecLink(0,txt,rpcdata.EqName,"#0","MESSAGE",
            (short)(-((int)strlen(txt))),CF_TEXT,1);
}
int verifyFecAddress(int n)
{
  int i,first=0,last=RPCNumFecsRead-1;
  if (n >= RPCNumFecsRead) return out_of_range;
  if (n >= 0) first = last = n;
  for (i=first; i<=last; i++) resolveFecAddress(i);
  return 0;
}
#define FILESPACE 100
#define TMPSPACE  10 

RPCFecExStruct *RPCFecExt = NULL,rpcfecex;
short *RPCDataDisabled = NULL, *RPCFecDisabled = NULL;
UINT32 *RPCDataRebootCount = NULL;

#define dbcmp(a,b,n,flg) {\
  if (memcmp(a,b,n)) { memcpy(a,b,n); flg = 1; }\
}
short numRemoteProps;
char *getCsvString(char *s,int length)
{
  char sbuf[128];
  char *c, *start;
  int len;

  if (s == NULL) return s; /* null */
  if (length > 126) length = 126; /* this won't happen */
  if ((c=strchr(s,'\n')) != NULL) *c = 0; /* no new-line char */
  if ((len=strlen(s)) == 0) return s; /* empty */
  if (strchr(s,',') == NULL) return s; /* nothing to worry about */
  for (start=s; isspace(*start) && len > 0; start++, len--); 
  if (*start == '"') return s; /* already a csv string */
  if (len > length-2) len = length-2; /* hardly possible */
  sbuf[0] = '"'; start[len] = 0;
  strncpy(&sbuf[1],start,len+1);
  strcat(sbuf,"\"");
  strncpy(s,sbuf,length);
  return s;
}
int queryPending = -1,fecTargetId = -1;
void querycallback(int id,int cc) { queryPending = cc;}
enum { check_idle=0,
       check_tag,
       check_tagaddr,
       check_taginfo,
       check_fec,
       check_fecaddr,
       check_pending,
       check_fecinfo,
       check_finished };
int lastRPCindex, lastFECindex;
void checkExport(void)
{
  static int i,idx;
  static int nextStep;
  static time_t interim = 0;
  static char errstr[64];
  UINT32 lIPX1,lIP1,lIP2;
  char fulldevname[64],prpname[36];
  DTYPE dout;
  int n;

  if (checkExportFlag == check_idle) return;
  while(checkExportFlag != check_idle)
  {
   switch (checkExportFlag)
   {
    case check_idle: return;
    case check_tag:
      if (debug) 
      {
        if (strlen(rpcdata.Context) > 0)
          sprintf(dbgbuf,"check_tag: /%s/%s\n>",rpcdata.Context,rpcdata.ExportName);
        else
          sprintf(dbgbuf,"check_tag: %s\n>",rpcdata.ExportName);
        dbgoutput(dbgbuf,tnull);
      }
      sprintf(logbuf,"ENS: check export tag %s",rpcdata.ExportName);
      nslog(logbuf);
      for (i=0; i<RPCNumRecsRead; i++)
      {
        if (strlen(rpcdata.Context) != 0 && 
            strlen(RPCData[i].Context) != 0 &&
            strncmp(rpcdata.Context,RPCData[i].Context,16)) continue;
        if (!strncmp(RPCData[i].ExportName,rpcdata.ExportName,16)) break;
      }
      lastRPCindex = i;
      if (i == RPCNumRecsRead) /* it's a new one */
      {
        strncpy(RPCData[i].ExportName,rpcdata.ExportName,16);
        strncpy(RPCData[i].FecName,rpcdata.FecName,16);
        strncpy(RPCData[i].EqName,rpcdata.EqName,6);
        strncpy(RPCData[i].Context,rpcdata.Context,16);
        RPCData[i].alarmSystem = rpcdata.alarmSystem;
        RPCDataDisabled[i] = 0;
        dumpEqpFileFlag = 1;
        sprintf(logbuf,"ENS: new DATA entry - %s",rpcdata.ExportName);
        nslog(logbuf);
        if (debug) dbgoutput(logbuf,tnull);
        strncpy(prpname,RPCData[i].ExportName,16); prpname[16] = 0;
        if (RegisterProperty("ENSEQM",prpname,
              sizeof(RPCDataStruct)+sizeof(RPCFecStruct),CF_STRUCT,
              0x01,"registered EQM") < 0)
        {
          sprintf(logbuf,"ENS: couldn't register %s for export",RPCFec[i].Name);
          nslog(logbuf);
        }
        RPCNumRecsRead++;
        checkExportFlag = check_fec;
      }
      else                  /* it's in the list */
      {
        if (strlen(rpcdata.Context) == 0) /* not given, copy it back */
        {
          strncpy(rpcdata.Context,RPCData[i].Context,16);
          if (debug)
          {
            sprintf(dbgbuf,"supply context %s to %s\n",RPCData[i].Context,RPCData[i].ExportName);
            dbgoutput(dbgbuf,tnull);
          }
        }
        checkExportFlag = check_tagaddr;
      }
      RPCDataRebootCount[i]++;
      if (!dumpEqpFileFlag &&
          strlen(RPCData[i].Context) > 0 && 
          strncmp(RPCData[i].Context,"TEST",4) != 0) dumpEqpFileFlag = -1;
      break;
    case check_tagaddr:
      if (debug) 
      {
        sprintf(dbgbuf,"check_tagaddr: FEC %s\n>",rpcdata.FecName);
        dbgoutput(dbgbuf,tnull);
      }
      checkExportFlag = check_taginfo;
      if (strncmp(RPCData[i].FecName,rpcdata.FecName,16))
      {                         /* same tag, new FEC !? */
        if (debug) 
        {
          sprintf(dbgbuf,"check_tagaddr: found %s vs %s\n>",RPCData[i].FecName,rpcdata.FecName);
          nslog(dbgbuf);
          dbgoutput(dbgbuf,tnull);
        }
        if (strlen(RPCData[i].Context) > 0)
          sprintf(fulldevname,"/%s/%s/#0",RPCData[i].Context,RPCData[i].ExportName);
        else
          sprintf(fulldevname,"%s/#0",RPCData[i].ExportName);
        dout.dFormat = CF_SHORT;
        dout.dArrayLength = 1;
        dout.data.sptr = (UINT16 *)&numRemoteProps;
        fecTargetId = AttachLink(fulldevname,"NPROPS",&dout,NULL,CA_READ,1000,querycallback,CM_SINGLE);
        if (fecTargetId >= 0) 
        {
          idx = connTable[fecTargetId]->FECindex;
          if (debug) 
          {
            sprintf(dbgbuf,"check_tagaddr: check addr of %s vs %s\n>",RPCFec[idx].Name,rpcfec.Name);
            dbgoutput(dbgbuf,tnull);
          }
          if (!memcmp(RPCFec[idx].IPh_addr,rpcfec.IPh_addr,4) &&
              GetPortOffset(&rpcfec) == GetPortOffset(&RPCFec[idx])) 
          {
            if (debug) 
            {
              sprintf(dbgbuf,"check_tagaddr: FEC name change only\n>");
              nslog(dbgbuf);
              dbgoutput(dbgbuf,tnull);
            }
            break; /* but same address ! */
          }
          interim = time(NULL);
          queryPending = -1;
          sprintf(errstr,"Tag name %s in use by FEC %s",
                  RPCData[i].ExportName,RPCData[i].FecName);
          checkExportFlag = check_pending;
          nextStep = check_taginfo;
	}
      }
      break;
    case check_taginfo:
      if (debug) 
      {
        sprintf(dbgbuf,"check_taginfo: %s %s %s alarmSystem %d\n>",
                RPCData[i].Context,RPCData[i].FecName,
                RPCData[i].EqName,RPCData[i].alarmSystem);
        dbgoutput(dbgbuf,tnull);
      }
      dbcmp(RPCData[i].FecName,rpcdata.FecName,sizeof(rpcdata.FecName),dumpEqpFileFlag);
      dbcmp(RPCData[i].EqName,rpcdata.EqName,sizeof(rpcdata.EqName),dumpEqpFileFlag);
      dbcmp(&RPCData[i].alarmSystem,&rpcdata.alarmSystem,2,dumpEqpFileFlag);
      dbcmp(&RPCData[i].Context,&rpcdata.Context,sizeof(rpcdata.Context),dumpEqpFileFlag);
      checkExportFlag = check_fec;
      break;
    case check_fec:
      if (debug) 
      {
        sprintf(dbgbuf,"check_fec: %s\n>",rpcfec.Name);
        dbgoutput(dbgbuf,tnull);
      }
      for (i=0; i<RPCNumFecsRead; i++)
      {
        if (!strncmp(RPCFec[i].Name,rpcfec.Name,16))
        {
          memcpy(&lIP1,rpcfec.IPh_addr,4);
          if (lIP1 == 0) strcpy(rpcfec.IP,"0.0.0.0");
          if (debug)
          {
            sprintf(dbgbuf,"check_fec: %02x%02x%02x%02x%02x%02x vs %02x%02x%02x%02x%02x%02x\n>",
                    RPCFec[i].Node[0],RPCFec[i].Node[1],RPCFec[i].Node[2],
                    RPCFec[i].Node[3],RPCFec[i].Node[4],RPCFec[i].Node[5],
                    rpcfec.Node[0],rpcfec.Node[1],rpcfec.Node[2],
                    rpcfec.Node[3],rpcfec.Node[4],rpcfec.Node[5]);
            dbgoutput(dbgbuf,tnull);
            memcpy(&lIP2,RPCFec[i].IPh_addr,4);
            sprintf(dbgbuf,"check_fec: %lx (%s) <%d> vs %lx (%s) <%d>\n>",
                   lIP2,RPCFec[i].IP,GetPortOffset(&RPCFec[i]),
                   lIP1,rpcfec.IP,GetPortOffset(&rpcfec));
            dbgoutput(dbgbuf,tnull);
          }
          break;
        }
      }
      lastFECindex = i;
      if (i == RPCNumFecsRead) /* it's a new one */
      {
        strncpy(RPCFec[i].Name,rpcfec.Name,16);
        memcpy(RPCFec[i].Network,rpcfec.Network,4);
        memcpy(RPCFec[i].Node,rpcfec.Node,6);
        memcpy(&lIPX1,RPCFec[i].Network,4);
        if (lIPX1) RPCFec[i].Prot = 0x01;
        memcpy(RPCFec[i].IPh_addr,rpcfec.IPh_addr,4);
        memcpy(&lIP2,RPCFec[i].IPh_addr,4);
        if (lIP2) RPCFec[i].Prot |= 0x02;
        memcpy(&lIP1,rpcfec.IPh_addr,4);
        if (lIP1 == 0) strcpy(rpcfec.IP,"0.0.0.0");
        else strncpy(RPCFec[i].IP,rpcfec.IP,16);
        RPCFec[i].PortOffset = rpcfec.PortOffset;
        memcpy(RPCFec[i].Port,rpcfec.Port,2);
        getCsvString(rpcfecex.os,12);
        memcpy(RPCFecExt[i].os,rpcfecex.os,12);
        getCsvString(rpcfecex.desc,64);
        memcpy(RPCFecExt[i].desc,rpcfecex.desc,64);
        getCsvString(rpcfecex.loc,32);
        memcpy(RPCFecExt[i].loc,rpcfecex.loc,32);
        getCsvString(rpcfecex.hdw,32);
        memcpy(RPCFecExt[i].hdw,rpcfecex.hdw,32);
        getCsvString(rpcfecex.sub,16);
        memcpy(RPCFecExt[i].sub,rpcfecex.sub,16);
        getCsvString(rpcfecex.resp,32);
        memcpy(RPCFecExt[i].resp,rpcfecex.resp,32);
        RPCFecDisabled[i] = 0;
        dumpFecFileFlag = 1;
        sprintf(logbuf,"ENS: new FEC entry - %s",rpcfec.Name);
        nslog(logbuf);
        strncpy(prpname,RPCFec[i].Name,16); prpname[16] = 0;
        if (RegisterProperty("ENSEQM",prpname,
                             sizeof(RPCFecStruct),CF_STRUCT,
                             0x01,"registered FEC") < 0)
        {
          sprintf(logbuf,"FILER: couldn't register %s for export",RPCFec[i].Name);
          nslog(logbuf);
          if (debug) dbgoutput(logbuf,tnull);
        }
        RPCNumFecsRead++;
        checkExportFlag = check_idle;   /* finished */
        if (debug)
        {
          sprintf(dbgbuf,"reload eqp tbl %d, reload fec tbl %d\n",
            dumpEqpFileFlag,dumpFecFileFlag);
          dbgoutput(dbgbuf,tnull);
        }
      }
      else checkExportFlag = check_fecaddr;
      break;
    case check_fecaddr:
      if (debug) 
      {
        sprintf(dbgbuf,"check_fecaddr: %s %s\n>",rpcfec.Name,rpcfec.IP);
        dbgoutput(dbgbuf,tnull);
      }
      checkExportFlag = check_fecinfo;
      memcpy(&lIP2,RPCFec[i].IPh_addr,4);
      memcpy(&lIPX1,rpcfec.Network,4);
      if ((lIP2 != 0 && memcmp(RPCFec[i].IPh_addr,rpcfec.IPh_addr,4) ) ||
          (lIPX1 != 0 && lIP2 == 0 &&
           (memcmp(RPCFec[i].Network,rpcfec.Network,4) || 
            memcmp(RPCFec[i].Node,rpcfec.Node,6)) )    || 
          GetPortOffset(&RPCFec[i]) != GetPortOffset(&rpcfec))  /* different address */
      {
        for (n=0; n<RPCNumRecsRead; n++)
          if (!strncmp(RPCData[n].FecName,rpcfec.Name,16)) break;
        if (n < RPCNumRecsRead)          /* same name, different address */
        {                                /* and there are eqm pointers to FEC */
          if (debug) 
          {
            sprintf(dbgbuf,"check_fecaddr: found %s vs %s\n>",RPCFec[i].IP,rpcfec.IP);
            nslog(dbgbuf);
            dbgoutput(dbgbuf,tnull);
          }
          if (strlen(RPCData[n].Context) > 0)
            sprintf(fulldevname,"/%s/%s/#0",RPCData[n].Context,RPCData[n].ExportName);
          else
            sprintf(fulldevname,"%s/#0",RPCData[n].ExportName);
          dout.dFormat = CF_SHORT;
          dout.dArrayLength = 1;
          dout.data.sptr = (UINT16 *)&numRemoteProps;
          fecTargetId = AttachLink(fulldevname,"NPROPS",&dout,NULL,CA_READ,1000,querycallback,CM_SINGLE);
          if (fecTargetId >= 0)
          {
            interim = time(NULL);
            queryPending = -1;
            sprintf(errstr,"FEC name %s in use by %s [%d]",
                    RPCFec[i].Name,RPCFec[i].IP,GetPortOffset(&RPCFec[i]));
            checkExportFlag = check_pending;
            nextStep = check_fecinfo;
	  }
        }
      }
      break;
    case check_pending:
      if (debug) 
      {
        sprintf(dbgbuf,"check_pending: waiting for reply\n>");
        dbgoutput(dbgbuf,tnull);
      }
      while (queryPending < 0 && time(NULL) < interim + 10) SystemDelay(500);
      if (queryPending != 0 ) /* this is okay (old addres doesn't answer or non existent elem */
      {
        checkExportFlag = nextStep;
        break;
      }
      sprintf(logbuf,"ENS: address in use");
      nslog(logbuf);
      if (debug) dbgoutput(logbuf,tnull);
      nsSend(errstr);
      dumpFecFileFlag = 0; dumpEqpFileFlag = 0;
      checkExportFlag = check_idle;      /* old address still in use */
      break;
    case check_fecinfo:
      if (debug) 
      {
        sprintf(dbgbuf,"check_fecinfo: final step\n>");
        dbgoutput(dbgbuf,tnull);
      }
      memcpy(&lIPX1,rpcfec.Network,4);
      if (lIPX1 != 0)
      {
        dbcmp(RPCFec[i].Network,rpcfec.Network,4,dumpFecFileFlag);
        dbcmp(RPCFec[i].Node,rpcfec.Node,6,dumpFecFileFlag);
      }
      dbcmp(RPCFec[i].IPh_addr,rpcfec.IPh_addr,4,dumpFecFileFlag);
      dbcmp(RPCFec[i].IP,rpcfec.IP,16,dumpFecFileFlag);
      dbcmp(&RPCFec[i].PortOffset,&rpcfec.PortOffset,1,dumpFecFileFlag);
      dbcmp(RPCFec[i].Port,&rpcfec.Port,2,dumpFecFileFlag);
      dbcmp(RPCFecExt[i].os,rpcfecex.os,12,dumpFecFileFlag);
      if (rpcfecex.desc[0] != '[' && RPCFecExt[i].desc[0] == '[')
      {  /* keep assigned sub system and description ! */
        strncpy(rpcfecex.desc,RPCFecExt[i].desc,64);
      }
      if (!strncmp(RPCFecExt[i].desc,"[XXX]",5))
      {  /* renegade FEC developer has mis-categorized his FEC ! */
        strncpy(rpcfecex.desc,RPCFecExt[i].desc,64);
      }
      getCsvString(rpcfecex.desc,64);
      dbcmp(RPCFecExt[i].desc,rpcfecex.desc,64,dumpFecFileFlag);
      if (rpcfecex.loc[0] == 0) strncpy(rpcfecex.loc,RPCFecExt[i].loc,32);
      getCsvString(rpcfecex.loc,32);
      dbcmp(RPCFecExt[i].loc,rpcfecex.loc,32,dumpFecFileFlag);
      getCsvString(rpcfecex.hdw,32);
      dbcmp(RPCFecExt[i].hdw,rpcfecex.hdw,32,dumpFecFileFlag);
      getCsvString(rpcfecex.sub,16);
      dbcmp(RPCFecExt[i].sub,rpcfecex.sub,16,dumpFecFileFlag);
      if (rpcfecex.resp[0] == 0) strncpy(rpcfecex.resp,RPCFecExt[i].resp,32);
      getCsvString(rpcfecex.resp,32);
      dbcmp(RPCFecExt[i].resp,rpcfecex.resp,32,dumpFecFileFlag);
      checkExportFlag = check_idle;      /* completely finished */
      if (debug)
      {
        sprintf(dbgbuf,"reload eqp tbl %d, reload fec tbl %d\n",
          dumpEqpFileFlag,dumpFecFileFlag);
        dbgoutput(dbgbuf,tnull);
      }
      break;
   default:
     checkExportFlag = check_idle;  /* turn off the check flag */
  }
  SystemDelay(20);
 }
 if (debug) 
 {
   sprintf(dbgbuf,"check_finished: %s\n>",rpcdata.ExportName);
   dbgoutput(dbgbuf,tnull);
 }
 sprintf(logbuf,"ENS: finished checking %s",rpcdata.ExportName);
 nslog(logbuf);
}
void fillinFecIndexTable(void)
{
  int i,j;

  for (i=0; i<RPCNumRecsRead; i++) for (j=0; j<RPCNumFecsRead; j++)
    if (!strncmp(RPCData[i].FecName,RPCFec[j].Name,16)) 
    {
      fecIndex[i] = j;
      break;
    }
}
#define MAX_SUBSYSTEMS 30
int nSubSystems = 0;
NAME16 subSystems[MAX_SUBSYSTEMS];
void appendSubSystemList(char *substr)
{
  int i;
  if (nSubSystems == MAX_SUBSYSTEMS) return;
  for (i=0; i<nSubSystems && strncmp(substr,subSystems[i].name,16); i++);
  if (i < nSubSystems) return;
  strncpy(subSystems[i].name,substr,16);
  nSubSystems++; 
}
char *getSubsystemFromDescription(char *str)
{
  static char substr[16];
  char *c,*c1,s[64];
  substr[0] = 0;
  strncpy(s,str,64);
  if (s[0] != '[' || (c=strchr(s,']')) == NULL) goto err;
  *c = 0; c = &s[1];
  while (*c == ' ') c++;
  if ((c1=strchr(c,' ')) != NULL) *c1 = 0;
  if (strlen(c) == 0) goto err;
  strncpy(substr,c,16);
err:
  return substr;
}
int loadDatabaseFiles(void)
{
  FILE *fp = NULL;
  DWORD haddr;
  int i,po=0,cc=0,done=0;
  int name_col,fecname_col,eqpmodule_col,alarm_col,context_col,reboot_col;
  int fecnetwork_col,fecnode_col,ipaddr_col,
      os_col,description_col,location_col,hardware_col,
      responsible_col,portoffset_col,sub_col;
  char str[256],hdr[256],scratch[128],s[128];
  char prpname[36];

  if (strlen(fecfname) == 0 || strlen(eqpfname) == 0) ccerr(un_allocated);

  if ((fp=fopen(eqpfname,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((name_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_line);
      if ((fecname_col=findcol(hdr,"FECNAME")) < 0) ccerr(no_such_line);
      if ((eqpmodule_col=findcol(hdr,"EQPMODULE")) < 0) ccerr(no_such_line);
      alarm_col=findcol(hdr,"ALARMSTATUS");
      context_col=findcol(hdr,"CONTEXT");
      reboot_col=findcol(hdr,"REBOOTCOUNT");
      done = 1;
      continue;
    }
    strncpy(RPCData[i].ExportName,colptr(name_col,str),ELEM_NAME_SIZE);
    strncpy(RPCData[i].FecName,colptr(fecname_col,str),FEC_NAME_SIZE);
    strncpy(RPCData[i].EqName,colptr(eqpmodule_col,str),EQP_NAME_SIZE);
    if (context_col >= 0)
    {
      strncpy(RPCData[i].Context,colptr(context_col,str),16);
    }
    if (alarm_col >= 0)
    {
      strncpy(scratch,colptr(alarm_col,str),16);
      RPCData[i].alarmSystem = atoi(scratch);
    }
    else RPCData[i].alarmSystem = 0;
    if (reboot_col >= 0)
    {
      strncpy(scratch,colptr(reboot_col,str),16);
      RPCDataRebootCount[i] = atoi(scratch);
    }
    if (NGdebug) 
    { 
      sprintf(s,"ENS: %s [FEC %s, EQP %s] added\n>",
              RPCData[i].ExportName,RPCData[i].FecName,RPCData[i].EqName);
      dbgoutput(s,tnull);
    }

    if (GetPropertyId("ENSEQM",RPCData[i].ExportName) < 0)
    {
      strncpy(prpname,RPCData[i].ExportName,16); prpname[16] = 0;
      if (RegisterProperty("ENSEQM",prpname,
                           sizeof(RPCDataStruct)+sizeof(RPCFecStruct),CF_STRUCT,
                           0x01,"registered EQM") < 0)
      {
        sprintf(logbuf,"ENS: couldn't register %s for export",RPCData[i].ExportName);
        nslog(logbuf);
      }
    }
    i++;
  }
  fclose(fp); fp = NULL;
  RPCNumRecsRead = i;

  if ((fp=fopen(fecfname,"r")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((fecname_col=findcol(hdr,"FECNAME")) < 0) ccerr(no_such_line);
      if ((fecnetwork_col=findcol(hdr,"FECNETWORK")) < 0) ccerr(no_such_line);
      if ((fecnode_col=findcol(hdr,"FECNODE")) < 0) ccerr(no_such_line);
      if ((ipaddr_col=findcol(hdr,"IPADDR")) < 0) ccerr(no_such_line);
      if ((os_col=findcol(hdr,"OS")) < 0) ccerr(no_such_line);
      if ((description_col=findcol(hdr,"DESCRIPTION")) < 0) ccerr(no_such_line);
      if ((location_col=findcol(hdr,"LOCATION")) < 0) ccerr(no_such_line);
      if ((hardware_col=findcol(hdr,"HARDWARE")) < 0) ccerr(no_such_line);
      if ((responsible_col=findcol(hdr,"RESPONSIBLE")) < 0) ccerr(no_such_line);
      portoffset_col = findcol(hdr,"PORTOFFSET");
      sub_col = findcol(hdr,"SUBSYSTEM");
      done = 1;
      continue;
    }
    strncpy(RPCFec[i].Name,colptr(fecname_col,str),FEC_NAME_SIZE);
    strncpy(scratch,colptr(fecnetwork_col,str),16);
    HexToBytes(scratch,RPCFec[i].Network,8);
    strncpy(scratch,colptr(fecnode_col,str),16);
    HexToBytes(scratch,RPCFec[i].Node,12);
    strncpy(RPCFec[i].IP,colptr(ipaddr_col,str),16);
    haddr = inet_addr(RPCFec[i].IP);
    if (portoffset_col >= 0)
    {
      strncpy(scratch,colptr(portoffset_col,str),16);
      po = atoi(scratch);
    }
    else
    {
      po = 0;
    }
    RPCFec[i].Port[0] = po&0xff;
    RPCFec[i].Port[1] = (po>>8)&0xff;
    RPCFec[i].PortOffset = (BYTE)po;
    memcpy(RPCFec[i].IPh_addr,&haddr,4);
    /* NB: colptr() strips the leading and trailing '"' if there are any */
    /* so we have to put them back if they're necessary */
    strncpy(scratch,colptr(os_col,str),12);
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%s\"",scratch); else strncpy(s,scratch,128);
    strncpy(RPCFecExt[i].os,s,12);

    strncpy(scratch,colptr(description_col,str),64);
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%s\"",scratch); else strncpy(s,scratch,128);
    strncpy(RPCFecExt[i].desc,s,64);
    strncpy(RPCFecExt[i].sub,getSubsystemFromDescription(RPCFecExt[i].desc),16);

    strncpy(scratch,colptr(location_col,str),32);
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%s\"",scratch); else strncpy(s,scratch,128);
    strncpy(RPCFecExt[i].loc,s,32);

    strncpy(scratch,colptr(hardware_col,str),32);
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%s\"",scratch); else strncpy(s,scratch,128);
    strncpy(RPCFecExt[i].hdw,s,32);

    if (sub_col >= 0)
    {
      strncpy(scratch,colptr(sub_col,str),16);
      if (strlen(scratch) > 0) strncpy(RPCFecExt[i].sub,scratch,16);
    }
    appendSubSystemList(RPCFecExt[i].sub);

    strncpy(scratch,colptr(responsible_col,str),32);
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%s\"",scratch); else strncpy(s,scratch,128);
    strncpy(RPCFecExt[i].resp,s,32);

    if (NGdebug) 
    { 
       sprintf(s,"ENS: FEC %s [%s %s] added\n>",
               RPCFec[i].Name,scratch,RPCFec[i].IP);
       dbgoutput(s,tnull);
    }

    if (GetPropertyId("ENSEQM",RPCFec[i].Name) < 0)
    {
      strncpy(prpname,RPCFec[i].Name,16); prpname[16] = 0;
      if (RegisterProperty("ENSEQM",prpname,
                           sizeof(RPCFecStruct),CF_STRUCT,
                           0x01,"registered FEC") < 0)
        printf("ENS: couldn't register %s for export\n>",RPCFec[i].Name);
    }

    i++;
  }
  fclose(fp); fp = NULL;
  RPCNumFecsRead = i;
  fillinFecIndexTable();
  return 0;

err: if (fp) fclose(fp);
  return cc;
}

int scanTables(void)
{
  BYTE *tmp;
  int cc;
  int rpcallocnum,fecallocnum;

  if (debug) printf("ENS: Scanning database files\n>");
  if (GetAllocSizesFromNameFile(&rpcallocnum,&fecallocnum))
  {
    rpcallocnum = 10; fecallocnum = 10;  /* no file found -> give it some space */
  }

  if (rpcallocnum > RPCallocNum)
  {
    RPCallocNum = rpcallocnum + FILESPACE; 
    if (!dbaseloaded || RPCData == NULL)
    {
      if (!(RPCData=(RPCDataStruct *)calloc(RPCallocNum+TMPSPACE,sizeof(RPCDataStruct))))
        return out_of_server_memory;
      if (!(fecIndex=(short *)calloc(RPCallocNum+TMPSPACE,sizeof(short))))
        return out_of_server_memory;
      if (!(RPCDataDisabled=(short *)calloc(RPCallocNum+TMPSPACE,sizeof(short))))
        return out_of_server_memory;
      if (!(RPCDataRebootCount=(UINT32 *)calloc(RPCallocNum+TMPSPACE,sizeof(UINT32))))
        return out_of_server_memory;
    }
    else
    {
      if (debug) printf("ENS: reallocating memory\n>");
      if (!(tmp=(BYTE *)realloc(RPCData,(RPCallocNum+TMPSPACE)*sizeof(RPCDataStruct))))
        return out_of_server_memory;
      RPCData = (RPCDataStruct *)tmp;
      if (!(tmp=(BYTE *)realloc(fecIndex,(RPCallocNum+TMPSPACE)*sizeof(short))))
        return out_of_server_memory;
      fecIndex = (short *)tmp;
      if (!(tmp=(BYTE *)realloc(RPCDataDisabled,(RPCallocNum+TMPSPACE)*sizeof(short))))
        return out_of_server_memory;
      RPCDataDisabled = (short *)tmp;
      memset(RPCDataDisabled,0,(RPCallocNum+TMPSPACE)*sizeof(short));
      if (!(tmp=(BYTE *)realloc(RPCDataRebootCount,(RPCallocNum+TMPSPACE)*sizeof(UINT32))))
        return out_of_server_memory;
      RPCDataRebootCount = (UINT32 *)tmp;
      memset(RPCDataRebootCount,0,(RPCallocNum+TMPSPACE)*sizeof(UINT32));
    }
  }

  if (fecallocnum > FECallocNum || RPCFec == NULL)
  {
    FECallocNum = fecallocnum + FILESPACE;
    if (!dbaseloaded || RPCFec == NULL)
    {
      if (!(RPCFec=(RPCFecStruct *)calloc(FECallocNum+TMPSPACE,sizeof(RPCFecStruct))))
        return out_of_server_memory;
      if (!(RPCFecExt=(RPCFecExStruct *)calloc(FECallocNum+TMPSPACE,sizeof(RPCFecExStruct))))
        return out_of_server_memory;
      if (!(RPCFecDisabled=(short *)calloc(FECallocNum+TMPSPACE,sizeof(short))))
        return out_of_server_memory;
    }
    else
    {
      if (!(tmp=(BYTE *)realloc(RPCFec,(FECallocNum+TMPSPACE)*sizeof(RPCFecStruct))))
        return out_of_server_memory;
      RPCFec = (RPCFecStruct *)tmp;
      if (!(tmp=(BYTE *)realloc(RPCFecExt,(FECallocNum+TMPSPACE)*sizeof(RPCFecExStruct))))
        return out_of_server_memory;
      RPCFecExt = (RPCFecExStruct *)tmp;
      if (!(tmp=(BYTE *)realloc(RPCFecDisabled,(FECallocNum+TMPSPACE)*sizeof(short))))
        return out_of_server_memory;
      RPCFecDisabled = (short *)tmp;
      memset(RPCFecDisabled,0,(FECallocNum+TMPSPACE)*sizeof(short));
    }
  }

  if (!RPCallocNum || !FECallocNum) return un_allocated;

  /* Now we have enough room: */
  if (debug) printf("ENS: Loading database files\n>");
  if ((cc=loadDatabaseFiles()) != 0)
  {
      sprintf(logbuf,"ENS: error loading database - %s",erlst[cc]);
      nslog(logbuf);
      return database_not_loaded;
  }
  dbaseloaded = 1;
  getContexts();
  return 0;
}
char fechdr[] = "FecName,FecNetwork,FecNode,IPaddr,PortOffset,OS,SubSystem,Description,Location,Hardware,Responsible\n";
char eqphdr[] = "Name,FecName,EqpModule,AlarmStatus,Context,RebootCount\n";

int dumpFECTable(void)
{
  FILE *fp = NULL;
  int i,cc = 0;
  char tmpFecName[32];

  if ((fp=fopen(fecfname,"w+t")) == NULL) ccerr(file_error);
  fprintf(fp,fechdr);
  for (i = 0; i<RPCNumFecsRead; i++)
  {
    if (RPCFecDisabled[i]) continue;
    strncpy(tmpFecName,RPCFec[i].Name,16); tmpFecName[16] = 0;
    fprintf(fp,"%s,%02x%02x%02x%02x,%02x%02x%02x%02x%02x%02x,%s,%d,%s,%s,%s,%s,%s,%s\n",
            tmpFecName,
            RPCFec[i].Network[0],RPCFec[i].Network[1],RPCFec[i].Network[2],RPCFec[i].Network[3],
            RPCFec[i].Node[0],RPCFec[i].Node[1],RPCFec[i].Node[2],RPCFec[i].Node[3],RPCFec[i].Node[4],RPCFec[i].Node[5],
            RPCFec[i].IP,GetPortOffset(&RPCFec[i]),RPCFecExt[i].os,
            RPCFecExt[i].sub,RPCFecExt[i].desc,RPCFecExt[i].loc,
            RPCFecExt[i].hdw,RPCFecExt[i].resp);
  }
  dumpFecFileFlag = 0;
err: if (fp) fclose(fp);
  if (cc) printf("ENS: write FEC Table - %s\n>",erlst[cc]);
  return cc;
}
int dumpEQPTable(void)
{
  FILE *fp = NULL;
  int i,cc = 0;
  char tmpExportName[32],tmpFecName[32],tmpContext[32];

  if ((fp=fopen(eqpfname,"w+t")) == NULL) ccerr(file_error);
  fprintf(fp,eqphdr);
  for (i = 0; i<RPCNumRecsRead; i++)
  {
    if (RPCDataDisabled[i]) continue;
    strncpy(tmpExportName,RPCData[i].ExportName,16); tmpExportName[16]=0;
    strncpy(tmpFecName,RPCData[i].FecName,16); tmpFecName[16]=0;
    strncpy(tmpContext,RPCData[i].Context,16); tmpContext[16]=0;
    fprintf(fp,"%s,%s,%s,%d,%s,%lu\n",
      tmpExportName,tmpFecName,RPCData[i].EqName,
      RPCData[i].alarmSystem,tmpContext,RPCDataRebootCount[i]);
  }
  dumpEqpFileFlag = 0;
err: 
  if (fp) fclose(fp);
  if (cc) feclog("ENS: write EQP Table - %s",erlst[cc]);
  return cc;
}
int RPCsize = 0, RECsize = 0;
#ifdef WINNT
# define lstat(f,s) _stat(f,s)
#endif
int loadENS(void) /* read name information from file */
{
  char *ptr;
  static int done = 0;
# ifdef WINNT
  struct _stat eqpstat,fecstat;
# else
  struct stat eqpstat,fecstat;
# endif
  static time_t OLDeqptime,OLDfectime;
  int n, i, j, cc;

  if (!done)
  {
    ControlDBpath[0] = 0;
    if ((ptr=getenv("CONTROLDB")) != NULL) strncpy(ControlDBpath,ptr,72);
    sprintf(eqpfname,"%s%s",ControlDBpath,EQPDB_FILE);
    sprintf(fecfname,"%s%s",ControlDBpath,FECDB_FILE);
    sprintf(gRPCDataBaseFullName,"%s%s",ControlDBpath,EQPDB_FILE);
    sprintf(gFECDataBaseFullName,"%s%s",ControlDBpath,FECDB_FILE);
    done = 1;
  }
  if (lstat(eqpfname,&eqpstat) < 0)
  {
    sprintf(logbuf,"ENS: cannot find eqp database");
    nslog(logbuf);
    return no_such_file;
  }
  if (lstat(fecfname,&fecstat) < 0)
  {
    sprintf(logbuf,"ENS: cannot find fec database");
    nslog(logbuf);
    return no_such_file;
  }
  if (OLDeqptime != eqpstat.st_mtime || OLDfectime != fecstat.st_mtime)
  {
    memset(RPCDataDisabled,0,RPCNumRecsRead*sizeof(short)); 
    memset(RPCDataRebootCount,0,RPCNumRecsRead*sizeof(UINT32)); 
    memset(RPCFecDisabled,0,RPCNumFecsRead*sizeof(short));
    if ((cc=scanTables()) != 0) printf("ENS: %s\n>",erlst[cc]);
    for (n=0; n<nConnectionTableEntries; n++)
    {
      for (i=0; i<RPCNumRecsRead; i++)
      {
        if (!strncmp(connTable[n]->name,RPCData[i].ExportName,16))
	{
	  for (j=0; j<RPCNumFecsRead; j++)
	    if (!strncmp(RPCData[i].FecName,RPCFec[j].Name,16))
	    {
	      connTable[n]->FECindex = j;
	      fixFecProtocol(&RPCFec[j]);
	      break;
	    }
	  if (j == RPCNumFecsRead)
          {
            sprintf(logbuf,"ENS: no FEC entry for %s link",RPCData[i].FecName);
            nslog(logbuf);
          }
	  connTable[n]->RPCindex = i;
          break;
        }
      }
      if (i == RPCNumRecsRead)
      {
        sprintf(logbuf,"ENS: no FEC entry for %s link",connTable[n]->name);
        nslog(logbuf);
      }
    }
    OLDeqptime = eqpstat.st_mtime;
    OLDfectime = fecstat.st_mtime;
    return -1;                              /* signal verification loop*/
  }
  return 0;                                 /* all's quiet */
}
typedef struct
{
  char pattern[16];
  char context[16];
} CONTEXTPATTERN;
CONTEXTPATTERN *contextPatterns = NULL;
#define CONTEXTS_FILE "contexts.csv"
int ncontextPatterns = 0;
NAME16 *contexts = NULL;
int ncontexts = 0;
int getContexts(void)
{
  int i,j,cc = 0;
  
  if (ncontexts >= RPCNumRecsRead)
  {
    free(contexts); contexts = NULL;
  }
  if (contexts == NULL)
  {
    ncontexts = 0;
    if ((contexts=(NAME16 *)calloc(RPCNumRecsRead,sizeof(NAME16))) == NULL)
      ccerr(out_of_server_memory);
  }
  for (i=0; i<RPCNumRecsRead; i++)
  {
    for (j=0; j<ncontexts; j++)
    {
      if (strncmp(contexts[j].name,RPCData[i].Context,16)) continue;
      break;
    }
    if (j == ncontexts && strlen(RPCData[i].Context)) /* not found */
    {
      strncpy(contexts[j].name,RPCData[i].Context,16);
      RegisterDeviceName("ENSEQM",contexts[j].name,4+j);
      ncontexts++;
    }
  }
err:
  return cc;
}
#define isENS(nam) (!strncmp(nam,"ENS",3))
#define ENSFWD_FILE "forward.csv"
int nforwards = 0;
NAME48 *forwards = NULL;
NAME64 *forwardContexts = NULL;
int getForwardsList(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int lst_col,con_col;
  char str[256],hdr[256];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (forwardContexts != NULL) 
  {
    free(forwardContexts);
    forwardContexts = NULL;
  }
  if (forwards != NULL)
  {
    free(forwards);
    forwards = NULL;
    nforwards = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((forwards=(NAME48 *)calloc(n,sizeof(NAME48))) == NULL) ccerr(out_of_server_memory);
  if ((forwardContexts=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
  nforwards = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<nforwards;)
  {
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((lst_col=findcol(hdr,"ENS_LIST")) < 0) ccerr(no_such_line);
      con_col = findcol(hdr,"CONTEXT");
      done = 1;
      continue;
    }
    strncpy(forwards[i].name,colptr(lst_col,str),48);
    if (con_col >= 0) strncpy(forwardContexts[i].name,colptr(con_col,str),64);
    if (forwardContexts[i].name[0] == '*') forwardContexts[i].name[0] = 0;
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
} 
int cmdForward(char *context,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int access)
{
  int i,cc=0;
  char devsrv[64];
  
  for (i=0; i<nforwards; i++)
  {
    if (strlen(forwardContexts[i].name) > 0 && 
        strstr(forwardContexts[i].name,context) == NULL) continue;
    sprintf(devsrv,"%s/%s",forwards[i].name,devName);
    cc = ExecLinkEx(devsrv,devProperty,dout,din,(short)access,200);
    feclog("Forward %s to %s\n",devProperty,devsrv);
  }
  return cc;
}
#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  for (i=0; i<nadmins && strncmp(nam,admins[i].name,16); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col;
  char str[256],hdr[256];
 
  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<nadmins;)
  {
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),16);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int getContextPatternInformation(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int pattern_col,context_col;
  char *c,str[256],hdr[256],scratch[16];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (contextPatterns != NULL)
  {
    free(contextPatterns);
    ncontextPatterns = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((contextPatterns=(CONTEXTPATTERN *)calloc(n,sizeof(CONTEXTPATTERN))) == NULL) ccerr(out_of_server_memory);
  ncontextPatterns = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<ncontextPatterns;)  /* scan file */
  {
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((pattern_col=findcol(hdr,"PATTERN")) < 0) ccerr(no_such_line);
      if ((context_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(pattern_col,str),16);
    if ((c=strchr(scratch,'*')) != NULL) *c = 0;
    strncpy(contextPatterns[i].pattern,scratch,16);
    strncpy(contextPatterns[i].context,colptr(context_col,str),16);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
void assignContext(char *expname,char *cntxt)
{
  char en[16];
  int i,len;
  
  strncpy(en,expname,16); strupr(en);
  for (i=0; i<ncontextPatterns; i++)
  {
    if ((len=strlen(contextPatterns[i].pattern)) > 16) len = 16;
    if (strncmp(en,contextPatterns[i].pattern,len)) continue;
    strncpy(cntxt,contextPatterns[i].context,16);
    return;
  }
  cntxt[0] = 0;
  return;
}
void initENS(void)
{
  int cc;

  debug = 1;
  printf("Initializing Equipment Name Server ...\n");
  if (!NameServerLoaded)
  {
    printf("Allocating Name Server space ...\n");
    if ((NameServer=(RPCFecStruct *)calloc(1,sizeof(RPCFecStruct))) == NULL)
    {
      printf("Cannot allocate Name Server space !\n");
      exit(1);
    }
  }
  getContextPatternInformation(CONTEXTS_FILE);
  getAdministrators(ADMINS_FILE);
  getForwardsList(ENSFWD_FILE);
  if ((cc=loadENS()) > 0)  { sprintf(logbuf,"ENS: %s",erlst[cc]); nslog(logbuf); }
  scanFlag = 0;
  printf("Equipment Name Server Initialized\n");
  nslog("ENS: initialized");
  debug = 0;
  return;
}
void taskENS(void)
{
  int cc;
  static int fec=0,counter = 0;
  static time_t interim = 0;
  time_t t = time(NULL);
  if (checkExportFlag) checkExport();
  else
  {
    if (dumpFecFileFlag) dumpFECTable();
    if (dumpEqpFileFlag > 0) dumpEQPTable();
  }
  if (t < interim + 60) return;
  if (dumpEqpFileFlag < 0) dumpEQPTable();
  interim = t;                          /* check for a new database each minute */
  counter++;                            /* count the minutes                    */

  if ((cc=loadENS()) > 0) feclog("ENS: loadENS -> %s",erlst[cc]);
  else if (cc) fec = 0;

  if (counter >= 60)
  {
    fec = 0;                            /* verify address each hour              */
    counter = 0;                        /* reset                                 */
  }
  return;
}
typedef struct
{
  char *elem;
  int   len;
} CSVINFO; 

CSVINFO fecexInfo[] = {
  { rpcfecex.os,12 },
  { rpcfecex.desc,64 },
  { rpcfecex.loc,32 },
  { rpcfecex.hdw,32 },
  { rpcfecex.sub,16 },
  { rpcfecex.resp,32 }
};
  
void checkCSVFormat(CSVINFO *inf,int num)
{
  char scratch[80];
  char *ptr,*loc;
  int i,j,slen,pos;

  for (i=0; i<num; i++)
  {
    loc = inf[i].elem; slen = inf[i].len;
    strncpy(scratch,loc,slen); scratch[slen] = 0;
    for (j=0; j<slen; j++) if (scratch[j] == '\n') scratch[j] = ' ';
    if ((ptr=strchr(scratch,'"')) != NULL)
    {
      if (!(ptr == scratch && scratch[strlen(scratch)-1] == '"'))
        loc[0] = 0;  /* don't know what's going on, and don't want to figure it out */
    }
    else if ((ptr=strchr(loc,',')) != NULL)
    {
      if ((pos=strlen(loc)) >= slen-2) pos = slen-2;
      loc[0] = '"'; strncpy(&loc[1],scratch,slen-2);
      loc[pos+1] = '"';
    }
  }
}
int getSubSystem(DTYPE *d,char *subs)
{
  int len;
  if (d->dArrayLength > 0)
  {
    memset(subs,0,32);
    switch (d->dFormat)
    {
      case CF_NAME8: strncpy(subs,d->data.cptr,8); break;
      case CF_NAME16:strncpy(subs,d->data.cptr,16); break;
      case CF_NAME32:strncpy(subs,d->data.cptr,32); break;
      case CF_TEXT: 
        len = MIN(32,d->dArrayLength);
        strncpy(subs,d->data.cptr,len); 
        break;
      default: return illegal_format;
    }
  }
  return 0;
}
int getRebootCount(char *cntxt,char *srv)
{
  int i,cntxtchk=0;
  if (cntxt != NULL &&  strlen(cntxt) > 0) cntxtchk = -1;
  for (i=0; i<RPCNumRecsRead; i++)
  {
    if (RPCDataDisabled[i]) continue;
    if (cntxtchk && strncmp(cntxt,RPCData[i].Context,16)) continue;
    if (strncmp(srv,RPCData[i].ExportName,16)) continue;
    return RPCDataRebootCount[i];
  }
  return -1;
}
int getNumExportTags(char *subsys,char *cntxt)
{
  int i,j,k=-1,n=0,subschk=0,cntxtchk=0,isFec=0;
   
  if (cntxt != NULL &&  strlen(cntxt) > 0) cntxtchk = -1;
  if (subsys != NULL && strlen(subsys) > 0)
  {
    if ((i=GetPropertyId("ENSEQM",subsys)) >= 0)
    {
      for (k=0; k<RPCNumFecsRead && strncmp(subsys,RPCFec[k].Name,16); k++);
      if (k < RPCNumFecsRead) isFec = -1;
    }
    if (!isFec)
    {
      if (!strncmp(subsys,"UNKNOWN",32)) subschk = -1; else subschk = 1;
    }
  }
  if (subschk == 0 && cntxtchk == 0 && isFec == 0) return RPCNumRecsRead;
  for (i=0; i<RPCNumRecsRead; i++)
  {
    if (RPCDataDisabled[i]) continue;
    if (cntxtchk && strncmp(cntxt,RPCData[i].Context,16)) continue;
    j = fecIndex[i];
    if (isFec && k != j) continue;
    if (subschk > 0 && strcmp(RPCFecExt[j].sub,subsys)) continue;
    if (subschk < 0 && strlen(RPCFecExt[j].sub) > 0) continue;
    n++;
  }
  return n;
}
int getNumKnownFecs(char *contxt,char *subsys)
{
  int i,j,k,n=0,chk=0;
  if (subsys == NULL || strlen(subsys) == 0) return RPCNumFecsRead;
  if (!strncmp(subsys,"UNKNOWN",32)) chk = -1; else chk = 1;
  if (contxt != NULL && strlen(contxt) > 0)
  {
    for (i=0; i<RPCNumRecsRead; i++)
    {
      if (RPCDataDisabled[i]) continue;
      if (strncmp(RPCData[i].Context,contxt,16)) continue;
      for (j=0; j<i && strncmp(RPCData[j].FecName,RPCData[i].FecName,16); j++);
      if (j < i) continue; 
      if (chk)
      {
        for (k=0; k<RPCNumFecsRead && strncmp(RPCData[i].FecName,RPCFec[k].Name,16); k++);
        if (k == RPCNumFecsRead) continue;
        if (RPCFecDisabled[k]) continue;
        if (chk > 0 && strcmp(RPCFecExt[k].sub,subsys)) continue;
        if (chk < 0 && strlen(RPCFecExt[k].sub) > 0) continue;
      }
      n++;
    }
  }
  else
  {
    for (i=0; i<RPCNumFecsRead; i++)
    {
      if (RPCFecDisabled[i]) continue;
      if (chk > 0 && strcmp(RPCFecExt[i].sub,subsys)) continue;
      if (chk < 0 && strlen(RPCFecExt[i].sub) > 0) continue;
      n++;
    }
  }
  return n;
}
int getKnownFecs(DTYPE *d,char *contxt,char *subsys)
{
  int i,j,k,m,n=-1,siz,dsiz,chk=0,isTag=0;
  dsiz = d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_NAME8: siz = 8; break;
    case CF_NAME16: siz = 16; break;
    case CF_NAME32: siz = 32; break;
    case CF_TEXT: siz = 16; dsiz /= 16; break;
    default: return illegal_format;
  }
  if (subsys != NULL && strlen(subsys) > 0) 
  {
    if ((i=GetPropertyId("ENSEQM",subsys)) >= 0)
    {
      for (n=0; n<RPCNumRecsRead && strncmp(subsys,RPCData[n].ExportName,16); n++);
      if (n < RPCNumRecsRead) isTag = -1;
    }
    if (!isTag)
    {
      if (!strncmp(subsys,"UNKNOWN",32)) chk = -1; else chk = 1;
    }
  }
  if (contxt != NULL && strlen(contxt) > 0)
  {
    for (i=0,k=0; i<RPCNumRecsRead && k<dsiz; i++)
    {
      if (RPCDataDisabled[i]) continue;
      if (strncmp(RPCData[i].Context,contxt,16)) continue;
      for (j=0; j<i; j++)
        if (!strncmp(RPCData[j].FecName,RPCData[i].FecName,16) &&
            !strncmp(RPCData[j].Context,RPCData[i].Context,16)) break;
      if (j < i) continue; 
      if (chk)
      {
        for (m=0; m<RPCNumFecsRead && strncmp(RPCData[i].FecName,RPCFec[m].Name,16); m++);
        if (m == RPCNumFecsRead) continue;
        if (RPCFecDisabled[m]) continue;
        if (chk > 0 && strcmp(RPCFecExt[m].sub,subsys)) continue;
        if (chk < 0 && strlen(RPCFecExt[m].sub) > 0) continue;
      }
      strncpy(&d->data.cptr[k*siz],RPCData[i].FecName,siz);
      k++;
    }
  }
  else
  {
    for (i=0,k=0; i<RPCNumFecsRead && k<dsiz; i++)
    {
      if (RPCFecDisabled[i]) continue;
      if (isTag && n != i) continue;
      if (chk)
      {
        if (chk > 0 && strstr(RPCFecExt[i].desc,subsys) == NULL) continue;
        if (chk < 0 && strchr(RPCFecExt[i].desc,']') != NULL) continue;
      } 
      strncpy(&d->data.cptr[k*siz],RPCFec[i].Name,siz);
      k++;
    }
  }
  if (k < dsiz) d->dArrayLength = k;
  return 0;
}
int getExportsForFec(int idx,BYTE *buf,int bufsiz)
{
  int i, n = 0, sizeleft=bufsiz;
  if (bufsiz < 16 || buf == NULL) return 0;
  for (i=0; i<RPCNumRecsRead && sizeleft >= 16; i++)
  {
    if (!strncmp(RPCData[i].FecName,RPCFec[idx].Name,16))
    {
      memcpy(&buf[n*16],RPCData[i].ExportName,16);
      sizeleft -= 16; n++;
    }
  }
  return n;
}
int getContextNames(DTYPE *d)
{
  int i,k,siz,dsiz;
  dsiz = d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_NAME8: siz = 8; break;
    case CF_NAME16: siz = 16; break;
    case CF_NAME32: siz = 32; break;
    case CF_TEXT: siz = 16; dsiz /= 16; break;
    default: return illegal_format;
  }
  for (i=0,k=0; i<ncontexts && k<dsiz; i++)
  {
    strncpy(&d->data.cptr[k*siz],contexts[i].name,siz);
    k++;
  }
  if (k < dsiz) d->dArrayLength = k;
  return 0;
}
int getExportTags(DTYPE *d,char *subsys,char *cntxt)
{
  int i,j,k,n=-1,siz,dsiz,subschk=0,cntxtchk=0,isFec=0;
  dsiz = d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_NAME8: siz = 8; break;
    case CF_NAME16: siz = 16; break;
    case CF_NAME32: siz = 32; break;
    case CF_TEXT: siz = 16; dsiz /= 16; break;
    default: return illegal_format;
  }
  if (cntxt != NULL && strlen(cntxt) > 0) cntxtchk = -1;
  if (subsys != NULL && strlen(subsys) > 0) 
  {
    if ((i=GetPropertyId("ENSEQM",subsys)) >= 0)
    {
      for (n=0; n<RPCNumFecsRead && strncmp(subsys,RPCFec[n].Name,16); n++);
      if (n < RPCNumFecsRead) isFec = -1;
    }
    if (!isFec)
    {
      if (!strncmp(subsys,"UNKNOWN",32)) subschk = -1; else subschk = 1;
    }
  }
  for (i=0,k=0; i<RPCNumRecsRead && k<dsiz; i++)
  {
    if (RPCDataDisabled[i]) continue;
    j = fecIndex[i];
    if (cntxtchk && strncmp(RPCData[i].Context,cntxt,16)) continue;
    if (isFec && n != j) continue;
    if (subschk)
    {
      if (subschk > 0 && strcmp(RPCFecExt[j].sub,subsys)) continue;
      if (subschk < 0 && strlen(RPCFecExt[j].sub) > 0) continue;
    } 
    strncpy(&d->data.cptr[k*siz],RPCData[i].ExportName,siz);
    k++;
  }
  if (k < dsiz) d->dArrayLength = k;
  return 0;
}
void trimname(char *nam,int len)
{
  int i, cplen=len;
  char tmp[64],*c = tmp;
  if (len > 62) len = 62;
  strncpy(tmp,nam,len); tmp[len] = 0;
  while (*c == ' ') { c++;  cplen--; }
  while (c[len-1] == ' ') { c[len-1] = 0; len--; cplen--; }
  strncpy(nam,c,cplen);
  /* trap this absurdity ... */
  for (i=0; i<(int)strlen(nam) && i<len; i++) if (nam[i] == '\n') nam[i] = ' ';
}
int checkNames(void)
{
# define CHECK_NAME(c) if (!isalnum((c)[0])) return -1;
  CHECK_NAME(rpcdata.FecName);
  CHECK_NAME(rpcdata.ExportName);
  CHECK_NAME(rpcdata.EqName);
  CHECK_NAME(rpcfec.Name);
  return 0;
}
void splitContextAndSubsystem(char *contxt,char *subsys)
{
  int i;
  char *c;
  if (contxt == NULL) return;
  if (subsys == NULL) return; /* no point in doing this .. */
  for (i=0; i<ncontexts && strncmp(contexts[i].name,contxt,16); i++);
  if (i < ncontexts) return;
  if ((c=strchr(contxt,'.')) == NULL) return;
  *c++ = 0;
  strncpy(subsys,c,8); subsys[8] = 0; /* divvy the 16-char devname up */
}
int enseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int i,j,cc,n,p,cntxtFlg = 0;
  UINT32 lIP,dsize,ssize;
  short EqNo,fmt;
  BYTE *dbuf = dout->data.bptr;
  static char tgt[20],subs[32],cntxt[20],usr[20],resp[32],ip[20];

  if ((EqNo=GetDeviceNumber("ENSEQM",devName)) < 0 || EqNo > 3) EqNo = 0;
  if (EqNo == 0 && devName[0] != '#' && 
      strlen(devName) > 0 && strcmp(devName,"DEFAULT"))
  {
    cntxtFlg = -1;
    memset(cntxt,0,16);
  }
  
  fmt = dout->dFormat;
  dsize = dout->dArrayLength;
  
  strncpy(usr,GetCaller(),16); strupr(usr);
  if (!strcmp(devProperty,"SIZES"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dsize != 2) return out_of_range;
    switch (fmt)
    {
      case CF_SHORT:
        dout->data.sptr[0] = RPCNumRecsRead;
        dout->data.sptr[1] = RPCNumFecsRead;
        break;
      case CF_LONG:
        dout->data.lptr[0] = RPCNumRecsRead;
        dout->data.lptr[1] = RPCNumFecsRead;
        break;
      default: return illegal_format;
    }
    return 0;
  }
  if (!strcmp(devProperty,"WHOAMI"))
  {
    tgt[0] = 0;
    if (din->dArrayLength > 0)
    {
      if (access & CA_WRITE) return illegal_read_write;
      switch (din->dFormat)
      {
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
          ssize = fmtsizeof(din->dFormat);
          break;
        case CF_TEXT:
          ssize = din->dArrayLength;
          break;
        default:
          return illegal_format;
      }
      if (ssize > EQP_NAME_SIZE) ssize = EQP_NAME_SIZE;
      strncpy(tgt,din->data.cptr,ssize);
    }
    i = 1;
    GetCallerInfo(NULL,NULL,&lIP,NULL,&i);
    for (i=0; i<RPCNumFecsRead; i++)
    {
      if (memcmp(&RPCFec[i].IPh_addr,&lIP,4)) continue;
      for (j=0; j<RPCNumRecsRead; j++)
      {
        if (strncmp(RPCData[j].FecName,RPCFec[i].Name,FEC_NAME_SIZE)) continue;
        if (strlen(tgt) > 0 && strncmp(tgt,RPCData[j].EqName,EQP_NAME_SIZE))
          continue;
        break;
      }
      if (j < RPCNumRecsRead) break;
    }
    if (i == RPCNumFecsRead) return non_existent_fec;
    if (j == RPCNumRecsRead) return non_existent_elem;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
        ssize = fmtsizeof(dout->dFormat);
        break;
      default:
        return illegal_format;
    }
    switch (dout->dArrayLength)
    {
      case 5:
        p = RPCFec[i].Port[0]+(RPCFec[i].Port[1]<<8);
        sprintf(dout->data.cptr+4*ssize,"%d",p);
      case 4:
        strncpy(dout->data.cptr+3*ssize,RPCData[j].Context,EQP_CONTEXT_SIZE);
      case 3:
        strncpy(dout->data.cptr+2*ssize,RPCData[j].FecName,FEC_NAME_SIZE);
      case 2:
        strncpy(dout->data.cptr+ssize,RPCData[j].EqName,EQP_NAME_SIZE);
      case 1:
        strncpy(dout->data.cptr,RPCData[j].ExportName,ELEM_NAME_SIZE);
        break;
      default:
        return dimension_error;
    }
    return 0;
  }

  if (!strcmp(devProperty,"CHECKFLAG"))
  {
    if (din->dArrayLength > 0)
    {
      if (!(access & CA_WRITE)) return illegal_read_write;
      if ((cc=getValuesAsLong(din,(UINT32 *)&checkExportFlag,1)) != 0) return cc;
    }
    if (dout->dArrayLength > 0)
    {
      if ((cc=putValuesFromLong(dout,(UINT32 *)&checkExportFlag,1)) != 0) return cc;
    }
    return 0;
  }
  if (!strcmp(devProperty,"REBOOTCOUNT"))
  {
    if (din == NULL || din->dArrayLength == 0) return illegal_read_write;
    memset(tgt,0,16);
    switch (din->dFormat)
    {
      case CF_TEXT: strncpy(tgt,din->data.cptr,MIN(16,din->dArrayLength)); break;
      case CF_NAME16: strncpy(tgt,din->data.cptr,16); break;
      default: return illegal_format;
    }
    if (strlen(tgt) == 0) return illegal_read_write;
    if (cntxtFlg) strncpy(cntxt,devName,16); else cntxt[0] = 0;
    if ((n=getRebootCount(cntxt,tgt)) < 0) return device_not_connected; 
    return putValuesFromLong(dout,(UINT32 *)&n,1);
  }
  if (!strcmp(devProperty,"REMOVE"))
  {
    if (din == NULL || din->dArrayLength == 0) return illegal_read_write;
    memset(tgt,0,16);
    switch (din->dFormat)
    {
      case CF_TEXT: strncpy(tgt,din->data.cptr,MIN(16,din->dArrayLength)); break;
      case CF_NAME16: strncpy(tgt,din->data.cptr,16); break;
      default: return illegal_format;
    }
    switch (EqNo)
    {
      case 1: /* FEC Name */
        sprintf(logbuf,"ENS: remove fec %s by %s",tgt,usr);
        for (i=0; i<RPCNumFecsRead; i++)
        {
          if (RPCFecDisabled[i]) continue;
          if (!strncmp(tgt,RPCFec[i].Name,16)) /* found it */
          {
            strncpy(resp,RPCFecExt[i].resp,32); strupr(resp);
            if (!isAdministrator(usr) && 
                 strlen(resp) && strstr(resp,usr) == NULL)
            {
              strcat(logbuf," -> access denied");
              nslog(logbuf);
              return access_denied;
            }
            RPCFecDisabled[i] = -1;
            for (n=0; n<RPCNumRecsRead; n++) 
              if (fecIndex[n] == i) RPCDataDisabled[n] = -1;
            dumpFecFileFlag = 1;
            dumpEqpFileFlag = 1;
            nslog(logbuf);
            if (cntxtFlg) strncpy(cntxt,devName,16); else cntxt[0] = 0;
            if (!isENS(usr)) cmdForward(cntxt,devName,devProperty,dout,din,access);
            return 0;
          }
        }
        strcat(logbuf," -> no such computer");
        nslog(logbuf);
        return no_such_computer;
      case 0:
        if (cntxtFlg == 0 && strcmp(devName,"DEFAULT")) break;
      case 2: /* EQP Name */
        sprintf(logbuf,"ENS: remove export tag %s by %s",tgt,usr);
        for (i=0; i<RPCNumRecsRead; i++)
        {
          if (RPCDataDisabled[i]) continue;
          if (cntxtFlg && strncmp(devName,RPCData[i].Context,16)) continue;
          if (!strncmp(tgt,RPCData[i].ExportName,16)) /* found it */
          {
            n = fecIndex[i];
            strncpy(resp,RPCFecExt[n].resp,32); strupr(resp);
            if (!isAdministrator(usr) &&
                strlen(resp) && strstr(resp,usr) == NULL) 
            {
              strcat(logbuf," -> access denied");
              nslog(logbuf);
              return access_denied;
            }
            RPCDataDisabled[i] = -1;
            dumpEqpFileFlag = 1;
            nslog(logbuf);
            if (!isENS(usr)) cmdForward(RPCData[i].Context,devName,devProperty,dout,din,access);
            return 0;
          }
        }
        strcat(logbuf," -> no such computer");
        nslog(logbuf);
        return no_such_computer;
      default: return illegal_equipment_number;
    }
  }
  if (!strcmp(devProperty,"EXPORT"))
  {
    if (debug) 
    {
      sprintf(dbgbuf,"ENS: incoming export ...\n>");
      dbgoutput(dbgbuf,tnull);
    }
    if (checkExportFlag) return operation_busy;
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if (din == NULL || din->dArrayLength != dsize) return illegal_read_write;
    if (dsize != ((int)(sizeof(RPCDataStruct)+sizeof(RPCFecStruct)+sizeof(RPCFecExStruct))))
      return out_of_range;
    memcpy(&rpcdata,din->data.vptr,sizeof(RPCDataStruct));
    memcpy(&rpcfec,&((char *)din->data.vptr)[sizeof(RPCDataStruct)],sizeof(RPCFecStruct));
    memcpy(&rpcfecex,&((char *)din->data.vptr)[sizeof(RPCDataStruct)+sizeof(RPCFecStruct)],sizeof(RPCFecExStruct));
    if (checkNames()) return illegal_string_set;
    checkCSVFormat(fecexInfo,5);
    if (EQPnconsumers == 1 && !isENS(usr) && !isAdministrator(usr))
    {
      inet_ntoa_b(EQPconsumerlist->client->IPaddress.sin_addr,(char *)rpcfec.IP);
      lIP = inet_addr(rpcfec.IP);
      memcpy(rpcfec.IPh_addr,&lIP,4);
      memcpy(&((char *)din->data.vptr)[sizeof(RPCDataStruct)],&rpcfec,sizeof(RPCFecStruct)); 
    }
    if (debug) 
    {
      sprintf(dbgbuf,"ENS: fec %s context %s tag %s\n>",rpcfec.Name,rpcdata.Context,rpcdata.ExportName);
      dbgoutput(dbgbuf,tnull);
    }
    if (strlen(rpcdata.Context) == 0 && strlen(devName) > 0)
    {
      strncpy(rpcdata.Context,devName,16);
      if (debug) 
      {
        sprintf(dbgbuf,"ENS: apply context from device name\n>");
        dbgoutput(dbgbuf,tnull);
      }
    }
    if (strlen(rpcdata.ExportName) &&
        strlen(rpcdata.EqName)  &&
        strlen(rpcdata.FecName) &&
        strlen(rpcfec.Name)) 
    {
      if (strlen(rpcdata.Context) < 3 || !isalpha(rpcdata.Context[0])) /* not allowed */
        assignContext(rpcdata.ExportName,rpcdata.Context);
      if (!strcmp(rpcdata.Context,"DEFAULT")) /* somebody claims context DEFAULT! */
        rpcdata.Context[0] = 0;               /* remove it ! */
      if (!strcmp(rpcdata.FecName,"ENS")) return 0; /* don't check yourself! */
      trimname(rpcdata.ExportName,16);
      trimname(rpcdata.EqName,6);
      trimname(rpcdata.FecName,16);
      trimname(rpcfec.Name,16);
      checkExportFlag = 1;
      lastRPCindex = lastFECindex = fecTargetId = -1; 
    }
    if (!isENS(usr)) cmdForward(rpcdata.Context,devName,devProperty,dout,din,access);
    return 0;
  }
  cntxt[0] = 0; subs[0] = 0;
  if (cntxtFlg) 
  {
    strncpy(cntxt,devName,16);
    splitContextAndSubsystem(cntxt,subs);
  }
  if (!strcmp(devProperty,"NFECS"))
  {
    if ((cc=getSubSystem(din,subs)) != 0) return cc;
    n = getNumKnownFecs(cntxt,subs);
    return putValuesFromLong(dout,(UINT32 *)&n,1);
  }
  if (!strcmp(devProperty,"FECS"))
  {
    if ((cc=getSubSystem(din,subs)) != 0) return cc;
    return getKnownFecs(dout,cntxt,subs);
  }
  if (!strcmp(devProperty,"NCONTEXTS"))
  {
    return putValuesFromLong(dout,(UINT32 *)&ncontexts,1);
  }
  if (!strcmp(devProperty,"CONTEXTS"))
  {
    return getContextNames(dout);
  }
  if (!strcmp(devProperty,"NTAGS"))
  {
    if ((cc=getSubSystem(din,subs)) != 0) return cc;
    n = getNumExportTags(subs,cntxt);
    return putValuesFromLong(dout,(UINT32 *)&n,1);
  }
  if (!strcmp(devProperty,"TAGS"))
  {
    if ((cc=getSubSystem(din,subs)) != 0) return cc;
    return getExportTags(dout,subs,cntxt);
  }
  if (!strcmp(devProperty,"NSUBSYSTEMS"))
  {
    return putValuesFromLong(dout,(UINT32 *)&nSubSystems,1);
  }
  if (!strcmp(devProperty,"SUBSYSTEMS"))
  {
    if (dout->dFormat != CF_NAME16) return illegal_format;
    if (dout->dArrayLength > (UINT32)nSubSystems) dout->dArrayLength = nSubSystems;
    for (i=0; i<(int)dout->dArrayLength; i++)
      strncpy(((NAME16 *)dout->data.vptr)[i].name,subSystems[i].name,16);
    return 0;
  }
  if (!strcmp(devProperty,"NALMTAGS"))
  {
    if (fmt != CF_INTEGER) return illegal_format;
    if (dsize != 1) return illegal_read_write;
    for (i=0,n=0; i < RPCNumRecsRead; i++) if (RPCData[i].alarmSystem) n++;
    *(short *)dbuf = n;
    return 0;
  }
  if (!strcmp(devProperty,"ALMTAGS"))
  {
    if (EqNo < 0) return illegal_equipment_number;
    if (access & CA_WRITE) return illegal_read_write;
    if (debug > 1) printf("ENS: request fec list\n>");
    if ((ssize=getFormatSize(fmt)) == 0) return illegal_format;
    for (i=EqNo,n=0; i<RPCNumRecsRead && n<(int)dsize-EqNo; i++) if (RPCData[i].alarmSystem)
    {
      strncpy(&dout->data.cptr[(n++)*ssize],RPCData[i].ExportName,ssize);
    }
    return 0;
  }

  switch (EqNo)
  {
    case 1: /* FEC Name query */
      if (NGdebug) dbglog("ENS: addr request for fec %s",devProperty);
      if (fmt != CF_STRUCT) return illegal_format;
      if (dsize != sizeof(RPCFecStruct)) return illegal_read_write;
      for (i=0; i<RPCNumFecsRead; i++)
      {
        if (RPCFecDisabled[i]) continue;
	if (!strncmp(devProperty,RPCFec[i].Name,16)) /* found it */
	{
          memcpy(dbuf,&RPCFec[i],sizeof(RPCFecStruct));
	  return 0;
        }
      }
      return no_such_computer;
    case 0:
      if (cntxtFlg == 0 && strcmp(devName,"DEFAULT")) 
      {
        if (NGdebug) 
          dbglog("ENS: addr request %s %s not well formed",devProperty,devName);
        break;
      }
    case 2: /* TAG Name query */
      if (NGdebug) 
        dbglog("ENS: addr request for srv %s in context %s",
          devProperty,cntxtFlg ? devName : "DEFAULT");
      if (debug > 1) printf("ENS: request info for tag %s\n>",devProperty);
      if (fmt != CF_STRUCT) return illegal_format;
      if (dsize != sizeof(RPCDataStruct)+sizeof(RPCFecStruct)) return illegal_read_write;
      for (n=0; n<RPCNumRecsRead; n++)
      {
        if (RPCDataDisabled[n]) continue;
        if (cntxtFlg && strncmp(devName,RPCData[n].Context,16)) continue;
        if (!strncmp(devProperty,RPCData[n].ExportName,16)) break;
      }
      if (n == RPCNumRecsRead) return non_existent;
      for (i=0; i<RPCNumFecsRead; i++)
      {
        if (RPCFecDisabled[i]) continue;
        if (!strncmp(RPCData[n].FecName,RPCFec[i].Name,16)) /* found it */
        {
          memcpy(dbuf,&RPCData[n],sizeof(RPCDataStruct));
          memcpy(&((char *)dbuf)[sizeof(RPCDataStruct)],&RPCFec[i],sizeof(RPCFecStruct));
          if (NGdebug) dbglog("ENS: return info for srv %s",devProperty);
          return 0;
        }
      }  
      return no_such_computer;
    case 3: /* FEC extended info query */
      if (NGdebug) dbglog("ENS: request extended info for fec %s",devProperty);
      if (fmt != CF_TEXT) return illegal_format;
      if (dsize < 188) return illegal_read_write;
      for (i=0; i<RPCNumFecsRead; i++)
      {
        if (RPCFecDisabled[i]) continue;
	if (!strncmp(devProperty,RPCFec[i].Name,16)) /* found it */
	{
          memcpy(dbuf,&RPCFecExt[i],188);
          if (dsize > 188) 
          { 
            if ((n=getExportsForFec(i,&dbuf[188],dsize-188)) > 0)
            {
              dout->dArrayLength = 188 + n*16;
            }
          }
	  return 0;
        }
      }
      return no_such_computer;
    default: break;
  }
  return illegal_property;
}
