/*
   =============================================================
   ==
   ==   Event archive script reader + GetAndStore()
   ==
   =============================================================
*/
#include "prolog.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "rpctypes.h"
#include "database.h"
#include "errors.h"
#include "servdef.h"
#if defined(UNIX)
#include "unixcmd.h"
#include "dirent.h"
#elif defined(WINNT)
#include "direct.h"
#include "ntcmd.h"
#elif defined(MSDOS)
#include "doscmd.h"
#endif
#include "arctools.h"
#include "arceqm.h"

char *archiveListFile = ARCHIVE_LIST_FILE;
ARB *arb = NULL;
int narb = 0;
int archivedebug = 1;
BYTE PreallocatedDataBuffer[PREALLOCSIZE];
UINT32 lastEventNumber = 0;
int EventNotificationPending = FALSE;
int RecordEventPending = FALSE;
int archiveDebugOutput = 0;
int GetARBIndex(char *srvName);
int RecordEvent(ulong eventNumber,ulong eventCode,char *devName);
int SetGlobalEvent(ulong eventCode,char *name);
int FillDeviceList(ARB *arb);
int FillPropertyAliasList(ARB *arb);
int getFiles(NAME8 *filebuf,char *fileext,time_t starttime,time_t stoptime,int bufsiz);
short ftoi(char *fmt);
int StoreWithStatus(int i,char *sts);
int registerDHSAlias(char *stag)
{
  char tag[10];
  if (stag == NULL) return argument_list_error;
  strncpy(tag,stag,8); tag[8] = 0;
  if (getStructFromTag(tag) != NULL) return 0; /* already registered */
  if (fillstruct(tag,OFFSETIN(DHS,channelID),32,CF_TEXT)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,eventcode),1,CF_LONG))  return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,arraysize),1,CF_SHORT)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,datatype),1,CF_SHORT)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,samplingrate),1,CF_FLOAT)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,pretrigger),1,CF_LONG)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,timeoffset),1,CF_LONG)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,lowcount),1,CF_LONG)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,highcount),1,CF_LONG)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,egu),16,CF_TEXT)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,egulow),1,CF_FLOAT)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,eguhigh),1,CF_FLOAT)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,datalength),1,CF_LONG)) return out_of_local_memory;
  if (fillstruct(tag,OFFSETIN(DHS,desc),4*64,CF_TEXT)) return out_of_local_memory;
  if (fillstruct(tag,sizeof(DHS),1,CF_NULL)) return out_of_local_memory;
  feclog("register DHS alias structure tag : %s",stag);
  return 0;
}
short acctoi(char *mod)
{
  short accessMode = 0;

  if (strstr(strupr(mod),"READ")) accessMode |= READ;
  if (strstr(strupr(mod),"WRITE")) accessMode |= WRITE;
  if (strstr(strupr(mod),"POLL")) accessMode |= POLL;
  if (strstr(strupr(mod),"WAIT")) accessMode |= WAIT;
  if (strstr(strupr(mod),"SETNUM")) accessMode |= SETNUM;
  if (strstr(strupr(mod),"PREALLOC")) accessMode |= PREALLOC;
  if (strstr(strupr(mod),"FORWARD")) accessMode |= FORWARD;

  return accessMode;
}

#define logerr(txt) { perror("open"); strcpy(s,txt); goto error; }

int MaxArchiveLines = 10000;
void AppendArchiveLog(char *str)
{
  time_t t = time(NULL);
  char s[80];
  FILE *fp = NULL;
  static int nlines = 0;

  if (nlines++ > MaxArchiveLines)
  {
    nlines = 0;
    remove("archive.bak");
    rename("archive.log","archive.bak");
  }
  s[0] = 0;
  if ((fp=fopen("archive.log","a")) == NULL) logerr("LOG: open");
  strncpy(s,str,78);
  if (s[strlen(s)-1] == '\n') s[strlen(s)-1] = 0;
  fprintf(fp,"%s : %s",s,ctime(&t));

error: if (fp) fclose(fp);
  return;
}
FILE *createAndOpen(char *fname)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WINNT)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];
  static char strdel[2] = { FS_DELIMITER , 0};

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,strdel); c != NULL; c=strtok(NULL,strdel))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return fopen(fname,"wb");
    }
    strcat(dirname,strdel);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH)) return NULL;
    }
#   elif defined(WINNT)
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return NULL;
    }
#   endif
  }
  return NULL;
}
/* ----------------------------------------------------------------- */
/*  *makefilename                                                    */
/* ----------------------------------------------------------------- */

char *makefilename(char *fname,char *src,char *ext,time_t t)
{
  struct tm *ts;
  int yr, mo;

  ts = localtime(&t);
  yr = ts->tm_year + 1900;
  mo = ts->tm_mon + 1;

  sprintf(fname,"%s/%d/%02d/%s/%lx.%s",src,yr,mo,ext,t,ext);
  return fname;
}
int getNumEntries(char *fn)
{
  FILE *fp = NULL;
  char str[256];
  int n=0;

  if ((fp=fopen(fn,"r")) == NULL) return -no_such_file;
  while (fgets(str,255,fp) != NULL)
  {
     if (strlen(str) == 0) continue;            /* carriage return */
     if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
     n++;
  }
  fclose(fp);
  return n-1;  /*  subtract off the header */
}
int getArchiveSets(char *fn)
{
  FILE *fp = NULL;
  int cc=0,done=0;
  int hdr_col,ext_col,name_col,rec_col,srv_col,dep_col;
  int fwd_srv_col,fwd_prp_col,fwd_dat_col;
  char str[256],hdr[256],scratch[16];
  int i=0,nsets;

  if ((nsets=getNumEntries(fn)) < 0) return -nsets;
  if ((arb=(ARB *)calloc(nsets,sizeof(ARB))) == NULL) ccerr(out_of_local_memory);

  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL && i<nsets)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str);
      if ((srv_col=findcol(hdr,"SERVER")) < 0) ccerr(no_such_line);
      if ((name_col=findcol(hdr,"NAME")) < 0) 
        name_col = findcol(hdr,"TRIGGER");
      if ((ext_col=findcol(hdr,"EXTENSION")) < 0) ccerr(no_such_line);
      rec_col = findcol(hdr,"RECORD");
      hdr_col = findcol(hdr,"HEADER");
      dep_col = findcol(hdr,"DEPTH");
      fwd_srv_col = findcol(hdr,"FWD_SERVER");
      fwd_prp_col = findcol(hdr,"FWD_PROPERTY");
      fwd_dat_col = findcol(hdr,"FWD_DATA");
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(arb[i].server,colptr(srv_col,str),16);
    if (name_col >= 0)
    {
      strncpy(arb[i].name,colptr(name_col,str),16);
    }
    else
    {
      strncpy(arb[i].name,arb[i].server,16);
      strlwr(arb[i].name);
    }
    if (hdr_col >= 0)
    {
      strncpy(arb[i].hdr,colptr(hdr_col,str),80);
    }
    else
    {
      sprintf(arb[i].hdr,"%s-header.csv",arb[i].name);
    }
    strncpy(arb[i].ext,colptr(ext_col,str),16);
    if (rec_col >= 0)
    {
      strncpy(scratch,colptr(rec_col,str),16);
      arb[i].recordEvents = atoi(scratch);
    }
    else
    {
      arb[i].recordEvents = 0;
    }
    if (dep_col >= 0)
    {
      strncpy(scratch,colptr(dep_col,str),16);
      arb[i].depthInMonths = atoi(scratch);
    }
    if (fwd_srv_col >= 0)
    {
      strncpy(arb[i].fwdsrv,colptr(fwd_srv_col,str),64);
      if (strlen(arb[i].fwdsrv) > 0)
      {
        if (fwd_prp_col >= 0)
        {
          strncpy(arb[i].fwdprp,colptr(fwd_prp_col,str),32);
        }
        else
        {
          strncpy(arb[i].fwdprp,"TRIGGER",32);
        }
        if (fwd_dat_col >= 0)
        {
          strncpy(scratch,colptr(fwd_dat_col,str),16);
          arb[i].fwddata = atoi(scratch);
        }
      }
    }
    i++;
  }
err:
  if (fp) fclose(fp);
  feclog("archive sets file %s : %s",fn, cc ? erlst[cc] : "found");
  return cc ? -cc : nsets;
}
int findDAS(DAS *das,int n)
{
  int i;
  
  for (i=0; i<n; i++)
  {
    if (!strncmp(das[i].tag,das[n].tag,16) &&
        !strncmp(das[i].prp,das[n].prp,32) &&
        !strncmp(das[i].dev,das[n].dev,32) &&
        das[i].dout.dArrayLength == das[n].dout.dArrayLength &&
        das[i].dout.dFormat == das[n].dout.dFormat) break;
  }
  return i;
}
int getDASInputData(DAS *das,char *fn,DTYPE *din)
{
  FILE *fp = NULL;
  int fmtsiz,cc=0,done=0;
  int data_col,lock_col = -1;
  char str[256],hdr[256],scratch[16];
  int i=0,n;

  if (din == NULL) ccerr(illegal_read_write);
  if ((fmtsiz=fmtsizeof(din->dFormat)) <= 0) ccerr(illegal_format);
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(no_such_file);
  if ((din->data.vptr=calloc(n,fmtsiz)) == NULL) ccerr(out_of_local_memory);
  if ((das->dinlock=(BYTE *)calloc(n,1)) == NULL) ccerr(out_of_local_memory);
  din->dArrayLength = n;

  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;            /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str);
      if ((data_col=findcol(hdr,"DATA")) < 0) ccerr(no_such_line);
      lock_col = findcol(hdr,"LOCK");
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(scratch,colptr(data_col,str),16);
    switch (din->dFormat)
    {
      case CF_NAME16: strncpy(&din->data.cptr[i*16],str,16); break;
      case CF_TEXT: strncpy(din->data.cptr,str,16); break;
      case CF_BYTE: din->data.bptr[i] = atoi(scratch); break;
      case CF_SHORT: din->data.sptr[i] = atoi(scratch); break;
      case CF_LONG: din->data.lptr[i] = atoi(scratch); break;
      case CF_FLOAT: din->data.fptr[i] = (float)strtod(scratch,NULL); break;
      case CF_DOUBLE: din->data.dptr[i] = strtod(scratch,NULL); break;
      default: ccerr(illegal_format);
    }
    if (lock_col >= 0)
    {
      strncpy(scratch,colptr(lock_col,str),16);
      das->dinlock[i]  = atoi(scratch);
    }
    i++;
  }
  err:
  if (fp) fclose(fp);
  if (cc) printf("Data Input File %s : %s\n",fn,erlst[cc]);
  return cc;
}
void getPathFromFileName(char *fn,char *path)
{
  char *c0, *c1;
  strncpy(path,fn,80);
  for (c1=c0=path; (c1=strchr(c0+1,'/')) != NULL; c0=c1);
  *c0 = 0;
}
int getMaxCallsForDeviceName(char *dev)
{
  int i,n = 1,pstart,pstop;
  char scratch[32],*substr[5],*c;
  strncpy(scratch,dev,32);

  if ((substr[0]=strtok(scratch,".")) == NULL) substr[0] = scratch;
  for (i=0; substr[i] != NULL && i<4; i++,substr[i]=strtok(NULL,"."))
  {
    pstart = pstop = 0;
    if ((c=strchr(substr[i],'#')) != NULL) pstart = atoi(&c[1]);
    if ((c=strrchr(substr[i],'#')) != NULL) pstop = atoi(&c[1]);
    n *= pstop - pstart + 1;
  }
  return n;
}
int getDAS(char *fn,DAS *das)
{
  FILE *fp = NULL;
  int dsiz,fmtsiz,cc=0,done=0;
  int tag_col,prp_col,dev_col,tagArc_col,prpArc_col,
      siz_col,fmt_col,acc_col,wait_col,value_col,to_col,
      scale_col,shift_col,file_col = -1,input_format_col = -1;
  char str[256],hdr[256],fname[960],scratch[64],path[80];
  char *c,stag[16];
  int i=0,j;

  if (das == NULL) ccerr(un_allocated);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  getPathFromFileName(fn,path);
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;            /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str);
      if ((tag_col=findcol(hdr,"TAG")) < 0) ccerr(no_such_line);
      if ((prp_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((dev_col=findcol(hdr,"DEVICE")) < 0) ccerr(no_such_line);
      if ((tagArc_col=findcol(hdr,"ARCHIVETAG")) < 0) ccerr(no_such_line);
      if ((prpArc_col=findcol(hdr,"ARCHIVEPROPERTY")) < 0) ccerr(no_such_line);
      if ((siz_col=findcol(hdr,"SIZE")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(hdr,"FORMAT")) < 0) ccerr(no_such_line);
      if ((acc_col=findcol(hdr,"ACCESS")) < 0) ccerr(no_such_line);
      if ((wait_col=findcol(hdr,"WAIT")) < 0) ccerr(no_such_line);
      if ((value_col=findcol(hdr,"VALUE")) < 0) ccerr(no_such_line);
      if ((to_col=findcol(hdr,"TIMEOUT")) < 0) ccerr(no_such_line);
      if ((scale_col=findcol(hdr,"SCALE")) < 0) ccerr(no_such_line);
      if ((shift_col=findcol(hdr,"SHIFT")) < 0) ccerr(no_such_line);
      file_col = findcol(hdr,"INPUTFILE");
      input_format_col = findcol(hdr,"INPUTFORMAT");
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(das[i].tag,colptr(tag_col,str),16);
    strncpy(das[i].prp,colptr(prp_col,str),32);
    strncpy(das[i].dev,colptr(dev_col,str),32);
    strncpy(das[i].tagArc,colptr(tagArc_col,str),16);
    strncpy(das[i].prpArc,colptr(prpArc_col,str),32);
    /* output DTYPE: */
    strncpy(scratch,colptr(siz_col,str),16);
    das[i].dout.dArrayLength = atoi(scratch);
    strncpy(scratch,colptr(fmt_col,str),64);
    stag[0] = 0;
    if ((c=strchr(scratch,'/')) != NULL)
    {
      *c++ = 0;
      strncpy(stag,c,8);
      registerDHSAlias(stag);
    }
    das[i].dout.dFormat = ftoi(scratch);
    strncpy(scratch,colptr(acc_col,str),16);
    das[i].acc = acctoi(scratch);
    if (das[i].dout.dFormat == CF_STRUCT && 
        das[i].dout.dArrayLength == 1) /* must be tagged struct */
    {
      /* if not given: tag = "DHS" */
      if (strlen(stag) == 0) strncpy(stag,"DHS",8);
      strncpy(das[i].dout.dTag,stag,8);
    }
    else das[i].dout.dTag[0] = 0;
    /* input DTYPE: */
    if (file_col >= 0)
    {
      if (input_format_col >= 0)
      {
        strncpy(scratch,colptr(input_format_col,str),16);
        das[i].din.dFormat = ftoi(scratch);
      }
      else das[i].din.dFormat = CF_LONG; /* default; */
      strncpy(scratch,colptr(file_col,str),16);
      sprintf(fname,"%s/%s",path,scratch);
      getDASInputData(&das[i],fname,&das[i].din);
    }
    strncpy(scratch,colptr(wait_col,str),16);
    das[i].wait = atoi(scratch);
    strncpy(scratch,colptr(value_col,str),16);
    das[i].value = atoi(scratch);
    strncpy(scratch,colptr(to_col,str),16);
    das[i].timeout = atoi(scratch);
    if (das[i].acc != POLL && (j=findDAS(das,i)) == i)
    {
      if (das[i].acc & PREALLOC)
      {
        das[i].dout.data.vptr = PreallocatedDataBuffer;
      }
      else
      {
        fmtsiz = fmtsizeof(das[i].dout.dFormat);
        dsiz   = das[i].dout.dArrayLength;
        if (das[i].dout.dFormat == CF_STRUCT && das[i].dout.dArrayLength == 1)
          fmtsiz = sizeof(DHS);
        /*if (dsiz > UDP_BUFFER_SIZE) <- this was surely a mistake ! */
        if ((dsiz*fmtsiz) > PREALLOCSIZE)
        {
          if ((das[i].dout.data.vptr=(BYTE *)calloc(dsiz,fmtsiz)) == NULL)
            ccerr(out_of_local_memory);
        }
        else
        { 
           das[i].dout.data.vptr = PreallocatedDataBuffer;
        }
      }
    }
    else das[i].dout.data.vptr = das[j].dout.data.vptr;
    strncpy(scratch,colptr(scale_col,str),16);
    das[i].scale = (float)atof(scratch);
    das[i].askRemote = strcmp(strupr(scratch),"ASK REMOTE") ? 0 : 1;
    strncpy(scratch,colptr(shift_col,str),16);
    das[i].shift = (float)atof(scratch);

    i++;
  }
  err:
  if (fp) fclose(fp);
  if (cc) printf("%s Data Access Structure : %s\n",fn,erlst[cc]);
  return cc ? -cc : i;
}

int readArchiveSets(void)
{
  char fn[128],s[128];
  int i,cc;

  if ((narb=getArchiveSets(archiveListFile)) < 0) return -narb;

  for (i=0; i<narb; i++)
  {
    sprintf(fn,"%s/triggerFiles/%s.csv",stdArchiveDirectory,arb[i].name);
    if ((arb[i].nsteps=getNumEntries(fn)) < 0)
    {
      sprintf(s,"readArchiveSet %s : %s\n",fn,erlst[-arb[i].nsteps]);
      feclog(s);
      continue;
    }
    if ((arb[i].das=(DAS *)calloc(arb[i].nsteps,sizeof(DAS))) == NULL) return out_of_local_memory;
    if ((cc=getDAS(fn,arb[i].das)) < 0) return -cc;
    FillDeviceList(&arb[i]);
    FillPropertyAliasList(&arb[i]);
  }
  return 0;
}
void initArchiveTask(void)
{
  int i,cc;

  RegisterUserCommand("ArchiveOutput",NULL,&archiveDebugOutput,NULL,CA_WRITE|CA_READ);   

  if ((cc=readArchiveSets()) != 0) printf("Init Read error: %s\n>",erlst[cc]);
  for (i=0; i<narb; i++) 
  {
    RegisterDeviceName("STREQM",arb[i].name,i);
  }
  AssignNumModules("STREQM",narb);
  return;
}

typedef struct
{
  char tag[16];
  char prp[32];
  char dev[16];
  short siz;
  short fmt;
  float scale;
  float shift;
  BYTE *data;
  FILE *fp;
  int  linkid;
} DCS; /* Data Callback structure */
#define MAXCALLBACKS 50
DCS dcs[MAXCALLBACKS];
int maxcallbacks = MAXCALLBACKS;
int ncallbacks;
int GetAndStoreCallbackError = 0;
void GetAndStoreCallback(int id,int cc)
{
  int i,fmtsiz;
  short sval;
  float fval;
  char s[80];
  
  for (i=0; i<maxcallbacks && dcs[i].linkid != id; i++);
  if (i == maxcallbacks) ccerr(link_error);
  if (cc)
  {
    sprintf(s,"%s",cc < numErr ? erlst[cc] : GetRPCLastError((short)cc));
    StoreWithStatus(i,s);
    ccerr(cc);
  }
  if (dcs[i].data == NULL) ccerr(out_of_local_memory);

  strncpy(s,dcs[i].tag,16);
  if (fwrite(s,16,1,dcs[i].fp) != 1) ccerr(file_error);
  strncpy(s,dcs[i].prp,32);
  if (fwrite(s,32,1,dcs[i].fp) != 1) ccerr(file_error);
  strncpy(s,dcs[i].dev,16);
  if (fwrite(s,16,1,dcs[i].fp) != 1) ccerr(file_error);
  sval = dcs[i].siz; memcpy(&s[0],&sval,2);
  sval = dcs[i].fmt; memcpy(&s[2],&sval,2);
  if (fwrite(s,4,1,dcs[i].fp) != 1) ccerr(file_error);
  fval = dcs[i].scale; memcpy(&s[0],&fval,4);
  fval = dcs[i].shift; memcpy(&s[4],&fval,4);
  if (fwrite(s,8,1,dcs[i].fp) != 1) ccerr(file_error);
  
  fmtsiz = fmtsizeof(dcs[i].fmt);
  if (fwrite(dcs[i].data,(size_t)fmtsiz,(size_t)dcs->siz,dcs[i].fp) != (unsigned int)dcs->siz) ccerr(file_error);
err:
  if (cc)
  {
    if (archiveDebugOutput)
    {
      if (i < maxcallbacks)
        sprintf(s,"archive: %s %s %s",dcs[i].tag,dcs[i].dev,dcs[i].prp);
      else
        sprintf(s,"archive: %s %s",dcs[0].tag,dcs[0].prp);
      sprintf(&s[strlen(s)]," :: <%d> %s",cc, cc > 0 && cc < numErr ? erlst[cc] : "user defined");
      AppendArchiveLog(s);
    }
    GetAndStoreCallbackError = cc;
  }
  if (ncallbacks) ncallbacks--;
}
int StoreWithStatus(int i,char *sts)
{
  int cc=0;
  short sval;
  float fval;
  char s[80];
  
  strncpy(s,dcs[i].tag,16);
  if (fwrite(s,16,1,dcs[i].fp) != 1) ccerr(file_error);
  strncpy(s,dcs[i].prp,32);
  strcat(s,".STS");
  if (fwrite(s,32,1,dcs[i].fp) != 1) ccerr(file_error);
  strncpy(s,dcs[i].dev,16);
  if (fwrite(s,16,1,dcs[i].fp) != 1) ccerr(file_error);
  sval = 32; memcpy(&s[0],&sval,2);
  sval = CF_TEXT; memcpy(&s[2],&sval,2);
  if (fwrite(s,4,1,dcs[i].fp) != 1) ccerr(file_error);
  fval = 0; memcpy(&s[0],&fval,4); memcpy(&s[4],&fval,4);
  if (fwrite(s,8,1,dcs[i].fp) != 1) ccerr(file_error);
  strncpy(s,sts,32);
  if (fwrite(s,32,1,dcs[i].fp) != 1) ccerr(file_error);
err:
  return cc;
}

#define WaitForArchiveLinks(n) {\
  ncallbacks = n;\
  while (ncallbacks) SystemDelay(100);\
  n = 0;\
}  
  
int GetAndStore(FILE *fp,DAS *das)
{
  int id,cc = 0,first = 1,retries = 0;
  char s[80],*substr[5],*c;
  float datascale[2];
  char scratch[32],devname[16],fulldevname[32];
  DTYPE dout,din,*pdin = NULL;
  int n,lnk = 0,pstart[4],pstop[4],pnow[4],i[4],depth,numlinks;
  int blocksize;
  UINT32 stst[2];
  time_t tentry = time(NULL);

  if (archiveDebugOutput)
  {
    sprintf(s,"GetAndStore: %s %s",das->dev,das->prp);
    AppendArchiveLog(s);
  }
  memset(pstart,0,sizeof(pstart)); memset(pstop,0,sizeof(pstop));
  ncallbacks = 0;
  
  memset(scratch,0,32);
  strncpy(scratch,das->dev,32);  /* copy to scratch area for manipulation */
  if ((substr[0]=strtok(scratch,".")) == NULL) substr[0] = scratch;
  for (depth=0; substr[depth] != NULL && depth<4; depth++,substr[depth]=strtok(NULL,"."))
  {
    pstart[depth] = pstop[depth] = 0;
    if ((c=strchr(substr[depth],'#')) != NULL) pstart[depth] = atoi(&c[1]);
    if ((c=strrchr(substr[depth],'#')) != NULL) pstop[depth] = atoi(&c[1]);
  }

  if ((blocksize=das->dout.dArrayLength*fmtsizeof(das->dout.dFormat)) <= 0) ccerr(illegal_read_write);
  if (das->dout.dFormat == CF_STRUCT && das->dout.dArrayLength == 1) blocksize = sizeof(DHS);
  maxcallbacks = blocksize > UDP_BUFFER_SIZE ? 1 : UDP_BUFFER_SIZE/blocksize;
  
top:
  GetAndStoreCallbackError = 0;

  pnow[0] = pstart[0]; pnow[1] = pstart[1]; 
  pnow[2] = pstart[2]; pnow[3] = pstart[3];
  numlinks = 0;

  for (i[0]=pstart[0]; i[0] <= pstop[0]; i[0]++)
    for (i[1]=pstart[1]; i[1] <= pstop[1]; i[1]++)
      for (i[2]=pstart[2]; i[2] <= pstop[2]; i[2]++)
        for (i[3]=pstart[3]; i[3] <= pstop[3]; i[3]++)
  {
    if (strchr(scratch,'#') == NULL) strncpy(devname,scratch,16);
    else 
    {
      for (n=0,devname[0]=0; n<depth; n++)
        sprintf(&devname[strlen(devname)],n < depth-1 ? "#%d." : "#%d",i[n]);
    }
    for (n=0,devname[0]=0; n<depth; n++)
    {
      if (substr[n][0] == '#')
        sprintf(&devname[strlen(devname)], n < depth-1 ? "#%d." : "#%d",i[n]);
      else
        sprintf(&devname[strlen(devname)], n < depth-1 ? "%s." : "%s",substr[n]);
    }
    sprintf(fulldevname,"%s/%s",das->tag,devname);

    if (das->askRemote)
    {
      dout.dFormat = CF_FLOAT;
      dout.dArrayLength = 2;
      dout.data.fptr = datascale;
      if ((cc=ExecLink(fulldevname,"DATASCALE",&dout,NULL,CA_READ)) != 0) ccerr(cc);
      das->scale = datascale[0];
      das->shift = datascale[1];
    }  
    memcpy(&dout,&das->dout,sizeof(DTYPE));
    dout.data.bptr += numlinks*blocksize; /* move pointer along for multiple calls */

    if (das->numberToAcquire)
    {
      if (!(das->acc&PREALLOC)) ccerr(un_allocated);
      if (dout.dFormat == CF_STRUCT) dout.dArrayLength *= das->numberToAcquire;
      else dout.dArrayLength = das->numberToAcquire;
      blocksize = dout.dArrayLength*fmtsizeof(dout.dFormat);
      if (blocksize > PREALLOCSIZE) ccerr(buffer_too_small);
    }
    
    if (das->din.dArrayLength > 0) pdin = &das->din;
    else if (strstr(das->prp,".HIST") != NULL)
    {
      stst[1] = tentry;
      stst[0] = stst[1] - dout.dArrayLength;
      din.dFormat = CF_LONG;
      din.dArrayLength = 2;
      din.data.lptr = (UINT32 *)&stst;
      pdin = &din;
    }
    else pdin = NULL;
    
    if ((id=AttachLink(fulldevname,das->prp,&dout,pdin,CA_READ,
            3000,GetAndStoreCallback,CM_SINGLE)) < 0) ccerr(-id);
    numlinks++;

    strncpy(dcs[lnk].tag,das->tagArc,16);
    strncpy(dcs[lnk].prp,das->prpArc,32);
    strncpy(dcs[lnk].dev,devname,16);
    dcs[lnk].siz = (short)dout.dArrayLength;
    if (das->dout.dFormat == CF_STRUCT && dcs[lnk].siz == 1) dcs[lnk].siz = sizeof(DHS);
    dcs[lnk].fmt = dout.dFormat;
    dcs[lnk].scale = das->scale;
    dcs[lnk].shift = das->shift;
    dcs[lnk].data = dout.data.bptr;
    dcs[lnk].fp = fp;
    dcs[lnk].linkid = id;
    if (++lnk >= maxcallbacks)
    {
       WaitForArchiveLinks(lnk);
       if (GetAndStoreCallbackError == link_not_open)
       {
         if (retries < 3)
         {
           if (archiveDebugOutput)
           {
             sprintf(s,"%s : GetAndStore: trying again ...",fulldevname);
             AppendArchiveLog(s);
           }
           retries++;
           if (first) goto top;
           i[0] = pnow[0]; i[1] = pnow[1]; i[2] = pnow[2]; i[3] = pnow[3];
         }
         else
         {
           sprintf(s,"%s : Assuming server down",fulldevname);
           AppendArchiveLog(s);
           return link_not_open;
         }
       }
       else
       {
         pnow[0] = i[0]; pnow[1] = i[1]; pnow[2] = i[2]; pnow[3] = i[3];
         retries = 0;
       }
       first = 0;
       GetAndStoreCallbackError = 0;
       numlinks = 0;
    }
  }
  WaitForArchiveLinks(lnk);
  if (GetAndStoreCallbackError == link_not_open && retries < 3)
  {
    if (archiveDebugOutput)
    {
      sprintf(s,"%s : GetAndStore: trying again ...",fulldevname);
      AppendArchiveLog(s);
    }
    if (retries == 0)
    {
      pstart[0] = pnow[0]; pstart[1] = pnow[1]; 
      pstart[2] = pnow[2]; pstart[3] = pnow[3];
      if (pstart[3] < pstop[3]) pstart[3]++;
      else if (pstart[2] < pstop[2]) pstart[2]++;
      else if (pstart[1] < pstop[1]) pstart[1]++;
      else if (pstart[0] < pstop[0]) pstart[0]++;
    }
    retries++;
    goto top;
  }
err:
  if (cc)
  {
    if (archiveDebugOutput)
    {
      sprintf(s,"archive: %s %s",das->tag,das->prp);
      sprintf(&s[strlen(s)]," :: %s",erlst[cc]);
      AppendArchiveLog(s);
    }
    das->numberToAcquire = 0;
  }
  else
  {
    /* special flag to signal the next call ... */
    if (das->acc&SETNUM) 
    {
      switch (dcs[0].fmt)
      {
        case CF_SHORT:       
          das->numberToAcquire = *(UINT16 *)dcs[0].data;
          break;
        case CF_LONG:       
          das->numberToAcquire = *(UINT32 *)dcs[0].data;
          break;
        default:
          das->numberToAcquire = 1;
      }
      if (archiveDebugOutput)
      {
        sprintf(s,"%s %s number to acquire %d",das->tag,das->prp,das->numberToAcquire);
        AppendArchiveLog(s);
      }
    }
    else das->numberToAcquire = 0;
  }
  
  return cc;
}
int PollTarget(DAS *das)
{
  static short val;
  int cc;
  char devname[32];
  DTYPE dout;

  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = (UINT16 *)&val;
  sprintf(devname,"%s/%s",das->tag,das->dev);
  if ((cc=ExecLink(devname,das->prp,&dout,NULL,CA_READ)) != 0) return -cc;
  if (val == das->value) return 0;
  return not_ready;
}
int SendSignal(DAS *das)
{
  char devname[32];
  DTYPE din;

  din.dFormat = CF_SHORT;
  din.dArrayLength = 1;
  din.data.sptr = (UINT16 *)&das->value;
  sprintf(devname,"%s/%s",das->tag,das->dev);
  return ExecLink(devname,das->prp,NULL,&din,CA_WRITE);
}
int GetDataForNextStep(DAS *dasIn,DAS *dasOut)
{
  int i,j=0,cc=0;
  char devname[32];
  DTYPE *pdin=NULL,dout;

  if (dasIn->dout.dFormat != dasOut->din.dFormat) ccerr(illegal_format);
  if (dasOut->dinlock == NULL) ccerr(un_allocated);
  if (dasIn->din.dArrayLength > 0) pdin = &dasIn->din;
  memcpy(&dout,&dasIn->dout,sizeof(DTYPE));
  sprintf(devname,"%s/%s",dasIn->tag,dasOut->dev);
  if ((cc=ExecLinkEx(devname,dasIn->prp,&dout,pdin,CA_READ,3000)) != 0) ccerr(cc);
  for (i=0; i<(int)dasOut->din.dArrayLength; i++)
  {
    if (!dasOut->dinlock[i]) switch (dasOut->din.dFormat)
    {
      case CF_SHORT: 
           dasOut->din.data.sptr[i] = dasIn->dout.data.sptr[j]; break;
      case CF_LONG:  
           dasOut->din.data.lptr[i] = dasIn->dout.data.lptr[j]; break;
      case CF_FLOAT:  
           dasOut->din.data.fptr[i] = dasIn->dout.data.fptr[j]; break;
      case CF_DOUBLE:  
           dasOut->din.data.dptr[i] = dasIn->dout.data.dptr[j]; break;
      case CF_BYTE:  
           dasOut->din.data.bptr[i] = dasIn->dout.data.bptr[j]; break;
      default:
           ccerr(illegal_format);
    }
    if (j < (int)dasIn->dout.dArrayLength - 1) j++; 
  }

err:
  return cc;
}
void checkRecordEvent(int cc)
{
  if (cc) printf("RecordEvent: %s\n>",cc ? erlst[cc] : "Success");
  RecordEventPending = FALSE;
}
void checkSetEvent(ulong eventnumber, int cc)
{
  if (cc == 0) /* everything is okay */
  {
    lastEventNumber = eventnumber;
  }
  EventNotificationPending = FALSE;
  printf("SetEvent %d : %s\n>",eventnumber,cc ? erlst[cc] : "Success");
}
int isArchiveOngoing(char *filename)
{
  int i;
  char fn[80],s[256];

  if (filename == NULL) return FALSE;
  for (i=0; i<narb; i++)
  {
    if (arb[i].fp == NULL) continue;
    makefilename(fn,stdArchiveDirectory,arb[i].ext,arb[i].eventtime);
    if (strncmp(filename,fn,80)) continue;
    sprintf(s,"attempt to access on-going archive %s refused\n",fn);
    AppendArchiveLog(s);
    return TRUE;
  }
  return FALSE;
}
void bkgArchiveTask(void)
{
  char s[160],fn[80],scratch[16];
  int i,j,k,cc,newmonth = 0;
  time_t tnow=time(NULL),stop;
  struct tm *tm,tmref;
  static int errCount = 0, mon = -1;

  tm = localtime(&tnow);
  if (mon != tm->tm_mon)
  {
    newmonth = -1;
    mon = tm->tm_mon;
  }
  if (CommandLevel != 0)     /* for tests from the command line */
  {
    if (CommandLevel <= narb)
    {
      arb[CommandLevel-1].step = 1;
      archivedebug = 1;
    }
    CommandLevel = 0;
  }

  for (i=0; i<narb; i++)
  {
    if (arb[i].depthInMonths > 0 && newmonth)
    {
      tmref.tm_mon = (tm->tm_mon - arb[i].depthInMonths + 240) % 12;
      tmref.tm_year = tm->tm_year - arb[i].depthInMonths/12;
      if (tmref.tm_mon > tm->tm_mon) tmref.tm_year--;
      tmref.tm_mday = 1;
      tmref.tm_hour = 0;
      tmref.tm_min = 0;
      tmref.tm_sec = 0;
      stop = mktime(&tmref);
      getFiles(NULL,arb[i].ext,0,stop,-1);
    }
    if (arb[i].step < 0) /* pre-step: get trigger time */
    {
      if (!arb[i].resultPending)
      {
        lastEventNumber = time(NULL);
        arb[i].resultPending = TRUE;
        if ((cc=SetGlobalEvent((MARCHSRV_EVENTCODE|i),arb[i].name)) != 0) 
        {
          sprintf(s,"SetGlobalEvent : %s",erlst[cc]);
          AppendArchiveLog(s);
        }
        else EventNotificationPending = TRUE;
      }
      if (EventNotificationPending) continue;
      arb[i].eventtime = lastEventNumber;
      arb[i].step = 1;
      sprintf(s,"%s : assign event 0x%x",arb[i].name,lastEventNumber);
      AppendArchiveLog(s);
    }
    if ((j=arb[i].step-1) >= 0)
    {
      if (j >= arb[i].nsteps)
      {
        scratch[0] = 0;
        sprintf(s,"%s : close archive file",arb[i].name);
        AppendArchiveLog(s);
        arb[i].step = 0;
        arb[i].resultPending = FALSE;
        if (arb[i].fp != NULL) fclose(arb[i].fp);
        arb[i].fp = NULL;
        if (arb[i].aborted)
        {
          makefilename(fn,stdArchiveDirectory,arb[i].ext,arb[i].eventtime);
          sprintf(s,"%s : removing archive file %s",arb[i].name,fn);
          AppendArchiveLog(s);
          remove(fn);
        }
        else if (arb[i].recordEvents)
        {
          RecordEventPending = TRUE;
          for (k=0; k<arb[i].nsteps; k++)
          {
            if (strncmp(arb[i].das[k].tagArc,scratch,16))
            {
              strncpy(scratch,arb[i].das[k].tagArc,16);
              RecordEvent(arb[i].eventtime,MARCHSRV_EVENTCODE|i,scratch);
            }
          }
          /* worry about whether this was successful or not later */
        }
        continue;
      }

      /* 1. All archive procedures get initialized */
      if (j == 0 && arb[i].fp == NULL) /* first things first */
      {
        errCount = 0;
        sprintf(s,"%s : start archiving",arb[i].name);
        AppendArchiveLog(s);
        arb[i].interim = arb[i].start = tnow;
        arb[i].aborted = 0;
        /* use the event number !!! */
        makefilename(fn,stdArchiveDirectory,arb[i].ext,arb[i].eventtime);
        /* check: if this file exists, then I'm already doing or have done this! */
        if ((arb[i].fp = fopen(fn,"wb")) == NULL &&
            (arb[i].fp = createAndOpen(fn)) == NULL)
        {
          sprintf(s,"%s : can't open file %s",arb[i].name,fn);
          AppendArchiveLog(s);
          arb[i].step = arb[i].nsteps+1;
          continue;
        }
        sprintf(s,"%s : open file %s",arb[i].name,fn);
        AppendArchiveLog(s);
      }

      /* 2. Does the archive step involve polling for a specific return value ? */
      if (arb[i].das[j].acc&POLL)
      {
        if (tnow < arb[i].interim + arb[i].das[j].wait) continue;
        sprintf(s,"%s : polling %s %s against %d",
          arb[i].name,arb[i].das[j].tag,arb[i].das[j].prp,arb[i].das[j].value);
        if (archiveDebugOutput) AppendArchiveLog(s);
        cc = PollTarget(&arb[i].das[j]);
        switch (cc)
        {
          case 0: 
            arb[i].step++; /* target met, next step */
            arb[i].start = tnow;
            sprintf(s,"%s : target met",arb[i].name);
            AppendArchiveLog(s);
            break;
          case not_ready:
            if (tnow > arb[i].start + arb[i].das[j].timeout)
            {
              sprintf(s,"%s : %s %s timed out",
                arb[i].name,arb[i].das[j].tag,arb[i].das[j].prp);
              AppendArchiveLog(s);
              arb[i].step = arb[i].nsteps+1;
              if (archivedebug) dbglog("timed out");
            }
            errCount = 0;
            break; /* keep trying */
          default:
            if (errCount++ > 10)  arb[i].step = arb[i].nsteps + 1;
            sprintf(s,"%s : %s %s -> %s",
              arb[i].name,arb[i].das[j].tag,arb[i].das[j].prp,erlst[-cc]);
            AppendArchiveLog(s);
            break;
        }
        arb[i].interim = tnow;
        continue;
      }

      /* 3. Does the archive step involve getting archive data ? */
      if (arb[i].das[j].acc&READ)
      {
        if (j>0 && arb[i].das[j-1].numberToAcquire > 0) 
          arb[i].das[j].numberToAcquire = arb[i].das[j-1].numberToAcquire;
        else 
          arb[i].das[j].numberToAcquire = 0;
        sprintf(s,"%s : get and store %s %s",
          arb[i].name,arb[i].das[j].tag,arb[i].das[j].prp);
        AppendArchiveLog(s);
        cc = GetAndStore(arb[i].fp,&arb[i].das[j]);
        arb[i].step++;
        arb[i].interim = arb[i].start = tnow;
        continue;
      }

      /* 4. Does the archive step involve waiting a specified time? */
      if (arb[i].das[j].acc&WAIT)
      {
        if (tnow < arb[i].interim + arb[i].das[j].wait) continue;
        arb[i].interim = arb[i].start = tnow;
        arb[i].step++;
        sprintf(s,"%s : waited %d sec",arb[i].name,arb[i].das[j].wait);
        AppendArchiveLog(s);
        continue;
      }

      /* 5. Does the archive step involve sending a signal to the FEC? */
      if (arb[i].das[j].acc&WRITE)
      {
        sprintf(s,"%s : Send signal %s %s",
           arb[i].name,arb[i].das[j].tag,arb[i].das[j].prp);
        AppendArchiveLog(s);
        cc = SendSignal(&arb[i].das[j]);
        arb[i].step++;
        arb[i].interim = arb[i].start = tnow;
        continue;
      }

      /* 6. Does the archive step involve getting archive data to forward ? */
      if (arb[i].das[j].acc&FORWARD)
      {
        if (j < arb[i].nsteps-1) 
          GetDataForNextStep(&arb[i].das[j],&arb[i].das[j+1]);
        arb[i].step++;
        arb[i].interim = arb[i].start = tnow;
        continue;
      }

      /* empty step ? */
      sprintf(s,"%s : empty step",arb[i].name);
      AppendArchiveLog(s);
      arb[i].step++;
    }
  }
}
int FillPropertyAliasList(ARB *arb)
{
  char fname[80];
  int i,j,cc=0,done,name_col,alias_col,n;
  char str[256],hdr[256],tagname[16],prpname[32],alias[32];
  FILE *fp = NULL;
  tagname[0] = 0; prpname[0] = 0;
  for (j=0; j<arb->nsteps; j++) if (arb->das[j].acc&READ)
  {
    if (!strncmp(tagname,arb->das[j].tagArc,16)) continue;
    if (!strncmp(prpname,arb->das[j].prp,32)) continue;
    strncpy(tagname,arb->das[j].tagArc,16);
    sprintf(fname,"%s-properties.csv",tagname);
    if (fp != NULL) fclose(fp);
    if ((fp=fopen(fname,"r")) == NULL) ccerr(no_such_file);
scanfile:
    n = 0; done = 0;
    while (fgets(str,255,fp) != NULL)
    {
      if (strchr("%#;\n",str[0])) continue;
      if (!done)
      {
        str[255] = 0; strcpy(hdr,str);
        if ((name_col=findcol(hdr,"PROPERTY_NAME")) < 0) ccerr(no_such_column);
        if ((alias_col=findcol(hdr,"PROPERTY_ALIAS")) < 0) ccerr(no_such_column);
        done = 1;
        continue;
      }
      strncpy(prpname,colptr(name_col,str),32);
      if (strncmp(arb->das[j].prp,prpname,32)) continue;
      if (arb->das[j].prpAliasLst != NULL)
      {
        strncpy(alias,colptr(alias_col,str),32);
        if (n < arb->das[j].nprpAlias)
        {
          strncpy(arb->das[j].prpAliasLst[n].name,alias,32);
        }
      }
      n++;
    }
    if (n > 0 && arb->das[j].prpAliasLst == NULL)
    {
      arb->das[j].nprpAlias = n;
      if ((arb->das[j].prpAliasLst=(NAME32 *)calloc(n,sizeof(NAME32))) == NULL)
ccerr(out_of_local_memory);
      rewind(fp);
      goto scanfile;
    }
    for (i=j+1; i<arb->nsteps; i++)
      if (arb->das[i].acc&READ &&
          !strncmp(arb->das[i].tagArc,tagname,16) &&
          !strncmp(arb->das[i].prp,prpname,32))
    {
      arb->das[i].prpAliasLst = arb->das[j].prpAliasLst;
      arb->das[i].nprpAlias   = n;
    }
  }
err:
  if (fp) fclose(fp);
  return cc ? -cc : n;
}
int FillDeviceList(ARB *arb)
{
  char fname[80];
  int i, cc=0,done,name_col,num_col,alias_col,n;
  char *c,str[256],hdr[256],scratch[16],tagname[16];
  FILE *fp = NULL;
  int j,p = 0,found = 0;

  tagname[0] = 0;
  for (j=0; j<arb->nsteps; j++) if (arb->das[j].acc&READ)
  {
    if (!strncmp(tagname,arb->das[j].tagArc,16)) continue;
    strncpy(tagname,arb->das[j].tagArc,16);
    sprintf(fname,"%s/deviceFiles/%s-devices.csv",stdArchiveDirectory,tagname);
    if ((fp=fopen(fname,"r")) == NULL) ccerr(no_such_file);
scanfile:
    n = 0; done = 0;
    while (fgets(str,255,fp) != NULL)
    {
      if (strchr("%#;\n",str[0])) continue;
      if (!done)
      {
        str[255] = 0; strcpy(hdr,str);
        if ((name_col=findcol(hdr,"DEVICE_NAME")) < 0) ccerr(no_such_column);
        if ((num_col=findcol(hdr,"DEVICE_NUMBER")) < 0) ccerr(no_such_column);
        alias_col=findcol(hdr,"DEVICE_ALIAS");
        done = 1;
        continue;
      }
      strncpy(scratch,colptr(num_col,str),16);
      if (arb->das[j].devlst != NULL)
      {
        i = atoi(scratch);
        if (i < arb->das[j].ndev)
        {
          if (*(c=colptr(name_col,str)) == '\'') c++;
          strncpy(arb->das[j].devlst[i].name,c,16);
          if (arb->das[j].devAliasLst != NULL && alias_col >= 0)
          {
            if ((c=colptr(alias_col,str)) == NULL) continue;
            strncpy(arb->das[j].devAliasLst[i].name,c,16);
          }
        }
      }
      n++;
    }
    if (n > 0 && arb->das[j].devlst == NULL)
    {
      arb->das[j].ndev = n;
      if ((arb->das[j].devlst=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_local_memory);
      if (alias_col >= 0 && 
          (arb->das[j].devAliasLst=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL)
        ccerr(out_of_local_memory);
      rewind(fp);
      goto scanfile;
    }
    for (i=j+1; i<arb->nsteps; i++) 
      if (arb->das[i].acc&READ && !strncmp(arb->das[i].tagArc,tagname,16))
    {
      arb->das[i].devlst = arb->das[j].devlst;
      arb->das[i].devAliasLst = arb->das[j].devAliasLst;
      arb->das[i].ndev   = n;
    }
err:
    if (fp) fclose(fp);
  }
  return cc ? -cc : n;
}
int GetStoredDevices(NAME16 *devnames,char *tagname,char *context,int id,int num)
{
  int n,j,k,found=0,done=0,cc=0;
  DAS *das = NULL;
  int name_col,num_col;
  FILE *fp=NULL;
  char *c,fn[128],scratch[32],hdr[256],str[256],tmp[128];

  /* parse off possible identifier from the tagname */
  if ((c=strchr(tagname,'[')) != NULL)
  {
    *c++ = 0;
    sprintf(tmp,"%s-%s",tagname,c);
    if ((c=strchr(tmp,']')) == NULL) ccerr(syntax_error);
    *c = 0;
    strcat(tmp,"-devices.csv");
    sprintf(fn,"%s/deviceFiles/%s",stdArchiveDirectory,tmp);

    if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
    n = 0; done = 0;
    while (fgets(str,255,fp) != NULL)
    {
      if (strchr("%#;\n",str[0])) continue;
      if (!done)
      {
        str[255] = 0; strcpy(hdr,str);
        if ((name_col=findcol(hdr,"DEVICE_NAME")) < 0) ccerr(no_such_column);
        if ((num_col=findcol(hdr,"DEVICE_NUMBER")) < 0) ccerr(no_such_column);
        done = 1;
        continue;
      }
      strncpy(scratch,colptr(num_col,str),16);
      j = atoi(scratch);
      if (j > n) n = j;
      if (j >= num) continue;
      if (devnames != NULL) 
        strncpy(devnames[j].name,colptr(name_col,str),16);
    }
err:
    if (fp) fclose(fp);
    return cc ? -cc : n;
  }

  /* get ARB from context: */
  if (context != NULL && strlen(context) > 0)
  {
    if ((k=GetARBIndex(context)) < 0) return 0;
  }
  else
  {
    for (k=0; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
      if (!strncmp(tagname,arb[k].das[j].tagArc,16)) found = -1;
    if (!found) return -non_existent_elem;
    k--; /* back up ! */
  }
  /* get das from tagname : */

  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (strncmp(tagname,das->tagArc,16)) continue;
    if (das->devlst != NULL) 
    {
      for (n=0; devnames != NULL && n<das->ndev && n<num; n++)
      {
        if (das->devAliasLst != NULL)
          strncpy(devnames[n].name,das->devAliasLst[n].name,16);
        else 
          strncpy(devnames[n].name,das->devlst[n].name,16);
      }
      for (;n<num; n++) devnames[n].name[0] = 0;
      return das->ndev;
    }
    break;
  }
  return 0;
}
char *GetStoredDeviceName(int devnr,char *tagname)
{
  int j,k,p = 0,found = 0;
  DAS *das;
  static char lcldevnam[17];

  sprintf(lcldevnam,"#%d",devnr); /* default */
  /* get das from tagname : */
  for (k=0; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
    if (!strncmp(tagname,arb[k].das[j].tagArc,16)) found = -1;
  if (!found) return lcldevnam;
  k--; /* back up ! */
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (strncmp(tagname,das->tagArc,16)) continue;
    if (das->devlst != NULL) 
    {
      if (devnr < 0 || devnr >= das->ndev) return NULL;
      strncpy(lcldevnam,das->devlst[devnr].name,16); lcldevnam[16] = 0;
      break;
    }
  }
  return lcldevnam;
}
char *GetStoredDeviceNameFromAlias(char *alias,char *tagname)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  static char lcldevnam[17];

  lcldevnam[0] = 0; /* default */
  /* get das from tagname : */
  for (k=0; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
    if (!strncmp(tagname,arb[k].das[j].tagArc,16)) found = -1;
  if (!found) return lcldevnam;
  k--; /* back up ! */
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (strncmp(tagname,das->tagArc,16)) continue;
    if (das->devlst == NULL || das->devAliasLst == NULL) return NULL;
    for (i=0; i<das->ndev; i++) 
      if (!strncmp(das->devAliasLst[i].name,alias,16)) 
        return das->devlst[i].name;
  }
  return NULL;
}
char *GetStoredDeviceNameFromAliasEx(char *context,char *alias,char *tagname)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  static char lcldevnam[17];

  lcldevnam[0] = 0; /* default */
  /* get ARB from context: */
  if (context != NULL && strlen(context) > 0 && context[0] != '#')
  {
    if ((k=GetARBIndex(context)) < 0) return NULL;
  }
  else
  {
    for (k=0; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
      if (!strncmp(tagname,arb[k].das[j].tagArc,16)) found = -1;
    if (!found) return lcldevnam;
    k--; /* back up ! */
  }
  /* get das from tagname : */
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (strncmp(tagname,das->tagArc,16)) continue;
    if (das->devlst == NULL || das->devAliasLst == NULL) return NULL;
    for (i=0; i<das->ndev; i++)
      if (!strncmp(das->devAliasLst[i].name,alias,16))
        return das->devlst[i].name;
  }
  return NULL;
}
int GetStoredDeviceNumber(NAME16 *devname,char *tagname)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;

  /* get das from tagname : */
  for (k=0; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
    if (!strncmp(tagname,arb[k].das[j].tagArc,16)) found = -1;
  if (!found) return -non_existent_elem;
  if (devname->name[0] == '#') return atoi(&devname->name[1]);
  k--; /* back up ! */
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (strncmp(tagname,das->tagArc,16)) continue;
    if (das->devlst != NULL) 
    {
      for (i=0; i<das->ndev; i++) if (!strncmp(das->devlst[i].name,devname->name,16)) break;
      if (i < das->ndev) return i;
    }
  }
  return -1;
}
int GetStoredDeviceNumberEx(char *context,NAME16 *devname,char *tagname)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  /* get ARB from context: */
  if (context != NULL && strlen(context) > 0 && context[0] != '#')
  {
    if ((k=GetARBIndex(context)) < 0) return -1;
  }
  else
  {
    for (k=0; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
      if (!strncmp(tagname,arb[k].das[j].tagArc,16)) found = -1;
    if (!found) return -non_existent_elem;
    if (devname->name[0] == '#') return atoi(&devname->name[1]);
    k--; /* back up ! */
  }
  /* get das from tagname : */
  if (devname->name[0] == '#') return atoi(&devname->name[1]);
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (strncmp(tagname,das->tagArc,16)) continue;
    if (das->devlst != NULL)
    {
      for (i=0; i<das->ndev; i++) if (!strncmp(das->devlst[i].name,devname->name
,16)) break;
      if (i < das->ndev) return i;
    }
  }
  return -1;
}
int GetStoredProperties(NAME16 *props,char *tagname,char *dev,int num)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  char prp[32];

  prp[0] = 0;
  /* get ARB from context: */
  if (dev != NULL && strlen(dev) > 0 && dev[0] != '#')
  {
    if ((k=GetARBIndex(dev)) < 0) return -1;
  }
  else
  {
    for (k=0; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
      if (!strncmp(tagname,arb[k].das[j].tagArc,16)) found = -1;
    if (!found) return -non_existent_elem;
    k--; /* back up ! */
  }
  /* get das from tagname : */
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (strncmp(tagname,das->tagArc,16)) continue;
    /* find all the properties that go with it: */
    if (!strncmp(prp,das->prpArc,16)) continue;
    for (i=0; props != NULL && i<p; i++) if (!strncmp(props[i].name,das->prpArc,16)) break;
    if (props != NULL && i < p) continue;
    strncpy(prp,das->prpArc,16);
    if (strstr(prp,".HDR") != NULL) continue;
    if (props != NULL && p < num) strncpy(props[p].name,das->prpArc,16);
    p++;
  }
  return p;
} 
int GetStoredPropertiesEx(char *context,NAME16 *props,char *tagname,char *dev,int num)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  char prp[32];

  prp[0] = 0;
  /* get ARB from context: */
  if ((k=GetARBIndex(context)) < 0) return -1;
  /* get das from tagname : */
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (strncmp(tagname,das->tagArc,16)) continue;
    /* find all the properties that go with it: */
    if (!strncmp(prp,das->prpArc,16)) continue;
    for (i=0; props != NULL && i<p; i++) if (!strncmp(props[i].name,das->prpArc,
16)) break;
    if (props != NULL && i < p) continue;
    strncpy(prp,das->prpArc,16);
    if (strstr(prp,".HDR") != NULL) continue;
    if (props != NULL && p < num) strncpy(props[p].name,das->prpArc,16);
    p++;
  }
  return p;
}
char *GetStoredPropertyFromAlias(char *tagname,char *alias)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  static char prp[33];
 
  /* get das from tagname : */
  for (k=0; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
    if (!strncmp(tagname,arb[k].das[j].tagArc,16)) found = -1;
  if (!found) return NULL;
  k--; /* back up ! */
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (das->prpAliasLst == NULL) continue;
    if (strncmp(tagname,das->tagArc,16)) continue;
    for (i=0; i<das->nprpAlias; i++)
      if (!strncmp(das->prpAliasLst[i].name,alias,32))
        return das->prp;
  }
  return NULL;
}
char *GetStoredPropertyFromAliasEx(char *context,char *tagname,char *alias)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  static char prp[33];

  /* get ARB from context: */
  if ((k=GetARBIndex(context)) < 0) return NULL;
  /* get das from tagname : */
  for (j=0; j<arb[k].nsteps; j++) if (arb[k].das[j].acc&READ)
  {
    das = &arb[k].das[j];
    if (das->prpAliasLst == NULL) continue;
    if (strncmp(tagname,das->tagArc,16)) continue;
    for (i=0; i<das->nprpAlias; i++)
      if (!strncmp(das->prpAliasLst[i].name,alias,32))
        return das->prp;
  }
  return NULL;
}
