// This module fills in the methods for eventlst.h
// It should maintain its basic form but can be edited as needed.

#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include "errors.h"
#include "eventlst.h"
#include "srvlst.h"
#include "rpctypes.h"

void AppendArchiveLog(char *str)
{
  FILE *fp=NULL;
  time_t t = time(NULL);
  char s[80];

  s[0] = 0;
  if ((fp=fopen("archive.log","a")) == NULL) return;
  strncpy(s,str,78);
  if (s[strlen(s)-1] == '\n') s[strlen(s)-1] = 0;
  fprintf(fp,"%s : %s",s,ctime(&t));
  if (fp) fclose(fp);
  return;
}

enum seek_dir { beg, cur, end };

int getDeviceID(char *strArchiveServer,char *strDeviceServer);

ulong findMaximumEventCode(int cd1,int cd2)
{
  return cd1 > cd2 ? cd1 : cd2;  // kluge this for now
}
void makeEventFileName(ulong eventNumber,char *eventFileName)
{
  struct tm *lt=localtime((time_t *)&eventNumber);
  sprintf(eventFileName,"%d.%d",1900+lt->tm_year,lt->tm_mon+1);
}
long EventList::Locate(ulong &eventNumber,int &yr, int &mn, int precision)
{
  // gives best starting position (fills in either eventNumber or yr and mn
  time_t tnow = time(NULL);
  struct tm *lt = localtime(&tnow), ltnow;
  long pos, en;
  char eventFileName[80];
  int i;

  memcpy(&ltnow,lt,sizeof(struct tm));
  if (eventNumber != 0)
  {
    if (eventNumber > tnow) return -1;
    if (debug) cout << "looking for event " << eventNumber << endl;
    lt = localtime((time_t *)&eventNumber);
    yr = 1900 + lt->tm_year;
    mn = lt->tm_mon + 1;
  }
  else
  {
    if (yr == 0) yr = 1900 + ltnow.tm_year;
    if (mn == 0) mn = 1;
  }

  if (((ltnow.tm_year + 1900) < yr) ||
      (((ltnow.tm_year + 1900) == yr) && ((ltnow.tm_mon + 1) < mn))) return -1;

  if (yr < 1999) return -1;

  sprintf(eventFileName,"%s%d.%d",DBPATH,yr,mn);
  if (strcmp(eventFileName,elfn)) strncpy(elfn,eventFileName,80);

  ifstream elin(elfn,ios::in|ios::out);
  if (debug) cout << "open file " << eventFileName << endl;
  assert(elin);

  elin.seekg(pos=0,ios::beg);      // always rewind the file
  if (eventNumber == 0)  // request for the 1st event number
  {
    elin.read((BYTE *)&eventNumber,(int)sizeof(long));
    return pos;
  }
  // do simple scan for now, this should be replaced by binary search
  for (i=0; !elin.eof(); i++)
  {
    elin.seekg(pos=i*sizeof(AES),ios::beg);
    elin.read((BYTE *)&en,(int)sizeof(long));
    if (debug > 1) cout << "scanning event " << en << endl;
    if (abs(eventNumber-en) <= precision) return pos;
    if (en > eventNumber) return pos;
  }
  return -1;
}
long EventList::Locate(ulong &eventNumber,int &yr, int &mn)
{
  return Locate(eventNumber,yr,mn,0);
}
long EventList::Find(ulong eventNumber)
{
  int yr = 0,mn = 0;
  return Locate(eventNumber,yr,mn,5); // precision of 5 seconds ...
}
long EventList::Find(ulong eventNumber,AES &aes)
{
  long pos = Find(eventNumber);

  if (pos != -1)
  {
    ifstream elin(elfn,ios::in|ios::out);
    if (debug) cout << "open file " << elfn << endl;
    if (!elin) { error("Couldn't open file !!!"); return -1; }
    elin.seekg(pos,ios::beg);
    elin.read((BYTE *)&aes,(int)sizeof(AES));
  }
  return pos;
}
int EventList::SetComment(int eventNumber,char *comment, char *mode)
{
  AES aes;
  int len;
  long pos = Find(eventNumber);
  char *c;

  char s[80];
  sprintf(s,"Set Comment to event %d",eventNumber);
  AppendArchiveLog(s);

  if (pos == -1) return un_allocated;
  ifstream elin(elfn,ios::in|ios::out);
  if (debug) cout << "open file " << elfn << endl;
  if (!elin) return file_error;
  elin.seekg(pos,ios::beg);
  elin.read((BYTE *)&aes,(int)sizeof(AES));
  if (debug) cout << "found event at pos " << pos << endl;

  c = aes.comment[0]; len = strlen(c);

  len = strchr(mode,'a') == NULL ? 0 : strlen(c);
  if (len + strlen(comment) >= sizeof(aes.comment))
    return string_too_long;
  if (len > 0) c[len] = '\n';
  strcpy(&c[len],comment);

  ostream elout(elin.rdbuf());
  elout.seekp(pos,ios::beg);
  elout.write((BYTE *)&aes,sizeof(AES));

  return 0;
}

ulong EventList::Set(int idSource, int eventCode, char *comment)
{
  AES aes;
  int i;
  time_t interimEvent = time(NULL);
  long pos = Find(interimEvent);

  char s[80];
  sprintf(s,"received SET command from %d",idSource);
  AppendArchiveLog(s);

  ifstream elin(elfn,ios::in|ios::out);
  if (debug) cout << "open file " << elfn << endl;
  if (!elin) return (ulong)error("Couldn't open file !!!");
  if (pos == -1) // new event
  {
    memset(&aes,0,sizeof(AES));
    aes.eventNumber = interimEvent;
    nEvents++;
    elin.seekg(0,ios::end);
    pos = elin.tellg();
    if (debug) cout << "new event appended at pos " << pos << endl;
  }
  else                                // already in list
  {
    elin.seekg(pos,ios::beg);
    elin.read((BYTE *)&aes,(int)sizeof(AES));
    interimEvent = aes.eventNumber;
    if (debug) cout << "found event at pos " << pos << endl;
  }

  aes.eventCode = findMaximumEventCode(aes.eventCode, MAKECODE(idSource,eventCode));
  for (i=0; i<4 && strlen(aes.comment[i]); i++);
  if (i < 4) memcpy(aes.comment[i],comment,64);

  ostream elout(elin.rdbuf());
  elout.seekp(pos,ios::beg);
  elout.write((BYTE *)&aes,sizeof(AES));

  return interimEvent;
}
int EventList::Trigger(long eventNumber)
{
  AESnotify aesNotify;
  AES aes;
  NAME16 property;

  char s[80];
  sprintf(s,"received Trigger command for event %d",eventNumber);
  AppendArchiveLog(s);

  if (Find(eventNumber,aes) == -1) return un_allocated;
  strcpy(property.name,"TRIGGER");
  aesNotify.eventNumber = aes.eventNumber;
  aesNotify.eventCode   = aes.eventCode;
  memcpy(aesNotify.comment,aes.comment,64*4);
  if (debug) cout << "::Trigger event " << eventNumber << "(" << aes.eventNumber << ")" << endl;
printf("trigger event %d (%d)\n",eventNumber,aes.eventNumber);
  return ArchiveServerList.Send((SRV *)NULL,&property,&aesNotify);
}

int EventList::Record(long eventNumber,char *strArchiveServer,char *strDeviceServer)
{
  int idDeviceServer = getDeviceID(strArchiveServer,strDeviceServer);
  if (debug) cout << "Record " << strArchiveServer << " " << strDeviceServer << " in event " << eventNumber << endl;
  if (idDeviceServer < 0) return -idDeviceServer;
  return Record(eventNumber,idDeviceServer);
}
int EventList::Record(long eventNumber, int idDeviceServer)
{
  AES aes;
  long pos;

  char s[80];
  sprintf(s,"received Record for event %d from device id %d",eventNumber,idDeviceServer);
  AppendArchiveLog(s);

  if ((pos=Find(eventNumber,aes)) == -1) return un_allocated;

  if (idDeviceServer < 0 || idDeviceServer > 512) return out_of_range;
  aes.dsMask[idDeviceServer / BITSinBYTE] |= (0x01 << (idDeviceServer % BITSinBYTE));

  // write it back to disk

  ifstream elin(elfn,ios::in|ios::out);
  ostream elout(elin.rdbuf());
  if (!elout) { error("Couldn't open file !!!"); return -1; }
  elout.seekp(pos,ios::beg);
  elout.write((BYTE *)&aes,(int)sizeof(AES));
  if (debug) cout << "::Record event " << eventNumber << "(" << aes.eventNumber << ")" << endl;

  return 0;
}
int EventList::Remove(long eventNumber,char *strArchiveServer,char *strDeviceServer)
{
  int idDeviceServer = getDeviceID(strArchiveServer,strDeviceServer);
  if (debug) cout << "Remove " << strArchiveServer << strDeviceServer << " from event " << eventNumber << endl;
  if (idDeviceServer < 0) return -idDeviceServer;
  return Remove(eventNumber,idDeviceServer);
}
int EventList::Remove(long eventNumber, int idDeviceServer)
{
  AES aes;

  if (Find(eventNumber,aes) == -1) return un_allocated;

  if (idDeviceServer < 0 || idDeviceServer > 512) return out_of_range;
  aes.dsMask[idDeviceServer / BITSinBYTE] &= ~(0x01 << (idDeviceServer % BITSinBYTE));

  // write it back to disk

  if (debug) cout << "::Remove event " << eventNumber << "(" << aes.eventNumber << ")" << endl;

  return 0;
}
EventList::EventList(void)
{
  nEvents  = 0;
  debug    = 0;
}
int EventList::queryNumEvents(ulong starttime, ulong stoptime)
{
  return queryNumEvents(starttime,stoptime,0);
}
int EventList::queryNumEvents(ulong starttime, ulong stoptime, ulong evncode)
{
  ulong t = starttime;
  int n=0,in,yr=0,mn=0;
  long en = 0,pos,p0=Locate(t,yr,mn),p1,len;
  ifstream elin;

  // note: for event code filtering, this won't work
  //       need to scan each event found and compare against the code given

  while (p0 != -1)
  {
    elin.open(elfn,ios::in|ios::out);
    if (debug) cout << "open file " << elfn << endl;
    if (!elin) { error("Couldn't open file !!!"); return -1; }

    elin.seekg(0,ios::end);         // find the last event in the file
    p1 = elin.tellg();
    elin.seekg(p1-sizeof(AES),ios::beg);
    elin.read((BYTE *)&en,sizeof(long));
    if (debug) cout << "read " << en << " vs stoptime " << stoptime << endl;
    if (en < stoptime)
    {
      in = (p1-p0)/sizeof(AES);
      n += in;
      if (debug) cout << "interim count : " << n << " (+" << in << ")" << endl;
      t = 0; if (mn < 12) mn++; else { yr++, mn = 0; }
      elin.close();
      p0 = Locate(t,yr,mn);
    }
    else { en = 0; break; }
  }

  while (p0 != -1 && !elin.eof() && en < stoptime)
  {
    elin.seekg(p0,ios::beg);
    elin.read((BYTE *)&en,sizeof(long));
    if (debug) cout << "read " << en << " vs stoptime " << stoptime << endl;
    p0 += sizeof(AES);
    n++;
  }
  if (debug) cout << "events found : " << n << endl;
  return n;
}
int EventList::queryEvents(ulong starttime, ulong stoptime,AES *aes,int num)
{
  return queryEvents(starttime,stoptime,0,aes,num);
}
int EventList::queryEvents(ulong starttime,ulong stoptime,ulong evncode,AES *aes,int num)
{
  ulong t = starttime;
  int n=0,yr=0,mn=0;
  long en = 0,pos,p0=Locate(t,yr,mn),p1,len;
  int nscanned=0,scansize;
  ifstream elin;

  // note: for event code filtering, this won't work
  //       need to scan each event found and compare against the code given
  //       This is done correctly in the second while loop ...

  if (debug)
  {
    cout << "queryEvents from " << starttime << " to " << stoptime;
    cout << " (above event code " << evncode << ")";
    cout << " stop after " << num << " events" << endl;
  }
  while ( p0 != -1 )
  {
    elin.open(elfn,ios::in|ios::out);
    if (debug) cout << "open file " << elfn << endl;
    if (!elin) { error("Couldn't open file !!!"); return -1; }

    elin.seekg(0,ios::end);         // find the last event in the file
    p1 = elin.tellg();
    elin.seekg(p1-sizeof(AES),ios::beg);
    elin.read((BYTE *)&en,sizeof(long));
    if (debug) cout << "compare " << en << " vs " << stoptime << endl;
    if (en < stoptime)
    {
      elin.seekg(p0,ios::beg);
      n += (len = (p1-p0)/sizeof(AES));
      scansize = nscanned + len < num ? len : num - nscanned;
      if (scansize)
      {
        if (debug) cout << "read in " << scansize << " events" << endl;
        elin.read((BYTE *)&aes[nscanned],sizeof(AES)*scansize);  // read them all in
        nscanned += scansize;
      }
      t = 0; if (mn < 12) mn++; else { yr++, mn = 0; }
      elin.close();
      p0 = Locate(t,yr,mn);
    }
    else {en = 0; break; }
  }

  while (p0 != -1 && !elin.eof() && en < stoptime)
  {
    elin.seekg(p0,ios::beg);
    elin.read((BYTE *)&en,sizeof(long));
    if (nscanned < num)
    {
      if (debug) cout << "read in next event at " << en << endl;
      elin.seekg(p0,ios::beg);
      elin.read((BYTE *)&aes[nscanned],sizeof(AES));  // read it in
//      if (aes[nscanned].eventCode >= evncode)
        nscanned++;
    }
    p0 += sizeof(AES);
    n++;
  }
  return n;
}
int EventList::queryEvents(ulong starttime,ulong stoptime,ulong evncode,ulong *evn,int num)
{
  ulong t = starttime;
  int i,n=0,yr=0,mn=0;
  long en = 0,pos,p0=Locate(t,yr,mn),p1,len;
  int nscanned=0,scansize;
  ifstream elin;
  AES aes;

  // note: for event code filtering, this won't work
  //       need to scan each event found and compare against the code given
  //       This is done correctly in the second while loop ...

  if (debug)
  {
    cout << "queryEvents from " << starttime << " to " << stoptime;
    cout << " (above event code " << evncode << ")";
    cout << " stop after " << num << " events" << endl;
  }
  while ( p0 != -1 )
  {
    elin.open(elfn,ios::in|ios::out);
    if (debug) cout << "open file " << elfn << endl;
    if (!elin) { error("Couldn't open file !!!"); return -1; }

    elin.seekg(0,ios::end);         // find the last event in the file
    p1 = elin.tellg();
    elin.seekg(p1-sizeof(AES),ios::beg);
    elin.read((BYTE *)&en,sizeof(long));
    if (debug) cout << "compare " << en << " vs " << stoptime << endl;
    if (en < stoptime)
    {
      elin.seekg(p0,ios::beg);
      n += (len = (p1-p0)/sizeof(AES));
      scansize = nscanned + len < num ? len : num - nscanned;
      for (i=0; i<scansize; i++)
      {
        elin.read((BYTE *)&aes,sizeof(AES));  // read them all in
        evn[i+nscanned] = aes.eventNumber;
      }
      nscanned += scansize;
      t = 0; if (mn < 12) mn++; else { yr++, mn = 0; }
      elin.close();
      p0 = Locate(t,yr,mn);
    }
    else {en = 0; break; }
  }

  while (p0 != -1 && !elin.eof() && en < stoptime)
  {
    elin.seekg(p0,ios::beg);
    elin.read((BYTE *)&en,sizeof(long));
    if (nscanned < num)
    {
      if (debug) cout << "read in next event at " << en << endl;
      elin.seekg(p0,ios::beg);
      elin.read((BYTE *)&aes,sizeof(AES));  // read it in
      evn[nscanned] = aes.eventNumber;
//      if (aes[nscanned].eventCode >= evncode)
        nscanned++;
    }
    p0 += sizeof(AES);
    n++;
  }
  return n;
}

