/* ============   EVNEQM.C    ============================================


   An archived file contains the following information according the
	 following protocol:

       DeviceServerName - 16 bytes   char string
       Property         - 32 bytes   char string
       DeviceName       - 16 bytes   char string
       size of call     - 2  bytes   short
       data format      - 2  bytes   short
       scale            - 4  bytes   float
       shift            - 4  bytes   float
       data buffer - N bytes according to format and size


  The above is repeated N times ....

  ========================================================================  */
#include "prolog.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "errors.h"
#include "rpctypes.h"
#include "toolkit.h"
#include "swaplib.h"
#include "database.h"
#include "arctools.h"
#include "lpceqm.h"
#include "arceqm.h"
#include "eventlst.h"

extern int numArchiveTypes;
extern int narb;
FILE *createAndOpen(char *fname);
int getContributingArchiveServers(int event,NAME16 *svrs,int *len);
int getContributingDeviceServers(int event,NAME16 *svrs,int *len);
int evneqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
char *gExt = "EVENT";

#define dumpStructInfo(tag) {\
  if ((s=getStructFromTag(tag)) != NULL) {\
    printf("%s: num %d size %d rawlen %d\n",tag,s->num,s->size,s->rawlen);\
  } else printf("%s: not found\n",tag);\
}
#define quit(i) { printf("ABORT: %s\n",erlst[i]); exit(1); }
void registerStructs(void)
{
  /* this must follow the order of the structure explicitly! */

  if (fillstruct("AESNOT",OFFSETIN(AESnotify,version),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",OFFSETIN(AESnotify,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",OFFSETIN(AESnotify,eventCode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",OFFSETIN(AESnotify,reserved),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",OFFSETIN(AESnotify,comment),4*64,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",sizeof(AESnotify),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("AESCON",OFFSETIN(AEScontext,version),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESCON",OFFSETIN(AEScontext,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESCON",OFFSETIN(AEScontext,eventCode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESCON",OFFSETIN(AEScontext,reserved),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESCON",OFFSETIN(AEScontext,deviceName),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("AESCON",sizeof(AEScontext),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("AES",OFFSETIN(AES,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AES",OFFSETIN(AES,eventCode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AES",OFFSETIN(AES,comment),4*64,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("AES",OFFSETIN(AES,dsMask),32,CF_BYTE)) quit(out_of_local_memory);
  if (fillstruct("AES",sizeof(AES),100,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("ADRB",OFFSETIN(ADRB,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,starttime),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,stoptime),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.deviceServerName),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.deviceProperty),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.deviceName),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.dataFormat),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.dataDimension),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",sizeof(ADRB),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("AIB",OFFSETIN(AIB,numProperties),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("AIB",OFFSETIN(AIB,numDevices),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("AIB",OFFSETIN(AIB,eventCode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AIB",OFFSETIN(AIB,comment),64,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("AIB",sizeof(AIB),1,CF_NULL)) quit(out_of_local_memory);

 if (fillstruct("PIB",OFFSETIN(PIB,numPreTriggerValues),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,TimeOffset),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,LowCount),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,HighCount),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,Units),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,LowEGU),1,CF_DOUBLE)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,HighEGU),1,CF_DOUBLE)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,numDevices),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,reserved),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("PIB",sizeof(PIB),1,CF_NULL)) quit(out_of_local_memory);

  return;
}
int writeCommentToFile(char *fn,char *comment)
{
  FILE *fp = NULL;
  char *c,fullName[128],ext[16];

  if (comment == NULL) return -1;
  strncpy(fullName,fn,128);
  sprintf(ext,".%s",gExt);
  if ((c=strstr(fullName,ext)) == NULL) return -1;
  *c = 0;
  strcat(fullName,".CMT");
  if ((fp=fopen(fullName,"w")) == NULL) return -1;
  if (fwrite(comment,strlen(comment),1,fp) != 1) return -1;
  return 0;
}
#define pmHeaderSize 76
int writeEventHeaderToFile(FILE *fp,int id,int eventCode)
{
  BYTE s[512];
  short siz;
  short fmt;
  int dsiz;
  float shift = 0.0, scale = 1.0;
  if (id < 0 || id > numArchiveTypes) return -1;
  siz = 16; fmt = CF_TEXT; dsiz = siz*fmtsizeof(fmt);
  memset(s,0,128);
  memcpy(s,"EventServer",16);
  memcpy(&s[16],"TriggerSource",32);
  memcpy(&s[48],"#0",16);
  memcpy(&s[64],&siz,2);
  memcpy(&s[66],&fmt,2);
  memcpy(&s[68],&scale,4);
  memcpy(&s[72],&shift,4);
  memcpy(&s[76],ArchRecord[id].nam,dsiz);
  if (fwrite(s,pmHeaderSize+dsiz,1,fp) != 1) return -1;
  siz = 1; fmt = CF_LONG; dsiz = siz*fmtsizeof(fmt);
  memset(&s[16],0,32); 
  memcpy(&s[16],"EventCode",32);
  memcpy(&s[64],&siz,2);
  memcpy(&s[66],&fmt,2);
  memcpy(&s[76],&eventCode,dsiz);
  if (fwrite(s,pmHeaderSize+dsiz,1,fp) != 1) return -1;
  return 0;
}
int appendEventToFile(FILE *fp,char *strArchiveServer,char *strDeviceServer)
{
  BYTE s[128];
  short siz;
  short fmt;
  int dsiz;
  float shift = 0.0, scale = 1.0;
  siz = 16; fmt = CF_TEXT; dsiz = siz*fmtsizeof(fmt);
  memset(s,0,128);
  memcpy(s,strArchiveServer,16);
  memcpy(&s[16],"EventRecord",32);
  memcpy(&s[48],"#0",16);
  memcpy(&s[64],&siz,2);
  memcpy(&s[66],&fmt,2);
  memcpy(&s[68],&scale,4);
  memcpy(&s[72],&shift,4);
  strncpy(&s[76],strDeviceServer,siz);
  if (fwrite(s,pmHeaderSize+dsiz,1,fp) != 1) return -1;
  return 0;
}

int eventExists(int event)
{
  FILE *fp = NULL;
  char fullName[128];

  sprintf(fullName,"%x.%s",event,gExt);
  if (makeFullName(fullName,gExt) == 0) 
  {
    if ((fp=fopen(fullName,"rb")) == NULL) return 0;
    fclose(fp);
    return -1;
  }
  return 0;
}
int ArchiveEventSet(int idSource, int eventCode, char *comment)
{
  char fullName[128];
  time_t interimEvent = time(NULL);
  static time_t lastEvent = 0;
  FILE *fp = NULL;
  int cc=0;

  if (interimEvent < lastEvent + 5) interimEvent = lastEvent;
  lastEvent = interimEvent;

  if (!eventExists(interimEvent))
  {
    sprintf(fullName,"%x.%s",interimEvent,gExt);
    if (makeFullName(fullName,gExt) != 0) ccerr(file_error);
    if ((fp=fopen(fullName,"wb")) == NULL && 
        (fp=createAndOpen(fullName)) == NULL) ccerr(file_error);
    // put the first enry into the file ...
    writeEventHeaderToFile(fp,idSource,eventCode);
    writeCommentToFile(fullName,comment);
    fclose(fp);
  }
err:
  if (cc)  feclog("cannot open event file for event %d",interimEvent);
  return interimEvent;
}
void PostSystemInit()
{
  registerStructs();
  RegisterEquipmentModule(NULL,"EVNEQM",0,evneqm,NULL,NULL,100,NULL);
  RegisterEquipmentModule(NULL,"FILER",0,filer,filerinit,filertask,100,NULL);
}
int SetGlobalEvent(ulong eventCode,char *cmt)
{
  // just a dummy routine so that everything links ...
  return 0;
}
int RecordEvent(ulong eventNumber,ulong eventCode,char *devName)
{
  // just a dummy routine so that everything links ...
  return 0;
}
int getSetNumber(char *arcname)
{
  int i;
  for (i=0; i<narb && strcmp(arcname,arb[i].name); i++);
  return i == narb ? -1 :  i;
}
int idxSelectedARB = 0;
int GetARBIndex(char *srvName)
{
  int i,j;
  if (srvName == NULL || strlen(srvName) == 0) return -1;
  // try srvName = archive context first:
  i = getSetNumber(srvName);
  if (i >= 0 && i < narb) return i;

  // try the selected index
  for (j=0; j<arb[idxSelectedARB].nsteps; j++)
  {
    if (!strncmp(srvName,arb[idxSelectedARB].das[j].tagArc,16))
      return idxSelectedARB;
  }
  // didn't match up, so try everything else
  for (i=0; i<narb; i++) for (j=0; j<arb[i].nsteps; j++)
  {
    if (!strncmp(srvName,arb[i].das[j].tagArc,16) && arb[i].das[j].acc&READ)
    {
      return i;
    }
  }
  // wasn't there
  return -1;
}

int ArchiveEventRecord(long eventNumber,char *strArchiveServer,char *strDeviceServer)
{
  int cc = 0;
  char fullName[128];
  FILE *fp = NULL;

  feclog("received Record for event %d from device %s",eventNumber,strDeviceServer);

  if (!eventExists(eventNumber)) 
  {
    feclog("Attemp to record %s to non-existent event by %s",
      strDeviceServer,strArchiveServer);
    return -1;
  }
  sprintf(fullName,"%x.%s",eventNumber,gExt);
  if (makeFullName(fullName,gExt) == 0 && (fp=fopen(fullName,"ab")) != NULL) 
  {
    // append the event to the event file ...
    appendEventToFile(fp,strArchiveServer,strDeviceServer);
    fclose(fp);
    return 0;
  }
  return file_error;
}
int evneqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int i, siz, cc = 0;
  time_t starttime = 0,stoptime = time(NULL);
  char remSrvName[17];
  AESnotify  *aesNotify;
  AEScontext *aesContext;
  ulong evnnum;

  if (!strcmp(devProperty,"SET"))
  {
   if (!(access & CA_WRITE)) return illegal_read_write;
    if ( din->dFormat != CF_STRUCT ||
         strncmp(din->dTag,"AESNOT",8) ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    aesNotify = (AESnotify *)din->data.vptr;
    strncpy(remSrvName,GetCaller(),16);
    feclog("EVNEQM: received SET request from %s",remSrvName);
    evnnum = ArchiveEventSet(0,aesNotify->eventCode,aesNotify->comment[0]);
    if (dout->dArrayLength > 0) putValuesFromLong(dout,(UINT32 *)&evnnum,1);
    return 0;
  }
  if (!strcmp(devProperty,"RECORD"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if ( din->dFormat != CF_STRUCT ||
         strcmp(din->dTag,"AESCON") ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    aesContext = (AEScontext *)din->data.vptr;
    strncpy(remSrvName,GetCaller(),16);
    feclog("EVNEQM: received RECORD request from %s",remSrvName);
    cc = ArchiveEventRecord(aesContext->eventNumber,remSrvName,aesContext->deviceName);
    return cc;
  }
  if (!strcmp(devProperty,"SETCOMMENT")) 
  {
    return filer(devName,"ANNOTATE",dout,din,access);
  }
  if (!strcmp(devProperty,"REMOVE")) 
  {
  }
  if (!strcmp(devProperty,"ASERVERS"))
  {
    switch (dout->dFormat)
    {
      case CF_NAME16: siz = 16; break;
      case CF_NAME32: siz = 32; break;
      case CF_NAME48: siz = 48; break;
      case CF_NAME64: siz = 64; break;
      default: return illegal_format;
    }
    if (din->dArrayLength == 0)
    {
      for (i=0; i<numArchiveTypes && i<(int)dout->dArrayLength; i++)
      {
        strncpy(dout->data.cptr+i*siz,ArchRecord[i].nam,16);
      }
      dout->dArrayLength = i;
      return 0;
    }
    if ((cc=getValuesAsLong(din,&evnnum,1)) != 0) return cc;
    cc = getContributingArchiveServers(evnnum,(NAME16 *)dout->data.vptr,&dout->dArrayLength);
    return cc;
  }
  if (!strcmp(devProperty,"NASERVERS"))
  {
    return putValuesFromLong(dout,(UINT32 *)&numArchiveTypes,1);
  }
  if (!strcmp(devProperty,"NEVENTS"))
  {
    return filer(devName,"QUERY.NFILES",dout,din,access);
  }
  if (!strcmp(devProperty,"EVENTS"))
  {
    return filer(devName,"QUERY.FILES",dout,din,access);
  }
  if (!strcmp(devProperty,"NDSERVERS"))
  {
  }
  if (!strcmp(devProperty,"DSERVERS"))
  {
    switch (dout->dFormat)
    {
      case CF_NAME16: siz = 16; break;
      case CF_NAME32: siz = 32; break;
      case CF_NAME48: siz = 48; break;
      case CF_NAME64: siz = 64; break;
      default: return illegal_format;
    }
    if (din->dArrayLength == 0) return illegal_read_write;
    if ((cc=getValuesAsLong(din,&evnnum,1)) != 0) return cc;
    cc = getContributingDeviceServers(evnnum,(NAME16 *)dout->data.vptr,&dout->dArrayLength);
    return cc;
  }
  return illegal_property;
}
int getEventHandle(int event)
{
  int cc=0;
  UINT32 levent = event;
  short sid = -1;
  DTYPE dout,din;

  dout.dArrayLength = 1;
  dout.dFormat = CF_SHORT;
  dout.data.vptr = &sid;
  dout.dTag[0] = 0;
  din.dArrayLength = 1;
  din.dFormat = CF_LONG;
  din.data.lptr = &levent;
  din.dTag[0] = 0;
  if ((cc=filer("EVENT","OPEN",&dout,&din,CA_READ)) != 0) return -cc;
  return sid;
}
int closeEventHandle(short id)
{
  int cc=0;
  short sid = id;
  DTYPE dout,din;

  dout.dArrayLength = 0;
  dout.dFormat = CF_NULL;
  din.dArrayLength = 1;
  din.dFormat = CF_SHORT;
  din.data.sptr = &sid;
  din.dTag[0] = 0;
  return filer("EVENT","CLOSE",&dout,&din,CA_READ);
}
int getContributingArchiveServers(int event,NAME16 *svrs,int *len)
{
  int cc=0,i,n;
  short sid = getEventHandle(event);
  NAME16 names[100];
  DTYPE dout,din;

  if (sid < 0) return -sid;
  dout.dArrayLength = 100;
  dout.dFormat = CF_NAME16;
  dout.data.vptr = names;
  dout.dTag[0] = 0;
  din.dArrayLength = 1;
  din.dFormat = CF_SHORT;
  din.data.sptr = &sid;
  din.dTag[0] = 0;
  cc = filer("EVENT","QUERY.EQP",&dout,&din,CA_READ);
  for (i=0,n=0; cc == 0 && i<(int)dout.dArrayLength && n < *len; i++)
  {
    if (!strncmp(names[i].name,"EventServer",16)) continue;;
    strncpy(svrs[n++].name,names[i].name,16);
  }
  *len = n;
  closeEventHandle(sid);
  return cc;
}
int getContributingDeviceServers(int event,NAME16 *svrs,int *len)
{
  int cc=0,i,n,siz;
  short sid = getEventHandle(event);
  NAME16 names[100];
  char name[32];
  DTYPE dout,din;
  APCBLK apc;

  if (sid < 0) return -sid;
  dout.dArrayLength = 100;
  dout.dFormat = CF_NAME16;
  dout.data.vptr = names;
  dout.dTag[0] = 0;
  din.dArrayLength = 1;
  din.dFormat = CF_SHORT;
  din.data.sptr = &sid;
  din.dTag[0] = 0;
  if ((cc=filer("EVENT","QUERY.EQP",&dout,&din,CA_READ)) != 0) ccerr(cc);
  siz = dout.dArrayLength;
  dout.dFormat = CF_TEXT;
  dout.dArrayLength = 16;
  dout.data.cptr = name;
  din.dArrayLength = 1;
  din.dFormat = CF_STRUCT;
  din.data.vptr = &apc;
  strncpy(din.dTag,"APCBLK",8);
  apc.sessionID = sid;
  for (i=0,n=0; cc == 0 && i<siz && n < *len; i++)
  { 
    if (!strncmp(names[i].name,"EventServer",16)) continue;;
    strncpy(apc.eqpName,names[i].name,16);
    strncpy(apc.devProperty,"EventRecord",32);
    strncpy(apc.devName,"#0",16);
    if ((cc=filer("EVENT","READDATA",&dout,&din,CA_READ)) != 0) ccerr(cc);
    strncpy(svrs[n++].name,name,16);
  }
  *len = n;
err:
  closeEventHandle(sid);
  return cc;
}
