/* ============   LPCEQM.C    ============================================


   An archived file contains the following information according the
	 following protocol:

       DeviceServerName - 16 bytes   char string
       Property         - 32 bytes   char string
       DeviceName       - 16 bytes   char string
       size of call     - 2  bytes   short
       data format      - 2  bytes   short
       scale            - 4  bytes   float
       shift            - 4  bytes   float
       data buffer - N bytes according to format and size


  The above is repeated N times ....

  ========================================================================  */
#include "prolog.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#ifdef UNIX
#include <dirent.h>
#endif
#include "servdef.h"
#include "errors.h"
#include "rpctypes.h"
#include "swaplib.h"
#include "database.h"
#include "toolkit.h"
#include "arctools.h"
#include "lpceqm.h"
#include "arceqm.h"

void dumpSessions(void);
void freeSessions(void);
int cmdDumpSessions(int arg0,int arg1,int arg2,int arg3);
int cmdFreeSessions(int arg0,int arg1,int arg2,int arg3);
int isArchiveOngoing(char *fn);
int LoadEQP(int year,char *fname,char *ext);
int makeFullName(char *fname,char *ext);
int explode(char *fname);
#define quit(i) { dbglog("Register struct: out of memory"); exit(i); }
void registerAPCBLK(void)
{
  /* this must follow the order of the structure explicitly! */
  if (fillstruct("APCBLK",OFFSETIN(APCBLK,sessionID),1,CF_LONG)) quit(1);
  if (fillstruct("APCBLK",OFFSETIN(APCBLK,event),1,CF_LONG)) quit(1);
  if (fillstruct("APCBLK",OFFSETIN(APCBLK,ext),4,CF_TEXT)) quit(1);
  if (fillstruct("APCBLK",OFFSETIN(APCBLK,eqpName),16,CF_TEXT)) quit(1);
  if (fillstruct("APCBLK",OFFSETIN(APCBLK,devName),16,CF_TEXT)) quit(1);
  if (fillstruct("APCBLK",OFFSETIN(APCBLK,devProperty),32,CF_TEXT)) quit(1);
  if (fillstruct("APCBLK",OFFSETIN(APCBLK,dSize),1,CF_SHORT)) quit(1);
  if (fillstruct("APCBLK",OFFSETIN(APCBLK,dFormat),1,CF_SHORT)) quit(1);
  /* terminate the structure definition like this! */
  if (fillstruct("APCBLK",sizeof(APCBLK),1,CF_NULL)) quit(1);
}

char ArchiveDirectoryPath[128] = "../CACHE";
ARCHRECORD ArchRecord[MAXARCHIVES];
int numArchiveTypes;
int fileNo = 0;
void *LPCdata;
EQP EQPlist[MAXEQP];
SESSION session[MAXSESSION];
int numEQP = 0;
int archDebug = 0;

void set_yday(struct tm *t)
{
  int daysInMonth[12]={31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
  int i;
  t->tm_yday = 0;
  for (i = 0; i < t->tm_mon; i++ ) t->tm_yday += daysInMonth[i];
  if (( t->tm_mon > 1 ) && ( (t->tm_year % 4) == 0 )) t->tm_yday += 1;
  t->tm_yday += t->tm_mday - 1;
}
long tm2time(struct tm * t)
{
  long y, d, s;
  y = t->tm_year - 70;          /* Anzahl der vollen Jahre ab 1.1.1970 */
  d = y * 365 + t->tm_yday;     /* Anzahl der vollen Tage */
  d += (t->tm_year-1) / 4 - 17; /* Schaltjahre ab 1970 */
  s = ((d * 24 + t->tm_hour) * 60 + t->tm_min) * 60 + t->tm_sec;
  return s;
}
long unixtime(int yr,int mm,int da,int hr,int mn,int sc,int dl)
{
  long s;
  struct tm t;

  t.tm_sec = sc;
  t.tm_min = mn;
  t.tm_hour = hr;
  t.tm_mday = da;
  t.tm_mon = mm;
  t.tm_year = yr;
  t.tm_isdst = 0;
  set_yday(&t);

  s  = tm2time(&t);
  if (dl) s -= 3600L;
  s -= 3600L;        /* convert to continental Europe */
  return s;
}
void fileSort(NAME8 *fbuf,int start,int num) /* files are mostly in order, so this */
{                                            /* should be quick */
  int i;

  if (num-start==1) return;
  for (i=start+1; i<num; i++)
    if ( strncmp(fbuf[i].name,fbuf[start].name,8) < 0 ) fswap(fbuf,start,i);
  fileSort(fbuf,++start,num);
  return;
}
void fswap(NAME8 *fbuf,int i,int j)
{
  static char ftemp[13];
  strncpy(ftemp,fbuf[i].name,8); 
  strncpy(fbuf[i].name,fbuf[j].name,8); 
  strncpy(fbuf[j].name,ftemp,8); 
  return;
}
int getArchiveContracts(void)
{
  FILE *fp = NULL;
  int cc=0,done=0,i=0;
  int ext_col,nam_col,srv_col;
  char str[256],hdr[256];

  if ((fp=fopen(archiveListFile,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      done = 1;
      str[255] = 0; strcpy(hdr,str);
      if ((ext_col=findcol(hdr,"EXTENSION")) < 0) ccerr(no_such_column);
      if ((nam_col=findcol(hdr,"NAME")) < 0) 
        nam_col = findcol(hdr,"TRIGGER");
      srv_col = findcol(hdr,"SERVER");
      continue;
    }
    strncpy(ArchRecord[i].ext,colptr(ext_col,str),15);
    if (nam_col >= 0)
    {
      strncpy(ArchRecord[i].nam,colptr(nam_col,str),15);
    }
    else
    {
      if (srv_col < 0) ccerr(no_such_column);
      strncpy(ArchRecord[i].nam,colptr(srv_col,str),15);
      strlwr(ArchRecord[i].nam);
    }
    if (++i >= MAXARCHIVES) break;
  }
  err:
  if (fp) fclose(fp);
  feclog("Archive Contracts file %s : %s",archiveListFile, cc ? erlst[cc] : "found");
  if (i > 0) strncpy(ArchiveDirectoryPath,stdArchiveDirectory,79);
  return i;
}
int getFiles(NAME8 *filebuf,char *fileext,time_t starttime,time_t stoptime,int bufsiz)
{
  int i, n, len, fctr=0;
  char dirname[64],fname[32],fullname[128];
  time_t ftim;
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WINNT)
  struct _finddata_t fp;
  long hfp;
  char fn[128];
# endif
  struct tm *t;
  int logdone = 0;
  int startyr, startmo, stopyr, stopmo, yr, mo;
  time_t now = time(NULL);
  char str[256];

  t = localtime(&starttime);
  startyr = t->tm_year + 1900; startmo = t->tm_mon;
  t = localtime(&stoptime);
  stopyr = t->tm_year + 1900; stopmo = t->tm_mon;
  n = 12 * (stopyr-startyr) + stopmo - startmo + 1;
  
  for (fctr=0,i=0; i<n; i++)
  {
    yr = startyr + (startmo+i)/12;
    mo = (startmo+i)%12 + 1;
#   if defined(UNIX)
    sprintf(dirname,"%s/%d/%02d/%s",ArchiveDirectoryPath,yr,mo,fileext);
    if ((dp=opendir(dirname)) == NULL) continue;
    while ((dirp=readdir(dp)) != NULL)
    {
      if (strstr(dirp->d_name,fileext) == NULL) continue;
      len = strchr(dirp->d_name,'.') - dirp->d_name;
      strncpy(fname,dirp->d_name,8); fname[8] = 0;
      ftim = strtol(fname,0,16);
      if (ftim < starttime || ftim > stoptime) continue;
      if (starttime == 0 && filebuf == NULL && bufsiz == -1) /* signal for removal */
      {
        sprintf(fullname,"%s/%s",dirname,dirp->d_name);
        remove(fullname);
        if (!logdone)
        {
          logdone = -1;
          sprintf(str,"removing all %s files older than %s",fileext,ctime(&stoptime));
          AppendArchiveLog(str);
          feclog(str);
        }
      }
      if (filebuf != NULL && fctr < bufsiz) strncpy(filebuf[fctr].name,fname,8);
      fctr++;
    }
    closedir(dp);
#   elif defined(WINNT)
    sprintf(dirname,"%s\\%d\\%02d\\%s",ArchiveDirectoryPath,yr,mo,fileext);
    sprintf(fn,"%s\\*.*",dirname);
    if ((hfp=_findfirst(fn, &fp)) == -1) continue;
    do
    {
      if (strstr(fp.name,fileext) == NULL) continue;
      len = strchr(fp.name,'.') - fp.name;
      strncpy(fname,fp.name,8); fname[8] = 0;
      ftim = strtol(fname,0,16);
      if (ftim < starttime || ftim > stoptime) continue;
      if (starttime == 0 && filebuf == NULL && bufsiz == -1) /* signal for removal */
      {
        sprintf(fullname,"%s\\%s",dirname,fp.name);
        remove(fullname);
        if (!logdone)
        {
          logdone = -1;
          sprintf(str,"removing all %s files older than %s",fileext,ctime(&stoptime));
          AppendArchiveLog(str);
          feclog(str);
        }
      }
      if (filebuf != NULL && fctr < bufsiz) strncpy(filebuf[fctr].name,fname,8);
      fctr++;
    } while (_findnext(hfp,&fp) != -1);
#   endif
  }
  if (filebuf != NULL && fctr > 0) fileSort(filebuf,0,MIN(fctr,bufsiz));
  return fctr;
}
void filerinit(void)
{
  int i;
  if ((numArchiveTypes=getArchiveContracts()) > 0)
  {
    for (i=0; i<numArchiveTypes; i++) dbglog("Archive contract %s",ArchRecord[i].ext);
  }

  memset(EQPlist,0,sizeof(EQP)*MAXEQP);
  for (i=0; i<MAXEQP; i++) EQPlist[i].id = -1;

  registerAPCBLK();

  RegisterUserCommand("ArchiveDebug",NULL,&archDebug,NULL,CA_WRITE|CA_READ);
  RegisterUserCommand("sessions",cmdDumpSessions,NULL,NULL,CA_READ);
  RegisterUserCommand("freeSessions",cmdFreeSessions,NULL,NULL,CA_READ);
  return;
}
void filertask()
{
  static int count = 0;
  static time_t interim = 0;
  static int sessionTaskDone = 0, zipTaskDone = 0;
  int sessionTaskTrigger,zipTaskTrigger;
  int i;
  time_t t = time(NULL);

  if (t > interim)       /* count the seconds */
  {
    count = (count + t-interim)%3600;  /* hour grid */
    interim = t;
  }

  /* snuff out old sessions */
  if ((count%60) == 0) sessionTaskTrigger = 1;
  else sessionTaskDone = sessionTaskTrigger = 0;
  /* look for zip jobs */
  if ((count%10) == 0) zipTaskTrigger = 1;
  else zipTaskDone = zipTaskTrigger = 0;

  if (sessionTaskTrigger && !sessionTaskDone)
  {
    if (archDebug) dbglog("LPC: checking session activity");
    for (i=0; i<MAXSESSION; i++)
    {
      if (session[i].active)
      {
        if (archDebug)
          dbglog("LPC: session %d active since %s>",i,ctime(&session[i].active));
        if (t > session[i].active + 1800)
        {
          session[i].count = 0;
          freeSession(i);
        }
      }
    }
    sessionTaskDone = 1;
  }

  if (zipTaskTrigger && !zipTaskDone)
  {
/*    dozipTask(); */
    zipTaskDone = 1;
  }

  return;
}
void freeEQP(EQP *eqp)
{
  LPCstruct *lpc, *Next;

  if (archDebug) dbglog("LPC: freeing %s (%d)",eqp->name,eqp->id);
  for (lpc = eqp->lpc; lpc != NULL; lpc = Next)
  {
    Next = lpc->Next;
    if (archDebug > 1) dbglog("LPC: freeing link %s %s <%d> (Addr %lx) (Next %lx)",
                          lpc->prp,lpc->dev,lpc->eqn,lpc,Next);
    free(lpc);
  }
  eqp->lpc = NULL;
  eqp->id  = -1;
  eqp->name[0] = 0;
  return;
}
int cmdDumpSessions(int arg0,int arg1,int arg2,int arg3)
{
  arg0 = arg1 = arg2 = arg3 = 0;
  dumpSessions();
  return 0;
}
void dumpSessions(void)
{
  int i;
  dbglog("SESSION\tFile\tLast Access\n");
  for (i=0; i<MAXSESSION; i++) if (session[i].active)
  {
    dbglog("%d (%d)\t%s\t%s\n",
          i+1,session[i].count,session[i].fullName,ctime(&session[i].active));
  }
  return;
}
int cmdFreeSessions(int arg0,int arg1,int arg2,int arg3)
{
  arg0 = arg1 = arg2 = arg3 = 0;
  freeSessions();
  return 0;
}
void freeSessions(void)
{
  int i;
  for (i=0; i<MAXSESSION; i++)
  {
    session[i].count = 0;
    freeSession(i);
  }
  return; 
}
int GetLocalSessionID(ulong eventNumber,char *ext)
{
  struct tm *t;
  int i, t_year;
  char fullName[128];
    
  t = localtime((time_t *)&eventNumber); t_year = t->tm_year;
  if (strlen(ext) == 0) return -string_expected;
  sprintf(fullName,"%x.%s",eventNumber,ext);
  if (makeFullName(fullName,ext)) return -no_such_file;
  if (explode(fullName)) return -file_error;
  if ((i=LoadEQP(t_year,fullName,ext)) < 0) return i;
  session[i].active = time(NULL);
  return i;
}
int getSessionID(char *fname)
{
  int i;
  time_t t = time(NULL);

  for (i=0; i<MAXSESSION; i++)
    if (session[i].active && !strcmp(session[i].fullName,fname))
    {
      if (t > session[i].active + 3)  /*kluge to trap UDP retries */
        session[i].count++;
      return i;
    }
  return -1;
}
int makeSessionID(char *fname)
{
  int i,idxLeastRecentlyUsed = -1;
  time_t leastRecentlyUsed,t = time(NULL);

  for (i=0; i<MAXSESSION; i++) if (!session[i].active)
  {
    strncpy(session[i].fullName,fname,80);
    session[i].active = session[i].est = time(NULL);
    session[i].count = 1;
    return i;
  }

  leastRecentlyUsed = t;
  for (i=0; i<MAXSESSION; i++)
  {
    if (session[i].active < leastRecentlyUsed) 
    {
      leastRecentlyUsed = session[i].active;
      idxLeastRecentlyUsed = i;
    }
  }
  i = idxLeastRecentlyUsed;
  if (i >= 0 && t > session[i].active + 60)
  {
    session[i].count = 0;
    freeSession(i);
    strncpy(session[i].fullName,fname,80);
    session[i].active = session[i].est = time(NULL);
    session[i].count = 1;
    return i;
  }
  dbglog("LPC: unable to create session ID");
  dumpSessions();
  return -1;
}
void freeSession(int id)
{
  int i;
  if (archDebug) dbglog("LPC: free session %d (%d)",id,session[id].count);
  if (--session[id].count > 0) return;
  if (session[id].hFile) close(session[id].hFile);
  memset(&session[id],0,sizeof(SESSION));
  for (i=0; i<numEQP; i++) if (EQPlist[i].id == id) freeEQP(&EQPlist[i]);
  return;
}
#ifdef UNIX
void itoa(int ival,char *aval,int radix)
{
  switch (radix)
  {
    default:
    case 10: sprintf(aval,"%d",ival); break;
    case 16: sprintf(aval,"%x",ival); break;
  }
}
#endif
int makeFullName(char *fname,char *ext)
{
  char dst[128],name[32];
  int i,year,month;
  time_t now = time(NULL);
  struct tm *t;
  long ts;

  for (i=0; i<numArchiveTypes; i++) if (!stricmp(ArchRecord[i].ext,ext)) break;
  if (i == numArchiveTypes) 
  {
    if (archDebug) 
      dbglog("LPC: can't make name for ext %s amoung %d entries",
        ext,numArchiveTypes);
    return -1;
  }
  strncpy(dst,stdArchiveDirectory,80);
  strncpy(name,fname,8); name[8] = 0;
  ts = strtol(name,0,16);
  t = localtime(&ts);
  year = t->tm_year + 1900;
  month = t->tm_mon + 1;
  sprintf(fname,"%s/%d/%02d/%s/%s.%s",dst,year,month,ext,name,ext);
  return 0;
}
int explode(char *fname)
{
  return 0;
}

int RemoveEvents(UINT32 starttime, UINT32 stoptime, char *ext)
{
  char fullName[128];
  int i,n,xlen = 3;
  NAME8 evnnames[1024];

  n = getFiles(evnnames,ext,starttime,stoptime,1024);
  for (i=0; i<n; i++)
  {
    sprintf(fullName,"%s.%s",evnnames[i].name,ext);
    if (makeFullName(fullName,ext)) return -no_such_file;
    if (archDebug) dbglog("LPC: remove file %s",fullName);
    remove(fullName);
    xlen = strlen(ext);
    sprintf(&fullName[strlen(fullName)-xlen],"CMT");
    if (archDebug) dbglog("LPC: remove file %s",fullName);
    remove(fullName);
  }
  return i;
}

int RemoveEvent(UINT32 evnt, char *ext)
{
  char fullName[128];
  int i=0,xlen = 3;

  sprintf(fullName,"%x.%s",evnt,ext);
  if (makeFullName(fullName,ext)) return no_such_file;
  if (archDebug) dbglog("LPC: remove file %s\n",fullName);
  if (remove(fullName)) i++;
  xlen = strlen(ext);
  sprintf(&fullName[strlen(fullName)-xlen],"CMT");
  if (archDebug) dbglog("LPC: remove file %s\n",fullName);
  if (remove(fullName)) i++;
  return i < 2 ? 0 : file_error;
}

int LoadEQP(int year,char *fname,char *ext)
/* returns Session ID on success */
/* returns negative completion code on error */
{
  int n=0, cc = 0, i, ID;
  static char s[200], EQPname[16], workname[16];
  long placeholder = 0;
  LPCstruct *lpc;
  int TagLength;

  if (archDebug) dbglog("LPC: find session for %s (%d)",fname,year);
  if (isArchiveOngoing(fname)) return -operation_busy;
  if ((ID = getSessionID(fname)) >= 0)
  {
    if (archDebug) dbglog("LPC: %s attaching to session %d",fname,ID);
    return ID; /* session already open */
  }
  if ((ID = makeSessionID(fname)) < 0) ccerr(resources_exhausted);
  strncpy(session[ID].ext,ext,15);
  
  if ((session[ID].hFile=open(fname,O_RDONLY)) == -1 )
    ccerr(no_such_file);

  TagLength = year < 91 ? 8 : 16; /* hopefully we'll convert these */
  for (i=0,workname[0]=0; read(session[ID].hFile,EQPname,TagLength) > 0; i++)
  {
    for (n=0; n<numEQP; n++)
    {
      if (archDebug > 1) dbglog("LPC: compare %s vs %s",EQPlist[n].name,EQPname);
      if (EQPlist[n].id == ID && !strcmp(EQPlist[n].name,EQPname)) break;
    }
    if (n == numEQP) for (n=0; n<numEQP && EQPlist[n].lpc; n++);  /* find free slot */
    if (n == numEQP) 
    {
      if (numEQP == MAXEQP)
      {
        dbglog("LPC: unable to find EQP list slot (%d slots full)",numEQP);
        ccerr(resources_exhausted)
      }
      else numEQP++;
    }
    if (archDebug > 1) dbglog("LPC: %d stored Eqp calls",numEQP);
    if (archDebug > 1) dbglog("LPC: compare %s vs %s",EQPname,workname);
    if (strcmp(EQPname,workname))
    {
      strncpy(EQPlist[n].name,EQPname,16);
      EQPlist[n].id = ID;
      if (workname[0] == 0) session[ID].numEQP = 1; else session[ID].numEQP++;
      if (archDebug) dbglog("LPC: added eqp %s (%d) to session %d",EQPname,n,ID);
    }
    strncpy(workname,EQPname,16);

    /* Archive storage information is here: */
    if ((lpc=(LPCstruct *)calloc(1,sizeof(LPCstruct))) == NULL) ccerr(out_of_local_memory);
    lpc->Next = EQPlist[n].lpc;                /* start of list */
    EQPlist[n].lpc = lpc;

    if (year < 91)  /* hopefully we'll convert these */
    {
      read(session[ID].hFile,s,16);
      strncpy(lpc->prp,s,10); lpc->prp[10] = '\0';
      memcpy(&lpc->eqn,&s[10],2);
      memcpy(&lpc->siz,&s[12],2);
      memcpy(&lpc->fmt,&s[14],2);
      lpc->scale = 1.0; lpc->shift = 0.0;
      placeholder += 24;  /* 16 + 8 */
    }
    else
    {
      read(session[ID].hFile,s,60);
      strncpy(lpc->prp,s,31); lpc->prp[31] = '\0';
      strncpy(lpc->dev,&s[32],15); lpc->dev[15] = '\0';
      memcpy(&lpc->siz,&s[48],2);
      memcpy(&lpc->fmt,&s[50],2);
      memcpy(&lpc->scale,&s[52],4);
      memcpy(&lpc->shift,&s[56],4);
      lpc->eqn = lpc->dev[0] == '#' ? atoi(&lpc->dev[1]) : 0;
      placeholder += 76;  /* 60 + 16; */
    }
    if (lpc->fmt == CF_STRUCT && lpc->siz == 1) lpc->siz = sizeof(DHS);
    lpc->dataPos = placeholder;           /* note where the data are */
    placeholder += lpc->siz*getFormatSize(lpc->fmt);
 
    lseek(session[ID].hFile,placeholder,SEEK_SET); /* skip over the data */
  }
err:if (cc)
  {
    if (archDebug) dbglog("LPC: load %s failed - %s",fname,erlst[cc]);
    freeSession(ID);
    return -cc;
  }
  if (archDebug) dbglog("LPC: searched %d records",i);
  return ID;
}
int readCMT(char *s,char *fname,int size)
{
  FILE *fp = NULL;
  char cmtname[80],tmp[256];
  int len,bytesread=0,cc = 0;

  strcpy(s,"No comment\n");
  if ((len=strlen(fname)) > 79) ccerr(file_error);
  strncpy(cmtname,fname,79);
  strcpy(&cmtname[len-3],"CMT");
  if ((fp=fopen(cmtname,"rb")) == NULL)	ccerr(no_such_file);
  if (fgets(tmp,size,fp) == 0) ccerr(0);
  if ((len=strlen(tmp)) > 1) ccerr(file_error);
  strncpy(s,&tmp[len+1],size);
  for (bytesread=strlen(s); fgets(tmp,size-bytesread,fp); bytesread=strlen(s))
  {
    strcat(s,tmp);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc == no_such_file ? 0 : cc;
}
int writeCMT(char *s,char *fname,int size)
{
  FILE *fp = NULL;
  char cmtname[80],tmp[260];
  int len,cc = 0;

  if ((len=strlen(fname)) > 79) ccerr(file_error);
  strcpy(cmtname,fname);
  strcpy(&cmtname[len-3],"CMT");
  if ((fp=fopen(cmtname,"w")) == NULL)	ccerr(no_such_file);
  if ((len=strlen(s)) > 255)
  {
    tmp[0] = len = 255;
    s[255] = 0;
  }
  else *((unsigned char *)&tmp[0]) = len;
  tmp[1] = 0;
  strncpy(&tmp[2],s,len);
  if (fwrite(tmp,len+2,1,fp) != 1) ccerr(file_error);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
                          
int reformatArchiveData(void *dst,short dstFmt,void *src,short srcFmt,int size,float scale,float shift)
{
   int i;

   if (scale == 0) scale = 1;
   if (srcFmt == dstFmt && scale == 1 && shift == 0)
   {
     memcpy(dst,src,size*fmtsizeof(dstFmt));
     return 0;
   }
   if (srcFmt > CF_FLOAT && srcFmt < CF_FLTINT) return illegal_format;
   if (srcFmt > CF_DBLINT) return illegal_format;
   switch (dstFmt)
   {
     case CF_SHORT:
         for (i=0; i<size; i++) ((short *)dst)[i] = (short)(scale * SRCVAL(src,i,srcFmt) + shift);
         return 0;
     case CF_LONG:
         for (i=0; i<size; i++) ((UINT32 *)dst)[i] = (UINT32)(scale * SRCVAL(src,i,srcFmt) + shift);
         return 0;
     case CF_FLOAT:
	 for (i=0; i<size; i++) ((float *)dst)[i] = (float)(scale * SRCVAL(src,i,srcFmt) + shift);
         return 0;
     case CF_BYTE:
         for (i=0; i<size; i++) ((BYTE *)dst)[i] = (BYTE)(scale * SRCVAL(src,i,srcFmt) + shift);
         return 0;
     case CF_DOUBLE:
         for (i=0; i<size; i++) ((double *)dst)[i] = (double)(scale * SRCVAL(src,i,srcFmt) + shift);
         return 0;
     default:
         return illegal_format;
   }
}

int reformatArchiveDataEx(void *dst,short dstFmt,void *src,short srcFmt,int index, int size,float scale,float shift)
{
   int i;

   if (scale == 0) scale = 1;
   if (srcFmt == dstFmt && scale == 1 && shift == 0 && index == 0)
   {
     memcpy(dst,src,size*fmtsizeof(dstFmt));
     return 0;
   }
   if (srcFmt > CF_FLOAT && srcFmt < CF_FLTINT) return illegal_format;
   if (srcFmt > CF_DBLINT) return illegal_format;
   switch (dstFmt)
   {
     case CF_SHORT:
       for (i=0; i<size; i++)
         ((short *)dst)[i] = (short)(scale * SRCVAL(src,i+index,srcFmt) + shift);
         return 0;
     case CF_LONG:
       for (i=0; i<size; i++)
         ((UINT32 *)dst)[i] = (UINT32)(scale * SRCVAL(src,i+index,srcFmt) + shift);
         return 0;
     case CF_FLOAT:
       for (i=0; i<size; i++)
         ((float *)dst)[i] = (float)(scale * SRCVAL(src,i+index,srcFmt) + shift);
         return 0;
     case CF_BYTE:
       for (i=0; i<size; i++)
         ((BYTE *)dst)[i] = (BYTE)(scale * SRCVAL(src,i+index,srcFmt) + shift);
         return 0;
     case CF_DOUBLE:
       for (i=0; i<size; i++)
         ((double *)dst)[i] = (double)(scale * SRCVAL(src,i+index,srcFmt) + shift);
         return 0;
     default:
         return illegal_format;
   }
}

BYTE archiveBuffer[ARCH_BUF_SIZE];
int (*filterArchiveData)(char *prp,char *nam,BYTE *buf,int fmt,UINT32 len,int EqNo) = NULL;
int filer(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int i = 0, cc = 0, j, n, pass, dsize, propFound = 0;
  int napcdevs,apcstart,apcstop,apcidx,k,datapos,datalen;
  int t_year;
  char name[32],ext[16],fullName[128];
  LPCstruct *lpc;
  time_t starttime = 0,stoptime = time(NULL);
  short id;
  APCBLK *apc;
  int EqNo = devName[0] == '#' ? atoi(&devName[1]) : 0;
  struct tm *t;
  time_t ts;
  char *c,tmpdevnam[17],tmpprpnam[33],apcdevname[17];
  float shift = 0.0, scale = 1.0;

  if (!strcmp(devProperty,"QUERY.FILES") || /* files of this extension  */
      !strcmp(devProperty,"QUERY.NFILES"))  /* how many files of this extension */
  {
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_LONG: 
           switch (din->dArrayLength)
           {
             case 2: stoptime  = din->data.lptr[1];
             case 1: starttime = din->data.lptr[0];
                     break;
             default: return illegal_read_write;
           }
           break;
        case CF_SHORT:
           switch (din->dArrayLength)
           {
             case 2: stoptime  = unixtime((int)(din->data.sptr[1]%1900),11,31,23,59,59,0);
             case 1: starttime = unixtime((int)(din->data.sptr[0]%1900),0,0,0,0,0,0);
                     break;
             default: return illegal_read_write;
           }
           break;
        default: return illegal_format;
      }
    }
    strncpy(ext,devName,15); ext[15] = 0;
    if (strlen(ext) == 0) return string_expected;
    if (!strcmp(devProperty,"QUERY.FILES"))
    {
      switch (dout->dFormat)
      {
        case CF_NAME8:
          n = getFiles((NAME8 *)dout->data.vptr,ext,starttime,stoptime,(int)dout->dArrayLength);
          dout->dArrayLength = n;
          break;
        case CF_LONG:
          if (dout->dArrayLength > tmpWorkAreaSize/4) dout->dArrayLength = tmpWorkAreaSize/4;
          n = getFiles((NAME8 *)tmpWorkArea,ext,starttime,stoptime,(int)dout->dArrayLength);
          tmpdevnam[8] = 0;
          for (i=0; i<n && i<(int)dout->dArrayLength; i++)
          {
            strncpy(tmpdevnam,((NAME8 *)tmpWorkArea)[i].name,8);
            dout->data.lptr[i] = strtol(tmpdevnam,NULL,16);
          }
          dout->dArrayLength = i;
          break;
        default:
          return illegal_format;
      }
    }
    else
    {
      n = getFiles((NAME8 *)NULL,ext,starttime,stoptime,0);
      putValuesFromLong(dout,(UINT32 *)&n,1);
    }    
    return n < 0 ? -n : 0;
  }
  if (!strcmp(devProperty,"QUERY.EQP")) /* the eqp modules for this session id */
  {
    if (din->dArrayLength != 1) return dimension_error;
    if (din->dFormat != CF_SHORT) return illegal_format;
    if ((id=din->data.sptr[0]-1) < 0 || id > MAXSESSION-1) return out_of_range;
    
    if (dout->dFormat != CF_CHAR16) return illegal_read_write;
    for (i=0,n=0; i<numEQP && n<session[id].numEQP; i++)  if (EQPlist[i].id == id)
    {

      if (n < (int)dout->dArrayLength) strncpy(((NAME16 *)dout->data.vptr)[n].name,EQPlist[i].name,16);
      else return buffer_too_small;
      n++;
    }
    dout->dArrayLength = n;
    return 0;
  }
  if (!strcmp(devProperty,"QUERY.PROPS")) /* the properties for this session id and eqp */
  {
    if (din->dArrayLength != 1) return dimension_error;
    if (din->dFormat != CF_SHORT) return illegal_format;
    if ((id=din->data.sptr[0]-1) < 0 || id > MAXSESSION-1) return out_of_range;
    
    if (dout->dFormat != CF_CHAR32) return illegal_read_write;
    for (i=0,n=0; i<numEQP && n<session[id].numEQP; i++)
      if (EQPlist[i].id == id && !strcmp(EQPlist[i].name,devName))
    {
      for (n=0,lpc=EQPlist[i].lpc; lpc; n++,lpc=lpc->Next)
      {
        if (n < (int)dout->dArrayLength) strncpy(((NAME32 *)dout->data.vptr)[n].name,lpc->prp,32);
        else return buffer_too_small;
      }
    }
    dout->dArrayLength = n;
    return 0;
  }
  if (!strncmp(devProperty,"OPEN",4))
  {
    if (din->dArrayLength != 1) return no_file_space;
    switch (din->dFormat)
    {
      case CF_NAME8:
        strncpy(name,(char *)din->data.vptr,8); break;
      case CF_LONG:
        sprintf(name,"%x",din->data.lptr[0]); break;
      default: return illegal_format;
    }
    name[8] = 0;
    ts = strtol(name,0,16); t = localtime(&ts); t_year = t->tm_year;
    strncpy(ext,devName,15); ext[15] = 0;
    if (strlen(ext) == 0) return string_expected;
    sprintf(fullName,"%s.%s",name,ext);
    if (archDebug) printf("LPC: make name for class %s",ext);
    if (makeFullName(fullName,ext)) return no_such_file;
    if (strstr(devProperty,"CMT")) return readCMT((char *)dout->data.vptr,fullName,dout->dArrayLength);
    if (explode(fullName)) return file_error;
    if (archDebug) dbglog("LPC: loading %s",fullName);
    if ((i=LoadEQP(t_year,fullName,ext)) < 0) return -i;
    session[i].active = time(NULL);
    if (dout->dFormat != CF_SHORT) return illegal_format;
    switch (dout->dArrayLength)
    {
      case 2: dout->data.sptr[1] = session[i].numEQP;
      case 1: dout->data.sptr[0] = i + 1;
              break;
      default: return illegal_read_write;
    }
    setDataTimeStamp((double)session[i].est);
    if (archDebug) dbglog("LPC: return session %d (%d eqps)",i,session[i].numEQP);
    return 0;
  }
  if (!strcmp(devProperty,"ANNOTATE"))                       /* CF_STRUCT */
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_TEXT) return illegal_format;
    if (din->dArrayLength < 8) return no_file_space;
    strncpy(name,(char *)din->data.vptr,8);  name[8] = 0;
    strncpy(ext,devName,15); ext[15] = 0;
    if (strlen(ext) == 0) return string_expected;
    sprintf(fullName,"%s.%s",name,ext);
    if (makeFullName(fullName,ext)) return no_such_file;
    return writeCMT(&((char *)din->data.vptr)[8],fullName,din->dArrayLength-8);
  }
  if (!strcmp(devProperty,"REMOVE"))                       /* CF_STRUCT */
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_CHAR8) return illegal_format;
    if (din->dArrayLength != 1) return no_file_space;
    strncpy(name,(char *)din->data.vptr,8);  name[8] = 0;
    strncpy(ext,devName,15); ext[15] = 0;
    if (strlen(ext) == 0) return string_expected;
    sprintf(fullName,"%s.%s",name,ext);
    if (makeFullName(fullName,ext)) return no_such_file;
    if (archDebug) dbglog("LPC: remove file %s\n",fullName);
    if (remove(fullName)) i++;
    sprintf(&fullName[strlen(fullName)-strlen(ext)],"CMT");
    if (archDebug) dbglog("LPC: remove file %s\n",fullName);
    if (remove(fullName)) i++;
    return i < 2 ? 0 : file_error;
  }
  if (!strcmp(devProperty,"CLOSE"))
  {
    if (din->dArrayLength != 1) return dimension_error;
    if ((cc=getValuesAsShort(din,&id,1)) != 0) return cc;
    id--;
    if (id < 0 || id > MAXSESSION-1) return out_of_range;
    freeSession(id);
    return 0;
  }
  if (!strcmp(devProperty,"SESSION"))
  {
    dumpSessions();
    return 0;
  }
  if (!strcmp(devProperty,"FLUSH"))
  {
    freeSessions();
    return 0;
  }
  if (!strcmp(devProperty,"DEBUG"))
  {
    if (din->dArrayLength == 1 && din->dFormat == CF_SHORT)
      archDebug = din->data.sptr[0];
    else if (dout->dArrayLength == 1 && dout->dFormat == CF_SHORT)
      dout->data.sptr[0] = archDebug;
    else return illegal_read_write;
    return 0;
  }
  if (!strncmp(devProperty,"READDATA",8))
  {
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if (strncmp(din->dTag,"APCBLK",8)) return invalid_structure_tag;
    if (din->dArrayLength != 1) return dimension_error;
    apc = (APCBLK *)din->data.vptr;
    
    if ((id=(short)apc->sessionID-1) < 0) ccerr(not_implemented); /* call open session (do later) */
    if (id > MAXSESSION - 1) ccerr(no_file_space);
    
    for (i=0; i<MAXEQP; i++) 
      if (EQPlist[i].id == id && !strncmp(EQPlist[i].name,apc->eqpName,16)) break;
    if (i == MAXEQP) ccerr(non_existent_elem);
err:
    if (cc != 0) /* try to get a new session slot if the request is faulty */
    {
      if (archDebug) dbglog("trying to recover from error %s",erlst[cc]);
      if ((id=GetLocalSessionID(apc->event,apc->ext)) < 0)
      {
        dbglog("unable to acquire a new session id : %s",erlst[-id]);
        return -id;
      }
    }
    setDataTimeStamp((double)session[id].est);
    if (!session[id].active) return file_not_open;
    session[id].active = time(NULL);  /* keep session alive */

    c = GetStoredPropertyFromAlias(apc->eqpName,apc->devProperty);
    if (c != NULL) strncpy(tmpprpnam,c,32); else tmpprpnam[0] = 0;

    strncpy(apcdevname,apc->devName,16); apcdevname[16] = 0;
    if (apcdevname[0] != '#') napcdevs = 1;
    else
    {
      apcstart = atoi(&apcdevname[1]);
      if ((c=strchr(&apcdevname[1],'#')) != NULL) apcstop = atoi(&c[1]);
      else apcstop = apcstart;
      napcdevs = apcstop - apcstart + 1;
      if ((c=strchr(apcdevname,':')) != NULL) apcidx = atoi(&c[1]);
      else apcidx = 0;
    }
    if (archDebug) dbglog("looking for %s %s %s",apc->eqpName,apc->devName,apc->devProperty);
    for (k=0,pass=0; k < napcdevs && pass < napcdevs; pass++)
    {
      if (napcdevs > 1) sprintf(apcdevname,"#%d",k+apcstart);
      for (lpc=EQPlist[i].lpc; lpc; lpc=lpc->Next)
      {
        if (strncmp(apc->devProperty,lpc->prp,32) &&
            strncmp(tmpprpnam,lpc->prp,32)) continue;
        propFound = 1;
        if (apcdevname[0] == '#')
        {  /* number asked for, prepare alternative devname as name */
           j = atoi(&apcdevname[1]);
           c = GetStoredDeviceName(j,apc->eqpName);
           if (c == NULL) return illegal_equipment_number; 
           strncpy(tmpdevnam,c,16); tmpdevnam[16]=0;
        }
        else /* name asked for, prepare alternative devname as number */
        {
           j = GetStoredDeviceNumber((NAME16 *)apcdevname,apc->eqpName);
           if (j >= 0) sprintf(tmpdevnam,"#%d",j);
           else tmpdevnam[0] = 0;
        }
        if (archDebug > 1) dbglog("alternate device names %s and %s",
          apcdevname,tmpdevnam);
        if (strlen(lpc->dev) && 
            strncmp(lpc->dev,apcdevname,16) && 
            strncmp(lpc->dev,tmpdevnam,16)) continue;
        if (archDebug) dbglog("found candidate");
        if ((dsize=fmtsizeof(lpc->fmt)*lpc->siz) > sizeof(archiveBuffer))
           return buffer_too_small;
        if (lpc->siz < dout->dArrayLength) dout->dArrayLength = lpc->siz;
        datapos = lpc->dataPos;
        datalen = dout->dArrayLength;
        if (napcdevs > 1) 
        {
          dsize = fmtsizeof(lpc->fmt);
          datapos += apcidx*dsize;
          datalen = 1;
        }
        if (lseek(session[id].hFile,datapos,SEEK_SET) == -1) return file_error;
        if (read(session[id].hFile,archiveBuffer,dsize) == -1) return file_error;
        if (filterArchiveData != NULL)
        {
          cc = (*filterArchiveData)(apc->devProperty,apcdevname,archiveBuffer,lpc->fmt,dout->dArrayLength,EqNo);
          if (cc != 0) return cc;
        }

        if (strstr(devProperty,"RAW")) { scale = 0; shift = 0; }
        else { scale = lpc->scale; shift = lpc->shift; }
        cc = reformatArchiveData(dout->data.bptr+k*dsize,dout->dFormat,
                                 archiveBuffer,lpc->fmt,datalen,scale,shift);
        if (cc) return cc;
        if (k++ > (int)dout->dArrayLength) return 0;
        pass = 0;
      }
    }
    if (k != 0) return 0;
    if (archDebug) dbglog("can't find %s %s %s\n",apc->eqpName,apc->devName,apc->devProperty);
    return un_allocated;
  }
  return illegal_property;
}

