/* gens.c - Group Equipment Name Server */

#include "prolog.h"
#include  <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include "rpcdefs.h"
#include "servdef.h"
#include "srvip.h"
#include "errors.h"
#include "database.h"
#include "brdcast.h"
#include "toolkit.h"
#include "srvip.h"
#if defined(UNIX)
# include "dirent.h"
# include "unixcmd.h"
#elif defined(WINNT)
# include "direct.h"
# include "ntcmd.h"
#endif

int scanTables(void);
int findcol(char *str,char *tag);
char *colptr(int col,char *str);
int scanFlag = 0;
void grpini(void);
void grptsk(void);
int grpeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);

typedef struct GrpDataXRefStruct
{
  char target[64];
  int offset;
  int index;
  int size;
  int disabled;
  NAME16 *devlist;
} GRPXREF;
typedef struct GrpDataStruct
{
  char context[16];
  char subsystem[16];
  char name[32];
  char module[8];
  int size;
  GRPXREF *xref; /* array xref of members */
} GRPDATA;
GRPDATA *GRPData=NULL;
int nGroupServers = 0;
#define GRP_FILE_NAME "grpdbase.csv"
#define GRP_RESERVED_SPACE 10
#define XREF_RESERVED_SPACE 10
#define MIN_GROUP_SPACE 100

#if defined(WINNT) || defined(SOLARIS)
int unsetenv(char *env)
{
  char envstr[128];
  strncpy(envstr,env,120);
  strcat(envstr,"=");
# ifdef WINNT
  return _putenv(envstr);
# else
  return putenv(envstr);
# endif
}
#endif
void PreSystemInit(void)
{
  int cc;
  char *c=NULL;
  int keepEnvSettings=FALSE;
  UINT32 haddr;
  struct in_addr ip;
  char ipaddr[16];

  if ((c=getenv("USE_TINE_HOME")) != NULL && strcmp(c,"TRUE") == 0)
     keepEnvSettings = TRUE;
  if (!keepEnvSettings)
  {
    unsetenv("TINE_HOME");
    unsetenv("FEC_HOME");
    /* legacy env variable */
    unsetenv("CONTROLDB");
    unsetenv("FECDB");
  }
  haddr = getmyipaddr();
  memcpy(&ip,&haddr,4);
  inet_ntoa_b(ip,ipaddr);
  SetNameServerAddress(ipaddr);
  feclog("assign ENS address : %s",ipaddr);

  gUseLoopback = TRUE;
  MaxNumContracts = 1000;

  cc = RegisterFecNameEx("GENS","ENS Group Server",OS_TAG,"ENS companion process","none","admin",101,"SERVICE");
  if (cc) feclog("RegisterFecNameEx : %s",erlst[cc]);

}
void PostSystemInit(void)
{
  int cc;
  cc = RegisterEquipmentModule("GENS","GRPEQM",100,grpeqm,grpini,grptsk,200,NULL);
  if (cc) 
  {
    feclog("RegisterEquipmentModule : %s",erlst[cc]);
    return;
  }
  cc = RegisterProperty("GRPEQM","GROUPS",100,CF_NAME16,CA_READ,"device group list");
  if (cc < 0) feclog("RegisterProperty GROUPS: %s",erlst[-cc]);
  cc = RegisterProperty("GRPEQM","MEMBERS",100,CF_NAME16,CA_READ,"device group member list");
  if (cc < 0) feclog("RegisterProperty MEMBERS: %s",erlst[-cc]);
  cc = RegisterProperty("GRPEQM","JOIN",1,CF_NAME16,CA_WRITE,"join device group member list");
  if (cc < 0) feclog("RegisterProperty JOIN: %s",erlst[-cc]);
  cc = RegisterProperty("GRPEQM","SEPARATE",1,CF_NAME16,CA_WRITE,"join device group member list");
  if (cc < 0) feclog("RegisterProperty SEPARATE: %s",erlst[-cc]);

}
int makeDirPath(char *fname)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WINNT)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];
  int len;

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,"/\\"); c != NULL; c=strtok(NULL,"/\\"))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return 0;
    }
    len = strlen(dirname);
    dirname[len] = FS_DELIMITER; dirname[len+1] = 0;
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH)) return 0;
    }
#   elif defined(WINNT)
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return 0;
    }
#   endif
  }
  return -1;
}

int isFec(char *nam)
{
  if (locateFecName(nam) >= 0) return TRUE;
  return FALSE;
}
#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  for (i=0; i<nadmins && strnicmp(nam,admins[i].name,16); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),16);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int GetDeviceNamesFromCache(char *srv,NAME16 *devbuffer,int *size)
{
  FILE *fp=NULL;
  char *c,fn[132],srvpath[64];
  int i,k,cc=0,done=0;
  int n, p;
  int dev_col=-1;
  char str[256],hdr[256];

  if (srv == NULL || devbuffer == NULL || size == NULL) ccerr(argument_list_error);
  p = srv[0] == '/' ? 1 : 0;
  strncpy(srvpath,&srv[p],64);
  if ((c=strstr(srvpath,"/#")) != NULL) *c = 0;
  sprintf(fn,"GROUPS/%s/devices.csv",srvpath);
  if ((n=GetDataBaseSize(fn)) < 0) ccerr(-n);
  if (*size < n) 
  {
    feclog("truncating size of device list !");
    n = *size;
  }
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  for (done=0,i=0,k=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((dev_col=findcol(hdr,"DEVICES")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(devbuffer[i].name,colptr(dev_col,str),16);
    i++;
  }
  *size = n;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int WriteDeviceNamesToCache(char *srv,NAME16 *devbuffer,int size)
{
  FILE *fp=NULL;
  char *c,fn[132],srvpath[64];
  int i,p,cc=0,done=0;

  if (srv == NULL || devbuffer == NULL || size < 1) ccerr(argument_list_error);
  p = srv[0] == '/' ? 1 : 0;
  strncpy(srvpath,&srv[p],64);
  if ((c=strstr(srvpath,"/#")) != NULL) *c = 0;
  sprintf(fn,"GROUPS/%s/devices.csv",srvpath);
  makeDirPath(fn);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Devices\n");
  for (i=0; i<size;i++)
    fprintf(fp,"%s\n",devbuffer[i].name);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int findServerGroup(char *group,char *context)
{
  int i;
  char *cntxt = context;
  if (context == NULL || strlen(context) == 0) cntxt = "DEFAULT";
  for (i=0; i<nGroupServers; i++)
  {
    if (strncmp(cntxt,GRPData[i].context,16)) continue;
    if (strncmp(group,GRPData[i].name,16)) continue;
    break;
  }
  return i < nGroupServers ? i : -1;
}
#define DEV_BUFFER_SIZE 1000
NAME16 devbuffer[DEV_BUFFER_SIZE];
int writeMembersFile(char *filename,GRPXREF *xref,int listsize)
{
  int i,cc=0;
  FILE *fp=NULL;

  if (filename == NULL || strlen(filename) == 0) ccerr(un_allocated);
  makeDirPath(filename);
  if ((fp=fopen(filename,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Members, Index\n"); /* write the header */
  for (i=0; i<listsize; i++)
  {
    if (xref[i].disabled) continue;
    fprintf(fp,"%s, %d\n",xref[i].target,xref[i].index);
  }
err:
  if (cc) feclog("writeMembersFile %s : %s",filename == NULL ? "[empty]" : filename,erlst[cc]);
  if (fp != NULL) fclose(fp);
  return cc;
}
void organizeMembers(GRPXREF *xref,int listsize)
{
  int i,k;
  GRPXREF tref;

  if (listsize <= 0) return;
  /* these will usually be ordered */
  /* the worst case (exactly backwards) can only happen once, so just do it this way */
  for (i=0; i<listsize; i++)
  {
    for (k=0; k<i; k++)
    {
      if (xref[i].index < xref[k].index)
      { /* then swap them */
        tref = xref[i];
        xref[i] = xref[k];
        xref[k] = tref;
      }
    }
  }
}
GRPXREF *loadMembersFile(char *filename,int *listsize)
{
  int i,idx,n=0,cc=0,done=0;
  int mem_col=-1,idx_col=-1;
  char str[256],hdr[256],scratch[128];
  GRPXREF *xref=NULL;
  FILE *fp=NULL;

  if (filename == NULL || strlen(filename) == 0) ccerr(un_allocated);

  if ((n=GetDataBaseSize(filename)) < 0) ccerr(-n);
  if ((xref=(GRPXREF *)calloc(n+XREF_RESERVED_SPACE,sizeof(GRPXREF))) == NULL) ccerr(out_of_server_memory);
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<n;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((mem_col=findcol(hdr,"MEMBERS")) < 0) ccerr(no_such_line);
      idx_col = findcol(hdr,"INDEX");
      done = 1;
      continue;
    }
    idx = i;
    if (idx_col >= 0) 
    {
      strncpy(scratch,colptr(idx_col,str),16);
      if ((idx=atoi(scratch)) < 0) ccerr(invalid_index);
    }

    xref[i].index = idx;
    strncpy(scratch,colptr(mem_col,str),16);
    if (strlen(scratch) == 0) continue;
    strncpy(xref[i].target,scratch,16);
    i++;
  }
  organizeMembers(xref,n);
err:
  if (listsize != NULL) *listsize = n;
  if (cc) feclog("loadMembersFile %s : %s",filename == NULL ? "[empty]" : filename,erlst[cc]);
  if (fp != NULL) fclose(fp);
  return xref;
}
int writeDatabaseFile(char *filename)
{
  FILE *fp = NULL;
  int i, cc=0;

  if (filename == NULL || strlen(filename) == 0) ccerr(un_allocated);
  if ((fp=fopen(filename,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Group, GroupModule, Context, SubSystem\n"); /* write the header */
  for (i=0; i<nGroupServers; i++)
  {
    fprintf(fp,"%s, %s, %s, %s\n",
      GRPData[i].name,GRPData[i].module,GRPData[i].context,GRPData[i].subsystem);
  }

err: 
  if (fp != NULL) fclose(fp);
  if (cc) feclog("writeDatabaseFile %s : %s",filename == NULL ? "[empty]" : filename, erlst[cc]);
  return cc;
}
int dumpDatabase(void)
{
  int i, len, rc = 0, cc = 0;
  char fn[128];

  for (i=0; i<nGroupServers; i++)
  {
    strcpy(fn,"GROUPS"); 
    len = strlen(fn); fn[len] = FS_DELIMITER; fn[len+1] = 0;
    strcat(fn,GRPData[i].context);
    len = strlen(fn); fn[len] = FS_DELIMITER; fn[len+1] = 0;
    strcat(fn,GRPData[i].name); strcat(fn,".csv");
    cc = writeMembersFile(fn,GRPData[i].xref,GRPData[i].size);
    if (cc != 0) rc = cc;
  }
  cc = writeDatabaseFile(GRP_FILE_NAME);
  if (cc != 0) rc = cc;
  /* can't re-read the files immediately ! */
  /* the call to get device names will fail ! */
  /* scanTables(); */
  return rc;
}
void checkDeviceNames(void)
{
  int i,k,len,n,nmembers,size,grpsize,grpspace,cc=0,rc;
  char srv[128];
  GRPXREF *xref;
  ExportListStruct *el;

  for (i=0; i<nGroupServers; i++)
  {
    /* TODO:  check to see if anything has changed (more devices, etc.) */
    grpsize = 0; nmembers = 0;
    for (k=0; k<GRPData[i].size; k++)
    {
      xref = &GRPData[i].xref[k];
      if (xref->size > 0) continue; /* already acquired */
      size = DEV_BUFFER_SIZE;
      sprintf(srv,"/%s/%s/#0",GRPData[i].context,xref->target);
      while ((rc=GetDeviceNames(srv,devbuffer,&size)) == invalid_transport_size)
      {
        size /= 2;
      }
      if (rc != 0)
      { /* read from file */
        rc = GetDeviceNamesFromCache(srv,devbuffer,&size);
      }
      else
      { /* write to file */
        for (n=0; n<size; n++)
        { /* have a look at the supplied device names first */
          if (devbuffer[n].name[0] == '#')
          {
            strncpy(devbuffer[n].name,xref->target,16);
            if (size > 1) 
            {
              devbuffer[n].name[12] = 0; /* 1st 12 chars from the srv name */
              len = strlen(devbuffer[n].name);
              sprintf(devbuffer[n].name,"#%3d",n);
            }
          }
        }
        WriteDeviceNamesToCache(srv,devbuffer,size);
      }
      if (rc) 
      {
        feclog("cannot acquire devices for %s : %s",srv,erlst[rc]);
        continue;
      }
      xref->size = size;
      if ((xref->devlist=(NAME16 *)calloc(xref->size,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);

      memcpy(xref->devlist,devbuffer,size*sizeof(NAME16));
      /* determine total size and device list offset */
      xref->offset = grpsize;
      grpsize += xref->size;
    }
    if (grpsize == 0) continue; /* database problems ? */
    /* register equipment module with no equipment function, leave room for additional devices */
    if ((el=getExportListItem(GRPData[i].module)) == NULL) 
    { /* if not already registered */
      strncpy(gDeviceContext,GRPData[i].context,EQP_CONTEXT_SIZE);
      strncpy(gDeviceSubSystem,GRPData[i].subsystem,16);
      /* leave some room for others to come: (if the total number exceeds this -> restart required!) */
      grpspace = MAX(grpsize*4,MIN_GROUP_SPACE);
      RegisterEquipmentModule(GRPData[i].name,GRPData[i].module,grpspace,NULL,NULL,NULL,0,NULL);
      /* register a nominal property */
      RegisterProperty(GRPData[i].name,"EMPTY",1,CF_SHORT,CA_READ,"empty property");
    }
    AssignNumModules(GRPData[i].module,grpsize);
    makeEqpModuleHashTable(el);

    /* register the device names */
    /* redirect the calls for each device */
    for (k=0; k<GRPData[i].size; k++)
    {
      xref = &GRPData[i].xref[k];
      for (n=0; n<xref->size; n++)
      {
        RegisterDeviceName(GRPData[i].module,xref->devlist[n].name,n + xref->offset);
        RedirectDeviceName(GRPData[i].module,xref->devlist[n].name,xref->target);
      }
    }
  }
err:
  if (cc) feclog("checkDeviceNames: %s",erlst[cc]);
}

int loadDatabaseFile(char *filename)
{
  FILE *fp = NULL;
  int i,len,nmembers,cc=0,done=0;
  int nam_col=-1,con_col=-1,sub_col=-1;
  int grp_col=-1,grpmod_col=-1,grpidx_col=-1;
  char str[256],hdr[256],fn[128];

  if (filename == NULL || strlen(filename) == 0) ccerr(un_allocated);

  nGroupServers = 0;
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      con_col=findcol(hdr,"CONTEXT");
      sub_col=findcol(hdr,"SUBSYSTEM");
      if ((grp_col=findcol(hdr,"GROUP")) < 0) ccerr(no_such_line);
      if ((grpmod_col=findcol(hdr,"GROUPMODULE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(GRPData[i].name,colptr(grp_col,str),16);
    if (strlen(GRPData[i].name) == 0) continue;
    strncpy(GRPData[i].module,colptr(grpmod_col,str),8);
    if (strlen(GRPData[i].module) == 0) continue;
    strncpy(GRPData[i].context,con_col >= 0 ? colptr(con_col,str) : "DEFAULT",16);
    if (sub_col >= 0)
    {
      strncpy(GRPData[i].subsystem,colptr(sub_col,str),16);
    }
    /* read the members list for this group */
    strcpy(fn,"GROUPS"); 
    len = strlen(fn); fn[len] = FS_DELIMITER; fn[len+1] = 0;
    strcat(fn,GRPData[i].context);
    len = strlen(fn); fn[len] = FS_DELIMITER; fn[len+1] = 0;
    strcat(fn,GRPData[i].name); strcat(fn,".csv");
    if ((GRPData[i].xref=loadMembersFile(fn,&nmembers)) == NULL) continue;
    GRPData[i].size = nmembers;
    i++;
  }
  fclose(fp); fp = NULL;
  nGroupServers = i;

  checkDeviceNames();

err: 
  if (fp != NULL) fclose(fp);
  if (cc) feclog("loadDatabaseFile %s : %s",filename == NULL ? "[empty]" : filename, erlst[cc]);
  return cc;
}
int scanTables(void)
{
  int i, n, cc = 0;

  if (GRPData != NULL) 
  {
    for (i=0; i<nGroupServers; i++)
    {
      if (GRPData[i].xref != NULL) free(GRPData[i].xref);
    }
    free(GRPData);
  }

  if ((n=GetDataBaseSize(GRP_FILE_NAME)) < 0) ccerr(-n);

  if ((GRPData=(GRPDATA *)calloc(n+GRP_RESERVED_SPACE,sizeof(GRPDATA))) == NULL)
    ccerr(out_of_server_memory);

  cc = loadDatabaseFile(GRP_FILE_NAME);
err:
  return cc;
}
void grpini(void)
{
  /* anything to do here ? */
}
#ifdef WINNT
# define lstat(f,s) _stat(f,s)
#endif
void grptsk(void)
{
  int cc;
  static time_t lastfiletime = 0;
# ifdef WINNT
  struct _stat fnstat;
# else
  struct stat fnstat;
# endif

  if (lstat(GRP_FILE_NAME,&fnstat) < 0)
  {
    feclog("panic: cannot find gqp database !");
    return;
  }
  if (lastfiletime != fnstat.st_mtime)
  {
    cc = scanTables();
    feclog("scanTables: %s",erlst[cc]);
    lastfiletime = fnstat.st_mtime;
  }

  checkDeviceNames();

  return;
}
char *makeModuleName(char *grpName)
{
  int i, k, len, idx = -1;
  char mod[8];
  static char modName[32];

  if (grpName == NULL || strlen(grpName) == 0) return NULL;
  if (strlen(grpName) < 7) 
  {
    strncpy(modName,grpName,6);
    for (i=0; i<nGroupServers && strncmp(modName,GRPData[i].module,6); i++);
    if (i == nGroupServers) return modName; /* simple name is available */
  }
  for (i=0; i<nGroupServers; i++)
  {
    if (GRPData[i].module[0] == 'G')
    { /* possibly assigned */
      strncpy(mod,&GRPData[i].module[1],5); mod[5] = 0;
      len = strlen(mod);
      for (k=0; k<len && isdigit(mod[k]); k++);
      if (k == len) /* all digits */
      { /* definitely assigned! (find the highest group index) */
        k = atoi(mod);
        if (idx < k) idx = k;
      }
    }
  }
  idx++;
  sprintf(modName,"G%05d",idx);
  return modName;
}
int getNextIndexFromGroup(int n)
{
  int i, idx=0;
  GRPXREF *xref;
  for (i=0; i<GRPData[n].size; i++)
  {
    if ((xref = &GRPData[n].xref[i]) == NULL) continue;
    if (xref->index >= idx) idx = xref->index + 1;
  }
  return idx;
}
int grpeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int i,n,len, index;
  char *c,*cs,str[64],cntName[32],grpName[32],srvName[32],subName[32],usr[20];
  GRPXREF *xref;

  strncpy(usr,GetCaller(),16); usr[16] = 0;
  if (!strcmp(devProperty,"GROUPS"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dout->dArrayLength == 0) return dimension_error;
    for (i=0; i<nGroupServers && i<(int)dout->dArrayLength; i++)
    {
      switch (dout->dFormat)
      {
        case CF_NAME16:
          strncpy(((NAME16 *)dout->data.vptr)[i].name,GRPData[i].name,16);
          break;
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          len = fmtsizeof(dout->dFormat);
          sprintf(str,"/%s/%s",GRPData[i].context,GRPData[i].name);
          strncpy(dout->data.cptr+i*len,str,len);
          break;
        default:
          return illegal_format;
      }
    }
    dout->dArrayLength = (UINT32)i;
    return 0;
  }
  if (!strcmp(devProperty,"MEMBERS"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dout->dArrayLength == 0) return dimension_error;
    c = NULL;
    if (din->dArrayLength > 0)
    {
      memset(cntName,0,32);
      switch (din->dFormat)
      {
        case CF_TEXT:
          len = MIN(din->dArrayLength,16);
          break;
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          len = 16;
          break;
        default:
          return illegal_format;
      }
      strncpy(cntName,din->data.cptr,len);
      c = cntName;
    }
    if ((n=findServerGroup(devName,c)) < 0) return non_existent_elem;
    for (i=0; i<GRPData[n].size; i++)
    {
      if ((xref = &GRPData[n].xref[i]) == NULL) continue;
      switch (dout->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          len = fmtsizeof(dout->dFormat);
          strncpy(dout->data.cptr+i*len,xref->target,len);
          break;
        default:
          return illegal_format;
      }
    }
    dout->dArrayLength = (UINT32)i;
    return 0;
  }
  if (!strcmp(devProperty,"JOIN"))
  { /* be strict as to the format and order of the input */
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isFec(usr) && !isAdministrator(usr)) return access_denied;
    if (din->dFormat != CF_NAME16) return illegal_format;
    index = 0; c = NULL; cntName[0] = 0; subName[0] = 0;
    switch (din->dArrayLength)
    {
      case 5:
        strncpy(subName,((NAME16 *)din->data.vptr)[4].name,16);
      case 4:
        strncpy(str,((NAME16 *)din->data.vptr)[3].name,16);
        index = atoi(str);
        if (index < 0) return invalid_index;
      case 3:
        strncpy(cntName,((NAME16 *)din->data.vptr)[2].name,16);
        c = cntName;
        if (strlen(subName) == 0 && (cs=strchr(cntName,'.')) != NULL)
        {
          *cs++ = 0;
          strncpy(subName,cs,15-strlen(c));
        }
      case 2:
        strncpy(grpName,((NAME16 *)din->data.vptr)[1].name,16);
        strncpy(srvName,((NAME16 *)din->data.vptr)[0].name,16);
        break;
      default: 
        return dimension_error;
    }
    if (strlen(grpName) == 0 || strlen(srvName) == 0) return argument_list_error;
    /* now join the group */
    if ((n=findServerGroup(grpName,c)) < 0) 
    { /* new group ! */
      n = nGroupServers; /* there will be enough reserved space ! */
      strncpy(GRPData[n].name,grpName,16);
      strncpy(GRPData[n].module,makeModuleName(grpName),8); /* grpName definitely contains a name ! */
      strncpy(GRPData[n].context,strlen(cntName) > 0 ? cntName : "DEFAULT",16);
      strncpy(GRPData[n].subsystem,subName,16);
      if ((GRPData[n].xref=(GRPXREF *)calloc(XREF_RESERVED_SPACE,sizeof(GRPXREF))) == NULL) 
        return out_of_server_memory;
      nGroupServers++;
    }
    for (i=0; i<GRPData[n].size; i++)
    {
      if ((xref = &GRPData[n].xref[i]) == NULL) continue;
      if (strncmp(xref->target,srvName,16)) continue;
      break;
    }
    if (GRPData[n].size > 0 && i < GRPData[n].size) 
    { /* nothing to do, it's already there (remote server is simply restarting) */
      return 0;
    }
    /* add member to the group */
    i = GRPData[n].size;
    xref = &GRPData[n].xref[i];
    strncpy(xref->target,srvName,16);
    if (i > 0 && index == 0) index = getNextIndexFromGroup(n);
    xref->index = index;
    GRPData[n].size++;
    /* now dump it out (causes immediate readback */
    return dumpDatabase();
  }
  if (!strcmp(devProperty,"SEPARATE"))
  { /* be strict as to the format and order of the input */
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isAdministrator(usr)) return access_denied;
    if (din->dFormat != CF_NAME16) return illegal_format;
    c = NULL; cntName[0] = 0;
    switch (din->dArrayLength)
    {
      case 3:
        strncpy(cntName,((NAME16 *)din->data.vptr)[2].name,16);
        c = cntName;
      case 2:
        strncpy(grpName,((NAME16 *)din->data.vptr)[1].name,16);
        strncpy(srvName,((NAME16 *)din->data.vptr)[0].name,16);
        break;
      default: return dimension_error;
    }
    if ((n=findServerGroup(grpName,c)) < 0) return non_existent;
    for (i=0; i<GRPData[n].size; i++)
    {
      if ((xref = &GRPData[n].xref[i]) == NULL) continue;
      if (strncmp(xref->target,srvName,16)) continue;
      break;
    }
    if (i == GRPData[n].size) return non_existent_fec;
    xref->disabled = TRUE;
    return dumpDatabase();
  }
 return not_implemented;
}

