#include "prolog.h"
#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "rpctypes.h"
#include "database.h"
#include "errors.h"
#include "archive.h"
#include "toolkit.h"
#if defined(UNIX)
# include "dirent.h"
# include "unixcmd.h"
#elif defined(WINNT)
# include "direct.h"
# include "ntcmd.h"
#endif

#define ccerr(i) { cc = i; goto err; }
char ArchiveDBpath[] = "../database/";
char *ArchiveDatapath = "../DATA"; 
extern FEC *fec;
extern char cbuff[NETMEX_BUF_SIZE];
extern int nfecs, nparam, nmki;
extern char cbuff[NETMEX_BUF_SIZE];
extern int archPending,totalNrFECs;
extern int archdebug, ramping, running, beam;
extern int g_isTerminating;
int debugon = 0;
void AppendProblemsLog(char *str);
int OpenDataFile(int i);
int makeDirPath(char *fname);
int allocActiveOnly = FALSE;

int nSubSystems = 0;
SUB_SYS *subSystems = NULL;
int nPrpAlias = 0;
PRP_ALIAS *prpAlias = NULL;

#define fielderr(str) {\
  printf("%s field too long: %s (%d)\n>",(str),fec[i].eqm,l);\
  printf("Hit any key to continue ...\n");\
  getchar(); }

int getEntryFromKeyword(char *keyword,int *rectyp,int *param)
{
  FECDATA *fecdata;
  int i,j;

  for (i=0; i<nfecs; i++)
  {
    if (!fec[i].active) continue;
    fecdata = fec[i].fecdata;
    for (j=0; j<fec[i].nparam; j++)
    {
      if (!strcmp(fecdata[j].param,keyword))
      {
        if (rectyp != NULL) *rectyp = fec[i].rectyp;
        if (param != NULL) *param  = j;
        return i;
      }
    }
  }
  return -1;
}
int getNumLines(FILE *fp)
{
  int i;
  char str[256];

  for (i=0; fgets(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    i++;
  }
  rewind(fp);
  return i-1;   /* subtract the header */
}
int OpenDataFile(int i)
{
  char filenm[80];
  int hDataFile = -1;
  int depth;
  sprintf(filenm,"%s/%04d/%02d/%s",
    ArchiveDatapath,fec[i].archYear,fec[i].archMonth,fec[i].fname);
  hDataFile = open(filenm,O_RDONLY);
  for (depth=0; hDataFile == -1 && depth < ARCHIVEDATAPATHDEPTH; depth++)
  {
    sprintf(filenm,"%s/%04d/%02d.%d/%s",
      ArchiveDatapath,fec[i].archYear,fec[i].archMonth,depth,fec[i].fname);
    hDataFile = open(filenm,O_RDONLY);
  }
  return hDataFile;
}
int inputDataLoad(int ntotal, int offset)
{
  int i, n;
  char str[256],scratch[16];
  int done = 0,cc = 0;
  int dat_col;
  short *indata;
  FILE *fp;

  for (i=offset; i<ntotal+offset; i++)
  {
    if (strlen(fec[i].infile) == 0) continue;
    sprintf(str,"%sinput/%s",ArchiveDBpath,fec[i].infile);
    if ((fp=fopen(str,"rt")) == NULL)
    {
      printf("%s : Can't open input file %s\n",fec[i].eqm,fec[i].infile);
      continue;
    }
    if ((fec[i].nindata=getNumLines(fp)) <= 0)
    {
      printf("%s : no records in input file !!\n",fec[i].eqm);
      exit(1);
    }
    indata = (short *) calloc(fec[i].nindata,sizeof(short));
    done = 0;
    for (n=0; fgets(str,255,fp) != NULL && n<fec[i].nindata; )
    {
      strtrm(str);
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        if ((dat_col=findcol(str,"DATA")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }
      strncpy(scratch,colptr(dat_col,str),16); 
      indata[n] = atoi(scratch);
      n++;
    }
    fec[i].indata = indata;
err: fclose(fp);
  }
  return cc;
}
void fillinArrayTypes(int num)
{
  int i,j,n,m,len;
  char tgt[64];

  for (i=0,tgt[32]=0; i<num; i++)
  {
    for (n=0; n<fec[i].nparam; n++)
    {
      fec[i].fecdata[n].arrayType = AT_SPECTRUM; /* default */
      if (fec[i].fecdata[n].size == 1) 
      {
        fec[i].fecdata[n].arrayType = AT_SCALAR;
      }
      else
      {
        strncpy(tgt,fec[i].fecdata[n].param,32);
        if ((len=strlen(tgt)) > 4 && !strcmp(&tgt[len-4],".NAM"))
        {
          fec[i].fecdata[n].arrayType = AT_CHANNEL;
          /* find its partner ... */
          tgt[len-4] = 0;
          if ((j=getEntryFromKeyword(tgt,NULL,&m)) < 0) continue;
          fec[j].fecdata[m].arrayType = AT_CHANNEL;
        }
        else 
        {
          switch (fec[i].fecdata[n].format)
          {
            case CF_TEXT: 
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME48:
            case CF_NAME64:
              fec[i].fecdata[n].arrayType = AT_UNKNOWN;
              break;
            default:
              break;
          }
        }
      }
    }
  }
  for (i=0; i<nPrpAlias; i++)
  {
    strncpy(tgt,prpAlias[i].name,32);
    if ((len=strlen(tgt)) > 4 && !strcmp(&tgt[len-4],".NAM"))
    {
      /* find its partner ... */
      tgt[len-4] = 0;
      if ((j=getEntryFromKeyword(tgt,NULL,&m)) < 0) continue;
      fec[j].fecdata[m].arrayType = AT_CHANNEL;
    }
  }
  return;
}
int paramLoad(int ntotal, int offset)
{
  int i, j, n, num, recsize, fmt;
  char str[256],scratch[16],dnamePrp[32];
  int done = 0,cc = 0;
  int key_col,fmt_col,siz_col,uni_col,max_col,min_col,
      tol_col,plt_col,cmt_col,dep_col,scale_col,offset_col,
      dev_col,bnd_col,sub_col;
  FECDATA *fecdata;
  FILE *fp;

  for (i=offset, num=0; i<ntotal+offset; i++)
  {
    sprintf(str,"%sinfo/%s",ArchiveDBpath,fec[i].sfile);
    if ((fp=fopen(str,"rt")) == NULL)
    {
      printf("%s : Can't open structure file %s\n",fec[i].eqm,fec[i].sfile);
      continue;
    }
    if ((fec[i].nparam=getNumLines(fp)) <= 0)
    {
      printf("%s : no records in structure file !!\n",fec[i].eqm);
      exit(1);
    }
    fecdata = (FECDATA *) calloc(fec[i].nparam,sizeof(FECDATA));
    recsize = 0; done = 0;
    for (n=0; fgets(str,255,fp) != NULL; )
    {
      strtrm(str);
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        if ((key_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line);
        if ((fmt_col=findcol(str,"FORMAT")) < 0) ccerr(no_such_line);
        if ((siz_col=findcol(str,"SIZE")) < 0) ccerr(no_such_line);
        if ((uni_col=findcol(str,"UNITS")) < 0) ccerr(no_such_line);
        if ((max_col=findcol(str,"MAXIMUM")) < 0) ccerr(no_such_line);
        if ((min_col=findcol(str,"MINIMUM")) < 0) ccerr(no_such_line);
        if ((tol_col=findcol(str,"TOLERANCE")) < 0) ccerr(no_such_line);
        if ((plt_col=findcol(str,"PLOT")) < 0) ccerr(no_such_line);
        if ((cmt_col=findcol(str,"COMMENT")) < 0) ccerr(no_such_line);
        scale_col = findcol(str,"SCALE");
        offset_col = findcol(str,"OFFSET");
        dep_col = findcol(str,"DEPTH");
        dev_col = findcol(str,"DEVICES");
        bnd_col = findcol(str,"BIND");
        sub_col = findcol(str,"SUBSYSTEM");
        done = 1;
        continue;
      }
      strncpy(fecdata[n].param,colptr(key_col,str),32);
      strncpy(scratch,colptr(fmt_col,str),16); strupr(scratch);
      if ((fmt=getFormatType(scratch)) == CF_NULL) fmt = atoi(scratch);
      fecdata[n].format = fmt;
      strncpy(scratch,colptr(siz_col,str),16);
      fecdata[n].size = atoi(scratch);
      strncpy(fecdata[n].units,colptr(uni_col,str),32);
      strncpy(scratch,colptr(max_col,str),16);
      fecdata[n].maximum = atof(scratch);
      strncpy(scratch,colptr(min_col,str),16);
      fecdata[n].minimum = atof(scratch);
      strncpy(scratch,colptr(tol_col,str),16);
      if (strchr(scratch,'%')) fecdata[n].pTolerance = (float)(atof(scratch) / 100.0);
      else fecdata[n].aTolerance = atof(scratch);
      strncpy(fecdata[n].desc,colptr(cmt_col,str),48);
      strncpy(fecdata[n].plot,colptr(plt_col,str),16);
      fecdata[n].offset = recsize;
      recsize += fecdata[n].size * fmtsizeof(fecdata[n].format);
      if (dep_col > 0) 
      {
        strncpy(scratch,colptr(dep_col,str),16);
        fecdata[n].depth = atoi(scratch);
        fecdata[n].data = (float *)calloc(fecdata[n].size*fecdata[n].depth,sizeof(float));
        if (fecdata[n].data == NULL) printf("%s : no history memory\n",fec[i].eqm);
      }
      fecdata[n].scaleFactor = 1;
      if (scale_col > 0)
      {
        strncpy(scratch,colptr(scale_col,str),16);
        fecdata[n].scaleFactor = (float)atof(scratch);
      }
      if (offset_col > 0)
      {
        strncpy(scratch,colptr(offset_col,str),16);
        fecdata[n].scaleOffset = (float)atof(scratch);
      }
      if (dev_col > 0 && nPrpAlias < totalNrFECs)
      {
        /*
          in case "ABC" gets its device names from "DEF"
          fecdata[n].param = "ABC", dev_col -> "DEF.NAM"
          then "ABC.NAM" is formed and matched to the alias "DEF.NAM"
        */
        strncpy(prpAlias[nPrpAlias].alias,colptr(dev_col,str),32);
        if (strlen(prpAlias[nPrpAlias].alias) > 0)
        {
          strncpy(dnamePrp,fecdata[n].param,27); strcat(dnamePrp,".NAM");
          strncpy(prpAlias[nPrpAlias].name,dnamePrp,32);
          nPrpAlias++;
        }
      }
      if (bnd_col > 0)
      {
        strncpy(fecdata[n].boundTo,colptr(bnd_col,str),32);
      }
      if (sub_col > 0)
      {
        strncpy(fecdata[n].subsys,colptr(sub_col,str),16);
      }
      else
      {
        strncpy(fecdata[n].subsys,"Diagnostics",16);
      }
      for (j=0; j<nSubSystems; j++) 
        if (!strncmp(subSystems[j].name,fecdata[n].subsys,16)) break;
      if (j == nSubSystems && nSubSystems < totalNrFECs)
      {
        subSystems[j].name = fecdata[n].subsys; /* the pointer won't disappear */
        nSubSystems++;
      }
      if (nSubSystems > 0)
      {
        subSystems[j].number++;
      }
      n++;
    }
err: 
    fclose(fp);
    if (recsize != fec[i].reclen)
    {
      feclog("[%d] %s %s %s (%s) : record length mis-match",
        i,fec[i].eqm,fec[i].property,fec[i].devname,fec[i].sfile); 
      exit(1);
    }
    fec[i].fecdata = fecdata;
    num += n;
  }
  return num;
}
int getNrActiveFecs(char *fname)
{
  FILE *fp = NULL;
  char fn[80];
  char str[256],scratch[16];
  int done = 0,i,cc = 0;
  int active_col;

  sprintf(fn,"%s%s",ArchiveDBpath,fname);
  if ((fp = fopen(fn,"rt")) == NULL)
  {
    printf("\n Can't open data base file : %s\n",fname);
    exit(1);
  }

  for (i=0; fgets(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((active_col=findcol(str,"ACTIVE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(active_col,str),16);
    if (atoi(scratch)) i++;
  }
err:fclose(fp);
  allocActiveOnly = TRUE;
  return i;
}
int archLoad(char *fname, int offset) /* returns number of equipment function calls found */
{
  FILE *fp;
  char fn[80];
  char str[256],scratch[48];
  int done = 0,i,fmt,cc = 0;
  int active_col,name_col,property_col,device_col,datalen_col,fmt_col,
      heartbeat_col,poll_col,amethod_col,sfile_col,resp_col,rectype_col,
      cntxt_col;
  int inf_col = -1;

  sprintf(fn,"%s%s",ArchiveDBpath,fname);
  if ((fp = fopen(fn,"rt")) == NULL)
  {
    printf("\n Can't open data base file : %s\n",fname);
    exit(1);
  }

  for (i=offset; fgets(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((rectype_col=findcol(str,"RECORD TYPE")) < 0) ccerr(no_such_line);
      if ((name_col=findcol(str,"NAME")) < 0) ccerr(no_such_line);
      if ((property_col=findcol(str,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((device_col=findcol(str,"DEVICE")) < 0) ccerr(no_such_line);
      if ((datalen_col=findcol(str,"DATA LENGTH")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(str,"RPC FORMAT")) < 0) ccerr(no_such_line);
      if ((heartbeat_col=findcol(str,"HEARTBEAT")) < 0) ccerr(no_such_line);
      if ((poll_col=findcol(str,"POLLING RATE")) < 0) ccerr(no_such_line);
      if ((amethod_col=findcol(str,"ARCHIVE")) < 0) ccerr(no_such_line);
      if ((sfile_col=findcol(str,"STRUCTURE")) < 0) ccerr(no_such_line);
      if ((resp_col=findcol(str,"RESPONSIBILITY")) < 0) ccerr(no_such_line);
      if ((active_col=findcol(str,"ACTIVE")) < 0) ccerr(no_such_line);
      inf_col = findcol(str,"DATA INPUT");
      cntxt_col=findcol(str,"CONTEXT");
      done = 1;
      continue;
    }
    if (allocActiveOnly) 
    {
      if (atoi(colptr(active_col,str)) == 0) continue;
      fec[i].active = 1;
    }
    else fec[i].active = atoi(colptr(active_col,str));
    strncpy(scratch,colptr(rectype_col,str),32);
    fec[i].rectyp = atoi(scratch);
    strncpy(fec[i].eqm,colptr(name_col,str),16);
    strncpy(fec[i].property,colptr(property_col,str),32);
    strncpy(scratch,colptr(device_col,str),32);
    strncpy(fec[i].devname,scratch,32);
    if (scratch[0] == '#') fec[i].EqNo = atoi(&scratch[1]);
    strncpy(scratch,colptr(datalen_col,str),32);
    fec[i].len = atoi(scratch);
    strncpy(scratch,colptr(fmt_col,str),32); strupr(scratch);
    if ((fmt=getFormatType(scratch)) == CF_NULL) fmt = atoi(scratch);
    fec[i].fmt = fmt;
    fec[i].reclen = fmtsizeof(fmt)*fec[i].len;
    if ((fec[i].data=(char *)calloc(1,fec[i].reclen)) == NULL)
    {
      printf("No room to allocate %d bytes\n>",fec[i].reclen);
    }
    /* allocate archive data storage */
    if ((fec[i].archdata=(char *)calloc(1,fec[i].reclen)) == NULL)
    {
      printf("No room to allocate %d bytes\n>",fec[i].reclen);
    }
    strncpy(scratch,colptr(heartbeat_col,str),32);
    fec[i].heartbeat = atoi(scratch);
    strncpy(scratch,colptr(poll_col,str),32);
    fec[i].poll = atoi(scratch);
    strncpy(scratch,colptr(amethod_col,str),48);
    if (strstr(scratch,"BEAM")) fec[i].option |= BEAM;
    if (strstr(scratch,"RUNNING")) fec[i].option |= RUNNING;
    if (strstr(scratch,"NEVER")) fec[i].option |= NEVER;
    if (strstr(scratch,"ALWAYS")) fec[i].option |= ALWAYS;
    if (strstr(scratch,"LUMI")) fec[i].option |= LUMI;
    if (strstr(scratch,"RUNNUM")) fec[i].option |= RUNNUM;
    if (strstr(scratch,"SYSTEMATIC")) fec[i].option |= SYSTEMATIC;
    if (strstr(scratch,"ONCE")) fec[i].option |= ONCE;
    if (strstr(scratch,"RAMPING")) fec[i].option |= RAMPING;
    if (strstr(scratch,"ELECTRONS")) fec[i].option |= ELECTRONS;
    if (strstr(scratch,"PROTONS")) fec[i].option |= PROTONS;
    if (strstr(scratch,"FASTRAMP")) fec[i].option |= FASTRAMP;
    if (strstr(scratch,"INJECTION")) fec[i].option |= INJECTION;
    if (strstr(scratch,"FASTINJ")) fec[i].option |= FASTINJ;
    if (strstr(scratch,"SLOW")) fec[i].option |= SLOW;
    if (strstr(scratch,"FAST") &&
        !(fec[i].option & (FASTRAMP|FASTINJ))) fec[i].option |= FAST;
    if (strstr(scratch,"FIXTIME")) fec[i].option |= FIXTIMESTAMP;
    strncpy(fec[i].sfile,colptr(sfile_col,str),32); /*strlwr(fec[i].sfile); */
    strncpy(fec[i].responsible,colptr(resp_col,str),40);
    if (inf_col >= 0)
    {
      strncpy(fec[i].infile,colptr(inf_col,str),32); /*strlwr(fec[i].infile); */
    }
    if (cntxt_col >= 0)
    {
      strncpy(fec[i].cntxt,colptr(cntxt_col,str),16); 
    }
    i++;
  }
err:fclose(fp);
  return (i-offset);
}
void logRestart(void)        /* establish file names (1 file per month) */
{
  time_t tnow=time(NULL);
  struct tm *lt=localtime(&tnow);
  int i, mm, yy;
  char fileName[128];

  mm = lt->tm_mon+1; yy = lt->tm_year;

  for (i=0; i<totalNrFECs; i++)
  {
    sprintf(fec[i].fname,"ar%02d%02d01.%0x",yy,mm,fec[i].rectyp);
    fec[i].archYear = 1900 + yy;
    fec[i].archMonth = mm;
  }
  /* make sure the directory structure is there ... */
  sprintf(fileName,"%s/%04d/%02d/%s",
    ArchiveDatapath,fec[0].archYear,fec[0].archMonth,fec[0].fname);
  makeDirPath(fileName);
}
int makeDirPath(char *fname)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WINNT)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,"/"); c != NULL; c=strtok(NULL,"/"))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return 0;
    }
    strcat(dirname,"/");
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH)) return 0;
    }
#   elif defined(WINNT)
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return 0;
    }
#   endif
  }
  return -1;
}
int WriteRecord(int i)  /* write record #i to file */
{
  int hFile,trecsiz,fsiz,rsiz;
  char fileName[128],s[128];

  if (g_isTerminating) return 0;
  fec[i].archPending = 0;
  fec[i].setArchPending = 0;
  if (archPending > 0) archPending--;
  if (fec[i].lastarchtime == fec[i].archtime) return 0; /* no reason */
  sprintf(fileName,"%s/%04d/%02d/%s",
    ArchiveDatapath,fec[i].archYear,fec[i].archMonth,fec[i].fname);
  trecsiz = sizeof(time_t) + fec[i].reclen;
# if defined(UNIX)
  if ( (hFile=open(fileName,O_RDWR|O_CREAT|O_APPEND,S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH)) == -1 )
# elif defined(WINNT)
  if ( (hFile=open(fileName,O_RDWR|O_CREAT|O_APPEND)) == -1 )
# endif
  {
    if (archdebug) 
    {
      perror(fileName);
      sprintf(s,"error opening file %s\n",fileName);
      ttyoutput(s);
    }
    return -1;
  } 
  fsiz = lseek(hFile,0L,SEEK_END);
  if ((rsiz=(fsiz%trecsiz)) != 0)
  {
    sprintf(s,"re-align archive file %s %s (corrupt?) <recno %d>",
            fec[i].eqm,fec[i].property,fec[i].rectyp);
    AppendProblemsLog(s);
    lseek(hFile,fsiz-rsiz,SEEK_SET);
  } 
  write(hFile,(char *)&fec[i].archtime,sizeof(time_t));
  write(hFile,fec[i].archdata,fec[i].reclen);
  close(hFile);
  fec[i].lastarchtime = fec[i].archtime;
  if (archdebug)
  {
    sprintf(s,"archiving %s (%ld)",fec[i].eqm,fec[i].archtime);
    ttyoutput(s);
  }
  return 0;
}
int ReadRecord(int i)  /* read most recent record #i from file */
{
  int hFile, recsize = fec[i].reclen+sizeof(time_t);
  long nrecs,p2;
  char s[128];

  if ( (hFile=OpenDataFile(i)) == -1 )
  {
    if (archdebug) 
    {  
      perror(fec[i].fname);
      sprintf(s,"error opening file %s for reading\n",fec[i].fname);
      ttyoutput(s);
    }
    return -1;
  }

  nrecs = lseek(hFile,0,SEEK_END)/recsize;
  p2 = (nrecs-1)*recsize;
  lseek(hFile,p2,SEEK_SET);
  read(hFile,(char *)&fec[i].tstamp,sizeof(time_t));
  read(hFile,fec[i].archdata,fec[i].reclen);
  if (fec[i].data != NULL)
    memcpy(fec[i].data,fec[i].archdata,fec[i].reclen);
  fec[i].archtime = fec[i].tstamp;

  close(hFile);
  return 0;

}
long locateRecord(int i,time_t t)
{
  time_t t0,t1,t2;
  long nrecs,p0,p1,p2;
  double f;
  int recsize = fec[i].reclen+sizeof(time_t);
  int hFile;
  char s[128];
  int it=0;

  if ( (hFile=OpenDataFile(i)) == -1 ) 
  {
    if (archdebug) 
    {
      perror(fec[i].fname);
      sprintf(s,"error opening file %s\n",fec[i].fname);
      ttyoutput(s);
    }
    return -1;
  }

  nrecs = lseek(hFile,0,SEEK_END)/recsize;
  p0 = p1 = 0; p2 = (nrecs-1)*recsize;
  lseek(hFile,p0,SEEK_SET); read(hFile,&t0,sizeof(time_t));
  if (t < t0) goto out;
  lseek(hFile,p2,SEEK_SET); read(hFile,&t2,sizeof(time_t));

  while (p2 > p0 + recsize)
  {
    if (t > (t1=t2)) { p1 = p2; break; }
    while (t1 != t0)   /* seek <-- */
    {
      f = ((double)(t-t0)/(double)(t1-t0))*((double)(p2-p0)/recsize);
      if ((p1=p0+recsize*(long)f) >= p2) break;
      lseek(hFile,p1,SEEK_SET); read(hFile,&t1,sizeof(time_t));
      if (t1 < t) break;
      p2 = p1;
      t2 = t1;
      it++;
    }
    t0 = t1; p0 = p1;
    while (t2 != t1)  /* seek --> */
    {
      f = ((double)(t2-t)/(double)(t2-t1))*((double)(p2-p0)/recsize);
      if ((p1=p2-recsize*(long)f) <= p0) break;
      lseek(hFile,p1,SEEK_SET); read(hFile,&t1,sizeof(time_t));
      if (t1 > t) break;
      p0 = p1; t0 = t1;
      it++;
    }
    p2 = p1; t2 = t1;
  }
  if (debugon) printf("%d iterations\n",it);
out: close(hFile);
  return p1;
}
