#ifndef _ARCHIVE_H_
#define _ARCHIVE_H_

#define ARCHIVEDATAPATHDEPTH 3
extern char ArchiveDBpath[];
extern char *ArchiveDatapath;
#define MAX_ADBFILES 20
extern char adbfile[MAX_ADBFILES][32];
extern int nadbfiles;

#define NDRIVES 4
#define u16 unsigned short
#define u32 unsigned long
#define MAX_FECS  50
#define NETMEX_BUF_SIZE 8192

#define NEVER      0x0001
#define BEAM       0x0002
#define RUNNING    0x0004
#define ALWAYS     0x0008
#define RUNNUM     0x0010
#define SYSTEMATIC 0x0020
#define FAST       0x0020
#define LUMI       0x0040
#define ONCE       0x0080
#define GLOBAL     0x0100
#define RAMPING    0x0200
#define ELECTRONS  0x0400
#define PROTONS    0x0800
#define FASTRAMP   0x1000
#define INJECTION  0x2000
#define FASTINJ    0x4000
#define SLOW       0x8000
#define FIXTIMESTAMP 0x10000

int getEntryFromKeyword(char *keyword,int *rectyp,int *param);
int archLoad(char *fname, int offset);
int paramLoad(int ntotal, int offset);
void fillinArrayTypes(int num);
int getparam(char *s, char *param, u16 *fmt, int *size);
extern int nfecs, nparam;
int handler(int deverr,int doserr, int bp,int si);
int addADBtoList(char *fn);

typedef struct
{
  char param[32];       /* parameter name */
  char boundTo[32];     /* bound to parameter */
  char subsys[16];      /* parameter subsystem */
  u16 format;           /* allowable format */
  u16 arrayType;        /* array type */
  int size;             /* data array size assuming given format */
  int offset;           /* array starting offset in bytes */
  double maximum;       /* maximum value for parameter */
  double minimum;       /* minimun value for parameter */
  double aTolerance;    /* absolute tolerance in units given */
  float  pTolerance;    /* percent tolerance */
  char units[32];       /* units */
  char desc[48];        /* description */
  char plot[16];        /* plot style (LOG, LIN, etc.) */
  time_t timestamp;     /* most recent timestamp */
  short depth;
  short ptr;
  int rate;
  float scaleFactor;
  float scaleOffset;
  float *data;
} FECDATA;

typedef struct
{
  u16  rectyp;            /* entry in record table */
  u32  tstamp;            /* timestamp corresponding to data */
  int  dataFlg;           /* 1 = new data; 0 = old data; < 0 = error */
  u16  fecStatus;         /* 0 = okay; -1 = no response; else hardware status */
  u16  errCounter;        /* number current consecutive errors */
  char cntxt[16];         /* calling context */
  char eqm[16];           /* calling eqpipment module */
  char property[32];      /* calling property */
  char devname[32];       /* calling device name */
  char *data;             /* pointer to data; */
  char *archdata;         /* pointer to most recent archived data */
  short *indata;          /* pointer to input data (if any); */
  short nindata;          /* number of input data points */
  u16  EqNo,              /* Equipment Number of RPC call */
       len,               /* length of RPC call */
       reclen,            /* total number of bytes in call */
       fmt,               /* RPC call format */
       heartbeat,         /* archiving heartbeat (not to be exceeded) */
       poll;              /* polling rate */
  char sfile[32];         /* structure file name */
  char infile[32];        /* input file name */
  int  nparam;            /* number of parameters in structure */
  u32  option;            /* option word */
  u16  active;            /* is table entry active */
  u16  pbeam;             /* last p beam flag for filter */
  u16  ebeam;             /* last e beam flag for filter */
  char responsible[40];   /* persons responsible for the FEC application */
  FECDATA *fecdata;       /* pointer to data structure information */
  time_t archtime;        /* timestamp of archived data */
  time_t lastarchtime;    /* timestamp of last written archived data */
  int  archPending;       /* flag to flush archive data to file */
  int  setArchPending;    /* flag used by bound keywords */
  int  archYear;          /* the archive year */
  int  archMonth;         /* the archive month */
  char fname[16];
} FEC;

extern int nfecs, nparam;

typedef struct
{
  char alias[32];
  char name[32];
} PRP_ALIAS;
extern int nPrpAlias;
extern PRP_ALIAS *prpAlias;

typedef struct
{
  int number;
  char *name;
} SUB_SYS;
extern int nSubSystems;
extern SUB_SYS *subSystems;
typedef struct
{
  char dbfile[10][16];
  int  ndbfiles;
  int  acquire;
  int  port;
  char name[16];
} TSRVINFO;

typedef struct   /* Archive Request Structure */
{
  time_t starttime;
  short  index,
         minutes;
  char   keyword[16];
} ARS;

typedef struct
{
  char name[16];
  char beamtype[16];
  int  fmt;
  void *ptr;
  int fecindex;
  int typeindex;
} CHECKNAME;

extern int p_stateindex;
extern int e_stateindex;

extern char gTargetStateServer[64];

typedef struct
{
  char typestr[32];
  char procstr[32];
  char trigstr[32];
} DECLPROC;

extern DECLPROC *eDeclaredProcedures;
extern int nEDeclaredProcedures;
extern DECLPROC *pDeclaredProcedures;
extern int nPDeclaredProcedures;

#endif  /* _ARCHIVE_H_ */

