/*  =============================================================
   ==
   ==   Dataserver routine: archlib.c
   ==
   ============================================================= */

#include "prolog.h"
#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "rpctypes.h"
#include "archive.h"
#if defined(UNIX)
# include "unixcmd.h"
#elif defined(WINNT)
# include "ntcmd.h"
#endif

extern float HEenergy;
extern FEC *fec;
int nfecs = 0, nparam = 0, nmki = 0;
int MKIoffset;
extern char cbuff[NETMEX_BUF_SIZE];
extern char dpath[];
int hMKIfile = -1;                                    /* Handle to MKI file */
int hPKTRfile = -1;                                   /* Handle to PKTR file */
char HEfileName[80];
char HPfileName[80];
extern int archPending,totalNrFECs;
extern int archdebug, ramping, running, beam;
extern u16 Eramping,Pramping,Erunning,Prunning,Ebeam,Pbeam;
extern int DclEramping, DclPramping, DclLumi;
extern int DclEInj, DclPInj;
char *_strdate(char *buf);
char *_strtime(char *buf);
int WriteRecord(int i);

int getEntryFromKeywordXX(char *keyword,int *rectyp,int *param)
{
  FECDATA *fecdata;
  int i,j;
 
  for (i=0; i<nfecs; i++)
  {
    if (!fec[i].active) continue;
    fecdata = fec[i].fecdata;
    for (j=0; j<fec[i].nparam; j++)
    {
      if (!strcmp(fecdata[j].param,keyword))
      {
        if (rectyp != NULL) *rectyp = fec[i].rectyp;
        if (param != NULL) *param  = j;
        return i;
      }
    }
  }
  return -1;
}
char *makedate(char *s)  /* Make a date string in the unusual NETMEX format */
{
  char timebuf[9], datebuf[9];

  _strdate(datebuf);
  sprintf(s,"19YY-MM-DD-%s",_strtime(timebuf));
  strncpy(&s[2],&datebuf[6],2); strncpy(&s[5],&datebuf[0],2);
  strncpy(&s[8],&datebuf[3],2);

  return s;
}
void outputData(void)
{
  int i;
  for (i=0; i<totalNrFECs; i++) if (fec[i].archPending) WriteRecord(i);
}
#define BYTE unsigned char
int checkData(int n) /* write data to archive file for given fec entry */
{
  time_t timstamp = time(NULL);
  int i,j,k,arch = 0,dsize;
  UINT16 opt = (UINT16)fec[n].option;
  double t;
  char *d,*a,s[128];
  u16 EbeamL, PbeamL;
  int isTransition = 0;
  SINT32 transitionTimestamp;

  if (opt&NEVER) return 0;                            /* no archive option */
  if (fec[n].tstamp <= (UINT32)fec[n].archtime) 
  {
    if (archdebug)
    {
      sprintf(s,"archive timestamp unchanged for %s/%s %s (%d vs %d)\n",fec[n].eqm,fec[n].devname,fec[n].property,fec[n].tstamp,fec[n].archtime);
      ttyoutput(s);
    }
    return 0;     			    /* data timestamp hasn't changed */
  }
  if (fec[n].setArchPending)
  {
    arch = 1;
    goto checkArchiveFlag;
  }
  if (fec[n].archPending)
  {
    if (archdebug)
    {
      sprintf(s,"archive still pending for %s/%s %s\n",fec[n].eqm,fec[n].devname,fec[n].property);
      ttyoutput(s);
      return -1;                	   	    /* archive is still pending */
    }
  }
  if (timstamp >= fec[n].archtime + fec[n].heartbeat) 
  {
    if (archdebug)
    {
       sprintf(s,"archive heartbeat for %s/%s %s\n",fec[n].eqm,fec[n].devname,fec[n].property);
       ttyoutput(s);
    }
    arch = 1;/* archive no matter what */
  }
  else
  {
    /* archive only if maching is running ... */
    if (!running && (opt&RUNNING)) return 0;  
    if (!Erunning && (opt&RUNNING) && (opt&ELECTRONS)) return 0;  
    if (!Prunning && (opt&RUNNING) && (opt&PROTONS)) return 0;  
    /* archive only if beam ... */
    EbeamL = fec[n].ebeam; PbeamL = fec[n].pbeam;
    if (!beam && !EbeamL && !PbeamL && (opt&BEAM)) return 0;
    if (!Ebeam && !EbeamL && (opt&BEAM) && (opt&ELECTRONS)) return 0;
    if (!Pbeam && !PbeamL && (opt&BEAM) && (opt&PROTONS)) return 0;
    if ((Ebeam != EbeamL) || (Pbeam != PbeamL)) isTransition = 1;
    fec[n].ebeam = Ebeam; fec[n].pbeam = Pbeam;
    /* archive only if ramping ... */
    if (!DclEramping && (opt&RAMPING) && (opt&ELECTRONS)) return 0;
    if (!DclPramping && (opt&RAMPING) && (opt&PROTONS)) return 0;
    /* archive only if injection */
    if (!DclEInj && (opt&INJECTION) && (opt&ELECTRONS)) return 0;
    if (!DclPInj && (opt&INJECTION) && (opt&PROTONS)) return 0;
  }

  if (archdebug)
  {
       sprintf(s,"checking tolerance for %s/%s %s\n",fec[n].eqm,fec[n].devname,fec[n].property);
       ttyoutput(s);
  }
  for (i=0; !arch && i<fec[n].nparam; i++)
  {
    d = fec[n].data + fec[n].fecdata[i].offset;
    a = fec[n].archdata + fec[n].fecdata[i].offset;
    switch (fec[n].fecdata[i].format)
    {
      case CF_DOUBLE:
           for (j=0; !arch && j<fec[n].fecdata[i].size; j++)
           {
             t = fec[n].fecdata[i].aTolerance +
                 fec[n].fecdata[i].pTolerance * fabs(((double *)a)[j]);
             if ( fabs(((double *)d)[j] - ((double *)a)[j]) > t) arch = 1;
           }
           break;
      case CF_FLOAT:
           for (j=0; !arch && j<fec[n].fecdata[i].size; j++)
           {
             t = fec[n].fecdata[i].aTolerance +
                 fec[n].fecdata[i].pTolerance * fabs(((float *)a)[j]);
             if ( fabs(((float *)d)[j] - ((float *)a)[j]) > t) arch = 1;
           }
           break;
      case CF_INTEGER:
           for (j=0; !arch && j<fec[n].fecdata[i].size; j++)
           {
             t = fec[n].fecdata[i].aTolerance +
                 fec[n].fecdata[i].pTolerance * abs(((short *)a)[j]);
             if ( abs(((short *)d)[j] - ((short *)a)[j]) > t) arch = 1;
           }
           break;
      case CF_LONG:
           for (j=0; !arch && j<fec[n].fecdata[i].size; j++)
           {
             t = fec[n].fecdata[i].aTolerance +
                 fec[n].fecdata[i].pTolerance * labs(((SINT32 *)a)[j]);
             if ( labs(((SINT32 *)d)[j] - ((SINT32 *)a)[j]) > t) arch = 1;
           }
           break;
      case CF_BYTE:
           for (j=0; !arch && j<fec[n].fecdata[i].size; j++)
           {
             t = fec[n].fecdata[i].aTolerance +
                 fec[n].fecdata[i].pTolerance * abs(((BYTE *)a)[j]);
             if ( abs(((BYTE *)d)[j] - ((BYTE *)a)[j]) > t) arch = 1;
           }
           break;
      default:
           dsize = fmtsizeof(fec[n].fecdata[i].format)*fec[n].fecdata[i].size;
           if (!arch && memcmp(d,a,dsize)) arch = 1;
           break;
    }
  }
checkArchiveFlag:
  if (arch)
  {
    if (archdebug)
    {
      sprintf(s,"signaling reason to archive %s/%s %s\n",fec[n].eqm,fec[n].devname,fec[n].property);
      ttyoutput(s);
    }
    if (isTransition && fec[n].archtime != 0)
    {
      transitionTimestamp = fec[n].tstamp - (fec[n].poll/1000);
      if (fec[n].archtime != transitionTimestamp)
      {
        fec[n].archtime = transitionTimestamp;
        WriteRecord(n);
        if (archdebug)
        {
          sprintf(s,"inserting transition record\n");
          ttyoutput(s);
        }
      }
    }
    memcpy(fec[n].archdata,fec[n].data,fec[n].reclen);
    fec[n].archtime = fec[n].tstamp;
    fec[n].archPending = 1;
    fec[n].setArchPending = 0;
    /* is this bound to another keyword ? */
    if (strlen(fec[n].fecdata[0].boundTo) > 0)
    {
      if ((k=getEntryFromKeyword(fec[n].fecdata[0].boundTo,NULL,NULL)) >= 0 &&
          !fec[k].archPending)
      {
        fec[k].archPending = 1;
        fec[k].setArchPending = 1;
        if (fec[k].archtime != (SINT32)fec[k].tstamp) /* it was dropped, put it back */
        {
          fec[k].archtime = fec[k].tstamp;
          memcpy(fec[k].archdata,fec[k].data,fec[k].reclen);
        }
      }
    }
    archPending++;
  }
  return 0;
}

