#include "prolog.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include "fecerr.h"
#include "rpctypes.h"
#include "database.h"
#include "brdcast.h"
#include "almsrv.h"
#include "errors.h"
#include "archive.h"
#include "toolkit.h"
#include "srvip.h"
#include "statebase.h"
#if defined(UNIX)
# include <dirent.h>
# include "unixcmd.h"
#elif defined(WINNT)
# include <direct.h>
# include "ntcmd.h"
#endif

int inputDataLoad(int ntotal, int offset);
int reacquireStaticDataSet(int i);
void AppendProblemsLog(char *str);
void updateStateInfo(int i,int j,int cc);
int overloadEqpmod(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
extern void getEPenergy(int id,int cc);
extern int NGdebug;
extern int MaxRPCAllocation;
extern int daylite, hPKTRfile,hMKIfile, RequireAcknowledgments;
FEC *fec;
int g_isTerminating = FALSE;
int checkData(int n);
int writedata(int i, int len);
int ReadRecord(int i);
void flcopy(long  *dst, long *src, int num);
char *_strdate(char *buf);
char *_strtime(char *buf);
int findDaylight(void);
int prepdata(char *buf);
int getNrActiveFecs(char *fname);
void writeTask(void);
void getFecInfo(void);
void flushdata(void);
void magnetinit(void);
void ReadMagnetFile(void);
void outputData(void);
int wrtFlg,archdebug=0;
int totalNrFECs = 0;

#define WRITE     -1
#define READ       1
#define MAXFILES   100

/* state variables: */
int ramping = 0, running = 0, beam = 0, lumi = 0;
u16 Pramping = 0, Eramping = 0, Pbeam = 0, Ebeam = 0;
u16 Prunning = 0, Erunning = 0;
int DclRun = 0;
int DclEramping = 0,DclPramping = 0,DclLumi = 0,DclEInj = 0,DclPInj = 0;

int checkStates = FALSE;

typedef struct
{
  int id;
  int startPending;
} LINKSTRCT;
LINKSTRCT *fid;

#define logerr(txt) { perror("open"); strcpy(s,txt); goto error; }

/*============= General archiving globals =========================== */

#define GLOBALS_FILENAME "nglobals.csv"
int registerGlobals = 0, brdcstRegistered = FALSE;
int globalsServer = FALSE, globalsWatchdog = FALSE;
int archPending = 0;

FILE *wfp, *rfp;
int Hwfp;
void pollTask(void);
void logRestart(void);
int readNETMEX(char *fname,int FECoffset, int num, time_t *lasttime);
int _filer(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
int GlobalsRegistrationEx(void);

char sbuf[20];
char timebuf[9], datebuf[9];

int fileNo = 0;

GlobalListStruct *globalList;
int ngloballist = 0;

void cbackBrdsrv(int i,int cc)
{
}
int getNetGlobalList(char *fname)
{
  char fn[80];
  char str[256],scratch[16];
  int done = 0,i,cc = 0;
  int key_col,fmt_col,siz_col,pmax_col,pmin_col,tol_col;
  FILE *fp = NULL;
 
  sprintf(fn,"%s",fname);
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (i=0; fgets(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    i++;
  } 
  if ((globalList=(GlobalListStruct *)calloc(i,sizeof(GlobalListStruct))) == NULL) ccerr(out_of_server_memory);
  rewind(fp);
  for (i=0; fgets(str,255,fp) != NULL; )   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((key_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(str,"FORMAT")) < 0) ccerr(no_such_line);
      if ((siz_col=findcol(str,"SIZE")) < 0) ccerr(no_such_line);
      if ((pmax_col=findcol(str,"PERIOD_MAX")) < 0) ccerr(no_such_line);
      if ((pmin_col=findcol(str,"PERIOD_MIN")) < 0) ccerr(no_such_line);
      if ((tol_col=findcol(str,"TOLERANCE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(globalList[i].keyword,colptr(key_col,str),16);
    strncpy(scratch,colptr(siz_col,str),16);
    globalList[i].size = atoi(scratch);
    strncpy(scratch,colptr(fmt_col,str),16); strupr(scratch);
    globalList[i].format = ftoi(scratch);
    strncpy(scratch,colptr(pmax_col,str),16);
    globalList[i].maxPeriod = atoi(scratch);
    strncpy(scratch,colptr(pmin_col,str),16);
    globalList[i].minPeriod = atoi(scratch);
    strncpy(scratch,colptr(tol_col,str),16);
    globalList[i].tolerance = strtod(scratch,NULL);
    i++;
  }
  ngloballist = i;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}  
NAME16 redirfile[10];
NAME16 redirsrv[10];
int nredirfiles = 0;
int getRedirInfo(void)
{
  FILE *fp;
  char str[256];
  int i=0,done = 0,cc = 0;
  int database_col,name_col;

  if ((fp = fopen("redir.csv","rt")) == NULL) return -1;
  while (fgets(str,255,fp) != NULL )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((database_col=findcol(str,"DATABASE")) < 0) ccerr(no_such_line);
      if ((name_col=findcol(str,"SERVERNAME")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(redirfile[i].name,colptr(database_col,str),16);
    strncpy(redirsrv[i].name,colptr(name_col,str),16);
    if (++i >= 10) break;
  }
err:
  return i;
}

TSRVINFO srvinfo;
int getStartupInfo(void)
{
  FILE *fp=NULL;
  char str[256],scratch[170],*c;
  int i,done = 0,cc = 0;
  int database_col,method_col,state_col;

  if ((fp=fopen("startup.csv","rt")) == NULL)
  {
    printf("\n Can't open startup data base file : startup.csv\n");
    exit(1);
  }

  while (fgets(str,255,fp) != NULL )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((database_col=findcol(str,"DATABASE")) < 0) ccerr(no_such_line);
      method_col = findcol(str,"METHOD");
      state_col = findcol(str,"CHECKSTATE");
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(database_col,str),170);
    for (i=0,c=strtok(scratch," "); i<10 && c != NULL; i++,c=strtok(NULL," "))
    {
      strncpy(srvinfo.dbfile[i],c,16);
    }
    srvinfo.ndbfiles = i;
    srvinfo.acquire = -1;
    if (method_col >= 0)
    {
      strncpy(scratch,colptr(method_col,str),16);
      srvinfo.acquire = strcmp(scratch,"PASSIVE");
    }
    if (state_col >= 0)
    {
      strncpy(scratch,colptr(state_col,str),16);
      strupr(scratch);
      if (strstr(scratch,"TRUE") != NULL) checkStates = TRUE;
    }
    done = 2;
    break;
  }      
err:
  if (fp != NULL) fclose(fp);
  return done == 2 ? 0 : -1;
}
void AppendProblemsLog(char *str)
{
  time_t t = time(NULL);
  char s[80];
  FILE *fp = NULL;

  if ((fp=fopen("problems.log","a")) == NULL) logerr("LOG: open");
  s[0] = 0;
  strncpy(s,str,78);
  if (s[strlen(s)-1] == '\n') s[strlen(s)-1] = 0;
  fprintf(fp,"%s : %s",s,ctime(&t));

error: 
  if (fp) fclose(fp);
  return;
}
void flcopy(long *dst, long *src, int num)
{
  while (num--) *dst++ = *src++;
  return;
}
void wrCopy(u16 *src,double *dst, int nwords)
{
  while (nwords--) *dst++ = *src++;
  return;
}
extern UINT32 TCPcount,UDPcount,SPXcount,IPXcount;
extern int nconsumer, ncontract;

extern NAME16 allowedNullStates[];
extern int nAllowedNullStates;

int pneedsToDeclare = FALSE;
int eneedsToDeclare = FALSE;
int p_stateindex = 0;
int e_stateindex = 0;

#define NULLVALUE 0.01
#define GRACE_PERIOD 600

int DeclareState(int stateindex, int typeindex)
{
  int i, cc=link_not_open, retries = 3;
  char dn[64];
  static time_t lastdecl = 0;
  time_t t = time(NULL);
  
  sprintf(dn,"%s/%s",gTargetStateServer,
    typeindex == 0 ? eDeclaredProcedures[stateindex].trigstr : 
                     pDeclaredProcedures[stateindex].trigstr );

  for (i=0; i<retries && cc != 0; i++)
  {
    cc = ExecLinkEx(dn,"STATECHANGE",NULL,NULL,CA_WRITE,400);
  }

  feclog("declare state %s : %s",dn,erlst[cc]);
  if (cc == 0)
  {
    lastdecl = t;
  }
  return cc;
}
void checkStateInfo(void)
{
  int i,j,k,nullcnt;
  static int done = 0;
  static int nparam;
  static int p_lastindex = 0, e_lastindex = 0;
  
  /* check null state consistency */
  if (!done)
  {
    done = TRUE;
    if (nCheckedStates == 0) nparam = 0;
    nparam = (nCheckedStates == 0) ? 0 : nCheckedParameters/nCheckedStates;
    for (k=0; k<nCheckedParameters; k++)
    {
      for (i=0; i<nfecs; i++) for (j=0; j<fec[i].nparam; j++)
        if (!strcmp(fec[i].fecdata[j].param,checkedParameters[k].name))
      {
        checkedParameters[k].ptr = fec[i].data + fec[i].fecdata[j].offset;
        checkedParameters[k].fmt = fec[i].fecdata[j].format;
        checkedParameters[k].fecindex = i;
        j = fec[i].nparam; i = nfecs; /* found it, get out of inner loops .. */
        break;
      }
    }
    for (k=0; k<nCheckedStates; k++)
    {
      for (i=0; i<nfecs; i++) for (j=0; j<fec[i].nparam; j++)
        if (!strcmp(fec[i].fecdata[j].param,checkedStates[k].name))
      {
        checkedStates[k].ptr = fec[i].data + fec[i].fecdata[j].offset;
        checkedStates[k].fecindex = i; 
        j = fec[i].nparam; i = nfecs; /* found it, get out of inner loops .. */
        break;
      }
    }
  }
  for (i=0; i<nCheckedStates; i++)
  {
    if (fec[checkedStates[i].fecindex].fecStatus != 0) continue;
    if (checkedStates[i].ptr == NULL) continue;
    for (j=0; j<nAllowedNullStates; j++)
    {
      if (!strncmp((char *)checkedStates[i].ptr,allowedNullStates[j].name,16)) break;
    }
    if (j < nAllowedNullStates) continue;  /* I'm in an allowed null state */
    nullcnt = 0;
    for (j=i*nparam; j<i*nparam+nparam; j++)
    {
      if (fec[checkedParameters[j].fecindex].fecStatus != 0) continue;
      switch (checkedParameters[j].fmt)
      {
        case CF_SHORT:
          if (*((short *)checkedParameters[j].ptr) <= NULLVALUE) nullcnt++;
          break;
        case CF_LONG:
          if (*((UINT32 *)checkedParameters[j].ptr) <= NULLVALUE) nullcnt++;
          break;
        case CF_FLOAT:
          if (*((float *)checkedParameters[j].ptr) <= NULLVALUE) nullcnt++;
          break;
        case CF_DOUBLE:
          if (*((double *)checkedParameters[j].ptr) <= NULLVALUE) nullcnt++;
          break;
        default: break;
      }
    }
    if (nCheckedStates > 0 && nullcnt == nparam)
    {
      DeclareState(0,checkedStates[i].typeindex); /* set to undefined */
    }
  }

  /* declare states from inferred machine values */
  if (p_stateindex != p_lastindex || pneedsToDeclare)
  {
    p_lastindex = p_stateindex;
    DeclareState(p_stateindex,1);
    pneedsToDeclare = FALSE;
  }
  if (e_stateindex != e_lastindex || eneedsToDeclare)
  {
    e_lastindex = e_stateindex;
    DeclareState(e_stateindex,0);
    eneedsToDeclare = FALSE;
  }
}

void filertask(void)
{
  pollTask(); /* check up on things */
  if (archPending) outputData();
  if (checkStates) checkStateInfo();
  return;
}
int getFecData(DTYPE *d,int i,int j,int k)
{
  int n,p,depth;
  int fmtsiz = fmtsizeof(d->dFormat);
  if (d->dFormat != fec[i].fecdata[j].format) return illegal_format;
  depth = fec[i].fecdata[j].depth;
  if (d->dArrayLength != (UINT32)depth) return illegal_read_write;
  p = fec[i].fecdata[j].ptr;
  for (n=0; n<depth; n++)
  {
     p = (++p)%depth;
     memcpy(d->data.fptr+n,&fec[i].fecdata[j].data[k*depth+p],(size_t)fmtsiz);
  }
  return 0;
}
void putFecdata(int i,int j)
{
  int n,k,p1,p2,nsteps,fmtsiz,depth;
  
  if (fec[i].fecdata[j].format != CF_FLOAT) return;
  fmtsiz = fmtsizeof(fec[i].fecdata[j].format);
  depth = fec[i].fecdata[j].depth;
  nsteps = MIN(RPCtimestamp - fec[i].fecdata[j].timestamp,depth-1);
  p1 = fec[i].fecdata[j].ptr;
  fec[i].fecdata[j].ptr = (fec[i].fecdata[j].ptr+nsteps)%depth;
  p2 = fec[i].fecdata[j].ptr;
  fec[i].fecdata[j].timestamp = RPCtimestamp;
  for (n=0; n<fec[i].fecdata[j].size; n++)
  {
     memcpy(&fec[i].fecdata[j].data[n*depth+p2],fec[i].data+fec[i].fecdata[j].offset+fmtsiz*n,(size_t)fmtsiz);
     if (nsteps > 1)
     {
       for (k=(p2>p1 ? p2-1 : p2+depth-1); k>p1; k--)
       {
         memset(&fec[i].fecdata[j].data[n*depth + k%depth],0,(size_t)fmtsiz);       
       }
     }
  }
}
void addOffsetToData(char *data,FECDATA *f)
{
  int i;
  char *dptr;

  dptr = (char *)(data + f->offset);
  switch (f->format)
  {
    case CF_SHORT:
      for (i=0; i<f->size; i++) ((short *)dptr)[i] += (short)f->scaleOffset;
      return;
    case CF_LONG:
      for (i=0; i<f->size; i++) ((UINT32 *)dptr)[i] += (UINT32)f->scaleOffset;
      return;
    case CF_FLOAT:
      for (i=0; i<f->size; i++) ((float *)dptr)[i] += (float)f->scaleOffset;
      return;
    case CF_DOUBLE:
      for (i=0; i<f->size; i++) ((double *)dptr)[i] += (double)f->scaleOffset;
      return;
  }
  return;
}
void scaleFactorToData(char *data,FECDATA *f)
{
  int i;
  char *dptr;

  dptr = (char *)(data + f->offset);
  switch (f->format)
  {
    case CF_SHORT:
      for (i=0; i<f->size; i++) ((short *)dptr)[i] *= (short)f->scaleFactor;
      return;
    case CF_LONG:
      for (i=0; i<f->size; i++) ((UINT32 *)dptr)[i] *= (UINT32)f->scaleFactor;
      return;
    case CF_FLOAT:
      for (i=0; i<f->size; i++) ((float *)dptr)[i] *= (float)f->scaleFactor;
      return;
    case CF_DOUBLE:
      for (i=0; i<f->size; i++) ((double *)dptr)[i] *= (double)f->scaleFactor;
      return;
  }
  return;
}
void dataCallback(int i,int cc)
{
  time_t t=time(NULL);
  int j, k;
  int checkArchive = 0;
  DTYPE din;
  UINT16 opt;
  static int lastPEnergy;
  static float lastEEnergy;
  /* monitor crucial parameters */
  static int PEpanic = 0, EEpanic = 0,PIpanic = 0; 
  static int STpanic = 0;
  static int globalsRegistered = FALSE;      
  char s[128];
  char namProperty[32];

  i--;
  if (i < 0 || i >= nfecs) ccerr(invalid_index);
  opt = (UINT16)fec[i].option;
  
  if (cc & CE_SENDDATA) cc = 0;
  if (opt&ONCE) 
  {
    if (cc == 0) 
    {
      Cancel(fid[i].id);
      fid[i].id = 0; /* recycle it */
    }
    else return;
  }
  fid[i].startPending = FALSE;

  if (archdebug > 1)
  {
    sprintf(s,"LINK (%d) %s %s : %s <%d> (%ld)\n>",
                  i,fec[i].eqm,fec[i].property,
                  cc?"not responding":"incoming",cc,RPCtimestamp);
    ttyoutput(s);
  }
  if (cc == 0 && fec[i].fecStatus == link_not_open && fec[i].errCounter > 2)
  {                                      /* wasn't okay but now it is ... */
    for (j=0; fec[i].len > 0 && j<fec[i].nparam; j++)
    {
      if (fec[i].fecdata[j].size <= 1) continue;
      if (strstr(fec[i].fecdata[j].param,".NAM") != NULL) continue;
      strncpy(namProperty,fec[i].fecdata[j].param,27);
      strcat(namProperty,".NAM");
      if ((k=getEntryFromKeyword(namProperty,NULL,NULL)) >= 0)
      {
        reacquireStaticDataSet(k);
      }
    }
  }
  fec[i].fecStatus = cc;
  fec[i].dataFlg = cc ? -cc : 1;
  if (cc) fec[i].errCounter++; else fec[i].errCounter = 0;
  for (j=0; j<fec[i].nparam; j++)
  {
    if (fec[i].fecdata[j].scaleOffset != 0) addOffsetToData(fec[i].data,&fec[i].fecdata[j]);
    if (fec[i].fecdata[j].scaleFactor != 1) scaleFactorToData(fec[i].data,&fec[i].fecdata[j]);
    if (fec[i].fecdata[j].data) /* keep history buffer */
    {
      if (!cc) putFecdata(i,j);
    }
    updateStateInfo(i,j,cc);
  }
  if (fec[i].option&FIXTIMESTAMP) fec[i].tstamp = t;
  else 
  {
    if (RPCtimestamp < 100) return;
    fec[i].tstamp = RPCtimestamp;
  }

  /* systematic entries are checked at their polling rates: */
  if (cc) return;
  if      (DclRun && (opt&RUNNUM)) checkArchive = 1;
  else if (DclEramping && (opt&FASTRAMP) && (opt&ELECTRONS)) checkArchive = 1;
  else if (DclPramping && (opt&FASTRAMP) && (opt&PROTONS)) checkArchive = 1;
  else if (DclEInj && (opt&FASTINJ) && (opt&ELECTRONS)) checkArchive = 1;
  else if (DclPInj && (opt&FASTINJ) && (opt&PROTONS)) checkArchive = 1;
  else if (opt&SYSTEMATIC) checkArchive = 1;
  else if (opt&FAST) checkArchive = 1;       /* otherwise : */
  else if (t < fec[i].archtime + 10) return; /* never more often than every 10 sec */
  if ((t < fec[i].archtime + 60) && (opt&SLOW)) return;
  if (beam && (opt&BEAM)) checkArchive = 1;
  if (lumi && (opt&LUMI)) checkArchive = 1;
  if (ramping) checkArchive = 1;
  if (t > fec[i].archtime + 60) checkArchive = 1;
  if (checkArchive) checkData(i);      /* decide whether to archive or not */
  cc = 0;
  if (registerGlobals && !globalsRegistered) 
  {
    /* make sure everything is there */
    for (j=0; j<nfecs; j++) if (fid[j].startPending) return;
    globalsRegistered = TRUE;
    /* anything extra to do? */
    GlobalsRegistrationEx();
    /* check what I know about: */
    for (i=0; i<nfecs; i++) for (j=0; j<fec[i].nparam; j++)
      for (k=0; k<ngloballist; k++)
       if (!strcmp(fec[i].fecdata[j].param,globalList[k].keyword))
    {
      /* tie it directly to the incoming data: */
      din.dFormat = fec[i].fecdata[j].format;
      din.dArrayLength = fec[i].fecdata[j].size;
      din.data.vptr = fec[i].data + fec[i].fecdata[j].offset;
      cc = sendNetGlobal(fec[i].fecdata[j].param,&din,cbackBrdsrv,
                        globalList[k].minPeriod,globalList[k].maxPeriod,
                        globalList[k].tolerance);
      feclog("register global %s : %s",
        fec[i].fecdata[j].param,cc < 0 ? erlst[-cc] : "success");
    } 
  }
err:
  if (archdebug && cc)
  {
    sprintf(s,"Callback : %s\n",erlst[cc]);
    ttyoutput(s);
  }
  return;
}
void dumpFecs(void)
{
  int i,j;
  
  for (i=0; i<nfecs; i++)
  {
    printf("%d: %s %s %s %d %d\n",i,fec[i].eqm,fec[i].property,fec[i].devname,
           fec[i].len,fec[i].reclen);
    for (j=0; j<fec[i].nparam; j++)
      printf("\t%s %d %d\n",fec[i].fecdata[j].param,fec[i].fecdata[j].format,fec[i].fecdata[j].size);
  }
}
int redirDataLink(char *fname,char *srv)
{
  int n, fmt;
  float maxval,minval;
  char str[256],scratch[16],dnamePrp[32],desc[128];
  int done = 0,cc = 0;
  int key_col,fmt_col,siz_col,uni_col,max_col,min_col,
      cmt_col;
  FILE *fp;
  DTYPE dout;

  sprintf(str,"%sinfo/%s",ArchiveDBpath,fname);
  if ((fp=fopen(str,"rt")) == NULL)
  {
    printf("%s : Can't open structure file %s\n",str);
    return -1;
  }
  for (n=0; fgets(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((key_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(str,"FORMAT")) < 0) ccerr(no_such_line);
      if ((siz_col=findcol(str,"SIZE")) < 0) ccerr(no_such_line);
      if ((uni_col=findcol(str,"UNITS")) < 0) ccerr(no_such_line);
      if ((max_col=findcol(str,"MAXIMUM")) < 0) ccerr(no_such_line);
      if ((min_col=findcol(str,"MINIMUM")) < 0) ccerr(no_such_line);
      if ((cmt_col=findcol(str,"COMMENT")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(dnamePrp,colptr(key_col,str),32);
    strncpy(scratch,colptr(fmt_col,str),16); strupr(scratch);
    if ((fmt=getFormatType(scratch)) == CF_NULL) fmt = atoi(scratch);
    dout.dFormat = fmt;
    strncpy(scratch,colptr(siz_col,str),16);
    dout.dArrayLength = atoi(scratch);
    strncpy(scratch,colptr(min_col,str),16);
    minval = (float)atof(scratch);
    strncpy(scratch,colptr(max_col,str),16);
    maxval = (float)atof(scratch);
    sprintf(desc,"[%g:%g ",minval,maxval);
    strncat(desc,colptr(uni_col,str),32);
    strcat(desc,"]");
    strncat(desc,colptr(cmt_col,str),48);
    RegisterPropertyInformation("FILER",dnamePrp,&dout,NULL,CA_READ,AT_UNKNOWN,0,desc,-1,srv);
  }
err:
  fclose(fp);
  return 0;
}
int redirDataFile(char *fname,char *srv)
{
  FILE *fp;
  char fn[80];
  char str[256],scratch[48];
  int done = 0,cc = 0;
  int active_col,sfile_col,rectype_col;

  sprintf(fn,"%s%s",ArchiveDBpath,fname);
  if ((fp = fopen(fn,"rt")) == NULL)
  {
    printf("\n Can't open data base file : %s\n",fname);
    return -1;
  }
  while (fgets(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((rectype_col=findcol(str,"RECORD TYPE")) < 0) ccerr(no_such_line);
      if ((sfile_col=findcol(str,"STRUCTURE")) < 0) ccerr(no_such_line);
      if ((active_col=findcol(str,"ACTIVE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    if (atoi(colptr(active_col,str)) == 0) continue;
    strncpy(scratch,colptr(rectype_col,str),32);
    if (atoi(colptr(rectype_col,str)) < 2000) continue;
    redirDataLink(colptr(sfile_col,str),srv);
  }
err:
  fclose(fp);
  return 0;
}
int redirInit(NAME16 *dbfile,NAME16 *dbsrv,int ndbfiles)
{
  FILE *fp = NULL;
  char fn[80];
  int done = 0,i,cc = 0;

  for (i=0; i<ndbfiles; i++)
  {
    sprintf(fn,"%s%s",ArchiveDBpath,dbfile[i].name);
    redirDataFile(fn,dbsrv[i].name);
  }
  return 0;
}
int reacquireStaticDataSet(int i)
{
  DTYPE dout, din, *dptr;
  int id, cc=0;
  int accmode;
  char devname[64],s[80];

  accmode = CM_POLL;
  fid[i].startPending = TRUE;
  dout.dFormat = fec[i].fmt;
  dout.dArrayLength = fec[i].len;
  dout.data.vptr = fec[i].data;
  if (strlen(fec[i].cntxt) > 0)
    sprintf(devname,"/%s/%s/%s",fec[i].cntxt,fec[i].eqm,fec[i].devname);
  else
    sprintf(devname,"%s/%s",fec[i].eqm,fec[i].devname);
  if (fec[i].nindata > 0)
  {
    din.dFormat = CF_SHORT;
    din.dArrayLength = fec[i].nindata;
    din.data.sptr = (UINT16 *)fec[i].indata;
    dptr = &din;
  }
  else dptr = NULL;

  if ((id=AttachLinkEx(devname,fec[i].property,&dout,dptr,CA_READ,
                     fec[i].poll,dataCallback,accmode,1+i)) < 0)
  {
    sprintf(s,"FILER: can't attach to %s %s : %s",
      devname,fec[i].property,erlst[-id]);
    AppendProblemsLog(s);
  }
  else
  {
    fid[i].id = id;
    feclog("reacquire %s %s",devname,fec[i].property);
  }
  return cc;
}
int linksEstablished = 0;
int pollInit(char *dbfile,int ndbfiles)
{
  DTYPE dout, din, *dptr, dprp;
  int i, n, id, cc=0;
  FECDATA *fecdata;
  int p, rectyp;
  int accmode;
  char devname[64],s[80];

  if (linksEstablished) return 0;

  for (i=0; i<ndbfiles; i++)
    totalNrFECs += getNrActiveFecs(&dbfile[i*16]);

  if ((fec=(FEC *)calloc(totalNrFECs,sizeof(FEC))) == NULL) ccerr(out_of_server_memory);
  if ((prpAlias=(PRP_ALIAS *)calloc(totalNrFECs,sizeof(PRP_ALIAS))) == NULL) ccerr(out_of_server_memory);
  if ((subSystems=(SUB_SYS *)calloc(totalNrFECs,sizeof(SUB_SYS))) == NULL) ccerr(out_of_server_memory);
  if ((fid=(LINKSTRCT *)calloc(totalNrFECs,sizeof(LINKSTRCT))) == NULL) ccerr(out_of_server_memory);

  for (i=0; i<ndbfiles; i++)
  {
    n = archLoad(&dbfile[i*16],nfecs);/* number of fecs being archived */
    cc = inputDataLoad(n,nfecs);      /* input data? */
    if (cc) ccerr(cc);
    nparam = paramLoad(n,nfecs);      /* number of parameters archived */
    nfecs += n;
  }
  fillinArrayTypes(nfecs);
  logRestart();                 /* establish archive file names */

  for (i=0; i<nfecs; i++)       /* start data links */
  {
    accmode = CM_POLL;
    fid[i].startPending = TRUE;
    dout.dFormat = fec[i].fmt;
    dout.dArrayLength = fec[i].len;
    dout.data.vptr = fec[i].data;
    if (strlen(fec[i].cntxt) > 0)
      sprintf(devname,"/%s/%s/%s",fec[i].cntxt,fec[i].eqm,fec[i].devname);
    else
      sprintf(devname,"%s/%s",fec[i].eqm,fec[i].devname);
    if (fec[i].nindata > 0)
    {
      din.dFormat = CF_SHORT;
      din.dArrayLength = fec[i].nindata;
      din.data.sptr = (UINT16 *)fec[i].indata;
      dptr = &din;
    }
    else dptr = NULL;

    if ((id=AttachLinkEx(devname,fec[i].property,&dout,dptr,CA_READ,
                       fec[i].poll,dataCallback,accmode,1+i)) < 0)
    {
      sprintf(s,"FILER: can't attach to %s %s : %s",
        devname,fec[i].property,erlst[-id]);
      AppendProblemsLog(s);
    }
    else
    {
      fid[i].id = id;
      sprintf(s,"FILER: established link %s %s",devname,fec[i].property);
      fec[i].fecStatus = link_not_open; /* let this be cleared normally */
    }
    feclog(s);
  }
  for (i=0; i<nfecs; i++)       /* register properties */
  {
    for (n=0; n<fec[i].nparam; n++)
    {
      sprintf(s,"[%g:%g %s]%s",
        fec[i].fecdata[n].minimum,fec[i].fecdata[n].maximum,
        fec[i].fecdata[n].units,fec[i].fecdata[n].desc);

      dprp.dFormat = fec[i].fecdata[n].format;
      dprp.dArrayLength = fec[i].fecdata[n].size;
      dprp.dTag[0] = 0;
      if ((id=RegisterPropertyInformation("FILER",fec[i].fecdata[n].param,&dprp,NULL,CA_READ,fec[i].fecdata[n].arrayType,0,s,-1,NULL)) < 0)
      {
        feclog("Couldn't register property %s: %s",
          fec[i].fecdata[n].param,erlst[-id]);
      }

      if (fec[i].fecdata[n].data) /* history buffer available */
      {
        strncpy(s,fec[i].fecdata[n].param,16); s[15] = 0;
        strcat(s,".HISTORY");
        if ((id=RegisterProperty("FILER",s,
                  fec[i].fecdata[n].depth,fec[i].fecdata[n].format,CA_READ,
                  fec[i].fecdata[n].desc)) < 0)
        {
          feclog("Couldn't register property %s: %s",
            fec[i].fecdata[n].param,erlst[-id]);
        }
      }
    }
  }
  for (n=0; n<nPrpAlias; n++)
  {
    if ((i=getEntryFromKeyword(prpAlias[n].alias,&rectyp,&p)) < 0) continue;
    fecdata = fec[i].fecdata;
    if ((id=RegisterProperty("FILER",prpAlias[n].name,
              fecdata[p].size,fecdata[p].format,CA_READ,fecdata[p].desc)) < 0)
    {
      cc = -id;
      feclog("Couldn't register property %s: %s",prpAlias[n].name,erlst[cc]);
    }
  }
  linksEstablished = 1;
err:
  if (cc) 
  {
    sprintf(s,"Init error: %s",erlst[cc]);
    feclog(s);
  }    
  return cc;
}

void wdCallback(int id,int cc)
{
  static int failcnt = 0;
  if (cc) failcnt++; else failcnt = 0;
  if (failcnt > 2) registerGlobals = TRUE;
} 
void filerexit(void)
{
}

void filerinit(void)
{
  DTYPE dout;
  char *data=NULL;
  int cc = 0;

  remove("problems.log");
  RegisterUserCommand("ArchiveDebug",NULL,&archdebug,NULL,CA_READ+CA_WRITE);

  pollInit(srvinfo.dbfile[0],srvinfo.ndbfiles);
  nredirfiles = getRedirInfo();
  redirInit(redirfile,redirsrv,nredirfiles);

  getNetGlobalList(GLOBALS_FILENAME);
  if (ngloballist > 0)
  {
    feclog("found %d globals from list %s",ngloballist,GLOBALS_FILENAME);
  }
  if (globalsServer) registerGlobals = TRUE;
  if (globalsWatchdog)
  {
    if (ngloballist > 0)
    {
      dout.dFormat = globalList[0].format;
      dout.dArrayLength = globalList[0].size;
      dout.data.vptr = data;
      if ((cc=recvNetGlobal(globalList[0].keyword,&dout,wdCallback)) < 0)
      {
        feclog("recvNetGlobal  : %s",erlst[-cc]);
      } 
    }
  }
  return;
}
void term_handler(int signum)
{
  g_isTerminating = TRUE;
  feclog("caught SIGTERM");
}
void PostSystemInit(void)
{
  int cc;
  signal(SIGTERM,term_handler);
  signal(SIGFPE,SIG_IGN); /* catch floating point domain errors */
  RequireAcknowledgments = 0; /* turn off ack. requirments for REFRESH mode */
  if ((cc=RegisterEquipmentModule(NULL,"FILER",0,_filer,filerinit,filertask,1000,filerexit)) != 0)
    feclog("Can't register FILER: %s",erlst[cc]);
  feclog("Check incoming state information : %s",checkStates ? "TRUE" : "FALSE");
  feclog("Monitoring %d checked states",nCheckedStates);
  feclog("Monitoring %d checked state parameters",nCheckedParameters);
  feclog("Monitoring %d declared e procedures",nEDeclaredProcedures);
  feclog("Monitoring %d declared p procedures",nPDeclaredProcedures);
  feclog("Is Globals Server : %s", globalsServer ? "TRUE" : "FALSE");
  feclog("Is Globals Watchdog : %s", globalsWatchdog ? "TRUE" : "FALSE");
  feclog("Using %s as targeted State Server",gTargetStateServer);
}
void pollTask(void)
{
  int i;
  time_t t = time(NULL);
  struct tm *tm = localtime(&t);
  static int mday = -1;

  if (mday != tm->tm_mday) /* new day */
  {
    mday = tm->tm_mday;
    logRestart();
    for (i=0; i<nfecs; i++)
    {
      if (fec[i].dataFlg < 1) continue;
      if (fec[i].option&ONCE)
      {
        fec[i].archtime = fec[i].tstamp = t;
        fec[i].archPending = 1;
      }
    }
  }
  return;
}
#define DATABLOB_SIZE 16384
BYTE g_datablob[DATABLOB_SIZE];
#define MAXNAMES (DATABLOB_SIZE/16)
NAME16 *name16blob = (NAME16 *)g_datablob;

int getIndexFromDeviceName(char *devName,char *devProperty)
{
  FECDATA *fecdata;
  int i, siz, n, p, rectyp, ptr;
  char namProperty[32] = "";
 
  if (devName[0] == 0) return 0;
  if (devName[0] == '#')
  {
    if ((i=atoi(&devName[1])) > 0) i--;   /* count from 0 */
    return i;
  }
  strncpy(namProperty,devProperty,27);
  strcat(namProperty,".NAM");
  for (i=0; i<nPrpAlias; i++)
  {
    if (!strncmp(prpAlias[i].name,namProperty,32))
    {
      strncpy(namProperty,prpAlias[i].alias,32);
      break;
    }
  }
  if ((n=getEntryFromKeyword(namProperty,&rectyp,&p)) < 0) return -non_existent;
  fecdata = fec[n].fecdata;
  switch (fec[n].fecdata[p].format)
  {
    case CF_NAME16:
      siz = fecdata[p].size; ptr = fec[n].fecdata[p].offset;
      for (i=0; i<siz; i++)
        if (!strncmp(devName,(char *)&fec[n].data[ptr+16*i],16)) return i;
      break;
    case CF_NAME32:
      siz = fecdata[p].size; ptr = fec[n].fecdata[p].offset;
      for (i=0; i<siz; i++)
        if (!strncmp(devName,(char *)&fec[n].data[ptr+32*i],16)) return i;
      break;
    default:
      return -illegal_format;
  }
  return -illegal_equipment_number;
}
int _filer(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc, i, j, n ;
  char prop[32],dev[16], *c;
  short stateArray[7];
  short dsize,index = 0,fmt;
  int namsiz,fmtsiz,minsiz;

# ifdef NO_ASYNC
  if ((access & CA_FIRST) && (access & CA_LAST)) /* blow synchronous calls out of the water */
  {
    return not_accepted;
  }
# endif
  /* make local copies of the incoming information ... */
  strncpy(dev,devName,16);
  strncpy(prop,devProperty,32);
  /* maybe it's an alias property ? */
  for (i=0; i<nPrpAlias; i++)
  {
    if (!strncmp(prpAlias[i].name,devProperty,32))
    {
      strncpy(prop,prpAlias[i].alias,32);
      break;
    }
  }

  if ((cc=overloadEqpmod(dev,prop,dout,din,access)) != illegal_property) return cc;
  if ((index=getIndexFromDeviceName(dev,prop)) < 0) return -index;
  fmt = dout->dFormat;
  dsize = (int)dout->dArrayLength;
  
  if (!strcmp(prop,"STATE"))  /* machine state(s) */
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dsize < 1 || dsize > 7) return out_of_range;
    stateArray[6] = Eramping;
    stateArray[5] = Ebeam;
    stateArray[4] = Erunning;
    stateArray[3] = Pramping;
    stateArray[2] = Pbeam;
    stateArray[1] = Prunning;
    stateArray[0] = lumi;
    for (i=0; i<dsize; i++)
    {
      switch (dout->dFormat)
      {
        case CF_BYTE: dout->data.bptr[i] = (BYTE)stateArray[i]; break;
        case CF_SHORT: dout->data.sptr[i] = (short)stateArray[i]; break;
        case CF_LONG:  dout->data.lptr[i] = (UINT32)stateArray[i]; break;
        case CF_FLOAT: dout->data.fptr[i] = (float)stateArray[i]; break;
        case CF_DOUBLE: dout->data.dptr[i] = (double)stateArray[i]; break;
      }
    }
    return 0;
  }
  if (!strcmp(prop,"STATUS"))
  {
    if (dsize < 1 || dsize > totalNrFECs) return illegal_read_write;
    if (fmt != CF_SHORT) return illegal_format;
    for (i = 0; i<dsize; i++) (dout->data.sptr)[i] = fec[i].fecStatus;
    return 0;
  }
  if (!strcmp(prop,"NKEYWORD"))
  {
    if (dsize != 1) return illegal_read_write;
    for (i=0,n=0; i<nfecs; i++) if (fec[i].active) n += fec[i].nparam;
    putValuesFromLong(dout,(UINT32 *)&n,1);
    return 0;
  }
  if (!strcmp(prop,"KEYWORDS"))  /* use CF_STRING8 */
  {
    if (fmt != CF_STRING8) return illegal_format;
    for (i=0,j=0; j<dsize && i<nfecs; i++) if (fec[i].active)
    {
      for (n=0; j<dsize && n<fec[i].nparam; n++,j++)
      {
        strncpy((char *)&dout->data.dptr[j],fec[i].fecdata[n].param,8);
      }
    }
    for (;j<dsize; j++) memset((char *)&dout->data.dptr[j],' ',8);
    return 0;
  }
  if (strstr(prop,".HISTORY"))
  {
    *(c = strstr(prop,".HISTORY")) = 0;
    for (i=0; i<nfecs; i++) if (fec[i].active)
    {
      for (n=0; n<fec[i].nparam; n++) 
        if (fec[i].fecdata[n].data && !stricmp(fec[i].fecdata[n].param,prop))
          return getFecData(dout,i,n,index);
    }
    return illegal_property;
  }
  for (i=0; i<nfecs; i++) if (fec[i].active)
  {
    for (n=0; n<fec[i].nparam; n++)
    {
      if (!stricmp(fec[i].fecdata[n].param,prop))
      {
        if (fmt == CF_TEXT)
        {
          switch(fec[i].fecdata[n].format)
          {
            case CF_CHAR8:
            case CF_CHAR16:
            case CF_CHAR32:
            case CF_CHAR48:
            case CF_CHAR64:
              if (dsize < fmtsizeof(fec[i].fecdata[n].format))
                return dimension_error;
              dsize = 1; fmt = fec[i].fecdata[n].format;
              break;
            default: break;
          }
        }
        if (dsize < 0) return dimension_error;
        setDataTimeStamp((double)fec[i].tstamp);
        switch( fec[i].fecdata[n].format )
        {
          case CF_INTEGER:
            return putValuesFromShortEx(dout,(short *)&fec[i].data[fec[i].fecdata[n].offset],(short)fec[i].fecdata[n].size,(short)index);
          case CF_LONG:
            return putValuesFromLongEx(dout,(UINT32 *)&fec[i].data[fec[i].fecdata[n].offset],(short)fec[i].fecdata[n].size,(short)index);
          case CF_FLOAT:
            return putValuesFromFloatEx(dout,(float *)&fec[i].data[fec[i].fecdata[n].offset],(short)fec[i].fecdata[n].size,(short)index);
          case CF_DOUBLE:
            return putValuesFromDoubleEx(dout,(double *)&fec[i].data[fec[i].fecdata[n].offset],(short)fec[i].fecdata[n].size,(short)index);
          case CF_CHAR8:
          case CF_CHAR16:
          case CF_CHAR32:
          case CF_CHAR48:
          case CF_CHAR64:
            namsiz = (fmt == CF_NAME8)  ? 8  :
                     (fmt == CF_NAME16) ? 16 :
                     (fmt == CF_NAME32) ? 32 :
                     (fmt == CF_NAME48) ? 48 :
                     (fmt == CF_NAME64) ? 64 : 0;
            if (namsiz == 0) return illegal_format;
            fmtsiz = fmtsizeof(fec[i].fecdata[n].format);
            memset(dout->data.cptr,0,dsize*namsiz);
            minsiz = MIN(namsiz,fmtsiz);
            for (j=0; j<dsize && j<fec[i].fecdata[n].size-index; j++)
              strncpy(dout->data.cptr+namsiz*j,(char *)&fec[i].data[fec[i].fecdata[n].offset+fmtsiz*(j+index)],minsiz);
            return 0;
          case CF_TEXT:
            switch (fmt)
            {
              case CF_TEXT: break;
              case CF_NAME8:
              case CF_NAME16:
              case CF_NAME32:
              case CF_NAME48:
              case CF_NAME64: dsize = fmtsizeof(fmt); break;
              default: return illegal_format;
            }
            memset(dout->data.cptr,0,dsize);
            if (dsize > fec[i].fecdata[n].size) dsize = fec[i].fecdata[n].size;
            strncpy(dout->data.cptr,(char *)&fec[i].data[fec[i].fecdata[n].offset]
,dsize);
            return 0;
          default:
            if (fmt == fec[i].fecdata[n].format)
            {
              if (dsize > fec[i].fecdata[n].size) dsize = fec[i].fecdata[n].size;
              memcpy(dout->data.bptr,&fec[i].data[fec[i].fecdata[n].offset],dsize);
              return 0;
            }
            return illegal_format;
        }
      }
    }
  }
  return illegal_property;
}

