#include "prolog.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "errors.h"
#include "rpctypes.h"
#include "servdef.h"
#include "toolkit.h"

double blob[8000];
PropertyQueryStruct prpinfo;
BYTE *bval = (BYTE *)blob;
short *sval = (short *)blob;
UINT32 *lval = (UINT32 *)blob;
float *fval = (float *)blob;
double *dval = blob;
char *cval = (char *)blob;
NAME16 *n16val = (NAME16 *)blob;
NAME32 *n32val = (NAME32 *)blob;

void tmonitorCallback(int id,int cc)
{
  int i;
  printf("\n");
  if (cc) 
  {
    printf("error: %s\n>",RPCLastErrorString);
    return;
  }
  switch (LFMT(prpinfo.prpFormat))
  {
    case CF_BYTE:
      for (i=0; i<prpinfo.prpSize; i++) printf("%d ",bval[i]); break;
    case CF_SHORT:
      for (i=0; i<prpinfo.prpSize; i++) printf("%d ",sval[i]); break;
    case CF_LONG:
      for (i=0; i<prpinfo.prpSize; i++) printf("%d ",lval[i]); break;
    case CF_FLOAT:
      for (i=0; i<prpinfo.prpSize; i++) printf("%g ",fval[i]); break;
    case CF_DOUBLE:
      for (i=0; i<prpinfo.prpSize; i++) printf("%g ",dval[i]); break;
    case CF_TEXT:
      printf("%s",cval); break;
    case CF_NAME16:
      for (i=0; i<prpinfo.prpSize; i++) printf("%16s ",n16val[i].name); break;
    case CF_NAME32:
      for (i=0; i<prpinfo.prpSize; i++) printf("%32s ",n32val[i].name); break;
    default: printf("output format type %d not implemented\n",LFMT(prpinfo.prpFormat));
  }
}
int _tmonitor(void *buf,char *dev,char *prp,int fmt,int len,int tmr)
{
  int id;
  DTYPE dout;

  dout.dFormat = fmt;
  dout.dArrayLength = len;
  dout.dTag[0] = 0;
  dout.data.vptr = buf;
  id = AttachLink(dev,prp,&dout,NULL,CA_READ,tmr,tmonitorCallback,CM_POLL);
  return id < 0 ? -id : 0;
}
int _tput(void *buf,char *dev,char *prp,int fmt,int len,int tmr)
{
  DTYPE dout,din;

  din.dFormat = fmt;
  din.dArrayLength = len;
  din.dTag[0] = 0;
  din.data.vptr = buf;

  return ExecLinkEx(dev,prp,NULL,&din,CA_WRITE,tmr);
}
int _tget(void *buf,char *dev,char *prp,int fmt,int len,int tmr)
{
  DTYPE dout;

  dout.dFormat = fmt;
  dout.dArrayLength = len;
  dout.dTag[0] = 0;
  dout.data.vptr = buf;
  return ExecLinkEx(dev,prp,&dout,NULL,CA_READ,tmr);
}
int GetPropertyInformation(char *srv,char *prp,PropertyQueryStruct *srvProps)
{
  DTYPE dout,din;
  int cc = 0;
 
  din.dFormat = CF_NAME32;
  din.dArrayLength = 1;
  din.data.vptr = prp;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = srvProps;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,&din,CA_READ,500)) != 0) return cc;
  srvProps->prpSize = ISWAP(srvProps->prpSize);
  return 0;
}
char *fmtstr(int fmt)
{
  static char s[80];
  switch (LFMT(fmt))
  {
    case CF_BYTE: sprintf(s,"BYTE"); break;
    case CF_SHORT: sprintf(s,"short"); break;
    case CF_LONG: sprintf(s,"UINT32"); break;
    case CF_FLOAT: sprintf(s,"float"); break;
    case CF_DOUBLE: sprintf(s,"double"); break;
    case CF_TEXT: sprintf(s,"char"); break;
    case CF_NAME16: sprintf(s,"NAME16"); break;
    case CF_NAME32: sprintf(s,"NAME32"); break;
    default: sprintf(s,"type <%d>",LFMT(fmt)); break;
  }
  return s;
}
int tinfo(char *dev,char *prp)
{
  int i,cc;
  float max,min;
  char egu[80];
 
  if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
  if ((cc=GetDevicePropertyEGU(dev,prp,&max,&min,egu)) != 0) return cc;
  printf("\nDevice : %s\nProperty : %s\nDescription : %s\nSize : %d\nFormat %s\nMaximum : %f\nMinimum : %f\nUnits : %s\n",
         dev,prpinfo.prpName,prpinfo.prpDescription,prpinfo.prpSize,
         fmtstr(prpinfo.prpFormat),max,min,egu);
  return 0;
}
int tserversEx(char *cnt,char *sub)
{
  int i,k,cc,n = 256;
  NAME16 srvs[256];
  char prop[32],srv[64];
  DTYPE dout,din,*dinptr=NULL;

  strncpy(srv,"ENS",16);
  if (cnt != NULL && strlen(cnt) > 0)
  {
    strcat(srv,"/"); strncat(srv,cnt,16);
  }
  if (sub != NULL)
  {
    din.dFormat = CF_TEXT;
    din.dArrayLength = strlen(sub);
    din.data.cptr = sub;
    dinptr = &din;
  }
  dout.dFormat = CF_NAME16;
  dout.dArrayLength = n;
  dout.data.vptr = srvs;
  if ((cc=ExecLinkEx(srv,"TAGS",&dout,dinptr,CA_READ,200)) != 0) return cc;
  for (i=0,k=0; i<n; i++)
  {
    if (strlen(srvs[i].name) == 0) continue;
    printf("%16s ",srvs[i].name);
    if ((i+1)%4 == 0) printf("\n");
    k++;
  }
  printf("\n%d servers found\n",k);
  return 0;
}
int tservers(char *cnt)
{
  int i,cc,n = 256;
  NAME16 srvs[256];
  if ((cc=GetDeviceServersEx(cnt,srvs,&n)) != 0) return cc;
  printf("%d servers found :\n",n);
  for (i=0; i<n; i++) 
  {
    printf("%16s ",srvs[i].name);
    if ((i+1)%4 == 0) printf("\n");
  }
  return 0;
}
int tprops(char *dev)
{
  int i,cc,n = 1024;
  NAME32 *prps = n32val;
  if ((cc=GetDeviceProperties(dev,prps,&n)) != 0) return cc;
  printf("%d properties found :\n",n);
  for (i=0; i<n; i++) 
  {
    if (strlen(prps[i].name) == 0) continue;
    printf("%32s ",prps[i].name);
    if ((i+1)%4 == 0) printf("\n");
  }
  return 0;
}
int tdevsEx(char *dev,char *prp)
{
  int i,cc,n = 1024;
  NAME16 *devs = n16val;
  char devname[32];
  if ((cc=GetDeviceNamesEx(dev,prp,devs,&n)) != 0) return cc;
  if (n == 2 && !strcmp(devs[1].name,"XYZHASQUERYFCN"))
  {
    n = 1;
    if (!strcmp(devs[0].name,"#0")) strncpy(devs[0].name,prp,16);
  }
  printf("%d devices found :\n",n);
  for (i=0; i<n; i++) 
  {
    strncpy(devname,devs[i].name,16);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%32s ",devname);
    if ((i+1)%4 == 0) printf("\n");
  }
  return 0;
}
int tdevs(char *dev)
{
  int i,cc,n = 1024;
  NAME16 *devs = n16val;
  char devname[32];
  if ((cc=GetDeviceNames(dev,devs,&n)) != 0) return cc;
  if (n == 2 && !strcmp(devs[1].name,"XYZHASQUERYFCN"))
  {
    printf("Device List depends on targetted Property: use tdevsEx()\n",n);
    return ambiguous;
  }
  printf("%d devices found :\n",n);
  for (i=0; i<n; i++) 
  {
    strncpy(devname,devs[i].name,16);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%32s ",devname);
    if ((i+1)%4 == 0) printf("\n");
  }
  return 0;
}
int tdevlist(char *dev)
{
  int i,cc,n = 1024;
  NAME16 *devs = n16val;
  char devname[32];
  if ((cc=GetDeviceNames(dev,devs,&n)) != 0) return cc;
  printf("DEVICE_NAME,DEVICE_NUMBER\n",n);
  for (i=0; i<n; i++)
  {
    strncpy(devname,devs[i].name,16);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%s,%d\n",devname,i);
  }
  return 0;
}
/* #define OLDWAY */
extern int NGdebug;
int tput(char *dev,char *prp)
{
  return 0;
}
int tget(char *dev,char *prp)
{
  return tgetEx(dev,prp,CF_NULL,0);
}
int tgetEx(char *dev,char *prp,int fmt_,int siz_)
{
  int i,cc;
  int fmt,siz;
  char s[65];

  if (fmt_ != CF_NULL && siz_ != 0)
  {
    fmt = fmt_;
    siz = siz_;
    if (siz > 8000) siz = 8000;
    if ((cc=_tget(blob,dev,prp,fmt,siz,5000)) != 0) return cc;
  }
  else
  {
# ifdef OLDWAY
    if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
    if ((cc=_tget(blob,dev,prp,LFMT(prpinfo.prpFormat),prpinfo.prpSize,5000)) != 0) return cc;
    fmt = LFMT(prpinfo.prpFormat);
    siz = prpinfo.prpSize;
  }
# else
    siz = sizeof(blob);
    if ((cc=_tget(blob,dev,prp,CF_DEFAULT,siz,500)) != 0) 
    {
      while (cc == invalid_transport_size || cc == use_stream_transport)
      {
        siz /= 2; 
        if (siz < 1000) return cc;
        if ((cc=_tget(blob,dev,prp,CF_DEFAULT,siz,500)) == 0) goto get_done;
      }
      return cc;
    }
get_done:
    fmt = GetCompletionDataType(-1);
    siz = GetCompletionDataSize(-1);
  }
# endif
  switch (fmt)
  {
    case CF_BYTE:
      for (i=0; i<siz; i++) printf("%d ",bval[i]); break;
    case CF_SHORT:
      for (i=0; i<siz; i++) printf("%d ",sval[i]); break;
    case CF_LONG:
      for (i=0; i<siz; i++) printf("%d ",lval[i]); break;
    case CF_FLOAT:
      for (i=0; i<siz; i++) printf("%g ",fval[i]); break;
    case CF_DOUBLE:
      for (i=0; i<siz; i++) printf("%g ",dval[i]); break;
    case CF_TEXT:
      printf("%s",cval); break;
    case CF_NAME16:
      for (i=0; i<siz; i++) 
      {
        strncpy(s,n16val[i].name,16);
        printf("%s ",s);
      }
      break;
    case CF_NAME32:
      for (i=0; i<siz; i++)
      {
        strncpy(s,n32val[i].name,32);
        printf("%s ",s);
      }
      break;
    default: printf("output format type %d not implemented\n",fmt);
  }
  return 0;
}
int tmonitor(char *dev,char *prp)
{
  return tmonitorEx(dev,prp,CF_NULL,0);
}
int tmonitorEx(char *dev,char *prp,int fmt_,int siz_)
{
  int i,cc;
  int fmt,siz;

  if (fmt_ != CF_NULL && siz_ != 0)
  {
    fmt = fmt_;
    siz = siz_;
    prpinfo.prpFormat = BFMT(fmt);
    prpinfo.prpSize = siz;
  }
  else
  {
    if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
    fmt = LFMT(prpinfo.prpFormat);
    siz = prpinfo.prpSize;
  }
  if (siz > 8000) siz = 8000;
  if ((cc=_tmonitor(blob,dev,prp,fmt,siz,1000)) != 0) return cc;
  return 0;
}
time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[16];
  struct tm tmts, *tnow;

  tnow = localtime(&t);
  len = strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
    }
    memset(&tmts,0,sizeof(struct tm));
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    t = mktime(&tmts);
  }
  return t;
}
int thistory(char *context,char *keyword,char *devname,char *stoptime,char *depth)
{
  char devsrv[256];
  long stop = time(NULL), start, mf = 1, dp = 0;
  int i, num = 5000, cc, offset = 0;
  FLTINT *fiDataArray = (FLTINT *)blob;
  NAME32I *niDataArray = (NAME32I *)blob;

  if (keyword == NULL) return argument_list_error;
  if (devname == NULL) devname = "#0";
  if (context == NULL) context = "HERA";
  else strupr(context);
  if (!strcmp(context,"HERA")) sprintf(devsrv,"/HERA/HEHISTORY/");
  else if (!strcmp(context,"PETRA")) sprintf(devsrv,"/PETRA/PEHISTORY/");
  else if (!strncmp(context,"DESY",4)) sprintf(devsrv,"/DESY/D3HISTORY/");
  else if (!strncmp(context,"LINAC",5)) sprintf(devsrv,"/DESY/D3HISTORY/");
  else if (!strcmp(context,"DORIS")) sprintf(devsrv,"/DORIS/DOHISTORY/");
  else if (!strcmp(context,"TTF")) sprintf(devsrv,"/TTF/TTFHISTORY/");
  else strncpy(devsrv,context,36);

  strcat(devsrv,keyword);
  strcat(devsrv,"/");
  strcat(devsrv,devname);
 
  stop = parseTimeString(stoptime);
  strupr(depth);
  if (strstr(depth,"DAY")) mf = 24 * 3600;
  else if (strstr(depth,"HOUR")) mf = 3600;
  else if (strstr(depth,"MINUTE")) mf = 60;
  else if (strstr(depth,"WEEK")) mf = 24 * 3600 * 7;
  dp = mf * atoi(depth);
  start = stop - dp;
  printf("%s/%s, Timestamp\n",keyword,devname);
  if ((cc=GetArchivedDataAsFloat(devsrv,start,stop,fiDataArray,&num)) == 0) 
  {
    if (fiDataArray[0].ival <= num) offset = 1;
    for (i=offset; i<num; i++)
    {
      printf("%g, %s",fiDataArray[i].fval,ctime(&fiDataArray[i].ival));
    }
    return 0;
  }
  if (cc != illegal_format) return cc;
  if ((cc=GetArchivedDataAsText(devsrv,start,stop,niDataArray,&num)) == 0) 
  {
    if (niDataArray[0].ival <= num) offset = 1;
    for (i=offset; i<num; i++)
    {
      printf("%s, %s",niDataArray[i].name,ctime(&niDataArray[i].ival));
    }
    return 0;
  }
  
  return cc;
}
  
