/* =============================================================
   ==
   ==   Dataserver routine: statebase.c
   ==
   ============================================================= */
#include "prolog.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#ifdef UNIX
#include <dirent.h>
#endif
#include "fecerr.h"
#include "rpctypes.h"
#include "database.h"
#include "servdef.h"
#include "almsrv.h"
#include "errors.h"
#include "archive.h"
#include "statebase.h"
#include "toolkit.h"
#include "brdcast.h"

/* state variables: */
extern int ramping, running, beam, lumi;
extern int Pramping, Eramping, Pbeam, Ebeam;
extern int Prunning, Erunning;
extern int DclEramping, DclPramping, DclLumi,DclRun;
extern int DclEInj, DclPInj; 
extern int pneedsToDeclare;
extern int eneedsToDeclare;
extern FEC *fec;

extern int BroadcastServer;
extern TSRVINFO srvinfo;
extern int globalsWatchdog, globalsServer;
int getStartupInfo(void);
 
char Eproc[32], Pproc[32], Etype[32], Ptype[32];
char gTargetStateServer[64];
DECLPROC *eDeclaredProcedures = NULL;
DECLPROC *pDeclaredProcedures = NULL;
int nEDeclaredProcedures = 0;
int nPDeclaredProcedures = 0;
extern int ngloballist;
extern GlobalListStruct *globalList;

NAME16 *allowedNullStates = NULL;
int nAllowedNullStates = 0;
CSVCOLUMN csvNullStateColumns[] =
{
  { "STATE","",0,CF_TEXT,16,NULL,NULL },
};
CSVDB csvNullStateDb =
{
  csvNullStateColumns,
  sizeof(csvNullStateColumns)/sizeof(csvNullStateColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(NAME16)
};
int getAllowedNullStates(void)
{
  int cc = 0;
  
  if ((cc=csvReadFile(FecDBpath,NULLSTATE_NAME,
            &csvNullStateDb,(void **)&allowedNullStates)) == 0)
  {
    nAllowedNullStates = csvNullStateDb.siz;
    feclog("found %d nullstates",nAllowedNullStates);
  }
  else
  {
    feclog("no null states found : %s",erlst[cc]);
  }
  return cc;
}

CHECKNAME *checkedParameters = NULL;
int nCheckedParameters = 0;
CSVCOLUMN csvCheckStateColumns[] =
{
  { "PARAMETER","",0,CF_TEXT,16,(BYTE *)OFFSETIN(CHECKNAME,name),NULL },
  { "FORMAT","",0,CF_LONG,1,(BYTE *)OFFSETIN(CHECKNAME,fmt),NULL },
  { "PTYPE","",0,CF_LONG,1,(BYTE *)OFFSETIN(CHECKNAME,typeindex),NULL },
  { "BEAMTYPE","",0,CF_TEXT,16,(BYTE *)OFFSETIN(CHECKNAME,beamtype),NULL },
  { "MINVALID","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CHECKNAME,minValid),NULL },
  { "MAXVALID","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CHECKNAME,maxValid),NULL },
};
CSVDB csvCheckStateDb =
{
  csvCheckStateColumns,
  sizeof(csvCheckStateColumns)/sizeof(csvCheckStateColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(CHECKNAME)
};
/*
// read the beam and state parameters in
// this assumes they are cross entered in the 'database' file
// that is used for the links in filer.c
// Here, the entries have a 'second' assigned variable for use
// in 'globals'.
*/
int getBeamParameters(void)
{
  int i, cc = 0;

  if ((cc=csvReadFile(FecDBpath,BEAMPARAM_NAME,
            &csvCheckStateDb,(void **)&checkedParameters)) == 0)
  {
    nCheckedParameters = csvCheckStateDb.siz;
    feclog("found %d beam parameters",nCheckedParameters);
    for (i=0; i<nCheckedParameters; i++)
    {
      if (checkedParameters[i].minValid != 0) 
        checkedParameters[i].chkValid |= CHK_VALID_MIN;
      if (checkedParameters[i].maxValid != 0) 
        checkedParameters[i].chkValid |= CHK_VALID_MAX;
      if (!stricmp(checkedParameters[i].beamtype,"status"))
        checkedParameters[i].chkValid |= CHK_VALID_MIN|CHK_VALID_MAX;
    }
  }
  else
  {
    feclog("no beam parameters found : %s",erlst[cc]);
  }
  return cc;
}
CHECKNAME *checkedStates = NULL;
int nCheckedStates = 0;
int getStateParameters(void)
{
  int cc = 0;

  if ((cc=csvReadFile(FecDBpath,STATEPARAM_NAME,
            &csvCheckStateDb,(void **)&checkedStates)) == 0)
  {
    nCheckedStates = csvCheckStateDb.siz;
    feclog("found %d state parameters",nCheckedStates);
  }
  else
  {
    feclog("no state parameters found : %s",erlst[cc]);
  }
  return cc;
}
int getStateServer(char *fname,char *stateSrv)
{
  char fn[128];
  char str[256];
  int i, done = 0,cc = 0;
  int sta_col=-1;
  FILE *fp = NULL;

  if (stateSrv == NULL) ccerr(argument_list_error);
  sprintf(fn,"%s%s",FecDBpath,fname);
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (i=0; fgets(str,255,fp) != NULL; )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((sta_col=findcol(str,"SERVER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(stateSrv,colptr(sta_col,str),64);
    ccerr(0);
  }
  cc = database_not_loaded;
err:
  if (fp != NULL) fclose(fp);
  if (cc)
  {
    feclog("%s - %s",fn,erlst[cc]);
  }
  else
  {
    feclog("using %s as state server",stateSrv);
  }
  return cc;
}
int getDeclaredProcedures(char *fname,char *defproc,char *deftype,DECLPROC **declp,int *ndeclp)
{
  char fn[128];
  char str[256];
  int done = 0,i,n = 0,cc = 0;
  int typ_col=-1,prc_col=-1,trg_col=-1;
  FILE *fp = NULL;
  DECLPROC *dp = NULL;

  if (defproc == NULL || deftype == NULL || declp == NULL) ccerr(argument_list_error);
  sprintf(fn,"%s%s",FecDBpath,fname);
  if ((n=GetDataBaseSize(fn)) < 0) ccerr(-n);
  if ((dp=(DECLPROC *)SystemCalloc(n,sizeof(DECLPROC))) == NULL)
    ccerr(out_of_local_memory);
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (i=0; fgets(str,255,fp) != NULL; )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((typ_col=findcol(str,"TYPE")) < 0) ccerr(no_such_line);
      if ((prc_col=findcol(str,"PROCEDURE")) < 0) ccerr(no_such_line);
      if ((trg_col=findcol(str,"TRIGGER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(dp[i].trigstr,colptr(trg_col,str),32);
    if (!stricmp(dp[i].trigstr,"none"))
    {
      strncpy(deftype,colptr(typ_col,str),32);
      strncpy(defproc,colptr(prc_col,str),32);
      continue;
    }
    strncpy(dp[i].typestr,colptr(typ_col,str),32);
    strncpy(dp[i].procstr,colptr(prc_col,str),32);
    if (i++ >= n) break;
  }
  if (ndeclp != NULL) *ndeclp = i;
  *declp = dp;
err:
  if (fp != NULL) fclose(fp);
  if (cc)
  {
    feclog("%s - %s",fn,erlst[cc]);
  }
  else
  {
    feclog("read %d procedures for default %s %s",i,deftype,defproc);
  }
  return cc;
}
int loadStartupInfo(void)
{
  if (getStartupInfo()) exit(2);
  getAllowedNullStates();
  getStateParameters();
  getBeamParameters();
  getDeclaredProcedures(EPROC_NAME,Eproc,Etype,&eDeclaredProcedures,&nEDeclaredProcedures);
  getDeclaredProcedures(PPROC_NAME,Pproc,Ptype,&pDeclaredProcedures,&nPDeclaredProcedures);
  getStateServer(STATESRV_NAME,gTargetStateServer);
  return 0;
}

void PreSystemInit(void)
{
  loadStartupInfo();
  BroadcastServer = TRUE;
  GlobalHeartbeat = 3;
  if (srvinfo.acquire) globalsServer = TRUE; else globalsWatchdog = TRUE;
  FeclogDepth = 2000;
  MaxNumClients = 500;
  MaxNumContracts = 1000; 
  MaxNumConnections = 1000;
  ListenForMCasts = 0;
  MaxRPCTransportSize = MAX_TRANSPORTSIZE*4;
 
  return;
}
char *trimString(char *str)
{
  static char trimBuffer[256];
  char *pstr;
  if (str == NULL)
  {
    trimBuffer[0] = 0;
    goto err;
  }
  for (pstr=str; strchr(" \t\n",*pstr) != NULL; pstr++);
  strncpy(trimBuffer,str,256);
  strtok(trimBuffer," \t\n");
err:
  return trimBuffer;
}

void updateStateInfo(int i, int j,int cc)
{
  int k,p,fmtsiz,isValid;
  char scratch[128];
  BYTE *bptr = (BYTE *)(fec[i].data + fec[i].fecdata[j].offset);
  static float lastPEnergy = 0, lastEEnergy = 0;
  float thisValue,*lastValue;
  UINT32 thisStatus;
  int *pRunning,*pRamping,*pBeam,*pInj;
  char *pType, *pProc;
 
  for (k=0; k<ngloballist; k++)
  {
    if (!strcmp(fec[i].fecdata[j].param,globalList[k].keyword))
    {
      ClearAlarm("FILER",(short)i);
      if (cc) fec[i].errCounter++; else fec[i].errCounter = 0;
      if (fec[i].errCounter < 5) break;
      if (fec[i].errCounter == 5) feclog("missing global %s!",globalList[k].keyword);
      SetAlarm("FILER",(short)i,link_not_open,NULL);
      switch (fec[i].fecdata[j].format)
      {
        case CF_SHORT:
          for (p=0; p<fec[i].fecdata[j].size; p++) ((short *)(bptr))[p] = -1;
          break;
        case CF_LONG:
          for (p=0; p<fec[i].fecdata[j].size; p++) ((int *)(bptr))[p] = -1;
          break;
        case CF_FLOAT:
          for (p=0; p<fec[i].fecdata[j].size; p++) ((float *)(bptr))[p] = -1;
          break;
        case CF_DOUBLE:
          for (p=0; p<fec[i].fecdata[j].size; p++) ((double *)(bptr))[p] = -1;
          break;
        case CF_TEXT:
          strncpy((char *)bptr,"-1",fec[i].fecdata[j].size);
          break;
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          fmtsiz = fmtsizeof(fec[i].fecdata[j].format);
          for (p=0; p<fec[i].fecdata[j].size; p++) strcpy((char *)(bptr+p*fmtsiz),"-1");
          break;
      }
    }
  }
  for (k=0; k<nCheckedParameters; k++)
  {
    if (!strcmp(fec[i].fecdata[j].param,checkedParameters[k].name))
    {
      switch (checkedParameters[k].typeindex)
      {
        case 0:
          pBeam = &Ebeam; pRunning = &Erunning; 
          pRamping = &Eramping; lastValue = &lastEEnergy;
          pType = Etype; pProc = Eproc;
          break;
        case 1:
          pBeam = &Pbeam; pRunning = &Prunning; 
          pRamping = &Pramping; lastValue = &lastPEnergy;
          pType = Ptype; pProc = Pproc;
          break;
        default:
          return;
      } 
      if (!stricmp(checkedParameters[k].beamtype,"status"))
      {
        thisStatus = *(UINT32 *)bptr;
        isValid = TRUE;
        if (CHK_VALID_MIN &&
            thisStatus < (UINT32)checkedParameters[k].minValid) isValid = FALSE;
        if (CHK_VALID_MAX &&
            thisStatus > (UINT32)checkedParameters[k].maxValid) isValid = FALSE;
        *pRunning = isValid;
        goto fixstate;
      }
      if (!stricmp(checkedParameters[k].beamtype,"energy"))
      {
        thisValue = *(float *)bptr;
        isValid = TRUE;
        if ((checkedParameters[k].chkValid & CHK_VALID_MIN) &&
            thisValue < checkedParameters[k].minValid) isValid = FALSE;
        if ((checkedParameters[k].chkValid & CHK_VALID_MAX) &&
            thisValue > checkedParameters[k].maxValid) isValid = FALSE;
        *pRunning = isValid;
        if (fabs(thisValue - *lastValue) > 1)
        {
          *pRamping = (int)(thisValue-(*lastValue));
          *lastValue = thisValue;
        }
        else *pRamping = 0;
        goto fixstate;
      }
      if (!stricmp(checkedParameters[k].beamtype,"current"))
      {
        thisValue = *(float *)bptr;
        isValid = TRUE;
        if ((checkedParameters[k].chkValid & CHK_VALID_MIN) &&
            thisValue < checkedParameters[k].minValid) isValid = FALSE;
        if ((checkedParameters[k].chkValid & CHK_VALID_MAX) &&
            thisValue > checkedParameters[k].maxValid) isValid = FALSE;
        *pBeam = isValid;
        goto fixstate;
      }
      if (!stricmp(checkedParameters[k].beamtype,"type"))
      {
        strncpy(pType,(char *)bptr,32);
        goto declstate;
      }
      if (!stricmp(checkedParameters[k].beamtype,"procedure"))
      {
        strncpy(pProc,(char *)bptr,32);
        goto declstate;
      }
    }
  }  
  for (k=0; k<nCheckedStates; k++)
  {
    if (!strcmp(fec[i].fecdata[j].param,checkedStates[k].name))
    {
      switch (checkedStates[k].typeindex)
      {
        case 0:
          pRamping = &DclEramping; pInj = &DclEInj;
          break;
        case 1:
          pRamping = &DclPramping; pInj = &DclPInj;
          break;
        default:
          return;
      } 
      strncpy(scratch,(char *)bptr,128); strupr(scratch);
      DclRun = strstr(scratch,"RUN") == NULL ? 0 : -1;
      DclLumi = strstr(scratch,"LUMI") == NULL ? 0 : -1;
      *pInj = strstr(scratch,"INJECTION") == NULL ? 0 : -1;
      *pRamping = strstr(scratch,"RAMP") == NULL ? 0 : -1;
      goto fixstate;
    }
  }
fixstate:
  /* fix local machine state variable */
  lumi = DclLumi;
  running = Erunning | Prunning;
  beam = Ebeam | Pbeam;
  ramping = DclEramping | DclPramping;
  return;
declstate:
  /* send declared state triggers based on incoming information */
  /* note: only active if there are checked parameters with beamtype */
  /*       of 'type' or 'procedure' */
  for (i=0; i<nEDeclaredProcedures; i++)
  {
    if (strlen(eDeclaredProcedures[i].typestr) > 0 &&
        stricmp(Etype,eDeclaredProcedures[i].typestr) != 0) continue;
    if (!stricmp(Eproc,eDeclaredProcedures[i].procstr)) break;
  }
  e_stateindex = (i < nEDeclaredProcedures) ? i : 0;
  for (i=0; i<nPDeclaredProcedures; i++)
  {
    if (strlen(pDeclaredProcedures[i].typestr) > 0 &&
        stricmp(Ptype,pDeclaredProcedures[i].typestr) != 0) continue;
    if (!stricmp(Pproc,pDeclaredProcedures[i].procstr)) break;
  }
  p_stateindex = (i < nPDeclaredProcedures) ? i : 0;
  return;
}

int overloadEqpmod(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  return illegal_property;
}
int GlobalsRegistrationEx(void)
{
  return 0;
}

