#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "errors.h"
#include "eventlst.h"
#include "srvlst.h"
#include "swaplib.h"
#include "rpctypes.h"
#include "servdef.h"
#include "toolkit.h"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif
char *getClientName(void);
#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif
#include "arceqm.h"
#include "arcdefs.h"

#ifndef DEF_DEV_SRV_NAME
# define DEF_DEV_SRV_NAME "TTFDIAGADC"
#endif

#ifndef DEF_DEV_NAME
# define DEF_DEV_NAME "BUNCHCURRENT"
#endif

AESnotify  aesNotify;
AEScontext aesContext;
ADRB *g_adrb;    // archive data request block
ADRBX *g_adrbx;  // extended archive data request block
ulong g_eventNumber;

void registerStructs(void);
EXPORTTAG char *getClientTagName(void);

void arcini(void)
{
  static int initialized = 0;

  if (initialized) return;
  registerStructs();
  strncpy(gUserName,getClientTagName(),16);
  initialized = 1;
  return;
}
int GotTriggerEvent = 0;

/* asynchronous callback related routines: */

// many async requests can be outstanding at the same time
// so we keep a callback table

struct
{
  int id;
  BYTE *buffer;
} callbackTable[100];

int getCallbackSlot(void)
{
  int i;
  for (i=0; i<100; i++) if (callbackTable[i].id == 0) return i;
  return -1;
}
int findCallbackSlot(int id)
{
  int i;
  for (i=0; i<100; i++) if (callbackTable[i].id == id) return i;
  return -1;
}
void (*RecordEventCallback)(int cc) = NULL;
void RcdEventCallback(int id,int cc)
{
  int i = findCallbackSlot(id);
  if (i >= 0)
  {
    callbackTable[i].id = 0; 
  }
  if (RecordEventCallback) (*RecordEventCallback)(cc);
}
void (*SetEventCallback)(ulong evnnr,int cc) = NULL;
void SetGlobalEventCallback(int id,int cc)
{
  int i = findCallbackSlot(id);
  if (i >= 0)
  {
    callbackTable[i].id = 0; 
  }
  if (SetEventCallback) (*SetEventCallback)(g_eventNumber,cc);
}

/* for uniqueness, we need to specify a local device id and
   use it as the equipment number.  Although it is not used
   at the server end, it guarantees a unique contract request
   (necessary for asynchronous notification)
*/

#define MAXDEVICES 64 /* this should be plenty */
int ndevices = 0;
char devNameTags[MAXDEVICES][16];

int getDeviceIdFromName(char *devName)
{
  int i;
  for (i=0; i<ndevices && strcmp(devName,devNameTags[i]); i++);
  if (i == ndevices)
  {
    if (ndevices == MAXDEVICES) return -1;
    strncpy(devNameTags[i],devName,16);
    ndevices++;
  }
  return i;
}

EXPORTTAG int RecordEvent(ulong eventNumber,ulong eventCode,char *devName)
{
  int i,id;
  char devname[48];
  DTYPE din;

  aesContext.version = EVENTLST_VERSION; // put in constructor
  aesContext.eventNumber = eventNumber;
  aesContext.eventCode   = eventCode;    // just testing ..
  strncpy(aesContext.deviceName,devName,16);

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESCON");
  din.dArrayLength = 1;
  din.data.vptr = &aesContext;
  // this will be a method latter when I get time ...

  if ((i=getDeviceIdFromName(devName)) < 0) return resources_exhausted;
  sprintf(devname,"%s/#%d",EVENTSRVTAG,i);
  id = AttachLink(devname,"RECORD",NULL,&din,CA_WRITE,0,RcdEventCallback,CM_SINGLE);
  if (id >= 0 && ((i=getCallbackSlot()) >= 0))
  {
    callbackTable[i].id = id;
  }
  return id < 0 ? -id : 0;
}

int defaultTriggerCallback(ulong eventNumber,ulong eventCode)
{
  int cc = 0;
  // do something meaningful
  cout << "reacting to event" << eventNumber << " code " << eventCode << endl;
  cc = RecordEvent(eventNumber,eventCode,"NULL");
  return cc;
}
int (*triggerCallback)(ulong eventNumber,ulong eventCode) = defaultTriggerCallback;
EXPORTTAG void setTriggerCallback(int (*fcn)(ulong eventNumber,ulong eventCode))
{
  triggerCallback = fcn;
}
EXPORTTAG void setRecordEventCallback(void (*fcn)(int cc))
{
  RecordEventCallback = fcn;
}
EXPORTTAG void setSetEventCallback(void (*fcn)(ulong evnnr,int cc))
{
  SetEventCallback = fcn;
}

int (*queryArchiveInfoFunction)(AIB *aib,ulong eventNumber,char *deviceServer) = NULL;

EXPORTTAG void setQueryArchiveInfoFunction(int (*fcn)(AIB *aib,ulong eventNumber,char *deviceServer))
{
  queryArchiveInfoFunction = fcn;
}
int queryArchiveInfo(AEScontext *aescon,AIB *aib)
{
  int cc = not_implemented;

  if (aib == NULL) return un_allocated;
  if (queryArchiveInfoFunction) cc = (*queryArchiveInfoFunction)(aib,aescon->eventNumber,aescon->deviceName);

  return cc;
}
EXPORTTAG int GetArchiveInfo(char *archiveServer,char *deviceServer,ulong evnnum,AIB *aib)
{
  AEScontext aescon;
  DTYPE din,dout;
  if (aib == NULL) return un_allocated;

  aescon.eventNumber = evnnum;
  strncpy(aescon.deviceName,deviceServer,16);
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESCON");
  din.dArrayLength = 1;
  din.data.vptr = &aescon;
  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"AIB");
  dout.dArrayLength = 1;
  dout.data.vptr = aib;
  return ExecLink(archiveServer,"ARCHINFO",&dout,&din,CA_READ);
}

int (*queryPropertyInfoFunction)(PIB *pib,ulong eventNumber,char *deviceServer,char *devProperty) = NULL;

EXPORTTAG void setQueryPropertyInfoFunction(int (*fcn)(PIB *pib,ulong eventNumber,char *deviceServer,char *devProperty))
{
  queryPropertyInfoFunction = fcn;
}
int queryPropertyInfo(AEScontext *aescon,char *property,PIB *pib)
{
  int cc = not_implemented;
  if (pib == NULL) return un_allocated;
  if (queryPropertyInfoFunction) 
    cc = (*queryPropertyInfoFunction)(pib,aescon->eventNumber,aescon->deviceName,property);

  return cc;
}
EXPORTTAG int GetPropertyInfo(char *archiveServer,char *deviceServer,char *deviceProperty,ulong evnnum,PIB *pib)
{
  DTYPE din,dout;
  AEScontext aescon;
  char fullProperty[32];
  if (pib == NULL) return un_allocated;

  if (strlen(deviceProperty) > 22) return string_too_long;
  aescon.eventNumber = evnnum;
  strncpy(aescon.deviceName,deviceServer,16);
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESCON");
  din.dArrayLength = 1;
  din.data.vptr = &aescon;
  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"PIB");
  dout.dArrayLength = 1;
  dout.data.vptr = pib;
  sprintf(fullProperty,"PROPINFO.%s",deviceProperty);
  return ExecLink(archiveServer,fullProperty,&dout,&din,CA_READ);
}

int (*queryNumLocalArchivedEventsFunction)(ulong starttime,ulong stoptime,ulong eventNumber) = NULL;

EXPORTTAG void setQueryNumLocalArchivedEventsFunction(int (*fcn)(ulong starttime,ulong stoptime,ulong eventNumber))
{
  queryNumLocalArchivedEventsFunction = fcn;
}
int queryNumLocalArchivedEvents(ulong starttime,ulong stoptime,ulong eventNumber)
{
  int num = 0;
  if (queryNumLocalArchivedEventsFunction) num = (*queryNumLocalArchivedEventsFunction)(starttime,stoptime,eventNumber);
  return num;
}
EXPORTTAG int GetNumArchiveServerEvents(char *archiveServer,ulong starttime,ulong stoptime,ulong eventNumber)
{
  ulong data[3],num;
  int cc;
  DTYPE dout,din;

  din.dFormat = CF_LONG;
  din.dArrayLength = 3;
  din.data.lptr = data;
  data[0] = starttime;
  data[1] = stoptime;
  data[2] = eventNumber;
  dout.dFormat = CF_LONG;
  dout.dArrayLength = 1;
  dout.data.lptr = &num;
  cc = ExecLink(archiveServer,"NEVENTS",&dout,&din,CA_READ);
  return cc ? -1 : (int)data[0];
}

int (*queryLocalArchivedEventsFunction)(ulong starttime,ulong stoptime,AES *aes,ulong num) = NULL;

EXPORTTAG void setQueryLocalArchivedEventsFunction(int (*fcn)(ulong starttime,ulong stoptime,AES *aes,ulong num))
{
  queryLocalArchivedEventsFunction = fcn;
}
int queryLocalArchivedEvents(ulong starttime,ulong stoptime,AES *aes,ulong num)
{
  int cc = 0;
  if (queryLocalArchivedEventsFunction) cc = (*queryLocalArchivedEventsFunction)(starttime,stoptime,aes,num);
  return cc;
}

int (*queryNumDeviceArchivedEventsFunction)(char *devsrv,char *device,ulong starttime,ulong stoptime) = NULL;

EXPORTTAG void setQueryNumDeviceArchivedEventsFunction(int (*fcn)(char *devsrv,char *device,ulong starttime,ulong stoptime))
{
  queryNumDeviceArchivedEventsFunction = fcn;
}
int queryNumDeviceArchivedEvents(char *devsrv,char *device,ulong starttime,ulong stoptime)
{
  int cc = 0;
  if (queryNumDeviceArchivedEventsFunction) cc = (*queryNumDeviceArchivedEventsFunction)(devsrv,device,starttime,stoptime);
  return cc;
}

int (*queryDeviceArchivedEventsFunction)(char *devsrv,char *device,ulong starttime,ulong stoptime,AES *aes,ulong num) = NULL;

EXPORTTAG void setQueryDeviceArchivedEventsFunction(int (*fcn)(char *devsrv,char *device,ulong starttime,ulong stoptime,AES *aes,ulong num))
{
  queryDeviceArchivedEventsFunction = fcn;
}
int queryDeviceArchivedEvents(char *devsrv,char *device,ulong starttime,ulong stoptime,AES *aes,ulong num)
{
  int cc = 0;
  if (queryDeviceArchivedEventsFunction) cc = (*queryDeviceArchivedEventsFunction)(devsrv,device,starttime,stoptime,aes,num);
  return cc;
}

int (*queryDeviceArchivedEventListFunction)(char *devsrv,char *device,ulong starttime,ulong stoptime,ulong *evn,ulong num) = NULL;

EXPORTTAG void setQueryDeviceArchivedEventListFunction(int (*fcn)(char *devsrv,char *device,ulong starttime,ulong stoptime,ulong *evn,ulong num))
{
  queryDeviceArchivedEventListFunction = fcn;
}
int queryDeviceArchivedEventList(char *devsrv,char *device,ulong starttime,ulong stoptime,ulong *evn,ulong num)
{
  int cc = 0;
  if (queryDeviceArchivedEventListFunction) cc = (*queryDeviceArchivedEventListFunction)(devsrv,device,starttime,stoptime,evn,num);
  return cc;
}

int (*queryNumLocalDeviceServersFunction)(void) = NULL;

EXPORTTAG void setQueryNumLocalDeviceServersFunction(int (*fcn)(void))
{
  queryNumLocalDeviceServersFunction = fcn;
}
int queryNumDeviceServers(void)
{
  int num = 0;
  if (queryNumLocalDeviceServersFunction) num = (*queryNumLocalDeviceServersFunction)();
  return num;
}

int (*queryNumLocalDevicePropertiesFunction)(char *devServer,char *devName) = NULL;

EXPORTTAG void setQueryNumLocalDevicePropertiesFunction(int (*fcn)(char *devServer,char *devName))
{
  queryNumLocalDevicePropertiesFunction = fcn;
}
int queryNumDeviceProperties(char *devServer,char *devName)
{
  int num = 0;
  if (queryNumLocalDevicePropertiesFunction) num = (*queryNumLocalDevicePropertiesFunction)(devServer,devName);
  return num;
}

int (*queryNumLocalDevicesFunction)(char *devServer,char *context,int id) = NULL;

EXPORTTAG void setQueryNumLocalDevicesFunction(int (*fcn)(char *devServer,char *context,int id))
{
  queryNumLocalDevicesFunction = fcn;
}
int queryNumDevices(char *devServer,char *context,int id)
{
  int num = 0;
  if (queryNumLocalDevicesFunction) num = (*queryNumLocalDevicesFunction)(devServer,context,id);
  return num;
}
EXPORTTAG int GetNumArchiveDeviceServers(char *archiveServer)
{
  int cc;
  ulong num;
  DTYPE dout;

  dout.dFormat = CF_LONG;
  dout.dArrayLength = 1;
  dout.data.lptr = &num;

  cc = ExecLink(archiveServer,"NDSERVERS",&dout,NULL,CA_READ);

  return cc ? -1 : (int)num;
}
EXPORTTAG int GetNumArchiveDevices(char *archiveServer,char *devServer)
{
  int cc;
  ulong num;
  DTYPE dout;
  char devnam[64];

  dout.dFormat = CF_LONG;
  dout.dArrayLength = 1;
  dout.data.lptr = &num;

  sprintf(devnam,"%s/%s",archiveServer,devServer);
  cc = ExecLink(devnam,"NADEVICES",&dout,NULL,CA_READ);

  return cc ? -1 : (int)num;
}

int (*queryLocalDeviceServerNamesFunction)(NAME16 *names,ulong start,ulong stop,int id,int num) = NULL;

EXPORTTAG void setQueryLocalDeviceServerNamesFunction(int (*fcn)(NAME16 *names,ulong start,ulong stop,int id,int num))
{
  queryLocalDeviceServerNamesFunction= fcn;
}
int queryDeviceServerNames(NAME16 *names,ulong start,ulong stop,int id,int num)
{
  int cc = 0;
  if (queryLocalDeviceServerNamesFunction) cc = (*queryLocalDeviceServerNamesFunction)(names,start,stop,id,num);
  return cc;
}
int (*queryLocalDevicesFunction)(NAME16 *names,char *devServer,char *context,int id,int num) = NULL;

EXPORTTAG void setQueryLocalDeviceNamesFunction(int (*fcn)(NAME16 *names,char *devServer,char *context,int id,int num))
{
  queryLocalDevicesFunction= fcn;
}
int queryDevices(NAME16 *names,char *devServer,char *context,int id,int num)
{
  int cc = 0;
  if (queryLocalDevicesFunction) cc = (*queryLocalDevicesFunction)(names,devServer,context,id,num);
  return cc;
}
int (*queryLocalDevicePropertiesFunction)(NAME16 *names,char *devServer,char *devName,int num) = NULL;

EXPORTTAG void setQueryLocalDevicePropertiesFunction(int (*fcn)(NAME16 *names,char *devServer,char *devName,int num))
{
  queryLocalDevicePropertiesFunction= fcn;
}
int queryDeviceProperties(NAME16 *names,char *devServer,char *devName,int num)
{
  int cc = 0;
  if (queryLocalDevicePropertiesFunction) cc = (*queryLocalDevicePropertiesFunction)(names,devServer,devName,num);
  return cc;
}

EXPORTTAG int GetArchiveDeviceServerNames(char *archiveServer,ulong start,ulong stop,NAME16 *names,int *num)
{
  int cc;
  UINT32 timebuf[2];
  DTYPE dout,din;

  dout.dFormat = CF_CHAR16;
  dout.dArrayLength = *num;
  dout.data.vptr = names;

  din.dFormat = CF_LONG;
  din.dArrayLength = 2;
  timebuf[0] = start; timebuf[1] = stop;
  din.data.lptr = timebuf;

  cc = ExecLink(archiveServer,"DSERVERS",&dout,&din,CA_READ);
  return cc ? -1 : (int)*num;
}
EXPORTTAG int GetArchiveDeviceNames(char *archiveServer,char *devSrvName,NAME16 *names,int *num)
{
  int cc;
  DTYPE dout;
  char dname[48];

  sprintf(dname,"%s/%s",archiveServer,devSrvName);

  dout.dFormat = CF_CHAR16;
  dout.dArrayLength = *num;
  dout.data.vptr = names;

  cc = ExecLink(dname,"ADEVICES",&dout,NULL,CA_READ);
  return cc ? -1 : (int)*num;
}
void arctsk(void)
{
  if (GotTriggerEvent)
  {
    if (triggerCallback) (*triggerCallback)(aesNotify.eventNumber,aesNotify.eventCode);
    GotTriggerEvent = 0;
  }
}
EXPORTTAG int SetGlobalEventComment(ulong eventCode,char *cmt,char *mode)
{
  AESnotify aesn;
  char devname[48];
  short modeNo = strchr(mode,'a') ? 1 : 0;
  DTYPE din;

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESNOT");
  din.dArrayLength = 1;
  din.data.vptr = &aesn;
  aesn.version = EVENTLST_VERSION;
  aesn.eventCode = eventCode;
  strncpy(aesn.comment[0],cmt,sizeof(aesn.comment));
  sprintf(devname,"%s/#%d",EVENTSRVTAG,modeNo);
  return ExecLink(devname,"SETCOMMENT",NULL,&din,CA_WRITE);
}
EXPORTTAG int SetGlobalEvent(ulong eventCode,char *cmt)
{
  int i,id;
  static AESnotify aesn;
  DTYPE din,dout;

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESNOT");
  din.dArrayLength = 1;
  din.data.vptr = &aesn;

  dout.dFormat = CF_LONG;
  dout.dArrayLength = 1;
  g_eventNumber = time(NULL); // set it to now in case the call fails!
  dout.data.lptr = &g_eventNumber;

  // make AESnotify a class and put this in the constructor:
  aesn.version = EVENTLST_VERSION;
  aesn.eventCode = eventCode;  // we'll have to work on this
  strncpy(aesn.comment[0],cmt,64);
  aesn.comment[1][0] = 0; aesn.comment[2][0] = 0; aesn.comment[3][0] = 0;
  // this will be a method latter when I get time ...
  id = AttachLink(EVENTSRVTAG,"SET",&dout,&din,CA_WRITE|CA_READ,0,SetGlobalEventCallback,CM_SINGLE);
  if (id >= 0 && ((i=getCallbackSlot()) >= 0))
  {
    callbackTable[i].id = id;
  }
  return id < 0 ? -id : 0;
}

int _AnnotateArchiveEventFunction(ulong eventNumber,char *dsrv,char *cmt) { return not_implemented; }
int (*AnnotateArchiveEventFunction)(ulong eventNumber,char *dsrv,char *cmt) = _AnnotateArchiveEventFunction;
EXPORTTAG void setAnnotateArchiveEventFunction(int (*fcn)(ulong eventNumber,char *dsrv,char *cmt))
{
  if (fcn != NULL) AnnotateArchiveEventFunction = fcn;
}

int _RemoveArchiveEventsFunction(ulong eventStart,ulong eventLast,char *dsrv) { return not_implemented; }
int (*RemoveArchiveEventsFunction)(ulong eventFirst,ulong eventLast,char *dsrv) = _RemoveArchiveEventsFunction;
EXPORTTAG void setRemoveArchiveEventsFunction(int (*fcn)(ulong eventFirst,ulong eventLast,char *dsrv))
{
  if (fcn != NULL) RemoveArchiveEventsFunction = fcn;
}
int _RemoveArchiveEventFunction(ulong eventNumber,char *dsrv) { return not_implemented; }
int (*RemoveArchiveEventFunction)(ulong eventNumber,char *dsrv) = _RemoveArchiveEventFunction;
EXPORTTAG void setRemoveArchiveEventFunction(int (*fcn)(ulong eventNumber,char *dsrv))
{
  if (fcn != NULL) RemoveArchiveEventFunction = fcn;
}
// old style:
int (*GetArchiveDataFunction)(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier) = NULL;
EXPORTTAG void setGetArchiveDataFunction(int (*fcn)(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier))
{
  GetArchiveDataFunction = fcn;
}
int _GetArchiveData(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier)
{
   int cc = 0;

   if (GetArchiveDataFunction == NULL) return not_implemented;
   if (NGdebug) 
   {
     cout << "Get data for event " << eventNumber << "(" << starttime ;
     cout << " to " << stoptime << ") " << devSpecifier->dataDimension ;
     cout << " as " << devSpecifier->dataFormat << endl;
   }
   return (*GetArchiveDataFunction)((BYTE *)data,eventNumber,starttime,stoptime,eventContext,devSpecifier);
}
// new style:
int (*GetArchiveDataFunctionEx)(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier) = NULL;
EXPORTTAG void setGetArchiveDataFunctionEx(int (*fcn)(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier))
{
  GetArchiveDataFunctionEx = fcn;
}
int _GetArchiveDataEx(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier)
{
   int cc = 0;

   if (GetArchiveDataFunctionEx == NULL) return not_implemented;
   if (NGdebug) 
   {
     cout << "Get extended data for event " << eventNumber << "(" << starttime ;
     cout << " to " << stoptime << ") " << devSpecifier->dataDimension ;
     cout << " entries as " << devSpecifier->dataFormat;
     cout << " for index " << devSpecifier->dataIndex << endl;
   }
   return (*GetArchiveDataFunctionEx)((BYTE *)data,eventNumber,starttime,stoptime,eventContext,devSpecifier);
}
//old style:
int (*GetArchiveHeaderFunction)(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier) = NULL;
EXPORTTAG void setGetArchiveHeaderFunction(int (*fcn)(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier))
{
  GetArchiveHeaderFunction = fcn;
}
int _GetArchiveHeader(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier)
{
   int cc = 0;

   if (GetArchiveHeaderFunction == NULL) return not_implemented;
   if (NGdebug) 
   {
     cout << "Get header for event " << eventNumber << "(" << starttime ;
     cout << " to " << stoptime << ") " << devSpecifier->dataDimension ;
     cout << " as " << devSpecifier->dataFormat << endl;
   }
   return (*GetArchiveHeaderFunction)((DHS *)dhs,eventNumber,starttime,stoptime,eventContext,devSpecifier);
}
// new style:
int (*GetArchiveHeaderFunctionEx)(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier) = NULL;
EXPORTTAG void setGetArchiveHeaderFunctionEx(int (*fcn)(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier))
{
  GetArchiveHeaderFunctionEx = fcn;
}
int _GetArchiveHeaderEx(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier)
{
   int cc = 0;

   if (GetArchiveHeaderFunctionEx == NULL) return not_implemented;
   if (NGdebug) 
   {
     cout << "Get extended header for event " << eventNumber << "(" << starttime ;
     cout << " to " << stoptime << ") " << devSpecifier->dataDimension ;
     cout << " as " << devSpecifier->dataFormat << endl;
   }
   return (*GetArchiveHeaderFunctionEx)((DHS *)dhs,eventNumber,starttime,stoptime,eventContext,devSpecifier);
}

// The following are general query routines for event information from the event
// server.  In practice they will be called from a console client and not from 
// an archive server.

EXPORTTAG int GetRemoteArchiveData(char *arcsrvTag,BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,DEVSPEC *devSpecifier)
{
  ADRB adrb;
  DTYPE dout,din;

  adrb.eventNumber = eventNumber;
  adrb.starttime = starttime;
  adrb.stoptime = stoptime;
  memcpy(&adrb.devSpecifier,devSpecifier,sizeof(DEVSPEC));
  if (NGdebug)
  {
    cout << "Get : " << adrb.devSpecifier.deviceName << " " << adrb.devSpecifier.deviceProperty ;
    cout << " " << adrb.devSpecifier.dataDimension << " elements as ";
    cout << adrb.devSpecifier.dataFormat << " at " << adrb.starttime;
    cout << " to " << adrb.stoptime << endl;
  }
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"ADRB");
  din.dArrayLength = 1;
  din.data.vptr = &adrb;

  dout.dFormat = devSpecifier->dataFormat;
  dout.dArrayLength = devSpecifier->dataDimension;
  dout.data.vptr = data;

  return ExecLink(arcsrvTag,"ARCHDATA",&dout,&din,CA_READ);
}

EXPORTTAG int GetRemoteArchiveHeader(char *arcsrvTag,DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier)
{
  ADRB adrb;
  DTYPE dout,din;
  char srvDevName[64];

  adrb.eventNumber = eventNumber;
  adrb.starttime = starttime;
  adrb.stoptime = stoptime;
  memcpy(&adrb.devSpecifier,devSpecifier,sizeof(DEVSPEC));
  if (NGdebug)
  {
    cout << "Get : " << adrb.devSpecifier.deviceName << " " << adrb.devSpecifier.deviceProperty ;
    cout << " data header " << endl;
  }
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"ADRB");
  din.dArrayLength = 1;
  din.data.vptr = &adrb;

  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"DHS");
  dout.dArrayLength = 1;
  dout.data.vptr = dhs;

  if (eventContext != NULL && strlen(eventContext) > 0)
    sprintf(srvDevName,"%s/%s",arcsrvTag,eventContext);
  else 
    sprintf(srvDevName,"%s",arcsrvTag);
  return ExecLink(srvDevName,"ARCHHEADER",&dout,&din,CA_READ);
}

EXPORTTAG int GetNumberOfEventsEx(time_t starttime,time_t stoptime,ulong eventCode)
{
  static ulong nevents[3],num;
  int cc,n;
  DTYPE din,dout;

  nevents[0] = starttime; nevents[1] = stoptime; nevents[2] = eventCode;
  n = eventCode > 0 ? 3 : 2;

  din.dFormat = CF_LONG;
  din.dArrayLength = n;
  din.data.lptr = nevents;

  dout.dFormat = CF_LONG;
  dout.dArrayLength = 1;
  dout.data.lptr = &num;

  cc = ExecLink(EVENTSRVTAG,"NEVENTS",&dout,&din,CA_READ);
  return cc ? -cc : (int)num;
}
EXPORTTAG int GetNumberOfEvents(time_t starttime,time_t stoptime)
{
  return GetNumberOfEventsEx(starttime,stoptime,0);
}

EXPORTTAG int GetNumberOfDeviceEventsEx(char *device)
{
  static ulong num;
  int cc;
  DTYPE din,dout;
  char deviceName[16] = "";

  strncpy(deviceName,device,16);

  din.dFormat = CF_CHAR16;
  din.dArrayLength = 1;
  din.data.cptr = deviceName;

  dout.dFormat = CF_LONG;
  dout.dArrayLength = 1;
  dout.data.lptr = &num;

  cc = ExecLink(EVENTSRVTAG,"NDEVEVENTS",&dout,&din,CA_READ);
  return cc ? -cc : (int)num;
}
EXPORTTAG int GetNumberOfDeviceEvents(char *device)
{
  return GetNumberOfDeviceEventsEx(device);
}
EXPORTTAG int GetArchivedEvents(char *serverName,time_t starttime,time_t stoptime,ulong eventCode,AES *aes,int *num)
{
  ulong evndata[3];
  int cc,n = *num;
  DTYPE din,dout;

  evndata[0] = starttime;
  evndata[1] = stoptime;
  evndata[2] = eventCode;

  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"AES");
  dout.dArrayLength = n;
  dout.data.vptr = aes;

  din.dFormat = CF_LONG;
  din.dArrayLength = 3;
  din.data.lptr = evndata;

  cc = ExecLink(serverName,"EVENTS",&dout,&din,CA_READ);

  *num = GetNumberOfEventsEx(starttime,stoptime,eventCode);
  return cc;
}

EXPORTTAG int GetEvents(time_t starttime,time_t stoptime,ulong eventCode,AES *aes,int *num)
{
  return GetArchivedEvents(EVENTSRVTAG,starttime,stoptime,eventCode,aes,num);
}
EXPORTTAG int GetDeviceArchivedEvents(char *archiveServer,char *devServer,char *device,ulong starttime,ulong stoptime,AES *aes,int *num)
{
  int cc,n = *num;
  DTYPE din,dout;
  NAME16I startstop[2];

  startstop[0].ival = starttime;
  startstop[1].ival = stoptime;
  strncpy(startstop[0].name,device,16);

  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"AES");
  dout.dArrayLength = n;
  dout.data.vptr = aes;

  din.dFormat = CF_NAME16I;
  din.dArrayLength = 2;
  din.data.vptr = startstop;

  cc = ExecLink(archiveServer,"DEVEVENTS",&dout,&din,CA_READ);

  /* *num = GetNumberOfDeviceEventsEx(deviceName); */
  return cc;
}

EXPORTTAG int GetDeviceEvents(char *device,AES *aes,int *num)
{
//  return GetDeviceArchivedEvents(EVENTSRVTAG,device,aes,num);
  return 0;
}
EXPORTTAG int GetNumberOfArchiveServers(void)
{
  short cc;
  unsigned short nsrv;
  DTYPE dout;
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = &nsrv;
  cc = ExecLink(EVENTSRVTAG,"NASERVERS",&dout,NULL,CA_READ);
  return cc ? -cc : (int)nsrv;
}
EXPORTTAG int GetArchiveServers(int idSrv,NAME16 *asrv,int *num)
{
  short cc;
  int n = *num;
  DTYPE dout;
  char devname[48];

  dout.dFormat = CF_CHAR16;
  dout.dArrayLength = n;
  dout.data.vptr = asrv;
  sprintf(devname,"%s/#%d",EVENTSRVTAG,idSrv);
  if (idSrv >= 0 && n != 1) return -out_of_range;
  cc = ExecLink(devname,"ASERVERS",&dout,NULL,CA_READ);
  *num = GetNumberOfArchiveServers();
  return cc;
}
EXPORTTAG int GetNumberOfDeviceServers(char *archiveServer)
{
  short cc,ndev;
  DTYPE dout;
  char devname[48];

  if (archiveServer != NULL)
    sprintf(devname,"%s/%s",EVENTSRVTAG,archiveServer);
  else
    sprintf(devname,"%s/ALL",EVENTSRVTAG);

  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = (UINT16 *)&ndev;
  cc = ExecLink(devname,"NDSERVERS",&dout,NULL,CA_READ);
  return cc ? -cc : (int)ndev;
}
EXPORTTAG int GetDeviceServersX(char *archiveServer,int idDev,NAME16 *dsrv,int *num)
{
  short cc;
  int n = *num;
  DTYPE dout;
  char devname[48];

  if (idDev >= 0 && n != 1) return -out_of_range;
  dout.dFormat = CF_CHAR16;
  dout.dArrayLength = n;
  dout.data.vptr = dsrv;
  if (archiveServer == NULL)
    sprintf(devname,"%s/#%d",EVENTSRVTAG,idDev);
  else
    sprintf(devname,"%s/%s",EVENTSRVTAG,archiveServer);

  cc = ExecLink(devname,"DSERVERS",&dout,NULL,CA_READ);  
  *num = GetNumberOfDeviceServers(archiveServer);
  return cc;
}

// This is the archive server equipment function.  This is where all of the
// server action is:
UINT32 archiveState = 0;
UINT32 archiveWindow = 30;
enum {
  TriggerNormal = 0,
  TriggerNext,
  TriggerWindow,
  TriggerRefuse
} archiveStates;
int isTriggerAllowed(void)
{
  static time_t lasttrigger = 0;
  time_t t = time(NULL);
  
  switch (archiveState)
  {
    case TriggerNormal: 
      lasttrigger = t;
      return TRUE;
    case TriggerNext:
      lasttrigger = t;
      archiveState = TriggerRefuse;
      return TRUE;
    case TriggerWindow:
      if (t-lasttrigger > (time_t)archiveWindow)
      {
        lasttrigger = t;
        return TRUE;
      }
      return FALSE;
    case TriggerRefuse:
      return FALSE;
  }
  return FALSE;
}
EXPORTTAG int arceqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char remSrvName[20];
  int devNr,eventNumber;
  int num,i,cc;
  static AES aes[20];
  NAME16 srv[10],dev[10];
  ulong starttime,stoptime,evncode;
  char devServerName[16];
  char device[16];

  if (!strcmp(devProperty,"TRIGGER"))  // respond to external trigger request
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isTriggerAllowed()) return not_accepted;
    if ( din->dFormat != CF_STRUCT ||
         strcmp(din->dTag,"AESNOT") ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    memcpy(&aesNotify,din->data.vptr,sizeof(AESnotify));
    if (aesNotify.eventCode == EVENT_IGNORE) return 0;
    strncpy(remSrvName,getClientName(),16);
    if (NGdebug)
      cout << "EVNEQM: received TRIGGER" << aesNotify.eventNumber << " from " << remSrvName << endl;
    // set trigger flag so that the call returns quickly
    GotTriggerEvent = 1;
    // handle the trigger asynchronously
    return 0;
  }
  if (!strcmp(devProperty,"ARCHINFO"))  // archive info data
  {
    if (access & CA_WRITE) return illegal_read_write;
    if ( din->dFormat != CF_STRUCT ||
         strcmp(din->dTag,"AESCON") ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    if ( dout->dFormat != CF_STRUCT ||
         strcmp(dout->dTag,"AIB") ) return illegal_format;
    if (dout->dArrayLength != 1) return dimension_error;
    strncpy(remSrvName,getClientName(),16);
    return queryArchiveInfo((AEScontext *)din->data.vptr,(AIB *)dout->data.vptr);
  }
  if (!strcmp(devProperty,"ARCHSTATE"))
  {
    if (access & CA_WRITE)
    {
      if ((cc=getValuesAsLong(din,&archiveState,1)) != 0) return cc;
    } 
    if (dout->dArrayLength > 0) cc = putValuesFromLong(dout,&archiveState,1);
    return cc;
  }
  if (!strcmp(devProperty,"ARCHWINDOW"))
  {
    if (access & CA_WRITE)
    {
      if ((cc=getValuesAsLong(din,&archiveWindow,1)) != 0) return cc;
    }
    if (dout->dArrayLength > 0) cc = putValuesFromLong(dout,&archiveWindow,1);
    return cc;
  }
  if (!strncmp(devProperty,"PROPINFO",8))  // property info data
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devProperty[8] != '.') return illegal_delimiter;
    if ( din->dFormat != CF_STRUCT ||
         strcmp(din->dTag,"AESCON") ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    if ( dout->dFormat != CF_STRUCT ||
         strcmp(dout->dTag,"PIB") ) return illegal_format;
    if (dout->dArrayLength != 1) return dimension_error;

    return queryPropertyInfo((AEScontext *)din->data.vptr,&devProperty[9],(PIB *)dout->data.vptr);
  }
  if (!strcmp(devProperty,"ARCHDATA"))  // respond to request for archive data
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    if (dout->dArrayLength < 1) return dimension_error;
    if (!strcmp(din->dTag,"ADRB"))
    {
      g_adrb = (ADRB *)din->data.vptr;
      g_adrb->devSpecifier.dataDimension = dout->dArrayLength;
      g_adrb->devSpecifier.dataFormat = dout->dFormat;
      // use devName as event context
      return _GetArchiveData((BYTE *)dout->data.vptr,g_adrb->eventNumber,g_adrb->starttime,g_adrb->stoptime,devName,&g_adrb->devSpecifier);
    }
    if (!strcmp(din->dTag,"ADRBX"))
    {
      g_adrbx = (ADRBX *)din->data.vptr;
      g_adrbx->devSpecifier.dataDimension = dout->dArrayLength;
      g_adrbx->devSpecifier.dataFormat = dout->dFormat;
      // use devName as event context
      return _GetArchiveDataEx((BYTE *)dout->data.vptr,g_adrbx->eventNumber,g_adrbx->starttime,g_adrbx->stoptime,devName,&g_adrbx->devSpecifier);
    }
    return illegal_format;
  }
  if (!strcmp(devProperty,"ARCHHEADER"))  // respond to request for archive data header
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dout->dFormat != CF_STRUCT || strcmp(dout->dTag,"DHS")) return illegal_format;
    if (dout->dArrayLength < 1) return dimension_error;
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    if (!strcmp(din->dTag,"ADRB"))
    {
      g_adrb = (ADRB *)din->data.vptr;
      g_adrb->devSpecifier.dataDimension = dout->dArrayLength;
      g_adrb->devSpecifier.dataFormat = dout->dFormat;
      // use devName as event context
      return _GetArchiveHeader((DHS *)dout->data.vptr,g_adrb->eventNumber,g_adrb->starttime,g_adrb->stoptime,devName,&g_adrb->devSpecifier);
    }
    if (!strcmp(din->dTag,"ADRBX"))
    {
      g_adrbx = (ADRBX *)din->data.vptr;
      g_adrbx->devSpecifier.dataDimension = dout->dArrayLength;
      g_adrbx->devSpecifier.dataFormat = dout->dFormat;
      // use devName as event context
      return _GetArchiveHeaderEx((DHS *)dout->data.vptr,g_adrbx->eventNumber,g_adrbx->starttime,g_adrbx->stoptime,devName,&g_adrbx->devSpecifier);
    }
    return illegal_format;
  }
  if (!strcmp(devProperty,"NEVENTS"))
  {
    if (din->dFormat != CF_LONG) return illegal_format;
    if (dout->dFormat != CF_LONG) return illegal_format;
    evncode = 0;
    switch (din->dArrayLength)
    {
      case 3: evncode   = din->data.lptr[2];
      case 2: starttime = din->data.lptr[0];
              stoptime  = din->data.lptr[1];
              break;
      default: return illegal_read_write;
    }
    *dout->data.lptr = queryNumLocalArchivedEvents(starttime,stoptime,evncode);
    return 0;
  }
  if (!strcmp(devProperty,"EVENTS"))
  {
    if (din->dFormat != CF_LONG) return illegal_format;
    if (dout->dFormat != CF_STRUCT ||
        strcmp(dout->dTag,"AES") ) return illegal_format;
    starttime = din->data.lptr[0];
    stoptime  = din->data.lptr[1];
    evncode   = din->data.lptr[2];
    return queryLocalArchivedEvents(starttime,stoptime,(AES *)dout->data.vptr,dout->dArrayLength);
  }
  if (!strcmp(devProperty,"NDSERVERS"))
  {
    if (dout->dArrayLength != 1) return illegal_read_write;
    switch (dout->dFormat)
    {
      case CF_INTEGER: *dout->data.sptr = queryNumDeviceServers(); break;
      case CF_LONG: *dout->data.lptr = queryNumDeviceServers(); break;
      default: return illegal_format;
    }
    return 0;
  }
  if (!strcmp(devProperty,"DSERVERS"))
  {
    starttime = 0; stoptime = time(NULL);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 2) return illegal_read_write;
      if (din->dFormat != CF_LONG) return illegal_format;
      starttime = din->data.lptr[0];
      stoptime = din->data.lptr[1];
    }
    if (dout->dFormat != CF_CHAR16) return illegal_format;
    if (devName[0] == '#') devNr = atoi(&devName[1]);
    else if (!strcmp(devName,"ALL")) devNr = -1;
    return queryDeviceServerNames((NAME16 *)dout->data.vptr,starttime,stoptime,devNr,dout->dArrayLength);
  }

  if (!strcmp(devProperty,"NAPROPS"))
  {
    char dev[16];
    if (strlen(devName) != 0) strncpy(devServerName,devName,16);
    else strncpy(devServerName,DEF_DEV_SRV_NAME,16);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      if (din->dFormat != CF_NAME16) return illegal_format;
      strncpy(dev,din->data.cptr,16);
    }
    if (dout->dArrayLength != 1) return illegal_read_write;
    switch (dout->dFormat)
    {
      case CF_INTEGER: *dout->data.sptr = queryNumDeviceProperties(devServerName,dev); break;
      case CF_LONG: *dout->data.lptr = queryNumDeviceProperties(devServerName,dev); break;
      default: return illegal_format;
    }
    return 0;
  }

  if (!strcmp(devProperty,"APROPS"))
  {
    char dev[16] = { "" };
    if (strlen(devName) != 0) strncpy(devServerName,devName,16);
    else strncpy(devServerName,DEF_DEV_SRV_NAME,16);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      if (din->dFormat != CF_NAME16) return illegal_format;
      strncpy(dev,din->data.cptr,16);
    }
    if (dout->dFormat != CF_CHAR16) return illegal_format;
    return queryDeviceProperties((NAME16 *)dout->data.vptr,devServerName,dev,dout->dArrayLength);
  }

  if (!strcmp(devProperty,"NADEVICES"))
  {
    int evnnum = 0;
    char context[16];
    context[0]  = 0;
    if (strlen(devName) != 0) strncpy(devServerName,devName,16);
    else strncpy(devServerName,DEF_DEV_SRV_NAME,16);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      switch (din->dFormat)
      {
        case CF_LONG: 
             evnnum = din->data.lptr[0];
             break;
        case CF_NAME16:
             strncpy(context,((NAME16 *)din->data.vptr)[0].name,16);
             break;
        case CF_NAME16I:
             strncpy(context,((NAME16I *)din->data.vptr)[0].name,16);
             evnnum = ((NAME16I *)din->data.vptr)[0].ival;
             break;
        default: return illegal_format;
      }
    }
    if (dout->dArrayLength != 1) return illegal_read_write;
    switch (dout->dFormat)
    {
      case CF_INTEGER: *dout->data.sptr = queryNumDevices(devServerName,context,evnnum); break;
      case CF_LONG: *dout->data.lptr = queryNumDevices(devServerName,context,evnnum); break;
      default: return illegal_format;
    }
    return 0;
  }

  if (!strcmp(devProperty,"ADEVICES"))
  {
    int evnnum = 0;
    char context[16];
    context[0]  = 0;
    if (strlen(devName) != 0) strncpy(devServerName,devName,16);
    else strncpy(devServerName,DEF_DEV_SRV_NAME,16);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      switch (din->dFormat)
      {
        case CF_LONG:
             evnnum = din->data.lptr[0];
             break;
        case CF_NAME16:
             strncpy(context,((NAME16 *)din->data.vptr)[0].name,16);
             break;
        case CF_NAME16I:
             strncpy(context,((NAME16I *)din->data.vptr)[0].name,16);
             evnnum = ((NAME16I *)din->data.vptr)[0].ival;
             break;
        default: return illegal_format;
      }
    }
    if (dout->dFormat != CF_CHAR16) return illegal_format;
    return queryDevices((NAME16 *)dout->data.vptr,devServerName,context,evnnum,dout->dArrayLength);
  }
  if (!strcmp(devProperty,"NDEVEVENTS"))
  {
    int start = 0, stop = time(NULL);
    if (strlen(devName) != 0) strncpy(devServerName,devName,16);
    else strncpy(devServerName,DEF_DEV_SRV_NAME,16);
    strncpy(device,DEF_DEV_NAME,16);
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_LONG: if (din->dArrayLength != 2) return illegal_format;
                      start = din->data.lptr[0];
                      stop  = din->data.lptr[1];
                      break;
        case CF_NAME16I: if (din->dArrayLength != 2) return illegal_format;
                      start = ((NAME16I *)din->data.vptr)[0].ival;
                      stop = ((NAME16I *)din->data.vptr)[1].ival;
                      strncpy(device,((NAME16I *)din->data.vptr)[0].name,16);
                      break;
      }
    }
    int nevn;
    nevn = queryNumDeviceArchivedEvents(devServerName,device,start,stop);
    switch (dout->dFormat)
    {
      case CF_SHORT: dout->data.sptr[0] = (UINT16)nevn; break;
      case CF_LONG: dout->data.lptr[0] = nevn; break;
      default: return illegal_format;
    }
    return nevn < 0 ? -nevn : 0;
  }
  if (!strcmp(devProperty,"DEVEVENTS"))
  {
    int start = 0, stop = time(NULL);
    if (strlen(devName) != 0) strncpy(devServerName,devName,16);
    else strncpy(devServerName,DEF_DEV_SRV_NAME,16);
    strncpy(device,DEF_DEV_NAME,16);
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_LONG: if (din->dArrayLength != 2) return illegal_format;
                      start = din->data.lptr[0];
                      stop  = din->data.lptr[1];
                      break;
        case CF_NAME16I: if (din->dArrayLength != 2) return illegal_format;
                      start = ((NAME16I *)din->data.vptr)[0].ival;
                      stop = ((NAME16I *)din->data.vptr)[1].ival;
                      strncpy(device,((NAME16I *)din->data.vptr)[0].name,16);
                      break;
      }
    }
    if (dout->dFormat == CF_STRUCT && !strcmp(dout->dTag,"AES") ) 
      return queryDeviceArchivedEvents(devServerName,device,start,stop,(AES *)dout->data.vptr,dout->dArrayLength);
    if (dout->dFormat == CF_LONG)
      return queryDeviceArchivedEventList(devServerName,device,start,stop,dout->data.lptr,dout->dArrayLength);

    return illegal_format;
  }

  if (!strcmp(devProperty,"REMOVE"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_LONG) return illegal_format;
    switch (din->dArrayLength)
    {
      case 2:
        cc = RemoveArchiveEventsFunction(din->data.lptr[0],din->data.lptr[1],devName);
        if (cc == not_implemented) cc = RemoveArchiveEventFunction(din->data.lptr[0],devName);
        break;
      case 1:
        cc = RemoveArchiveEventsFunction(din->data.lptr[0],din->data.lptr[0],devName);
        if (cc == not_implemented) cc = RemoveArchiveEventFunction(din->data.lptr[0],devName);
        break;
      default:
        return dimension_error;
    }
    return cc;
  }

  if (!strcmp(devProperty,"ANNOTATE"))
  {
    char *c,tmpstr[512];
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_TEXT) return illegal_format;
    if (din->dArrayLength == 0) return dimension_error;
    strncpy(tmpstr,din->data.cptr,511); tmpstr[511] = 0;
    if (tmpstr[0] != '[' || (c=strtok(tmpstr,"]")) == NULL) return illegal_delimiter;
    if ((c=strtok(NULL,"]")) == NULL) return illegal_delimiter;
    eventNumber = atoi(&tmpstr[1]);
    return AnnotateArchiveEventFunction(eventNumber,devName,c);
  }

  // below are simple test properties which exercise client calls to an event-server:

  if (!strcmp(devProperty,"TEST"))
  {
     return SetGlobalEvent(1234,"This is a test");
  }
  if (!strcmp(devProperty,"Q:NEVENTS"))
  {
    if ((cc=GetNumberOfEvents(0,time(NULL))) >= 0)
      printf("nevents: %d\n",cc);
    return cc < 0 ? -cc : 0;
  }
  if (!strcmp(devProperty,"Q:EVENTS"))
  {
    num = 20;
    if ((cc=GetEvents(0,time(NULL),0,aes,&num)) >= 0)
    {
      printf("%d events received: \n",cc);
      for (i=0; i<20 && i<cc; i++)
        printf(" %ld %ld %s %x%x%x%x%x%x%x%x%x\n",
             aes[i].eventNumber,aes[i].eventCode,aes[i].comment[0],
             aes[i].dsMask[0],aes[i].dsMask[1],aes[i].dsMask[3],
             aes[i].dsMask[4],aes[i].dsMask[5],aes[i].dsMask[6],
             aes[i].dsMask[7],aes[i].dsMask[8],aes[i].dsMask[9]);
    }
    return cc < 0 ? -cc : 0;
  }
  if (!strcmp(devProperty,"Q:NSRV"))
  {
    if ((cc = GetNumberOfArchiveServers()) >= 0) printf("num archive servers: %d\n",cc);
    return cc < 0 ? -cc : 0;
  }
  if (!strcmp(devProperty,"Q:SRV"))
  {
    num = 10;
    if ((cc = GetArchiveServers(-1,srv,&num)) >= 0)
    {
      printf("archive servers:\n");
      for (i=0; i<10; i++) printf("%s\n",srv[i].name);
    }
    return cc < 0 ? -cc : 0;
  }
  if (!strcmp(devProperty,"Q:NDEV"))
  {
    if ((cc = GetNumberOfDeviceServers(NULL)) >= 0) printf("num device servers: %d\n",cc);
    return cc < 0 ? -cc : 0;
  }
  if (!strcmp(devProperty,"Q:DEV"))
  {
    num = 10;
    if ((cc = GetDeviceServersX(NULL,-1,dev,&num)) >= 0)
    {
      printf("device servers:\n");
      for (i=0; i<10; i++) printf("%s\n",dev[i].name);
    }
    return cc < 0 ? -cc : 0;
  }

  return illegal_property;
}
EXPORTTAG int initArchiveServer(void)
{
  RegisterEquipmentModule(NULL,"ARCEQM",0,arceqm,arcini,arctsk,100,NULL);
  return 0;
}
EXPORTTAG int initArchiveClient(void)
{
  arcini();
  return 0;
}
