#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "errors.h"
#include "eventlst.h"
#include "srvlst.h"
#include "arceqm.h"
#include "swaplib.h"
#include "rpctypes.h"

/*
   The code section below illustrates how one registers a structure
   to be exported.  All network structures should come in and go out
   int INTEL byte order with 1-byte allignment.  INTEL to INTEL does
   not need to swap.

   Note the use of the short "reserved" in the structure below.
   Although not necessary, it is strongly recommended that you
   assume structures to be alligned on 4-byte boundaries and fill
   in any gaps with reserved or dummy tags.  This will make life
   much easier in Visual Basic 4.0 (32-bit) where structures are
   by definition alligned on 4-byte boundaries.  If you don't follow
   this suggestion, then you will have to (in Visual basic 4.0) read
   your structure into a byte array and parse it out into your
   desired VB structure.
*/

#define dumpStructInfo(tag) {\
  if ((s=getStructFromTag(tag)) != NULL) {\
    printf("%s: num %d size %d rawlen %d\n",tag,s->num,s->size,s->rawlen);\
  } else printf("%s: not found\n",tag);\
}
#define quit(i) { printf("ABORT: %s\n",erlst[i]); exit(1); }
void registerStructs(void)
{
  /* this must follow the order of the structure explicitly! */

  if (fillstruct("AESNOT",OFFSETIN(AESnotify,version),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",OFFSETIN(AESnotify,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",OFFSETIN(AESnotify,eventCode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",OFFSETIN(AESnotify,reserved),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",OFFSETIN(AESnotify,comment),4*64,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("AESNOT",sizeof(AESnotify),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("AESCON",OFFSETIN(AEScontext,version),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESCON",OFFSETIN(AEScontext,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESCON",OFFSETIN(AEScontext,eventCode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESCON",OFFSETIN(AEScontext,reserved),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AESCON",OFFSETIN(AEScontext,deviceName),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("AESCON",sizeof(AEScontext),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("AES",OFFSETIN(AES,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AES",OFFSETIN(AES,eventCode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AES",OFFSETIN(AES,comment),4*64,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("AES",OFFSETIN(AES,dsMask),32,CF_BYTE)) quit(out_of_local_memory);
  if (fillstruct("AES",sizeof(AES),100,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("ADRB",OFFSETIN(ADRB,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,starttime),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,stoptime),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.deviceServerName),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.deviceProperty),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.deviceName),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.dataFormat),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",OFFSETIN(ADRB,devSpecifier.dataDimension),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("ADRB",sizeof(ADRB),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("ADRBX",OFFSETIN(ADRBX,eventNumber),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,starttime),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,stoptime),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,devSpecifier.deviceContext),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,devSpecifier.deviceServerName),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,devSpecifier.deviceProperty),32,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,devSpecifier.deviceName),32,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,devSpecifier.dataFormat),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,devSpecifier.reserved),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,devSpecifier.dataIndex),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",OFFSETIN(ADRBX,devSpecifier.dataDimension),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("ADRBX",sizeof(ADRBX),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("AIB",OFFSETIN(AIB,numProperties),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("AIB",OFFSETIN(AIB,numDevices),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("AIB",OFFSETIN(AIB,eventCode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("AIB",OFFSETIN(AIB,comment),64,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("AIB",sizeof(AIB),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("PIB",OFFSETIN(PIB,SamplingRate),1,CF_FLOAT)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,numPreTriggerValues),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,TimeOffset),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,LowCount),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,HighCount),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,Units),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,LowEGU),1,CF_DOUBLE)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,HighEGU),1,CF_DOUBLE)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,numDevices),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("PIB",OFFSETIN(PIB,reserved),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("PIB",sizeof(PIB),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("DHS",OFFSETIN(DHS,channelID),32,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,eventcode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,arraysize),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,datatype),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,samplingrate),1,CF_FLOAT)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,pretrigger),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,timeoffset),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,lowcount),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,highcount),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,egu),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,egulow),1,CF_FLOAT)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,eguhigh),1,CF_FLOAT)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,datalength),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("DHS",OFFSETIN(DHS,desc),4*64,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("DHS",sizeof(DHS),1,CF_NULL)) quit(out_of_local_memory);

  if (fillstruct("HEADER",OFFSETIN(DHS,channelID),32,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,eventcode),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,arraysize),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,datatype),1,CF_SHORT)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,samplingrate),1,CF_FLOAT)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,pretrigger),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,timeoffset),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,lowcount),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,highcount),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,egu),16,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,egulow),1,CF_FLOAT)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,eguhigh),1,CF_FLOAT)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,datalength),1,CF_LONG)) quit(out_of_local_memory);
  if (fillstruct("HEADER",OFFSETIN(DHS,desc),4*64,CF_TEXT)) quit(out_of_local_memory);
  if (fillstruct("HEADER",sizeof(DHS),1,CF_NULL)) quit(out_of_local_memory);

  return;
}

