#include "prolog.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include "rpctypes.h"
#include "database.h"
#include "errors.h"
#include "srvip.h"
#include "servdef.h"
#include "toolkit.h"
#include "arctools.h"
#include "arceqm.h"
#include "arctypes.h"
#include "lpceqm.h"

time_t getTimeStampFromLogEntry(char *str);
extern int gUseLoopback;
extern int RemoveEvent(UINT32 evnt, char *ext);
extern int RemoveEvents(UINT32 starttime, UINT32 stoptime, char *ext);
extern int EventNotificationPending;
extern UINT32 lastEventNumber;
void setGetArchiveDataFunctionEx(int (*fcn)(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier));
void setGetArchiveHeaderFunctionEx(int (*fcn)(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier));
int makeFullName(char *fname,char *ext);
int getSetNumber(char *arcname);
int writeCMT(char *s,char *fname,int size);
int readCMT(char *s,char *fname,int size);
char *makefilename(char *fname,char *src,char *ext,time_t t);
void AppendArchiveLog(char *str);
int getFiles(NAME8 *filebuf,char *fileext,time_t starttime,time_t stoptime,int bufsiz);
void freeSession(int id);
int GetStoredDeviceNumber(NAME16 *devname,char *tagname);
char *GetStoredDeviceNameFromAlias(char *alias,char *tagname);
int readComment(int setNr,UINT32 evnt,char *txt,int txtsiz)
{
  char fn[128];
  makefilename(fn,stdArchiveDirectory,arb[setNr].ext,evnt);
  return readCMT(txt,fn,txtsiz);
}
int writeComment(int setNr,UINT32 evnt,char *txt)
{
  char fn[128];
  makefilename(fn,stdArchiveDirectory,arb[setNr].ext,evnt);
  return writeCMT(txt,fn,strlen(txt));
}
void initArchiveTask(void);
void bkgArchiveTask(void);
extern int archivedebug;
extern int narb;
extern ARB *arb;
int streqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void filertask(void);
void filerinit(void);
int filer(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void checkSetEvent(ulong eventnumber, int cc);
void checkRecordEvent(int cc);
int GetLocalSessionID(ulong eventNumber,char *ext);
int GetStoredDevices(NAME16 *devnames,char *tagname,char *context,int id,int num);
int GetStoredProperties(NAME16 *props,char *tagname,char *dev,int num);
int idxSelectedARB = 0;

int respondToTrigger(ulong eventNumber,ulong eventCode)
{
  int i;
  char s[80];

  sprintf(s,"Received trigger from event %d",eventNumber);
  AppendArchiveLog(s);
  if (EventNotificationPending) return 0;  /* nothing to do */
  if (eventNumber == lastEventNumber) return 0;  /* nothing to do */
  if (!(eventCode & MARCHSRV_EVENTCODE)) return 0; /* until KRYK has finished testing */
  for (i=0; i<narb; i++) 
    if (arb[i].recordEvents && arb[i].step == 0) arb[i].step = -1;
  return 0;
}
LHS **lhs = NULL;
extern int narb;
LHS *fillLHS(FILE *fp)
{
  int cc = 0, done = 0;
  int prp_col,dev_col,siz_col,fmt_col,rat_col,
      lcnt_col,hcnt_col,egu_col,legu_col,hegu_col;
  char str[256],hdr[256],scratch[16];
  LHS *l,*list = NULL;
  
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str); 
      if ((prp_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((dev_col=findcol(hdr,"DEVICE")) < 0) ccerr(no_such_line);
      if ((siz_col=findcol(hdr,"ARRAYSIZE")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(hdr,"DATATYPE")) < 0) ccerr(no_such_line);
      if ((rat_col=findcol(hdr,"SAMPLINGRATE")) < 0) ccerr(no_such_line);
      if ((lcnt_col=findcol(hdr,"LOWCOUNT")) < 0) ccerr(no_such_line);
      if ((hcnt_col=findcol(hdr,"HIGHCOUNT")) < 0) ccerr(no_such_line);
      if ((egu_col=findcol(hdr,"EGU")) < 0) ccerr(no_such_line);
      if ((legu_col=findcol(hdr,"LOWEGU")) < 0) ccerr(no_such_line);
      if ((hegu_col=findcol(hdr,"HIGHEGU")) < 0) ccerr(no_such_line);
      done = 1;                              /* read header info */
      continue;
    }
    if ((l=(LHS *)calloc(1,sizeof(LHS))) == NULL) ccerr(out_of_server_memory);
    strncpy(l->devProperty,colptr(prp_col,str),32);
    strncpy(l->devServer,colptr(dev_col,str),16);
    strncpy(scratch,colptr(rat_col,str),16);
    l->samplingrate = (float)atof(scratch);
    strncpy(scratch,colptr(fmt_col,str),16); strupr(scratch);
    l->format = getFormatType(scratch);
    strncpy(scratch,colptr(siz_col,str),16);
    l->arraySize = atoi(scratch);
    strncpy(scratch,colptr(lcnt_col,str),16);
    l->lowcount = atoi(scratch);
    strncpy(scratch,colptr(hcnt_col,str),16);
    l->highcount = atoi(scratch);
    strncpy(l->egu,colptr(egu_col,str),16);
    strncpy(scratch,colptr(legu_col,str),16);
    l->egulow = (float)strtod(scratch,NULL);
    strncpy(scratch,colptr(hegu_col,str),16);
    l->eguhigh = (float)strtod(scratch,NULL);
    l->next = list;
    list = l;
  }
err:
  return cc ? NULL : list;  
}
int getLHS(char *ext)
{
  int i;
  FILE *fp;
  static int initialized = 0;
  char s[80];
  
  if (narb == 0) return -1;
  if (!initialized)
  {
    initialized = TRUE;
    if ((lhs=(LHS **)calloc(narb,sizeof(LHS *))) == NULL) return -1;
  }
  
  /* find the header information file: */
  for (i=0; i<narb; i++)
  {
    if (strncmp(arb[i].ext,ext,3)) continue;
    if (lhs[i] != NULL) return i;  /* found it */
    if (arb[i].hdr == NULL || strlen(arb[i].hdr) == 0) return -1;
    sprintf(s,"%s/headerFiles/%s",stdArchiveDirectory,arb[i].hdr);
    if ((fp=fopen(s,"r")) == NULL) return -1;
    lhs[i] = fillLHS(fp);
    fclose(fp);
    return lhs[i] != NULL ? i : -1;
  }
  return -1;  
}
int GetARBIndex(char *srvName)
{
  int i,j;
  if (srvName == NULL || strlen(srvName) == 0) return -1;
  /* try srvName = archive context first: */
  i = getSetNumber(srvName);
  if (i >= 0 && i < narb) return i;

  /* try the selected index  */
  for (j=0; j<arb[idxSelectedARB].nsteps; j++)
  {
    if (!strncmp(srvName,arb[idxSelectedARB].das[j].tagArc,16))
      return idxSelectedARB;
  }
  /* didn't match up, so try everything else */
  for (i=0; i<narb; i++) for (j=0; j<arb[i].nsteps; j++)
  {
    if (!strncmp(srvName,arb[i].das[j].tagArc,16) && arb[i].das[j].acc&READ)
    {
      return i;
    }
  }
  /* wasn't there */
  return -1;
}
int GetLocalArchiveHeader(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier)
{
  int cc=0,id=-1,i=-1,j,k,dsize,fmtsize;
  BYTE *sptr;
  float fdat;
  short sdat;
  UINT32 ldat;
  double ddat;
  LPCstruct *lpc;
  LHS *l;
  char *c,tmpdev[16];
  int devnr = -1;
  char lclprop[32];

  fmtsize = fmtsizeof(devSpecifier->dataFormat);
  dsize = devSpecifier->dataDimension*fmtsize;
  sptr = srcptr(devSpecifier->dataFormat);
  if ((c=GetStoredDeviceNameFromAliasEx(eventContext,devSpecifier->deviceName,devSpecifier->deviceServerName)) != NULL)
  {
    strncpy(tmpdev,c,16);
  }
  else
  {
    devnr = GetStoredDeviceNumberEx(eventContext,(NAME16 *)devSpecifier->deviceName,devSpecifier->deviceServerName);
    sprintf(tmpdev,"#%d",devnr);
  }
  if ((i=GetARBIndex(eventContext)) < 0  &&
      (i=GetARBIndex(devSpecifier->deviceServerName)) < 0) ccerr(un_allocated);
  if ((id=GetLocalSessionID(eventNumber,arb[i].ext)) < 0) ccerr(-id);
  if ((j=getLHS(session[id].ext)) != -1)  /* from config file: */
  {
    for (l=lhs[j]; l != NULL; l=l->next)
    {
      if (strlen(l->devServer) > 0 && 
          strncmp(devSpecifier->deviceServerName,l->devServer,16)) continue;
      if (strncmp(devSpecifier->deviceProperty,l->devProperty,32)) continue;
      strncpy(dhs->channelID,devSpecifier->deviceName,16);
      dhs->eventcode = 0; /* put/get this into session structure */
      dhs->arraysize = (UINT16)l->arraySize; 
      dhs->datalength = l->arraySize; 
      dhs->datatype = l->format;
      dhs->samplingrate = l->samplingrate;
      dhs->pretrigger = 0; /* ? */
      dhs->timeoffset = 0; /* ? */
      dhs->lowcount = l->lowcount;
      dhs->highcount = l->highcount;
      strncpy(dhs->egu,l->egu,16);
      dhs->egulow = l->egulow;
      dhs->eguhigh = l->eguhigh;
      /* copy comment into dhs->desc; */
      readComment(i,eventNumber,&dhs->desc[0][0],256);
      return 0;
    }
    ccerr(un_allocated);
  }
  else /* from archive file: */
  {
    for (k=0; k<MAXEQP; k++) 
      if (EQPlist[k].id == id && 
          !strncmp(EQPlist[k].name,devSpecifier->deviceServerName,16)) break;
    if (k == MAXEQP) ccerr(un_allocated);
    setDataTimeStamp((double)session[id].est);
    if (!session[id].active) ccerr(file_not_open);
    session[id].active = time(NULL);  /* keep session alive */
    for (lpc=EQPlist[k].lpc; lpc; lpc=lpc->Next)
    {
      strncpy(lclprop,devSpecifier->deviceProperty,27); lclprop[27] = 0;
      strcat(lclprop,".HDR");
      if (strncmp(lclprop,lpc->prp,32)) continue;
      if (strlen(lpc->dev) && 
          strncmp(lpc->dev,devSpecifier->deviceName,16) &&
          strncmp(lpc->dev,tmpdev,16)) continue;
      if ((dsize=sizeof(DHS)) > sizeof(archiveBuffer)) ccerr(buffer_too_small);
      if (lpc->siz < devSpecifier->dataDimension) ccerr(dimension_error);
      if (lseek(session[id].hFile,lpc->dataPos,SEEK_SET) == -1) ccerr(file_error);
      if (read(session[id].hFile,archiveBuffer,dsize) == -1) ccerr(file_error);
      memcpy(dhs,archiveBuffer,dsize);
      if (dhs->datatype < CF_DOUBLE) dhs->datatype += CF_DOUBLE;
      /* copy comment into dhs->desc; */
      readComment(i,eventNumber,&dhs->desc[0][0],256);
      return 0;
    }
  }
  ccerr(non_existent);
err:
  if (cc && i >= 0 && arb[i].step != 0) cc = operation_busy;
  if (cc && id >= 0) freeSession(id);
  return cc;
}

int GetLocalArchiveHeaderEx(DHS *dhs,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier)
{
  int cc=0,id=-1,i=-1,j,k,dsize,fmtsize;
  BYTE *sptr;
  float fdat;
  short sdat;
  UINT32 ldat;
  double ddat;
  LPCstruct *lpc;
  LHS *l;
  char *c,tmpdev[16];
  int devnr = -1;
  char lclprop[32];

  fmtsize = fmtsizeof(devSpecifier->dataFormat);
  dsize = devSpecifier->dataDimension*fmtsize;
  sptr = srcptr(devSpecifier->dataFormat);
  if ((c=GetStoredDeviceNameFromAliasEx(eventContext,devSpecifier->deviceName,devSpecifier->deviceServerName)) != NULL)
  {
    strncpy(tmpdev,c,16);
  }
  else
  {
    devnr = GetStoredDeviceNumberEx(eventContext,(NAME16 *)devSpecifier->deviceName,devSpecifier->deviceServerName);
    sprintf(tmpdev,"#%d",devnr);
  }
  if ((i=GetARBIndex(eventContext)) < 0  &&
      (i=GetARBIndex(devSpecifier->deviceServerName)) < 0) ccerr(un_allocated);
  if ((id=GetLocalSessionID(eventNumber,arb[i].ext)) < 0) ccerr(-id);
  if ((j=getLHS(session[id].ext)) != -1)  /* from config file: */
  {
    for (l=lhs[j]; l != NULL; l=l->next)
    {
      if (strlen(l->devServer) > 0 && 
          strncmp(devSpecifier->deviceServerName,l->devServer,16)) continue;
      if (strncmp(devSpecifier->deviceProperty,l->devProperty,32)) continue;
      strncpy(dhs->channelID,devSpecifier->deviceName,16);
      dhs->eventcode = 0; /* put/get this into session structure */
      dhs->arraysize = (UINT16)l->arraySize; 
      dhs->datalength = l->arraySize; 
      dhs->datatype = l->format;
      dhs->samplingrate = l->samplingrate;
      dhs->pretrigger = 0; /* ? */
      dhs->timeoffset = 0; /* ? */
      dhs->lowcount = l->lowcount;
      dhs->highcount = l->highcount;
      strncpy(dhs->egu,l->egu,16);
      dhs->egulow = l->egulow;
      dhs->eguhigh = l->eguhigh;
      /* copy comment into dhs->desc; */
      readComment(i,eventNumber,&dhs->desc[0][0],256);
      return 0;
    }
    ccerr(un_allocated);
  }
  else /* from archive file: */
  {
    for (k=0; k<MAXEQP; k++) 
      if (EQPlist[k].id == id && 
          !strncmp(EQPlist[k].name,devSpecifier->deviceServerName,16)) break;
    if (k == MAXEQP) ccerr(un_allocated);
    setDataTimeStamp((double)session[id].est);
    if (!session[id].active) ccerr(file_not_open);
    session[id].active = time(NULL);  /* keep session alive */
    for (lpc=EQPlist[k].lpc; lpc; lpc=lpc->Next)
    {
      strncpy(lclprop,devSpecifier->deviceProperty,27); lclprop[27] = 0;
      strcat(lclprop,".HDR");
      if (strncmp(lclprop,lpc->prp,32)) continue;
      if (strlen(lpc->dev) && 
          strncmp(lpc->dev,devSpecifier->deviceName,16) &&
          strncmp(lpc->dev,tmpdev,16)) continue;
      if ((dsize=sizeof(DHS)) > sizeof(archiveBuffer)) ccerr(buffer_too_small);
      if (lpc->siz < devSpecifier->dataDimension) ccerr(dimension_error);
      if (lseek(session[id].hFile,lpc->dataPos,SEEK_SET) == -1) ccerr(file_error);
      if (read(session[id].hFile,archiveBuffer,dsize) == -1) ccerr(file_error);
      memcpy(dhs,archiveBuffer,dsize);
      if (dhs->datatype < CF_DOUBLE) dhs->datatype += CF_DOUBLE;
      /* copy comment into dhs->desc; */
      readComment(i,eventNumber,&dhs->desc[0][0],256);
      return 0;
    }
  }
  ccerr(non_existent);
err:
  if (cc && i >= 0 && arb[i].step != 0) cc = operation_busy;
  if (cc && id >= 0) freeSession(id);
  return cc;
}

int GetLocalArchiveData(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPEC *devSpecifier)
{
  int cc=0,id=-1,i=-1,k,dsize,fmtsize;
  BYTE *sptr;
  float fdat;
  short sdat;
  UINT32 ldat;
  double ddat;
  LPCstruct *lpc;
  char *c,tmpdev[16];
  int devnr = -1;
  float scale = 1, shift = 0;
  DHS dhs;

  if (GetLocalArchiveHeader(&dhs,eventNumber,starttime,stoptime,eventContext,devSpecifier) == 0)
  {
     if (dhs.eguhigh && dhs.highcount) scale = dhs.eguhigh;
  }
scale = 1.0;  /* put this back for now, until we get the go-ahead */

  fmtsize = fmtsizeof(devSpecifier->dataFormat);
  dsize = devSpecifier->dataDimension*fmtsize;
  sptr = srcptr(devSpecifier->dataFormat);
  if ((c=GetStoredDeviceNameFromAliasEx(eventContext,devSpecifier->deviceName,devSpecifier->deviceServerName)) != NULL)
  {
    strncpy(tmpdev,c,16);
  }
  else
  {
    devnr = GetStoredDeviceNumberEx(eventContext,(NAME16 *)devSpecifier->deviceName,devSpecifier->deviceServerName);
    sprintf(tmpdev,"#%d",devnr);
  }
  if ((i=GetARBIndex(eventContext)) < 0  &&
      (i=GetARBIndex(devSpecifier->deviceServerName)) < 0) ccerr(un_allocated);
  if ((id=GetLocalSessionID(eventNumber,arb[i].ext)) < 0) ccerr(-id);
  for (k=0; k<MAXEQP; k++) 
    if (EQPlist[k].id == id && !strncmp(EQPlist[k].name,devSpecifier->deviceServerName,16)) break;
  if (k == MAXEQP) ccerr(non_existent_elem);
    
  setDataTimeStamp((double)session[id].est);
  if (!session[id].active) ccerr(file_not_open);
  session[id].active = time(NULL);  /* keep session alive */
  for (lpc=EQPlist[k].lpc; lpc; lpc=lpc->Next)
  {
    if (strncmp(devSpecifier->deviceProperty,lpc->prp,32)) continue;
    if (strlen(lpc->dev) && 
        strncmp(lpc->dev,devSpecifier->deviceName,16) &&
        strncmp(lpc->dev,tmpdev,16)) continue;
    if ((dsize=fmtsizeof(lpc->fmt)*lpc->siz) > sizeof(archiveBuffer)) ccerr(buffer_too_small);
    if (lpc->siz < devSpecifier->dataDimension) ccerr(dimension_error);
    if (lseek(session[id].hFile,lpc->dataPos,SEEK_SET) == -1) ccerr(file_error);
    if (read(session[id].hFile,archiveBuffer,dsize) == -1) ccerr(file_error);
    if (strstr(devSpecifier->deviceProperty,"RAW"))
    {
      return reformatArchiveData((void *)data,devSpecifier->dataFormat,
                                 archiveBuffer,lpc->fmt,devSpecifier->dataDimension,0,0);
    }
    else
    {
      if (scale == 1 && lpc->scale != 1) scale = lpc->scale;
      if (shift == 0 && lpc->shift != 0) shift = lpc->shift;
      return reformatArchiveData((void *)data,devSpecifier->dataFormat,
                                archiveBuffer,lpc->fmt,
                                devSpecifier->dataDimension,scale,shift);
     }
  }
  ccerr(non_existent_elem);
err:
  if (cc && i >= 0 && arb[i].step != 0) cc = operation_busy;
  if (cc && id >= 0) freeSession(id);
  return cc;
}

int GetLocalArchiveDataEx(BYTE *data,ulong eventNumber,time_t starttime,time_t stoptime,char *eventContext,DEVSPECX *devSpecifier)
{
  int cc=0,id=-1,i=-1,k,dsize,dindex,fmtsize;
  BYTE *sptr;
  float fdat;
  short sdat;
  UINT32 ldat;
  double ddat;
  LPCstruct *lpc;
  char *c,tmpdev[16];
  int devnr = -1;
  float scale = 1, shift = 0;
  DHS dhs;

  if (GetLocalArchiveHeaderEx(&dhs,eventNumber,starttime,stoptime,eventContext,devSpecifier) == 0)
  {
     if (dhs.eguhigh && dhs.highcount) scale = dhs.eguhigh;
  }
scale = 1.0;  /* put this back for now, until we get the go-ahead */

  fmtsize = fmtsizeof(devSpecifier->dataFormat);
  dsize = devSpecifier->dataDimension*fmtsize;
  dindex = devSpecifier->dataIndex;
  sptr = srcptr(devSpecifier->dataFormat);
  if ((c=GetStoredDeviceNameFromAliasEx(eventContext,devSpecifier->deviceName,devSpecifier->deviceServerName)) != NULL)
  {
    strncpy(tmpdev,c,16);
  }
  else
  {
    devnr = GetStoredDeviceNumberEx(eventContext,(NAME16 *)devSpecifier->deviceName,devSpecifier->deviceServerName);
    sprintf(tmpdev,"#%d",devnr);
  }
  if ((i=GetARBIndex(eventContext)) < 0  &&
      (i=GetARBIndex(devSpecifier->deviceServerName)) < 0) ccerr(un_allocated);
  if ((id=GetLocalSessionID(eventNumber,arb[i].ext)) < 0) ccerr(-id);
  for (k=0; k<MAXEQP; k++) 
    if (EQPlist[k].id == id && !strncmp(EQPlist[k].name,devSpecifier->deviceServerName,16)) break;
  if (k == MAXEQP) ccerr(non_existent_elem);
    
  setDataTimeStamp((double)session[id].est);
  if (!session[id].active) ccerr(file_not_open);
  session[id].active = time(NULL);  /* keep session alive */
  for (lpc=EQPlist[k].lpc; lpc; lpc=lpc->Next)
  {
    if (strncmp(devSpecifier->deviceProperty,lpc->prp,32)) continue;
    if (strlen(lpc->dev) && 
        strncmp(lpc->dev,devSpecifier->deviceName,16) &&
        strncmp(lpc->dev,tmpdev,16)) continue;
    if ((dsize=fmtsizeof(lpc->fmt)*lpc->siz) > sizeof(archiveBuffer)) ccerr(buffer_too_small);
    if (lpc->siz < devSpecifier->dataIndex+devSpecifier->dataDimension) ccerr(dimension_error);
    if (lseek(session[id].hFile,lpc->dataPos,SEEK_SET) == -1) ccerr(file_error);
    if (read(session[id].hFile,archiveBuffer,dsize) == -1) ccerr(file_error);
    if (strstr(devSpecifier->deviceProperty,"RAW"))
    {
      return reformatArchiveDataEx((void *)data,devSpecifier->dataFormat,
                                 archiveBuffer,lpc->fmt,
                                 devSpecifier->dataIndex,
                                 devSpecifier->dataDimension,0,0);
    }
    else
    {
      if (scale == 1 && lpc->scale != 1) scale = lpc->scale;
      if (shift == 0 && lpc->shift != 0) shift = lpc->shift;
      return reformatArchiveDataEx((void *)data,devSpecifier->dataFormat,
                                archiveBuffer,lpc->fmt,
                                devSpecifier->dataIndex,
                                devSpecifier->dataDimension,scale,shift);
     }
  }
  ccerr(non_existent_elem);
err:
  if (cc && i >= 0 && arb[i].step != 0) cc = operation_busy;
  if (cc && id >= 0) freeSession(id);
  return cc;
}

int AnnotateLocalArchiveEvent(ulong evntime,char *devName,char *cmt)
{
  int setNr;
 
  setNr = getSetNumber(devName);
  if (setNr < 0 || setNr >= narb) return invalid_index;
  if (evntime == (ulong)arb[setNr].tot)
  {
    if (arb[setNr].eventtime == 0) return operation_busy;
    return writeComment(setNr,arb[setNr].eventtime,cmt);
  }
  return writeComment(setNr,evntime,cmt);
}
int RemoveLocalArchiveEvents(ulong evnstart,ulong evnstop,char *dsrv)
{
  int cc,j,k=0,p = 0,found = 0;

  /* try the selected index  */
  for (j=0; j<arb[idxSelectedARB].nsteps; j++)
  {
    if (!strncmp(dsrv,arb[idxSelectedARB].das[j].tagArc,16))
    {
      found = -1;
      k = idxSelectedARB+1;
      break;
    }
  }
  /* if it didn't match up, try everything else */
  for (; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
    if (!strncmp(dsrv,arb[k].das[j].tagArc,16)) found = -1;
  if (!found) return no_such_file;

  /* this will have to match the first time or bail out ! */
  if ((cc=RemoveEvents(evnstart, evnstop, arb[k-1].ext)) < 0) return -cc;
  return 0;
}
int RemoveLocalArchiveEvent(ulong evntime,char *dsrv)
{
  int cc,j,k=0,p = 0,found = 0;

  /* try the selected index  */
  for (j=0; j<arb[idxSelectedARB].nsteps; j++)
  {
    if (!strncmp(dsrv,arb[idxSelectedARB].das[j].tagArc,16))
    {
      found = -1;
      k = idxSelectedARB+1;
      break;
    }
  }
  /* if it didn't match up, try everything else */
scan_arbs:
  for (; k<narb  && !found; k++) for (j=0; j<arb[k].nsteps && !found; j++)
    if (!strncmp(dsrv,arb[k].das[j].tagArc,16)) found = -1;
  if (!found) return no_such_file;

  if ((cc=RemoveEvent(evntime, arb[k-1].ext)) != 0) 
  {
    found = 0;
    goto scan_arbs;
  }
  return cc;
}
int GetLocalnumDevices(char *devSrvName,char *context,int evnnr)
{
  int n = GetStoredDevices(NULL,devSrvName,context,evnnr,0);
  return n > 0 ? n : 0;
}
int GetLocalnumDeviceProperties(char *devSrvName,char *devName)
{
  int n = GetStoredProperties(NULL,devSrvName,devName,0);
  return n > 0 ? n : 0;
}
NAME16 *dsNames = NULL;
int numDsNames = 0;
int GetLocalnumDeviceServers(void)
{
  int num,i,j,k,l;
  static int n = 0;

  if (n == 0)
  {
    for (i=0; i<narb; i++) for (j=0; j<arb[i].nsteps; j++) if (arb[i].das[j].acc&READ)
    {
      for (l=0; l<=i; l++)
      {
        for (k=0; k<j; k++) 
          if (arb[l].das[k].acc&READ && 
              !strncmp(arb[i].das[j].tagArc,arb[l].das[k].tagArc,16)) break;
        if (k < j) break;
      }
      if (k == j && l > i) n++;
    }
    /* worst case n, now allocate and fill in. */
    if ((dsNames=(NAME16 *)calloc(n,sizeof(NAME16))) != NULL)
    {
      numDsNames = n;
      num = n; n = 0;
      for (i=0; i<narb; i++) for (j=0; j<arb[i].nsteps; j++) if (arb[i].das[j].acc&READ)
      {
        for (k=0; k<n; k++) if (!strncmp(arb[i].das[j].tagArc,dsNames[k].name,16)) break;
        if (k == n)
        {
          if (n >= num) return 0;
          strncpy(dsNames[n++].name,arb[i].das[j].tagArc,16);
        }
      }      
    }
  }
  return n;
}
int GetLocalDeviceServers(NAME16 *names,ulong start,ulong stop,int id,int num)
{
  int i,j,k;
  int n = 0;  
  if (dsNames != NULL) 
  {
    if (num > numDsNames) n = numDsNames; else n = num;
    memcpy(names,dsNames,n*16);
    return 0;
  }
  for (i=0; i<narb; i++) for (j=0; j<arb[i].nsteps; j++) if (arb[i].das[j].acc&READ)
  {
    for (k=0; k<n; k++) if (!strncmp(arb[i].das[j].tagArc,names[k].name,16)) break;
    if (k == n)
    {
      if (n >= num) return 0;
      strncpy(names[n++].name,arb[i].das[j].tagArc,16);
    }
  }
  return 0;
}

int GetLocalArchiveInfo(AIB *aib,ulong eventNumber,char *deviceServer)
{
  return 0;
}
int GetLocalPropertyInfo(PIB *pib,ulong eventNumber,char *deviceServer,char *devProperty)
{
  return 0;
}
int GetNumLocalDeviceArchivedEvents(char *devServer,char *devName,ulong starttime,ulong stoptime)
{
  int i;

  /* get das from devServer : */
  if ((i=GetARBIndex(devServer)) < 0) return -non_existent_elem;
  /* find EXT from devServer */
  /* use local filebuf array to hold NAME8 event numbers */

  return getFiles(NULL,arb[i].ext,starttime,stoptime,10000);
}
int GetNumLocalArchivedEvents(ulong starttime,ulong stoptime,ulong eventNumber)
{
  /* scan the CACHE directory ? */
  return 0;
}
NAME8 evnbuf[10000];
int GetLocalDeviceArchivedEvents(char *devServer,char *devName,ulong starttime,ulong stoptime,AES *aes,ulong num)
{
  int i,k,n,p = 0,found = 0;
  char s[16];

  /* get das from devServer : */
  if ((k=GetARBIndex(devServer)) < 0) return non_existent_elem;
  /* find EXT from devServer */
  /* use local filebuf array to hold NAME8 event numbers */

  if (num > 10000) num = 10000;
  if ((n=getFiles(evnbuf,arb[k].ext,starttime,stoptime,num)) < 0) return -n;

  /* fill in aes with atoi(filebuf[]), local comment files, event mask */
  for (i=0,s[8]=0; i<n; i++) 
  {
    strncpy(s,evnbuf[i].name,8); s[8] = 0;
    aes[i].eventNumber = strtol(s,0,16);
  }
  return 0;
}
int GetLocalDeviceArchivedEventList(char *devServer,char *devName,ulong starttime,ulong stoptime,ulong *evnlst,ulong num)
{
  int i,k,n,p = 0,found = 0;
  char s[16];

  /* get das from devServer : */
  if ((k=GetARBIndex(devServer)) < 0) return non_existent_elem;
  /* find EXT from devServer */
  /* use local filebuf array to hold NAME8 event numbers */

  if (num > 10000) num = 10000;
  if ((n=getFiles(evnbuf,arb[k].ext,starttime,stoptime,num)) < 0) return -n;

  /* fill in aes with atoi(filebuf[]), local comment files, event mask */
  for (i=0,s[8]=0; i<n; i++) 
  {
    strncpy(s,evnbuf[i].name,8); s[8] = 0;
    evnlst[i] = strtol(s,0,16);
  }

  return 0;
}
int GetLocalArchivedEvents(ulong starttime,ulong stoptime,AES *aes,ulong num)
{
  /* scan the CACHE directory ? */
  return 0;
}
int GetLocalDevices(NAME16 *names,char *devSrvName,char *context,int id,int num)
{
  /* react to id = event number */
  int n = GetStoredDevices(names,devSrvName,context,id,num); 
  return n < 0 ? -n : 0;
}
int GetLocalDeviceProperties(NAME16 *props,char *devSrvName,char *devName,int num)
{
  /* react to id = event number */
  int n = GetStoredProperties(props,devSrvName,devName,num); 
  return n < 0 ? -n : 0;
}
void PreSystemInit(void)
{
  ListenForMCasts = FALSE;
  MaxNumConnections = 500;
  maxFdSets = 512;
  ListenForGlobals = 0;
  gRetardSingleContractRemoval = FALSE;
  gUseLoopback = TRUE;
  SystemPresetMemory = TRUE;
  MaxRPCTransportSize = MAX_TRANSPORTSIZE*8;
  return;
}
void PostSystemInit()
{
  int i;
  
  initArchiveServer();
  setSetEventCallback(checkSetEvent);
  setTriggerCallback(respondToTrigger);
  setRecordEventCallback(checkRecordEvent);
  setGetArchiveDataFunction(GetLocalArchiveData);
  setGetArchiveDataFunctionEx(GetLocalArchiveDataEx);
  setRemoveArchiveEventFunction(RemoveLocalArchiveEvent);
  setRemoveArchiveEventsFunction(RemoveLocalArchiveEvents);
  setAnnotateArchiveEventFunction(AnnotateLocalArchiveEvent);
  setQueryLocalDeviceServerNamesFunction(GetLocalDeviceServers);
  setQueryNumLocalDeviceServersFunction(GetLocalnumDeviceServers);

  setQueryArchiveInfoFunction(GetLocalArchiveInfo);
  setQueryPropertyInfoFunction(GetLocalPropertyInfo);
  setQueryNumLocalArchivedEventsFunction(GetNumLocalArchivedEvents);
  setQueryLocalArchivedEventsFunction(GetLocalArchivedEvents);

  setGetArchiveHeaderFunction(GetLocalArchiveHeader);
  setGetArchiveHeaderFunctionEx(GetLocalArchiveHeaderEx);

  setQueryNumLocalDevicesFunction(GetLocalnumDevices);
  setQueryLocalDeviceNamesFunction(GetLocalDevices);

  setQueryNumLocalDevicePropertiesFunction(GetLocalnumDeviceProperties);
  setQueryLocalDevicePropertiesFunction(GetLocalDeviceProperties);

  setQueryDeviceArchivedEventsFunction(GetLocalDeviceArchivedEvents);
  setQueryDeviceArchivedEventListFunction(GetLocalDeviceArchivedEventList);
  setQueryNumDeviceArchivedEventsFunction(GetNumLocalDeviceArchivedEvents);

  RegisterEquipmentModule(NULL,"STREQM",0,streqm,initArchiveTask,bkgArchiveTask,100,NULL);
  RegisterEquipmentModule(NULL,"FILER",0,filer,filerinit,filertask,100,NULL);

  for (i=0; i<narb; i++) getLHS(arb[i].ext);
}
int forwardTrigger(int setNr)
{
  DTYPE din, *dinptr = NULL;
  int i, cc = link_not_open, retries= 2;

  if (arb[setNr].fwddata > 0)
  {
    din.dFormat = CF_SHORT;
    din.dArrayLength = 1;
    din.data.lptr = (UINT32 *)&arb[setNr].fwddata;
    dinptr = &din;
  }
  for (i=0; i<retries && cc != 0; i++)
  {
    cc = ExecLinkEx(arb[setNr].fwdsrv,arb[setNr].fwdprp,NULL,dinptr,CA_WRITE,800);
  }
  feclog("%s -> forward trigger to %s %s (%d tries): %s",
    arb[setNr].server,arb[setNr].fwdsrv,arb[setNr].fwdprp,i,
    cc ? GetRPCLastError((short)cc) : "success");
  return cc;

}
int getSetNumber(char *arcname)
{
  int i;
  for (i=0; i<narb && strcmp(arcname,arb[i].name); i++);
  return i == narb ? -1 :  i;
}
int streqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int len,cc,setNr;
  short firststep;
  char name[9],fullName[80],s[256];
  time_t tot;
  char *c,*t;
  UINT32 evnt;
  static int lastSetNr = -1;
  static time_t lastTime = 0;

  if (!strcmp(devProperty,"TRIGGER"))
  {
    setNr = getSetNumber(devName);
    sprintf(s,"%s : TRIGGER %s (set nr %d)",GetCaller(),devName,setNr);
    AppendArchiveLog(s);
    if (setNr < 0 || setNr >= narb) return invalid_index;
    if (access&CA_WRITE && din->dArrayLength == 1)
    {
      if (arb[setNr].step || arb[setNr].fp) return operation_busy;
      if ((cc=getValuesAsShort(din,&firststep,1)) != 0) return cc;
      arb[setNr].tot = RPCtimestamp;
      if (firststep == 0) return unauthorised_action;
      tot = time(NULL);
      if (lastSetNr == setNr && tot == lastTime) return not_accepted;
      lastSetNr = setNr; lastTime = tot;
      arb[setNr].eventtime = tot;   /* set this temporarily */
      if (EventNotificationPending) /* I'm working on this event already */
      {
        arb[setNr].resultPending = TRUE;
      }
      arb[setNr].step = firststep;      
      if (strlen(arb[setNr].fwdsrv) > 0) forwardTrigger(setNr);
    }
    if (dout->dArrayLength == 1)
    {
      if (dout->dFormat == CF_SHORT)     /* return current archive step */
        dout->data.sptr[0] = arb[setNr].step; 
      else if (dout->dFormat == CF_LONG) /* return current time as identifier */
        dout->data.lptr[0] = arb[setNr].eventtime;  
      else return illegal_format;
    }
    return 0;
  }
  if (!strcmp(devProperty,"SELECT"))  /* select a context */
  {
    if (din->dArrayLength > 0)
    {
      setNr = getSetNumber(devName);
      if (setNr < 0 || setNr >= narb) return invalid_index;
      idxSelectedARB = setNr;
    }
    if ((len=dout->dArrayLength) > 0)
    {
      switch (dout->dFormat)
      {
        case CF_TEXT: strncpy(dout->data.cptr,arb[idxSelectedARB].name,len);
                      break;
        case CF_NAME16: strncpy(dout->data.cptr,arb[idxSelectedARB].name,16);
                      break;
        default: return illegal_format;
      }
    }
    return 0;
  }
  if (!strcmp(devProperty,"COMMENT"))  /* read a comment */
  {
    setNr = getSetNumber(devName);
    if (setNr < 0 || setNr >= narb) return invalid_index;
    if (din->dArrayLength == 0) return illegal_read_write;
    if (dout->dFormat != CF_TEXT) return illegal_format;
    if (dout->dArrayLength < 1) return dimension_error;
    switch (din->dFormat)
    {
      case CF_LONG: evnt = din->data.lptr[0]; break;
      case CF_TEXT: evnt = atoi(din->data.cptr); break;
      default: return illegal_format;
    }
    return readComment(setNr,evnt,dout->data.cptr,dout->dArrayLength);
  }
  if (!strcmp(devProperty,"ANNOTATE"))  /* add a comment */
  {
    setNr = getSetNumber(devName);
    if (setNr < 0 || setNr >= narb) return invalid_index;
    if (access&CA_WRITE && din->dArrayLength > 0)
    {
      if (din->dFormat != CF_TEXT) return illegal_format;
      if (strstr(din->data.cptr,"@") == NULL) return out_of_range;
      if ((c=strtok(din->data.cptr,"@")) == NULL) return out_of_range;
      if ((t=strtok(NULL,"@")) == NULL) out_of_range;
      tot = atoi(t);
      if (tot < 100) /* not UTC format */
      {
        tot = getTimeStampFromLogEntry(t);
      }
      if (tot == arb[setNr].tot)
      {
        if (arb[setNr].eventtime == 0) return operation_busy;
        return writeComment(setNr,arb[setNr].eventtime,c);
      }
      return writeComment(setNr,tot,c);
    }
    return illegal_read_write;
  }
  if (!strcmp(devProperty,"REMOVE"))                       /* CF_STRUCT */
  {
    setNr = getSetNumber(devName);
    if (setNr < 0 || setNr >= narb) return invalid_index;
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_LONG) return illegal_format;
    if (din->dArrayLength != 1) return no_file_space;
    tot = din->data.lptr[0];
    if (tot == arb[setNr].tot)
    { /* request removal on most recent archive for this set (on going?) */
      if (arb[setNr].eventtime == 0 || arb[setNr].step != 0)
      {
        arb[setNr].step = arb[setNr].nsteps+1;
        arb[setNr].aborted = TRUE;
        return 0;
      }
      sprintf(name,"%x",arb[setNr].eventtime);
      sprintf(fullName,"%s.%s",name,arb[setNr].ext);
      if (makeFullName(fullName,arb[setNr].ext)) return no_such_file;
      if (arb[setNr].step == 0) /* not actively archiving  */
      {
        if (archDebug) printf("STR: remove file %s\n",fullName);
        remove(fullName);
        sprintf(&fullName[strlen(fullName)-strlen(arb[setNr].ext)],"CMT");
        if (archDebug) printf("STR: remove file %s\n",fullName);
        remove(fullName);
        return 0;
      }
    }
    else
    { /* removal of stored event */
      sprintf(fullName,"%x.%s",tot,arb[setNr].ext);
      if (makeFullName(fullName,arb[setNr].ext)) return no_such_file;
      if (archDebug) printf("STR: remove file %s\n",fullName);
      remove(fullName);
      sprintf(&fullName[strlen(fullName)-strlen(arb[setNr].ext)],"CMT");
      if (archDebug) printf("STR: remove file %s\n",fullName);
      remove(fullName);
      return 0;
    }
  }
  return illegal_property;
}
